/*
	Alib -- Basic 2-D drawing routines with Z-buffer for polygons
*/

#ifndef _ALIB_H
#define _ALIB_H

#include <stdio.h>
#include "Vlibmath.h"
#include "VPoly.h"
#ifdef HAVE_X11_DBE
	#include <X11/Xlib.h>
	#include <X11/extensions/Xdbe.h>
#endif


#ifdef EXTERN_ALIB
	#define EXTERN
#else
	#define EXTERN extern
#endif


#ifdef WIN32

	#define HAS_FRAME_BUFFER
	#include <stdlib.h>
	/*#include <crtdbg.h>*/

	#define printf  acm_printf
	#define fprintf acm_fprintf

	#include <stdio.h>

	extern int _cdecl acm_printf( const char *format, ... );
	extern int _cdecl acm_fprintf( FILE *, const char *format, ... );
	extern void FrameBufferIOInit(), FrameBufferRun();

	typedef struct {
		short     x, y;
	} Point;

	typedef struct {
		short     x1, y1, x2, y2;
	} Segment;

	typedef struct {
		short     x, y;
		unsigned short width, height;
	} Rectangle;

	typedef Point XPoint;
	typedef Segment XSegment;
	typedef Rectangle XRectangle;
	typedef int Bool;

	#define DisplayHeightMM(x, s)  (168)
	#define DisplayWidthMM(x, s)   (224)
	#define RootWindow(x, s)	0
	#define BlackPixel(x, s) 0
	#define WhitePixel(x, s) 1

	#define XParseColor(d, c, name, color)	PCParseColor(name, color)

#else

	#define X11

	#undef HAS_FRAME_BUFFER			/* use direct frame buffer output */
	#define X11_DRAW_SEGMENTS		/* draw segments instead of rectangles */
	#include <X11/Xlib.h>

	typedef XPoint Point;
	typedef XSegment Segment;
	typedef XRectangle Rectangle;

#endif

#define MAXCOLORS 8     /* max number of colors available */
						/* when double buffering */

#if (MAXCOLORS==4)
	#define PLANECOUNT 2
#endif
#if (MAXCOLORS==8)
	#define PLANECOUNT 3
#endif
#if (MAXCOLORS==16)
	#define PLANECOUNT 4
#endif

/*
	(VWorkContext)->flags:
*/

#define ColorEnableDepthCueing  1

typedef struct {
	int       flags;
	VColor   *VColorList;
	int       usePixmaps;
	int       nextPixel;		/* next pixel cell to allocate */

	/* Set by AlibSetVisibility(): */
	int       depthCueSteps;	/* 0 = no haze */
	VColor   *depthCueColor;	/* haze color */
	double    visibility;		/* visibility (in feet) */
	double   *visTable;
} VWorkContext;


EXTERN VWorkContext _v_default_work_context;
EXTERN VWorkContext *_VDefaultWorkContext
	#ifdef EXTERN_ALIB
	= &_v_default_work_context
	#endif
	;

void SetPoint(Point *p, int x, int y);

/*
	A Rect is exactly (b.x - a.x) pixels wide and (b.y - a.y) pixels tall,
	being a,b two geometrical points.

	This library assumes a.x < b.x and a.y < b.y, otherwise the Rect
	is considered empty.

	Please note that the pixel "a" is inside, while "b" is outside.
*/

typedef struct {
	Point a, b;
} Rect;


void SetRect(Rect *r, int ax, int ay, int bx, int by);
/*
	Set the rectangle. For example, if your screen is W pixel wide
	and H pixel tall, it's area can be set with:

	Rect screen;
	SetRectangle(&screen, 0, 0, W, H);
*/

int IsEmptyRect(Rect *r);
/*
	Return 1 if the given rectangle is empty, 0 otherwise.
*/

void IntersectRect(Rect *r1, Rect *r2, Rect *res);
/*
	Return in res the intersection between r1,r2. If the result is empty,
	return (Rect){{0,0},{0,0}}. Rectangles may overlap in memory.
*/

void ExpandRect(Rect *r, int dx, int dy);
/*
	Expands the given rectangle around its center so that it becomes 2*dx
	pixels wider and 2*dy pixel taller.
*/

#define RectWidth(r) ((r).b.x - (r).a.x)
#define RectHeight(r) ((r).b.y - (r).a.y)
#define RectMiddleX(r) (((r).b.x + (r).a.x)/2)
#define RectMiddleY(r) (((r).b.y + (r).a.y)/2)


typedef unsigned short Color;

#define UnchangedColor	0xffff	/* special Color value */


typedef struct _ZInfo {
	Color     color;			/* color of this polygon */
	unsigned long depth;		/* depth of this polygon */
	struct _ZInfo *next;		/* next polygon in plane sweep set */
	struct _ZInfo *prev;		/* previous polygon in plane sweep set */
} ZInfo;


#define NotAnElement		((ZInfo *) -1)

#define MaxDepth		0xFFFFFFFF
#define MinDepth		0

/*
 *  Private data types
 */

typedef struct _Edge         Edge;
typedef struct _EdgeList     EdgeList;
typedef struct _ScanLine     ScanLine;
typedef struct _ColorSegment ColorSegment;


#ifdef X11

typedef struct {
#ifdef X11_DRAW_SEGMENTS
	XSegment *head;				/* pointer to a vector of BufferedSegment's */
#else
	XRectangle *head;			/* pointer to a vector of BufferedSegment's */
#endif
	int       count;			/* number of seg's currently in use */
} SegmentGroup;

#define AMaxPixelValue	255		/* we'll only handle up-to 8-bit Visuals */

#endif

/*
	(AWindow)->flags:
*/

#define VPMono            4  /* Monochrome environment */
#define VPPixmap          8  /* Use color Pixmap rather than double buffering */
#define VPFastAnimation  16  /* frame-by-frame comparison */
#define VPDepthCueing    32  /* Perform depth cueing */
#define VPDepthCueParsed 64  /* Pixel has been assigned to dc color */
#define VPDoubleBuffer  128  /* use double buffering, if available */


typedef struct _AWindow {
#ifdef X11
	Display  *display;			/* X display for this window */
	int       screen;			/* X screen number */
	Window    win;
	Colormap  cmap;
	Drawable  d;				/* dest drawable for this window */
	GC        gc;
#endif
	int       width, height;	/* dimensions of this window */
	Rect      rect;             /* (0,0)-(width,height) */
	Rect      clip;				/* clipping bounds (0,0)-(width,height) */
	int       flags;            /* (for internal use only) */
	int       set;				/* id of buffer currently in use */
	Color    *aPixel;
	Color    *pixel;            /* current pixel drawing values */
	Pixmap    monoPixmap;		/* Pixmap used to buffer drawing */
#ifdef HAVE_X11_DBE
	XdbeBackBuffer dbeDrawable;
	unsigned long lastPixel;
#else
	#warning "X Window DBE support not enabled, using slower routines"
#endif
	unsigned long mask;			/* current plane mask */
	int       bsegSize;			/* Count of elements in each bseg vec */
	SegmentGroup bseg[AMaxPixelValue + 1];
	unsigned long color_to_pixel_map[AMaxPixelValue+1];
	long color_to_pixel_map_top;
	Pixmap   *stipple;			/* pointer to a list of pixmaps */
	int       ymin, ymax;		/* range of EdgeList's with polygons */
	int       doubleBuffered;	/* 0=single video buffer; 1=double buffer */
	ScanLine *scanLine;			/* a vector with height elements */
	ScanLine *lastScanLine;		/* a vector with height elements */
	ScanLine *otherLastScanLine;	/* a vector with height elements */
	EdgeList *edges;			/* a vector with height elements */
	EdgeList *lines;			/* a vector with height elements */
	Edge     *edgePool;			/* a pool of polygon edges */
	unsigned int EPSize;		/* number of entries in edgePool */
	unsigned int EPTop;			/* index of first free edge */
	unsigned int curPool;		/* selects csPool (0 or 1 (or 2 when double buffering)) */
	ColorSegment *csPool0;
	unsigned int CSSize0;		/* number of entries in csPool0 */
	unsigned int CSTop0;		/* index of first free color seg */
	ColorSegment *csPool1;
	unsigned int CSSize1;		/* number of entries in csPool1 */
	unsigned int CSTop1;		/* index of first free color seg */
	ColorSegment *csPool2;
	unsigned int CSSize2;		/* number of entries in csPool2 */
	unsigned int CSTop2;		/* index of first free color seg */


	unsigned long depth;		/* polygon depth while plotting */
	int       colors;			/* color count */
	unsigned long aMask, bMask;
	Color bPixel[MAXCOLORS];
	XColor    aColor[MAXCOLORS * MAXCOLORS];
	XColor    bColor[MAXCOLORS * MAXCOLORS];
	ZInfo    *zpool;
	long      zsize;
	long      ztop;
	XColor    xdepthCueColor;
	Visual   *visual;

	void      (*colorTweakProc) (struct _AWindow *w,
	                        char *color_name, char *real_color);
	Status    (*AllocColor)(struct _AWindow *w,
							Colormap colormap, 
							XColor *c);
	void      (*SetClipRect)(struct _AWindow *w, Rect *r);
	void      (*DrawLine)(struct _AWindow *w, 
						  int x1, 
						  int y1, 
						  int x2, 
						  int y2,
						  Color color);
	void      (*DrawSegments)(struct _AWindow *w, 
							  Segment * seg, 
							  int nseg, 
							  Color color);
	void      (*FillRectangle)(struct _AWindow *w, Rect *r, Color color);
	void      (*FillPolygon)(struct _AWindow *w, Point *pts, int npts,
			ZInfo *z);
	void      (*ExposeBuffer)(struct _AWindow *w);
} AWindow;


#define AMonoMode	1			/* monochrome flag */


#ifndef X11
	typedef int Status;
	typedef struct {
		unsigned short red, green, blue;
		unsigned long rgb;
		unsigned short pixel;
		unsigned short flags;
	} XColor;

	typedef int Drawable;
	typedef int Display;
	typedef int Window;
	typedef int Pixmap;
	typedef int Colormap;
	typedef struct _visual {
		int       map_entries;
	} Visual;

	#define GC	long
	#define AllPlanes (-1)
	#define True	(1)
	#define False	(0)
#endif


EXTERN int _AlibDoubleBuffered
	#ifdef EXTERN_ALIB
	= 0
	#endif
	;



AWindow * AlibOpen(Display * display, int screen, Window win, Colormap cmap,
    Visual * visual, int width, int height);
/*
	Create a new window of the given size and set the clipping rectangle
	to the whole available space (0,0)-(width,height).
*/

void AlibResize(AWindow * w, int width, int height);
/*
	Resize the AWindow, intersection the current clipping rectangle
	with the new area.
*/

void AlibClose(AWindow *w);

VColor *VAllocColor(char *);
/*
	Allocate a non depth-cued color.
*/

int AlibSetDepthCueing(AWindow *w, int flag);
/*
	Enable (if flag=TRUE) or disbale (if flag=FALSE) depth-cued color on
	this window.
*/

void AlibSetVisibility(double visibility, VColor * haze_color, int steps);
/*
	Set visibility (meters), haze color and number of steps for color
	depth-cueing.
*/

VColor *VAllocDepthCueuedColor(char * color, int flag);
/*
	Generates a table of colors ranging from "color" to the current
	haze color as set by AlibSetVisibility().
*/

int VBindColors(AWindow *, char *);

void AlibSetClipRect(AWindow *w, Rect *r);
/*
	Set the clipping rectangle w->clip as the intersection between r and
	the whole available drawing area (0,0)-(w->width,w->height). Drawing
	will never be performed outside this rectangle.
*/

void DrawPoint(AWindow * w, int x, int y, ZInfo * zinfo);
void DrawLine(AWindow * w, int x1, int y1, int x2, int y2, ZInfo * zinfo);
void FillRectangle(AWindow * w, Rect *r, ZInfo * zinfo);
void FillPolygon(AWindow * w, Point * pts, int npts, ZInfo * zinfo);
/*
	Basic 2-D drawing routines with color z-buffer that draw on the given
	window and perform clipping to the current clipping rectangle as set by
	AlibSetClip().
*/

void ForceWindowRedraw(AWindow *);
/*
	Invalidate the previous frame's drawing information so that the
	animation is completely redrawn at the completion of the current
	frame.
*/

void AlibDrawLine(AWindow *w, int x1, int y1, int x2, int y2, Color color);
void AlibDrawArc(AWindow *w, int x, int y,
	int width, int height,
	int angle1, int angle2, Color color);
void AlibDrawSegments(AWindow *w, Segment * seg, int nseg, Color color);
void AlibDrawRect(AWindow *w, Rect *r, Color color);
void AlibFillRectangle(AWindow *w, Rect *r, Color color);
void AlibFillPolygon(AWindow *w, Point * pts, int npts, Color color);
/*
	Basic 2-D drawing routines that draw on the given window and perform
	clipping to the current clipping rectangle as set by AlibSetClip().
*/

Status AlibAllocColor(AWindow *w, Colormap colormap, XColor *c);
/*
	Allocate non depth cued color. Return 0 on success, non zero on
	failure.

	FIXME: client modules should call w->AllocColor() instead.
*/

void AlibExposeBuffer(AWindow *w);

unsigned long VConstantColor(AWindow *w, int cIndex);

int AlibComputePolygonColor(AWindow *w, VPolygon * poly);

void Alib_FreeAll();
/*
	Release internally allocated spare memory.
*/

#undef EXTERN_ALIB
#undef EXTERN

#endif /* _ALIB_H */
