#ifndef _VPOLY_H
#define _VPOLY_H

#include <stdio.h>
#include "Vlibmath.h"

#ifdef EXTERN_VPOLY
	#define EXTERN
#else
	#define EXTERN extern
#endif

#define VmaxVP  2048            /* max # of vertices in a polygon */
#define NUM_ASPECTS 10

typedef struct _vcolor {
	char     *color_name;       /* text name */
	short     flags;
	short     cIndex;           /* index into pixel value or halftone table */
	struct _vcolor *next;
} VColor;


/*
	(VPolygon)->flags:
*/

#define PolyClipBackface    0x01    /* Polygon is "one-sided" */
#define PolyUseBackColor    0x02    /* should be rendered using backColor */
#define PolyNormalValid     0x04    /* "normal" contains valid info */
#define PolyUseCullDistance 0x08    /* cullDistance should be tested before plotting */

typedef struct {
	short     flags;            /* (see flags above) */
	short     numVtces;         /* vertex count */
	short     assignedDepth;    /* hard-assigned depth cue or -1 */
	VPoint    normal;           /* normal vector plus vertex[0] */
	VPoint   *vertex;           /* pointer to array of vertices (& normal) */
	VColor   *color;            /* pointer to color descriptor */
	VColor   *backColor;        /* pointer to back's color (sometimes NULL) */
	double    cullDistance;     /* this polygon is ignored beyond this distance */
} VPolygon;


struct _euler {
	double    yaw;
	double    pitch;
};

struct entry {
	int       id;
	double    d;
	struct entry *next;
};


/*
	Views corresponding to each of the eight defined major aspects
*/

EXTERN struct _euler view[]
	#ifdef EXTERN_VPOLY
	= {
		{DEGtoRAD(90.0), DEGtoRAD(0.0)},        /* 0 - right */
		{DEGtoRAD(-90.0), DEGtoRAD(0.0)},       /* 1 - left */
		{DEGtoRAD(45.0), DEGtoRAD(45.0)},       /* 2 - front right bottom */
		{DEGtoRAD(45.0), DEGtoRAD(-45.0)},      /* 3 - front right top    */
		{DEGtoRAD(-45.0), DEGtoRAD(45.0)},      /* 4 - front left bottom  */
		{DEGtoRAD(-45.0), DEGtoRAD(-45.0)},     /* 5 - front left top     */
		{DEGtoRAD(135.0), DEGtoRAD(45.0)},      /* 6 - aft right bottom   */
		{DEGtoRAD(135.0), DEGtoRAD(-45.0)},     /* 7 - aft right top      */
		{DEGtoRAD(-135.0), DEGtoRAD(45.0)},     /* 8 - aft left bottom    */
		{DEGtoRAD(-135.0), DEGtoRAD(-45.0)},    /* 9 - aft left top       */
	}
	#endif
	;

char *VGetAspectName (int aspect);

VPolygon *VCreatePolygon (int npts, VPoint * pts, VColor *);

#define VDestroyPolygon(a)  {if((a)!=NULL){FREE(a->vertex); FREE(a);}}

/*
	VPolySet data type is a set that can contain zero or more VPolygon(s).
	Polygons can be added but cannot be removed. Typical usage:
*/

typedef struct _VPolySet VPolySet;

VPolySet * VPolySet_New();
/*
	Allocates a new set of polygons.
*/

void       VPolySet_Free(VPolySet *s, int release_polygons_too);
/*
	Release the set from memory. The flag says if also the VPolygon(s) the
	set points to have to be released with VDestroyPolygon().
*/

void       VPolySet_Empty(VPolySet *s, int release_polygons_too);
/*
	Empties the set to zero elements without actually releasing the memory,
	so the set can be re-used several times. The flag says if the polygons
	have to released with VDestroyPolygon().
*/

void       VPolySet_Add(VPolySet *s, VPolygon *p);
/*
	Add the polygon to the set. Polygons are not copied, only their pointer
	it is.
*/

VPolygon * VPolySet_First(VPolySet *s);
/*
	Start a new scanning sequence of the set, and return the first polygon
	of this sequence. Return NULL if the set is empty.
*/

VPolygon * VPolySet_Next(VPolySet *s);
/*
	Return the pointer to the next polygon in set, or NULL if the scanning
	sequence is terminated and all the polygons scanned.
*/

void       VPolySet_Set(VPolySet *s, VPolygon *poly);
/*
	Set the value of the current entry with a new one.
*/

int        VPolySet_Count(VPolySet *s);
/*
	Return the number of polygons in the set.
*/

VPolygon *VCreatePolygonFromTemplate (int npts, VPoint * pts, VPolygon * templ);

VPolygon *VCopyPolygon (VPolygon *);

void VPrintPolygon (FILE * file, VPolygon * p);

VPolygon * VTransformPolygon (VPolygon * poly, VMatrix * m);

VPolygon * VClipPolygon(VPolygon * poly, VPolygon * clipPoly);

VPolygon * VClipSidedPolygon(VPolygon * poly, VPolygon * clipPoly);
/*
	Clip polygon 'poly' against the view frustum 'clipPoly'. The original
	polygon gets destroyed and released and the new clipped one is
	returned. Return NULL if not visible at all.
*/

VPolygon * ScalePolygon(VPolygon * in, VPoint * offset, VPoint * scale,
	VPoint *e, double r);

void VPoly_FreeAll(void);
/*
	Release internally allocated spare memory of this module.
*/

#undef EXTERN_VPOLY
#undef EXTERN

#endif /* _VPOLY_H */
