/* Vlib -- 3-D aware graphical library */

#ifndef _VLIB_H
#define _VLIB_H

#include <stdio.h>

#ifdef WIN32
	#include <stdlib.h>
	#define SYSV
#else
	#include <X11/Xlib.h>
#endif

#if HAVE_STDLIB_H
	#include <stdlib.h>
#endif
#if !defined(__GNUC__)
	#include <memory.h>
#endif

#include "Vlibmath.h"
#include "../../src/mem_manager.h"
#include "Alib.h"

#ifdef EXTERN_VLIB
	#define EXTERN
#else
	#define EXTERN extern
#endif

#ifdef WIN32
extern int DisplayWidth();
extern int DisplayHeight();
extern int DisplayPlanes();
extern Pixmap XCreatePixmap(
    Display*            /* display */,
    Drawable            /* d */,
    unsigned int        /* width */,
    unsigned int        /* height */,
    unsigned int        /* depth */
);
extern XFreePixmap(
    Display*            /* display */,
    Pixmap              /* pixmap */
);
extern XFlush(
    Display*            /* display */
);
extern XCopyArea(
    Display*            /* display */,
    Drawable            /* src */,
    Drawable            /* dest */,
    GC                  /* gc */,
    int                 /* src_x */,
    int                 /* src_y */,
    unsigned int        /* width */,
    unsigned int        /* height */,
    int                 /* dest_x */,
    int                 /* dest_y */
);
extern XSetForeground(
    Display*            /* display */,
    GC                  /* gc */,
    unsigned long       /* foreground */
);
extern XFillRectangle(
    Display*            /* display */,
    Drawable            /* d */,
    GC                  /* gc */,
    int                 /* x */,
    int                 /* y */,
    unsigned int        /* width */,
    unsigned int        /* height */
);
extern Status XAllocColorCells(
    Display*            /* display */,
    Colormap            /* colormap */,
    Bool                /* contig */,
    unsigned long*      /* plane_masks_return */,
    unsigned int        /* nplanes */,
    unsigned long*      /* pixels_return */,
    unsigned int        /* npixels */
); 
extern XStoreColors(
    Display*            /* display */,
    Colormap            /* colormap */,
    XColor*             /* color */,
    int                 /* ncolors */
);

extern XSetPlaneMask(
    Display*            /* display */,
    GC                  /* gc */,
    unsigned long       /* plane_mask */
); 
#endif


/*
 *  Viewport flags (must be changed manually after VOpenViewport for now)
 */

#define	VPClip		1      /* polygons should be clipped before drawing */
#define VPPerspective	2      /* Z coordinate used for depth information */

typedef struct _Viewport {
	AWindow  *w;                /* performs 2-D drawing and z-buffer */
	Rect      rect;             /* view rectangle inside w */
	Point     focus;            /* infinite sight point (pixel) */
	double    dist;             /* distance from eye to focus (m) */
	Point     Middl;            /* focus point *4 (pixels) */
	double    xres;             /* x screen resolution (pixels/m) */
	double    yres;             /* y screen resolution (pixels/m) */
	VPoint    Scale;            /* scaling factor *4 */
	int       flags;            /* (see VP* constants above) */
	VMatrix   eyeSpace;         /* transforms from world to eyeSpace system */
	VPolygon *clipPoly;         /* planes to clip viewed polygons */
	VPoint    clipNormals[4];   /* normal vectors corresponding to clipPoly */
} Viewport;


/*
	Font
*/

typedef struct {
	short     path_count;
	short     path_start;
	short     glyph_width;
} VGlyph;

typedef struct {
	short     vertex_count;
	short     vertex_start;
} VGlyphPath;

typedef struct {
	short     x;
	short     y;
} VGlyphVertex;


Viewport *VOpenViewport (
	AWindow *w,
	Rect *view_rect,
	Point *focus,
	double dist);
/*
	Return a new 3-D rectangular area (view_rect) on the given window (w).
	The focus is the focus of the projection in screen coordinates (pixel).
	The distance of the eye from the focus of the view_rect (dist, meters)
	must also be given.

	Note that the AWindow must be already set, and that there may be
	several Viewport on the same AWindow.
*/

void VResizeViewport (
	Viewport *v,
	Rect *view_rect,
	Point *focus,
	double dist);
/*
	Change size and projection infos. Basically does the same as
	VOpenViewport(), with the only difference that it operates on an
	existing Viewport. Note that the AWindow is not affected.
*/

void VCloseViewport(Viewport * v);
/*
	Release memory of the Viewport. The AWindow is not released.
*/

void VSetClipRect(Viewport *v, Rect *r);
/*
	Set the clipping rectangular region in the AWindow to which this
	Viewport is linked. The resulting clipping rectangle is the
	intersection among the Viewport rectangle and the AWindow rectangle,
	possibly empty.
*/

void VExposeBuffer(Viewport *);

int VEyeToScreen(Viewport * v, VPoint * p, int *x, int *y);
/*
	Transform a 3-D point in the eye space system into viewable
	coordinates. The function returns 1 if the x,y information is
	displayable (probably displayable, that is).
*/

int VWorldToScreen(Viewport * v, VPoint * p, int *x, int *y);
/*
	Transform a 3-D point in the world system into viewable coordinates.
	The function returns 1 if the x,y information is displayable
	(probably displayable, that is).
*/

void VDrawSegments(Viewport * v, Segment * seg, int nseg, Color color);

void VFillRectangle(Viewport *v, Rect *r, Color c);

void VDrawPolygon(Viewport * v, VPolygon * poly);

void VFillPolygon(Viewport * v, VPolygon * poly);

void VDrawString(Viewport * v, VPoint * p, char *str, int len);

int VFontWidthPixels(Viewport * v, int scale);

void VDrawStrokeString(Viewport * v, int x, int y,
 		  char *str, int len, int scale, Color color);

void VGetStrokeString(Viewport * v, int x, int y, 
 					 Segment * seg, int *nseg,
 					 char *str, int len, int scale);

void VGetEyeSpace(Viewport * v, VPoint EyePt, VPoint CntrInt, VPoint up);

VPolygon *VGetPlanes(VPolygon * poly);

void VSetColorTweakProc(Viewport * v, void (*proc) ( /* ??? */ ));
/*
	We'll create the notion of a viewport-specific color tweaking
	procedure.  This procedure allows some user code to get control
	just before we hand the color name to XParseColor() so that we
	might alter the name a bit.  Why use it?  It allows us to shrink
	down the color space a bit on color-poor screens.
*/

void VForceWindowRedraw(Viewport * v);

void Vlib_FreeAll(void);
/*
	Release memory of all internal data structures and reset its internal
	status. All the Viewport(s) must be already closed before calling this
	function.
*/

#undef EXTERN_VLIB
#undef EXTERN

#endif
