#ifndef _VLIBMATH_H
#define _VLIBMATH_H

#include <math.h>

#ifdef EXTERN_VLIBMATH
	#define EXTERN
#else
	#define EXTERN extern
#endif

/*
 *  If M_PI has been cast to a long double, force the use of our
 *  own, double constants
 */

#ifdef M_PI_LONG_DOUBLE
#undef M_PI
#undef M_PI_2
#undef M_PI_4
#endif

/*
 *  M_PI undefined?
 */

#ifndef M_PI
#define M_PI            3.14159265358979323846
#define M_PI_2          1.57079632679489661923
#define M_PI_4          0.78539816339744830962
#endif

/*
 *  Angle Conversions
 */

#define DEGtoRAD(a)	((a) * M_PI / 180.0)
#define RADtoDEG(a)	((a) * 180.0 / M_PI)


typedef struct {
	//double    x, y, z;			/* A point or vector in 3-space */
	double x;
	double y;
	double z;
} VPoint;

#define mag(x,y,z) sqrt ( (x) * (x) + (y) * (y) + (z) * (z) )

#define Vmagnitude(pt)	sqrt((pt)->x*(pt)->x+(pt)->y*(pt)->y+(pt)->z*(pt)->z)

typedef struct {
	double    m[4][4];
} VMatrix;


EXTERN VPoint VUnitVectorI 
	#ifdef EXTERN_XYZ
	= {1.0, 0.0, 0.0}
	#endif
	;

EXTERN VPoint VUnitVectorJ
	#ifdef EXTERN_XYZ
	= {0.0, 1.0, 0.0}
	#endif
	;

EXTERN VPoint VUnitVectorK
	#ifdef EXTERN_XYZ
	= {0.0, 0.0, 1.0}
	#endif
	;

#define VPointToClipPlaneDistance(pt, pl) \
	(- VDotProd(pl, pt) / ((pl)->x * (pl)->x + (pl)->y * (pl)->y + \
	(pl)->z * (pl)->z))

/*
inline void VSetPoint(VPoint *p, double x, double y, double z);
*/

inline void VSetPoint(VPoint *p, double x, double y, double z)
{
	p->x = x;
	p->y = y;
	p->z = z;
}


void VAdd(VPoint *a, VPoint *b, VPoint *sum);
/*
	Return sum=(a.x+b.x, a.y+b.y, a.z+b.z).
*/

void VSub(VPoint *a, VPoint *b, VPoint *dif);
/*
	Return dif=(a.x-b.x, a.y-b.y, a.z-b.z).
*/

double VDotProd (VPoint *a, VPoint *b);
/*
	Return a.x*b.x + a.y*b.y + a.z*b.z.
*/

void VCrossProd (VPoint *a, VPoint *b, VPoint *p);
/*
	Return

	p.x = a.y*b.z - a.z*b.y
	p.y = a.z*b.x - a.x*b.z
	p.z = a.x*b.y - a.y*b.x

	The vectors may overlap.
*/

void VIdentMatrix(VMatrix * Mtx);
/*
	Set the matrix with the identity tranformation (i.e. no rotation,
	no translation).
*/

void VMatrixTranspose(VMatrix *ori, VMatrix *trans);
/*
	Transpose 'ori' into 'trans'. Matrices may overlap.
*/

void VMatrixMult (VMatrix *a, VMatrix *b, VMatrix *r);
/*
	Return r=b*a. The matrices may overlap.

	NOTE that the b matrix is applied NEXT to the a matrix, so that
	the algebrical expression r=a*b translates into VMatrixMult(b,a,r).
	Another way to remember the order of the arguments is thinking that
	the matrices a,b as apper in the formal arguments are applied in
	the order to a point.
*/

void VMatrixMultByRank (VMatrix *a, VMatrix *b, VMatrix *r, int rank);
/*
	Performs a sub-matrix multiplication by the given rank 1-4.
	Typically rank=3 to perform pure rotations. Matrices may overlap.
*/

void VTranslate (VMatrix *, double, double, double);
void VTranslatePoint (VMatrix *, VPoint);

/*
 *  VRotate options (see VRotate() below)
 */

#define XRotation	1			/* rotate about X axis */
#define YRotation	2			/* rotate about Y axis */
#define ZRotation	3			/* rotate about Z axis */

void VRotate (VMatrix *r, int axis, double theta);
/*
	Add the rotation around the specified axis to the right of
	the r matrix, so that the rotation will be applied first.
	The resulting matrix is returned in the r matrix itself.
*/

void ComputeRotationMatrix(double r, VPoint * e, VMatrix * m);
/*
	FIXME: missing description, but from the source emerges that r
	are DEG.
*/

void VScaleMatrix(VMatrix *m, double xscale, double yscale, double zscale);
/*
	Add a scale trasformation next to the given matrix. Example:
*/

void VTransform (VPoint *p, VMatrix *m, VPoint *q);
/*
	Apply the roto-translation matrix m to the vector p, giving
	q[i] = m[i,j]*p[j]+m[i,3] with i,j=0..2. p,q may overlap.
*/

void VTransform_ (VPoint *p, VMatrix *m, VPoint *q);
/*
	Apply the rotational matrix only. p,q may overlap.
*/

void VReverseTransform (VPoint *p, VMatrix *m, VPoint *q);
/*
	Apply the reverse roto-translation matrix m to the vector p, giving
	q[i] = m[j,i]*(p[j]+m[j,3]) with i,j=0..2. p,q may overlap.
*/

void VReverseTransform_ (VPoint *p, VMatrix *m, VPoint *q);
/*
	Apply the reverse rotational matrix only. p,q may overlap.
*/

typedef struct {
	double    s;
	VPoint    v;
} VQuaternion;

void VQuaternionToMatrix(VQuaternion * q, VMatrix * m);

void VMatrixToQuaternion(VMatrix * m, VQuaternion * q);

void VInterpolateQuaternion(VQuaternion * p, VQuaternion * q, double t, VQuaternion * qt);

VQuaternion * VQuaternionMult(VQuaternion * a, VQuaternion * b, VQuaternion * r);

VQuaternion * VQuaternionComplement(VQuaternion * a, VQuaternion * r);

#undef EXTERN_VLIBMATH
#undef EXTERN

#endif /* _VLIBMATH_H */
