#!/bin/sh
#:acm.sh -or- acmdebug.sh	
#:	Runs (via the gdb debugger if invoked as acmdebug.sh) src/acm as generated
#:	by the build.sh script in the current directory.  In the args,
#:	  '.' means do not include the arg in the script's internal src/acm command
#:	  '-' means include the arg's default value
#:	The stdout and stderr streams are piped into a gzip -c process that writes
#:	the potentially gigabytes of log/trace data to either ./$USER.$$.log.gz or
#:	./$USER.$$.debuglog.gz, depending upon the script name.
#:
#:	Arg 1 '-h' or '--help'	display help messages and exit
#:		- or -
#:	Arg 1 (window_init)	. | - | file_prefix; e.g., 800x600 for ./800x600.init
#:	Arg 2 (exercise_init)	. | - | file_prefix
#:	Arg 3 (plane)		. | - | AMX|B‑747|C‑172|F‑16|MD‑81|MiG‑29|P‑51A|Su‑30
#:	Arg 4 (team)		. | - | 1|2
#:	Other args		specified per the manual's options specifications--can
#:				override any previous args including those specified
#:				in the init files
#:
#:	Arg defaults:	1 (window_init)		800x600
#:			2 (exercise_init)	exercise_1
#:			3 (plane)		C-172
#:			4 (team)		1
#:
#:	Five window init files are provided for the 1st argument (window-init);
#:	  800x600   1024x768   1280x1024   1440x1024   2048x1024
#:	If none work, try making a copy of one and tweaking its options.  Running a
#:	diff command on these files will give you an idea of what to modify and how.
#:
#:	One init file is provided for the 2nd argument (exercise_init):
#:	  exercise_1
#:	and it should be copied and tweaked for different mission/arena setups.
#:
#:	If you want to start multiple instances of acm, you can add an ampersand at
#:	the end of the command line that will put the new process in background so
#:	you can start a second (or third) acm session, maybe for a different
#:	exercise (in which case, the two aircraft could be in different worlds).
#:	Avoid getting confused about which plane the keyboard is controlling, though.
#:
#:	Consult the manual for information about all the options and change their
#:	values as needed, add any new ones you need, or delete any with acceptable
#:	default values.  Remember many init file options can be overriden in the
#:	command line.
#:
#:	The allworlds.acmscene should not require customization, but you might want
#:	to tweak some of its settings.  Of course you are free to create your own
#:	acmscene files as well.
#:
#:	The acm.sh script may have design that you want to be different; e.g., you
#:	want to merge the window and exercose init files into one file.  If so,
#:	make a backup copy somewhere of your version in case any ACM updates change
#:	acm.sh (and remember acmdebug.sh is a link to acm.sh if you want the name of
#:	your script to be different).
#:
#:	Lastly, consider any command line options you might want to frequently
#:	specify when running this script, which should only be overrides as needed
#:	to the options in the init files or normal default values of options not
#:	coded in the init files.
set +xv
test    ".$1" = '.-h' \
     -o ".$1" = '.--help' \
&& { gawk '$0 ~ /^#:/ {print substr($0, 3)}' ./acmdebug.sh
     exit 1
   }
test    ! -d src \
     -o ! -d dis/lib \
     -o ! -d V/lib \
     -o ! -d doc \
&& { echo "The current directory <$PWD> does not look like an ACM directory!"
     exit 4
   }
test $# -lt 4 \
&& { echo "Only $# argument(s) were specified--at least 4 are required"
     exit 3
   }
acmdir="$PWD"
# Process arg 1: system init
if   test    ".$1" = '.' \
          -o ".$1" = '.-'
then sinit='./800x600.init'
elif test "$1" = .
then sinit=''
else sinit="./$1.init"
fi
test ! -f "$sinit" \
&& echo "Warning: The system init file <$sinit> does not exist"
sinit="-init $sinit"
shift
# Process arg 2: exercise init
if   test    ".$1" = '.' \
          -o ".$1" = '.-'
then einit='./exercise_1.init'
elif test "$1" = .
then einit=''
else einit="./$1.init"
fi
test ! -f "$einit" \
&& echo "Warning: The exercise init file <$einit> does not exist"
einit="-init $einit"
shift
# Process arg 3: plane
if   test    ".$1" = '.' \
          -o ".$1" = '.-'
then plane='-plane C-172'
elif test "$1" = .
then plane=''
else case "$1" in
       'AMX')    plane='-plane AMX'    ;;
       'B-747')  plane='-plane B-747'  ;;
       'C-172')  plane='-plane C-172'  ;;
       'F-16')   plane='-plane F-16'   ;;
       'MD-81')  plane='-plane MD-81'  ;;
       'MiG-29') plane='-plane MiG-29' ;;
       'P-51A')  plane='-plane P-51A'  ;;
       'Su-30')  plane='-plane Su-30'  ;;
       *) echo 'Arg 3 (plane) is invalid'
          exit 1
          ;;
     esac
fi
shift
# Process arg 4: team
if   test    ".$1" = '.' \
          -o ".$1" = '.-'
then team='-team 1'
elif test "$1" = .
then team=''
else team="-team $1"
fi
shift
if   test -x $acmdir/src/acm
then dbg=$(echo "$0" | gawk '$0 ~ /acmdebug\.sh$/ { print "gdb -e src/acm -d src -d V/lib -d dis/lib --args " }' )
#    test ".$dbg" = '.' && log='log' || log='debuglog'
     acmcmd="src/acm $sinit $einit -name $USER.$$ -dis-appl $$ $plane $team $@"
     echo "In directory $acmdir,"
     echo "launching $dbg $acmcmd"
#    2>&1 $acmcmd | tee | gzip -c >$USER.$$.$log.gz
     test ".$dbg" = '.' \
     && $acmcmd 2>&1 | tee | gzip -c >$USER.$$.log.gz \
     || $dbg $acmcmd
else echo 'no src/acm executable (did you run ./build.sh and was it successful?)'
     exit 6
fi
# If the $USER.$$.$log.gz files are logging too much information, you could
# comment out the fprintf calls in the ACM code that are producing trace
# information you or your ACM troubleshooter do not wish to capture.  Or
# you could just send stderr to /dev/null.
