#!/bin/sh
# build.sh: (Re)builds acm in an acm tree by:
#               autoconf    (if ./configure or ./Makefile do not exist)
#               ./configure
#               make clean
#               make
#
# To install in Linux Mint Xfce 19.2 (an Ubuntu derivative), the following
# packages and their dependencies needed to be installed (the "apt-file find"
# command can tell you what package you need to install to provide a missing
# file; e.g., 'apt-find X11/xpm.h' tells you libxpm-dev is needed):
# 
#   build-essential
#   autoconf (version 2.69)
#   libx11-dev
#   libxext-dev
#   libxpm-dev
# BTW, also needed if installing acm-6.0-ico:
#   libasound2-dev
#   libmotif-dev
#
# The following instructions presume you followed the instructions in the
# README for downloading, checksumming, and expanding the package tarball.
#
# Examine this script closely and ensure the CFLAGS environment variable
# is defined properly for your system.  View /proc/cpuinfo to determine
# the appropriate values for the target processor, and plug them into the
# case statement using your system's hostname as the data to set those
# values.  If you're building this on an ARM system and want to let the
# case logic take the default, you must change the default--early testing
# indicates CFLAGS='' works.
#
# Now you're ready to build ACM; e.g.,
#
#   $ ./build.sh 2>&1 | tee build.log
#
# This redirects the stdout and stderr output streams into a tee command that
# both displays the streams and copies them into the ~/acm-7.0-dlc/build.log
# file for possible troubleshooting.
# 
# Once the build completes with all zero return codes and there is not
# a nastygram at the end about the src/acm executable, it's time to tailor
# the init files and acm.sh script to actually run ACM (read the acm.sh
# script for more instructions).
#
set +xv
case `hostname` in
  ('trudy-dt') CFLAGS="-m64 -march=nocona"               # Intel Pentium 4 D    f:15 m:4  s:7  64-bit
  #            CFLAGS="-m32 -march=prescott"             # Intel Pentium 4 D    f:15 m:4  s:7  32-bit
               CFLAGS="$CFLAGS -mmmx -msse -msse2 -mssht -mpni -mcx16"
               ;;
  ('trudy-lt') CFLAGS="-m64 -march=westmere"             # Intel Core i5 M 520  f:6  m:36 s:2
               CFLAGS="$CFLAGS -mmmx -msse -msse2 -msse3 -mssse3 -mcx16 -msahf -maes"
               CFLAGS="$CFLAGS -mpclmul -mpopcnt -msse4.2 -msse4.1 -mfxsr --param"
               CFLAGS="$CFLAGS l1-cache-size=32 --param l1-cache-line-size=64"
               CFLAGS="$CFLAGS --param l2-cache-size=3072 -mtune=westmere"
               ;;
  ('able')     CFLAGS="-m64 -march=core2"		 # Intel Q9550		f:6  m:23 s:10
               CFLAGS="$CFLAGS -mcx16 -mfxsr -mmmx -msahf -msse -msse2 -msse3 -msse4.1 -mssse3"
               ;;
  ('mint')     CFLAGS="-m64 -march=athlon-fx"		 # AMD FX-9590		f:21 m:2  s:0
               CFLAGS="$CFLAGS -mabm -maes -mavx -mcx16 -mf16c -mfma -mfma4 -mfxsr -mlwp"
               CFLAGS="$CFLAGS -mmmx -mmmx -msse -msse2 -msse4a -mssse3 -mtbm -mxop -mxsave"
               ;;
  (*)          CFLAGS='-m64 -march=x86-64 -mtune=x86-64' # Generic Intel/AMD 64-bit processor
         #     CFLAGS=''                                 # Generic ARM              processor
               ;;
esac
export CFLAGS="-UNETAUDIO -DTRACE_ARGS -DTRACE_DISW -DTRACE_DISR -DDIS_DEBUG -g -O1 $CFLAGS"
# 
rd_rc() {
  cat /tmp/$USER.acm_build.rc
}
#
set -x
rm -f build.log
if   test    ! -f configure \
          -o ! -f Makefile
then { autoconf
       echo $? >/tmp/$USER.acm_build.rc
     } 2>&1 | tee -a build.log
     rc=`rd_rc`
     test $rc -ne 0 \
     && { set +xv
          echo "#######################################"
          echo "# autoconf failed! You must fix this! #"
          echo "#######################################"
          exit $rc
        }
fi
{ ./configure
  echo $? >/tmp/$USER.acm_build.rc
} 2>&1 | tee -a build.log
rc=`rd_rc`
test $rc -ne 0 \
&& { set +xv
     echo "########################################"
     echo "# configure failed! You must fix this! #"
     echo "########################################"
     exit $rc
   }
{ make clean
  echo $? >/tmp/$USER.acm_build.rc
} 2>&1 | tee -a build.log
test $rc -ne 0 \
&& { set +xv
     echo "#########################################"
     echo "# make clean failed! You must fix this! #"
     echo "#########################################"
     exit $rc
   }
rc=`rd_rc` ; test $rc -ne 0 && echo make clean incurred status $rc && exit $rc
{ make
  echo $? >/tmp/$USER.acm_build.rc
} 2>&1 | tee -a build.log
test $rc -ne 0 \
&& { set +xv
     echo "###################################"
     echo "# make failed! You must fix this! #"
     echo "###################################"
     exit $rc
   }
rm -f /tmp/$USER.acm_build.rc
/bin/ls -lh $PWD/src/acm
test -x src/acm \
&& { set +xv
     echo "#######################################################"
     echo "# Hurray!!!  The src/acm executable is ready to be    #"
     echo "# invoked.  Make sure the acmscene, init, and acm.sh  #"
     echo "# files are ready to go--read the acm.sh script file, #"
     echo "# or just use the ./acm.tk command to set up a flight #"
     echo "# using the GUI approach. Either way, happy landings! #"
     echo "#######################################################"
   } \
|| { set +xv
     echo "##########################################################"
     echo "# There is no executable src/acm file You must fix this! #"
     echo "##########################################################"
     exit 16
   } \
