/*
 *  acm : an aerial combat simulator for X
 *  Copyright (C) 1991-1998  Riley Rainey
 *  Copyright (C) 2007  Umberto Salsi
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 dated June, 1991.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program;  if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
 */

#include <math.h>
#include <string.h>
#include "draw.h"
#include "pm.h"
#include "navaid.h"
#include "vpath_gallery.h"

#define RECEPTION_CHECK_INTERVAL 1.0
/* repeat station repection check every ... (s) */

#define ID_DELAY 10.0
/* every station sends its Morse ID every ... (s) */


typedef struct _adf_data {
	struct _adf_data *next;

	_BOOL     enabled;     /* actually display or standby? */
	freq_t    frequency;   /* selected freq. */
	navaid_t *station;
	double    next_check;
	double    id_update_at;
	char     *id;
	int       hdg;         /* heading pointer (DEG) */
} adf_data;


static adf_data *free_list = NULL;


void adf_enable(viewer *u)
{
	adf_data *adf;

	if( u->adf == NULL ){

		/* Allocate memory for ADF status: */
		if( free_list == NULL ){
			adf = MALLOC( sizeof(adf_data) );
		} else {
			adf = free_list;
			free_list = adf->next;
		}

		/* Initialize: */
		adf->frequency = 200;
		adf->station = NULL;
		adf->next_check = curTime + RECEPTION_CHECK_INTERVAL;
		adf->id_update_at = curTime + ID_DELAY;
		adf->id = NULL;
		adf->hdg = 0;

		u->adf = adf;
	} else {
		adf = u->adf;
	}

	if( ! adf->enabled ){
		adf->enabled = TRUE;
		adf->station = NULL;
		adf->next_check = curTime + RECEPTION_CHECK_INTERVAL;
		adf->id_update_at = curTime + ID_DELAY;
		adf->id = NULL;
	}
}


void adf_disable(viewer *u)
{
	adf_data *adf;

	adf = u->adf;
	if( adf == NULL )
		return;
	
	adf->enabled = FALSE;
}


static void adf_retune(adf_data *adf)
{
	adf->station = NULL;
	adf->next_check = curTime + RECEPTION_CHECK_INTERVAL;
	adf->id_update_at = curTime + ID_DELAY;
	adf->id = NULL;
}


static void adf_update_reception(craft *c, adf_data *adf)
{
	navaid_t *n;

	if( curTime < adf->next_check )
		return;

	adf->next_check = curTime + RECEPTION_CHECK_INTERVAL;

	n = navaid_reception_check(c, adf->frequency);

	if( n == NULL ){
		/* no station found. */
		adf->station = NULL;
		adf->id_update_at = curTime + ID_DELAY;
		adf->id = NULL;
	
	} else if( adf->station != n ){
		/* Station change: */
		adf->station = n;
		adf->id_update_at = curTime + ID_DELAY;
		adf->id = NULL;
	} else {
		/* Still the same station. Update station ID: */
		if( adf->id == NULL && curTime > adf->id_update_at ){
			adf->id = n->id;
		}
	}
}


static _BOOL is_auto_repeat( double *timeout )
{
	_BOOL res;

	res = (curTime < *timeout);
	*timeout = curTime + 0.1;
	return res;
}


void adf_frq_inc(viewer * u, int step)
{
	adf_data *adf;
	static double timeout = 0.0;

	adf = u->adf;
	if( adf == NULL || ! adf->enabled )
		return;
	
	if( is_auto_repeat(&timeout) )
		step *= 2;
	adf->frequency += step;
	if( adf->frequency < NDB_CHANNEL_MIN )
		adf->frequency = NDB_CHANNEL_MIN;
	if( adf->frequency > NDB_CHANNEL_MAX )
		adf->frequency = NDB_CHANNEL_MAX;
	adf_retune(adf);
}


void adf_hdg_inc(viewer *u, int step)
{
	adf_data *adf;
	static double timeout = 0.0;

	adf = u->adf;
	if( adf == NULL || ! adf->enabled )
		return;
	
	if( is_auto_repeat(&timeout) )
		step *= 2;
	adf->hdg += step;
	while( adf->hdg < 0 )
		adf->hdg += 360;
	while( adf->hdg >= 360 )
		adf->hdg -= 360;
}


static void adf_panel_string(viewer *u, double x, double y, double fh,
	char *s1, char *s2, char *s3)
{
	double fw, fh2, fw2;

	fw = fh;

	fh2 = 1.5*fh;
	fw2 = fh2;

	VDrawStrokeString(u->v,
		(int) (x + 0.5), (int) (y + 0.5),
		s1, strlen(s1), (int) (fh + 0.5), u->rz.color);

	VDrawStrokeString(u->v,
		(int) (x + 13*fw - fw2*strlen(s2) + 0.5), (int) (y + 0.5),
		s2, strlen(s2), (int) (fh2 + 0.5), u->v->w->pixel[whiteColor->cIndex]);

	VDrawStrokeString(u->v,
		(int) (x + 14*fw + 0.5), (int) (y + 0.5),
		s3, strlen(s3), (int) (fh + 0.5), u->rz.color);
}


void adf_panel_draw(viewer * u)
{
	adf_data *adf;
	craft *c;
	AWindow *w;
	double x, y, fh, fw, il;
	char s[100];

	adf = u->adf;
	if( adf == NULL || ! adf->enabled )
		return;
	
	adf_update_reception(u->c, adf);

	w = u->v->w;
	AlibSetClipRect(w, &u->tuner);
	AlibFillRectangle(w, &u->tuner, w->pixel[radarBackgroundColor->cIndex]);

	c = u->c;

	fh = RectWidth(u->tuner) / 20.0;
	fw = fh;
	il = 1.2*fh*2.5; /* inter-line spacing */

	x = u->tuner.a.x + fw;
	y = u->tuner.a.y + il;

	/*
		Displays mode:
	*/
	adf_panel_string(u, x, y, fh, "Mode", "ADF", "");
	y += il;

	/*
		Displays frequency
	*/
	sprintf(s, "%d", adf->frequency);
	adf_panel_string(u, x, y, fh, "FRQ", s, "KHz");
	y += il;

	/*
		Displays station ID:
	*/
	if( adf->id != NULL )
		strcpy(s, adf->id);
	else
		strcpy(s, "----");
	adf_panel_string(u, x, y, fh, "STA", s, "");
	y += il;

	/*
		DIsplays signal level "LVL":
	*/
	if( ! adf->enabled )
		adf_panel_string(u, x, y, fh, "LVL", "OFF", "");
	if( adf->station == NULL )
		adf_panel_string(u, x, y, fh, "LVL", "no sgn", "");
	else if( adf->id == NULL )
		adf_panel_string(u, x, y, fh, "LVL", "low", "");
		/* FIXME: actually, station ID still not received */
	else
		adf_panel_string(u, x, y, fh, "LVL", "good", "");
	y += il;
}


void adf_draw(viewer * u)
{
	adf_data *adf;
	AWindow  *w;
	int       xc, yc, h, x, y;
	VMatrix   m;
	VPoint    rloc;
	double    scale, hdg, r, a;
	Color     color;

	adf = u->adf;
	if( adf == NULL || ! adf->enabled )
		return;
	
	w = u->v->w;

	AlibSetClipRect(w, &u->indicator);
	AlibFillRectangle(w, &u->indicator, w->pixel[radarBackgroundColor->cIndex]);

	color = w->pixel[whiteColor->cIndex];

	x = u->indicator.a.x;
	y = u->indicator.a.y;
	scale = RectWidth(u->indicator);
	r = 0.44*scale;
	xc = x + (int) (0.50*scale+0.5);
	yc = y + (int) (0.50*scale+0.5);

	if (u->c->showMag)
		hdg = navaid_mag_heading(u->c);
	else
		hdg = u->c->curHeading;

	h = (int) (0.045*scale + 0.5);

	/* Displays "TH" or "MH": */
	VDrawStrokeString(u->v,
		x + (int) (0.05*scale+0.5), y + (int) (0.06*scale),
		(u->c->showMag)? "MH":"TH", 2, h, color);

	/* Draw compass scale: */
	VIdentMatrix(&m);
	VRotate(&m, ZRotation, -hdg);
	VScaleMatrix(&m, r, r, r);
	VTranslate(&m, xc, yc, -1.0);
	vpath_stroke( vpath_gallery_get_compass_scale(), &m, w, whitePixel);

	/* Draw fixed scale: */
	VIdentMatrix(&m);
	VScaleMatrix(&m, r/0.90, r/0.90, r/0.90);
	VTranslate(&m, xc, yc, -1.0);
	vpath_stroke( vpath_gallery_get_compass_fixed_scale(), &m, w, whitePixel);

	/* Draw stylized aircraft: */
	VIdentMatrix(&m);
	VScaleMatrix(&m, 0.25*r, 0.20*r, 0.25*r);
	VTranslate(&m, xc, yc, -1.0);
	vpath_stroke( vpath_gallery_get_stylized_aircraft(), &m, w, whitePixel);

	/* Draw selected HDG: */
	VIdentMatrix(&m);
	VRotate(&m, ZRotation, -hdg + DEGtoRAD(adf->hdg));
	VScaleMatrix(&m, r, r, r);
	VTranslate(&m, xc, yc, -1.0);
	vpath_stroke( vpath_gallery_get_pointer_triangle(), &m, w,
		w->pixel[HSIMagentaColor->cIndex]);

	/* Draw pointer: */

	if( adf->station == NULL ){
		/* No signal -- pointer indicates right: */
		a = DEGtoRAD(90.0);
	} else {
		VTransform(&adf->station->Sg, &u->c->XYZtoNED, &rloc);
		VReverseTransform(&rloc, &u->c->trihedral, &rloc);
		a = atan2(rloc.y, rloc.x);
	}

	VIdentMatrix(&m);
	VRotate(&m, ZRotation, a);
	VScaleMatrix(&m, 0.90*r, 0.90*r, 0.90*r);
	VTranslate(&m, xc, yc, -1.0);
	vpath_stroke( vpath_gallery_get_pointer_arrow(), &m, w, whitePixel);

}


void adf_free(viewer *u)
{
	adf_data *adf;

	adf = u->adf;
	if( adf == NULL )
		return;
	
	adf->next = free_list;
	free_list = u->adf;
	u->adf = NULL;
}


void adf_free_all()
{
	adf_data *p;

	while( free_list != NULL ){
		p = free_list;
		free_list = free_list->next;
		FREE(p);
	}
}

/* End of the "adf" module */
