/*
 *  acm : an aerial combat simulator for X
 *  Air and wind properties module.
 *  Copyright (C) 1991-1998  Riley Rainey
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 dated June, 1991.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program;  if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
 */

#include <math.h>
#include "pm.h"
#include "damage.h"
#include "air.h"

#define	GM	1.4					/* ratio of specific heats for air */
#define P0	2116.22				/* sea-level pressure [slug/ft^2] */
#define TA	459.67				/* CONVERSION TEMPERATURE FROM F TO R */
#define T0	(TA + 59.0)			/* sea-level temperature [R=abs. F] */
#define	G	32.1741				/* gravitational acceleration [ft/sec^2] */
#define RU	(1545.31 * G)		/* universal gas constant */
#define MA	28.9644				/* molecular weight for air [lbm/lbm-mole] */
#define	RA	(RU / MA)			/* gas constant for air [ft^2/sec^2/R] */

/*
 *  airProperties :  Compute rho and the speed of sound for the standard
 *              atmosphere up to 60,000 feet.
 */

static void
airProperties(double h, double *rho, double *mach1)
{

	double    t, p;

	if (h < 36089.2) {
		t = T0 - 3.56616e-3 * h;
		p = P0 * pow(t / T0, 5.255912);  /* lb/ft^2 */
	}
	else {
		t = 389.97;
		p = 472.6773 * exp(-G * (h - 36089.2) / (RA * t));  /* lb/ft^2 */
	}

	*rho = p / (RA * t);  /* slug/ft^3 */
	*mach1 = sqrt(GM * RA * t);
}


/*
 *  We keep a table of atmospheric constants for different altitudes.
 */

static struct {
	double    alt;				/* altitude x1000 ft */
	double    rho;				/* rho value (air density) */
	double    mach1;			/* speed of sound (ft/s) */
}        *rhop, rhoTable[] = {

#ifdef THIS_PART_OF_THE_TABLE_NOT_USED
	{
		0.0, 23.77, 1116.9
	},
	{
		2.0, 22.41, 1109.2
	},
	{
		4.0, 21.11, 1101.4
	},
	{
		6.0, 19.87, 1093.6
	},
	{
		8.0, 18.68, 1085.7
	},
	{
		10.0, 17.55, 1077.8
	},
	{
		15.0, 14.96, 1057.7
	},
	{
		20.0, 12.66, 1037.3
	},
	{
		25.0, 10.65, 1016.4
	},
	{
		30.0, 8.89, 995.1
	},
	{
		35.0, 7.365, 973.3
	},
	{
		40.0, 5.851, 968.5
	},
	{
		50.0, 3.618, 968.5
	},
#endif
	{
		60.0, 2.238, 968.5
	},
	{
		80.0, 0.9065, 980.0
	},
	{
		100.0, 0.3371, 1015.0
	},
	{
		120.0, 0.1340, 1053.0
	},
	{
		160.0, 0.02622, 1083.0
	},
	{
		100000.0, 0.02622, 1083.0   /* a large value for alt at the end */
	}
};


/*
 *  calcRho : Calculate air density and the speed of sound.
 */

double
air_calc_rho(double alt, double *mach)
{
	double    deltaAlt, b;
	double    rho;

	/*
		Use analytic formula up to 60000 ft:
	*/

	if (alt <= 60000.0) {
		airProperties(alt, &rho, mach);
		return rho;
	}

	/*
		Interpolation above 60000 ft:
	*/

	alt = alt / 1000.0;

	for (rhop = rhoTable; alt > (rhop + 1)->alt; ++rhop);
	deltaAlt = (rhop + 1)->alt - rhop->alt;
	b = ((rhop + 1)->mach1 - rhop->mach1) / deltaAlt;
	*mach = rhop->mach1 + b * (alt - rhop->alt);
	b = ((rhop + 1)->rho - rhop->rho) / deltaAlt;
	return (rhop->rho + b * (alt - rhop->alt)) / 10000.0;
}



/*
	Wind has two components: a constant component const_wind[], and a gust
	component gust[] that varies in intensity and frequence. The resulting
	wind is
	
		wind[] = const_wind[] + gust[].
*/


static VPoint wind = {
	0.0 /* from N */,
	0.0 /* from E */,
	0.0 /* from zenith */
};
/* wind speed (ft/s) */


static VPoint const_wind = {
	0.0 /* from N */,
	0.0 /* from E */,
	0.0 /* from zenith */
};
/* Constant wind speed (ft/s) */


static double gust_max = 0.0;
/* Max gust speed (ft/s). */

static VPoint gust = {
	0.0 /* from N */,
	0.0 /* from E */,
	0.0 /* from zenith */
};
/* Current gust speed [ft/s] */


static VPoint target_gust = {
	0.0 /* from N */,
	0.0 /* from E */,
	0.0 /* from zenith */
};
/* Target gust speed (ft/s) */


void
air_set_wind(
	double wd, /* wind direction (deg) */
	double wv) /* wind velocity (kt) */
{
	wd = DEGtoRAD(wd);
	wv = KTtoFPS(wv);
	const_wind.x = wv * cos(wd);
	const_wind.y = wv * sin(wd);
	const_wind.z = 0.0;
}


void air_set_gust(double _gust_max)
{
	gust_max = _gust_max;
}


#define T1_2 0.5


static void
air_update_gust()
{
	static double t_prev = 0.0, t_next = 0.0;

	double dt, x;

	/*
		Update gust speed:
	*/

	dt = curTime - t_prev;
	t_prev = curTime;
	gust.x += dt / T1_2 * (target_gust.x - gust.x);
	gust.y += dt / T1_2 * (target_gust.x - gust.y);
	gust.z += dt / T1_2 * (target_gust.x - gust.z);

	/*
		Update target gust speed:
	*/

	if( curTime <= t_next )
		return;
	
	x = 2.0 * damage_frandom() - 1.0;  /* -1.0 <= x <= +1.0 */
	target_gust.x = gust_max * x * x * x * x * x;

	x = 2.0 * damage_frandom() - 1.0;  /* -1.0 <= x <= +1.0 */
	target_gust.y = gust_max * x * x * x * x * x;

	x = 2.0 * damage_frandom() - 1.0;  /* -1.0 <= x <= +1.0 */
	target_gust.z = gust_max * x * x * x * x * x;

	/*
		Evaluate time of the next gust evaluation.

		The idea behind this calculation is that the bigger is the
		gust, the more is the frequency of the recalculation. So we
		start with this (approximated) gust module calculation:
	*/

	double m = fabs(target_gust.x) + fabs(target_gust.y) + fabs(target_gust.z);

	t_next = curTime + 10.0 / (m + 5.0);
	
}


static double t_next_gust_upd = 20.0;


VPoint * air_get_wind(double altitude_meters)
{
	/*
		Update current gust speed:
	*/
	
	if( curTime > t_next_gust_upd ){
		air_update_gust();
		t_next_gust_upd = curTime + 0.1;
	}

	/*
		Calculate total wind speed. The gust intensity decreases as the
		altitude increases.
	*/

	double k = 3000.0 / (altitude_meters + 3000.0);

	wind.x = const_wind.x + k * gust.x;
	wind.y = const_wind.y + k * gust.y;
	wind.z = const_wind.z + k * gust.z;

	return &wind;
}
