/*
 *  acm : an aerial combat simulator for X
 *  Copyright (C) 1991-1994  Riley Rainey
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 dated June, 1991.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program;  if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
 */

/*
 *  2014-12-29 dlc - Refer to CHANGES text file for details of modifications
 */

#include "pm.h"
#include "init.h"
#include "audio.h"
#include "mem_manager.h"

#ifndef WIN32
#include <X11/Xos.h>
#endif

#define NUM_SOUNDS      11

static char *files[] =
{
	"engine.au",
	"crash.au",
	"gear_up.au",
	"gear_dn.au",
	"missile.au",
	"cannon.au",
	"crash.au",
	"screetch.au",
	"stall.au",
	"rwr.au",
	"apglock.au"
};

static char *paths[NUM_SOUNDS];

/*
	The "AudioData" structure is bound to the field (viewer)->audio
	of the viewer data structure.
*/

typedef struct _AudioData {

	struct _AudioData * next;

#if defined(HPAUDIO)

	Audio    *aserver;			/* audio server for this user */
	SBucket  *sound[NUM_SOUNDS];
	ATransID  flow[NUM_SOUNDS];

#elif defined(NETAUDIO)

	AuServer *aserver;			/* audio server for this user */
	AuBucketID sound[NUM_SOUNDS];
	AuFlowID  flow[NUM_SOUNDS];
	AuDeviceID audioOutput[1];	/* audio output device ID */

#else /* poor men audio */

	/* the sound to play once or continously: */
	SoundID in;          /* ID */
	_BOOL   in_cont;     /* play continously? (loop) */
	int     in_offset;   /* current buffer offset */

	/* the background sound (typically the engine): */
	SoundID in_bg;       /* ID */
	int     in_bg_vol;   /* volume [0,128] */
	int     in_bg_rpm;   /* RPM % [0,128] */
	int     in_bg_offset;  /* current buffer offset */

#endif

} AudioData;


static AudioData * freeList = NULL;


static AudioData *
AudioDataAlloc()
{
	AudioData * ad;

	if ( freeList == NULL ) {
		ad = MALLOC(sizeof(AudioData));
	}
	else {
		ad = freeList;
		freeList = freeList->next;
	}

	ad->next = NULL;
	return ad;
}


static void
AudioDataFree(AudioData * ad)
{
	if ( ad == NULL )
		return;
	ad->next = freeList;
	freeList = ad;
}


/*
   NETAUDIO
   ------------------------------------------------------------------------
*/

#if defined(NETAUDIO)

_BOOL
audio_init(viewer * v, char * display)
{
	int       i, n;
	AudioData * ad;
	AuBucketAttributes *ret_attr;
	AuDeviceAttributes *dev;
	AuStatus  status;
	char      auserver[256], *p;

	ad = AudioDataAlloc();
	v->audio = (void *) ad;

/*
 *  If the X display specification has a screen field, remove it so that we
 *  can use it as an audio server name.
 */

	strcpy(auserver, display);
	n = strlen(auserver);
	if (auserver[n] == '.') {   // FIXME: BUG
		auserver[n] = '\0';
	}

/*
 *  Lazy initialization of sound paths array
 */

	if (!paths[0]) {
		for (i = 0; i < NUM_SOUNDS; ++i) {
			paths[i] = STRDUP(acm_find_file(files[i]));
		}
	}

/*
 *  Open the audio server and send it our sounds
 */

	if ((ad->aserver = AuOpenServer(auserver, 0, (char *) NULL,
							0, (char *) NULL, (char **) NULL)) == NULL) {
		ad->aserver = (AuServer *) NULL;
		return FALSE;
	}

	for (i = 0; i < NUM_SOUNDS; ++i) {
		if ((ad->sound[i] = AuSoundCreateBucketFromFile(ad->aserver,
						 paths[i], AuAccessListMask | AuAccessImportMask,
										&ret_attr, &status)) == AuNone) {
			printf("Unable to create sound bucket: %s\n",
				   paths[i]);
			AuCloseServer(ad->aserver);
			ad->aserver = (AuServer *) NULL;
			return FALSE;
		}
		ad->flow[i] = 0;
	}

/*
 *  Select an output device
 */

	ad->audioOutput[0] = AuNone;
	for (i = 0; i < AuServerNumDevices(ad->aserver); ++i) {
		dev = AuServerDevice(ad->aserver, i);
		if (AuDeviceKind(dev) == AuComponentKindPhysicalOutput &&
			AuDeviceNumTracks(dev) == 1) {
			ad->audioOutput[0] = AuDeviceIdentifier(dev);
			break;
		}
	}

	return TRUE;
}

void
audio_shutdownAudio(viewer * v)
{
	AudioData * ad;
	int       i;
	AuStatus  status;

	ad = (AudioData *) v->audio;

	if (ad == NULL)
		return;

	if (ad->aserver != NULL) {
		for (i = 0; i < NUM_SOUNDS; ++i) {
			AuDestroyBucket(ad->aserver, ad->sound[i], &status);
		}
		AuCloseServer(ad->aserver);
		ad->aserver = (AuServer *) NULL;
		for (i = 0; i < NUM_SOUNDS; ++i) {
			FREE(paths[i]);
		}
	}

	AudioDataFree(ad);
	v->audio = NULL;
}

static void
done_callback( AuServer *audio, AuEventHandlerRec *which,
	AuEvent  *event, AuPointer data)
{
	viewer   *v = (viewer *) data;
	int       mult, mon;
	AuStatus  status;

	printf("event type = %d\n", event->type);
}


void
audio_playSound(craft * c, SoundID id, _BOOL loop)
{
	viewer   *v;
	AudioData * ad;
	AuStatus  status;

	for (v = c->vl; v; v = v->next) {
		ad = (AudioData *) v->audio;
		if (ad != NULL && ad->aserver != NULL) {
			if ( loop )
				AuSoundPlayFromBucket(ad->aserver, ad->sound[id],
								ad->audioOutput[0], AuFixedPointFromSum(1, 0),
									  done_callback, (AuPointer) v,
								10, &ad->flow[id], (int *) NULL, (int *) NULL,
									  &status);
			else
				AuSoundPlayFromBucket(ad->aserver, ad->sound[id],
								ad->audioOutput[0], AuFixedPointFromSum(1, 0),
								   NULL, NULL, 0, &ad->flow[id], (int *) NULL,
									  (int *) NULL, &status);
		}
	}
}

void
audio_stopSound(craft * c, SoundID id)
{
	viewer   *v;
	AudioData * ad;
	AuStatus  status;

	for (v = c->vl; v; v = v->next) {
		ad = (AudioData *) v->audio;
		if (ad != NULL && ad->aserver != (AuServer *) NULL && ad->flow[id] != 0) {
			AuStopFlow(ad->aserver, ad->flow[id], &status);
			AuDestroyFlow(ad->aserver, ad->flow[id], &status);
			ad->flow[id] = 0;
		}
	}
}

void
audio_setBackgroundSound(craft * c, double dThtlPercent,
				   _BOOL bABFlag, double dDynamicPressure)
{
	// FIXME: TODO
}


void
audio_update()
{
}


/*
	HPAUDIO
	---------------------------------------------------------------------
*/

#elif defined (HPAUDIO)


_BOOL
audio_init(viewer * v, char * display)
{
	AudioData * ad;
	int       i, n;
	Audio    *conn;
	AudioAttributes sb_attr;
	long      status;

	ad = AudioDataAlloc();
	v->audio = (void *) ad;

/*
 *  Lazy initialization of sound paths array
 */

	if (!paths[0]) {
		for (i = 0; i < NUM_SOUNDS; ++i) {
			paths[i] = STRDUP(acm_find_file(files[i]));
		}
	}

/*
 *  Open the audio server and send it our sounds
 */

	if ((ad->aserver = AOpenAudio(display, &status)) == NULL) {
		printf("AOpenAudio: failed, status = %d\n", status);
		ad->aserver = (Audio *) NULL;
		return FALSE;
	}

	for (i = 0; i < NUM_SOUNDS; ++i) {
		if ((ad->sound[i] = ALoadAFile(ad->aserver,
									  paths[i], AFFUnknown, 0, &sb_attr,
									  &status)) == NULL) {
			printf("Unable to create sound bucket: %s (status %d)\n",
				   paths[i], status);
			ACloseAudio(ad->aserver, &status);
			ad->aserver = (Audio *) NULL;
			return FALSE;
		}
		ad->flow[i] = 0;
	}

	return TRUE;
}

void
audio_shutdownAudio(viewer * v)
{
	AudioData * ad;
	int       i;
	long      status;

	ad = (AudioData *) v->audio;

	if (ad == NULL)
		return;

	if (ad->aserver != NULL) {
		for (i = 0; i < NUM_SOUNDS; ++i) {
			ADestroySBucket(ad->aserver, ad->sound[i], &status);
		}
		ACloseAudio(ad->aserver, &status);
		ad->aserver = (Audio *) NULL;
		for (i = 0; i < NUM_SOUNDS; ++i) {
			FREE(paths[i]);
		}
	}

	AudioDataFree(ad);
	v->audio = NULL;
}

void
audio_playSound(craft * c, SoundID id, _BOOL loop)
{
	viewer   *v;
	SBPlayParams pparams;
	AGainEntry gainEntry;
	long      status;
	AudioData * ad;

	gainEntry.u.o.out_ch = AOCTMono;
	gainEntry.gain = AUnityGain;
	gainEntry.u.o.out_dst = AODTMonoIntSpeaker;

	bzero(&pparams, sizeof(pparams));
	pparams.priority = APriorityNormal;
	pparams.gain_matrix.type = AGMTOutput;
	pparams.gain_matrix.num_entries = 1;
	pparams.gain_matrix.gain_entries = &gainEntry;
	pparams.play_volume = AUnityGain;
	pparams.pause_first = False;
	pparams.start_offset.type = ATTSamples;
	pparams.start_offset.u.samples = 0;
	pparams.duration.type = ATTFullLength;
	pparams.loop_count = loop? -1 : 0;
	pparams.previous_transaction = 0;
	pparams.event_mask = 0;

	for (v = c->vl; v; v = v->next) {
		ad = (AudioData *) v->audio;
		if (ad != NULL && ad->aserver != NULL) {
			ad->flow[id] = APlaySBucket(ad->aserver, ad->sound[id],
									   &pparams, &status);
			if (ad->flow[id] == NULL)
				printf("APlaySBucket: failed, status %d\n", status);
		}
	}
}

void
audio_stopSound(craft * c, SoundID id)
{
	viewer   *v;
	long      status;
	AudioData * ad;

	for (v = c->vl; v; v = v->next) {
		ad = (AudioData *) v->audio;
		if (ad != NULL && ad->aserver != (Audio *) NULL && ad->flow[id] != 0) {
			AStopAudio(ad->aserver, ad->flow[id], ASMEndLoop, NULL, &status);
			if (status)
				printf("AStopAudio: status %d\n", status);
			ad->flow[id] = 0;
		}
	}
}


void
audio_setBackgroundSound(craft * c, double dThtlPercent,
				   _BOOL bABFlag, double dDynamicPressure)
{
	// FIXME: TODO
}


void
audio_update()
{
}



/*
	Poor men audio...
	----------------------------------------------------------------------

	We try to play sounds sending *.au data to /dev/dsp.

	FIXME: only sounds of the viewer that has input focus should be played
	FIXME: every viewer should have his sound status: current backg. sound, etc.
*/

#else

#warning "No NETAUDIO support."
#warning "No HPAUDIO support."
#warning "Trying with the poor men sounds sending to /dev/dsp..."

#include <errno.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <termios.h>
#include <unistd.h>


#define AUDIO_DEV "/dev/dsp"

/*
	The state of our "poor men" audio server.
*/

static _BOOL   sound_initialized = FALSE;
static int     out;         /* AUDIO_DEV file descriptor */


/*
	Every sound file is loaded in this struct the first time it
	is requested.
*/
static struct {
	int size;
	unsigned char * buffer;
} sound_buffers[NUM_SOUNDS];


static off_t
file_size(char * fn)
{
	struct stat statbuf;

	if ( stat(fn, &statbuf) < 0 )
		return -1;
	return statbuf.st_size;
}


static void
load_sound(SoundID id)
/*
	Ensures the required sound be loaded into sound_buffers[id]
	Do nothing if it is already there.
*/
{
	FILE * f;
	int n, nr;
	unsigned char * b;

	if ( sound_buffers[id].size > 0 )
		return;

	n = file_size(paths[id]);
	if ( n < 0 ) {
		fprintf(stderr, "load_sound() Can't read size of the sound file `%s': %s\n",
			paths[id], strerror(errno));
		exit(1);
	}
	sound_buffers[id].size = n;

	f = fopen(paths[id], "r");
	if ( f == NULL ) {
		fprintf(stderr, "load_sound() Can't open sound file `%s': %s\n",
			paths[id], strerror(errno));
		exit(1);
	}

	b = MALLOC(n);
	sound_buffers[id].buffer = b;

	nr = fread(b, 1, n, f);
	if ( nr <= 0 ) {
		fprintf(stderr, "load_sound() Can't read sound file `%s': %s\n",
			paths[id], strerror(errno));
		exit(1);
	}
	if ( nr != n ) {
		fprintf(stderr, "load_sound() Can't read sound file `%s': incomplete read\n",
			paths[id]);
		exit(1);
	}

	fclose(f);
}


static int sound_chunk = 100;

static void
AudioHandleSub(AudioData * ad)
/*
	Send audio data to the AUDIO_DEV based on the current state of the
	audio server.
*/
{
	int n;
	fd_set out_set;
	struct timeval tout;
	unsigned char chunk[500];

	if ( ! sound_initialized )
		return;

	if ( ad->in == SoundNone && ad->in_bg == SoundNone ) {
		if ( out >= 0 )
			close(out);
		return;
	}

	if ( out < 0 ) {
		out = open(AUDIO_DEV, O_WRONLY);
		if ( out < 0 ) {
			fprintf(stderr, "AudioHandleSub() Can't open for writing the audio device `%s': %s\n",
				AUDIO_DEV, strerror(errno));
			exit(1);
		}
	}

	FD_ZERO(&out_set);
	FD_SET(out, &out_set);

	tout.tv_sec = 0;
	tout.tv_usec = 0;

	n = select(out + 1, NULL, &out_set, NULL, &tout);

	if ( n == -1 ) {
		perror("select()");
		return;
	}

	if ( ! FD_ISSET(out, &out_set) ) {
		sound_chunk = sound_chunk / 2;
		if ( sound_chunk < 50 )
			sound_chunk = 50;
		return;
	}

	sound_chunk += 1;
	if ( sound_chunk > 500 )
		sound_chunk = 500;

	if ( ad->in != SoundNone ) {
		n = sound_buffers[ad->in].size - ad->in_offset;
		if ( n > sound_chunk )
			n = sound_chunk;
		n = write(out, &sound_buffers[ad->in].buffer[ad->in_offset], n);
		if ( n < 0 ) {
			perror("Error writing to the audio device " AUDIO_DEV);
		}
		else {
			ad->in_offset += n;
			if ( ad->in_offset >= sound_buffers[ad->in].size ) {
				if ( ad->in_cont )
					ad->in_offset = 0;
				else
					ad->in = SoundNone;
			}
		}
	}
	else if ( ad->in_bg != SoundNone ) {
#ifdef not_def
		unsigned char *p_base, *p_top, *p;
		int i, b, d, d_inc;
		b = 255 * (128 - ad->in_bg_vol) / 2;
		p_base = sound_buffers[ad->in_bg].buffer;
		p_top = p_base + sound_buffers[ad->in_bg].size;
		p = p_base + ad->in_bg_offset;
		i = 0;
		d_inc = 128 + 1 - ad->in_bg_rpm;
		d_inc = 128;
		d = 0;
		while ( i < sound_chunk ) {
			//chunk[i] = (ad->in_bg_vol * *p + b) >> 7;
			chunk[i] = (unsigned char) (
				((*p * ad->in_bg_vol) >> 7 ) - ad->in_bg_vol + 128
			);
			i++;

			d += d_inc;

			p += (d >> 7);
			if ( p >= p_top )
				p -= sound_buffers[ad->in_bg].size;

		}
		ad->in_bg_offset = p - p_base;
#endif
		unsigned char *p_base, *p_top, *p;
		int i, b, d, d_inc;
		b = 255 * (128 - ad->in_bg_vol) / 2;
		p_base = sound_buffers[ad->in_bg].buffer;
		p_top = p_base + sound_buffers[ad->in_bg].size;
		p = p_base + ad->in_bg_offset;
		i = 0;
		d_inc = 128 + 1 - ad->in_bg_rpm;
		d_inc = 128;
		d = 0;
		while ( i < sound_chunk ) {
			chunk[i] = (unsigned char) (
				((*p * ad->in_bg_vol) >> 7 ) - ad->in_bg_vol + 128
			);
			i++;
			p++;
			if ( p >= p_top )
				p -= sound_buffers[ad->in_bg].size;

		}
		ad->in_bg_offset = p - p_base;
		n = write(out, chunk, sound_chunk);
		if ( n < 0 ) {
			perror("Error writing to the audio device");
		}
		else {
			ad->in_bg_offset += n;
			if ( ad->in_bg_offset >= sound_buffers[ad->in_bg].size ) {
				ad->in_bg_offset = 0;
			}
		}
	}

}


void
audio_update()
{
	if ( vl_head == NULL )
		return;

	AudioHandleSub( (AudioData *) vl_head->audio );
}


_BOOL
audio_init(viewer * v, char * display_IGNORED)
{
	AudioData * ad;
	int       i;

	ad = AudioDataAlloc();
	v->audio = (void *) ad;

	out = SoundNone;
	ad->in = SoundNone;
	ad->in_bg = SoundNone;

/*
 *  Lazy initialization of sound paths array
 */

	if (!paths[0]) {
		for (i = 0; i < NUM_SOUNDS; ++i) {
			paths[i] = STRDUP(acm_find_file(files[i]));
			sound_buffers[i].size = 0;
		}
	}

	sound_initialized = TRUE;

	return TRUE;
}


void
audio_shutdownAudio(viewer * v)
{
	AudioData * ad;

	ad = (AudioData *) v->audio;

	if ( ad == NULL )
		return;
	
	ad->in = SoundNone;
	ad->in_bg = SoundNone;
	AudioDataFree(ad);
	v->audio = NULL;
}


void
audio_playSound(craft * c, SoundID id, _BOOL loop)
{
	viewer   *v;
	AudioData * ad;

	if ( ! sound_initialized )
		return;

	for (v = c->vl; v; v = v->next) {
		ad = (AudioData *) v->audio;

		if ( ad != NULL ) {
			if ( ad->in != id ) {
				load_sound(id);
				ad->in = id;
				ad->in_offset = 0;
			}
			ad->in_cont = loop;
		}
	}
}


void
audio_stopSound(craft * c, int id)
{
	viewer   *v;
	AudioData * ad;

	for (v = c->vl; v; v = v->next) {

		ad = (AudioData *) v->audio;
		if ( ad != NULL ) {
			ad->in = SoundNone;
		}

	}
}


void
audio_setBackgroundSound(craft * c, double dThtlPercent,
				   _BOOL bABFlag, double dDynamicPressure)
{
	// FIXME: the arguments bABFlag,dDynamicPressure actually ignored

	viewer   *v;
	AudioData * ad;

	if ( ! sound_initialized )
		return;

	for (v = c->vl; v; v = v->next) {

		ad = (AudioData *) v->audio;
		if ( ad != NULL ) {

			ad->in_bg_vol = (int) (128.0 * dThtlPercent + 0.5);
			if ( ad->in_bg_vol < 0 )
				ad->in_bg_vol = 0;
			else if ( ad->in_bg_vol > 128 )
				ad->in_bg_vol = 128;

			ad->in_bg_rpm = ad->in_bg_vol;

			if ( ad->in_bg_vol == 0 || ad->in_bg_rpm == 0 ) {
				ad->in_bg = SoundNone;
				return;
			}

			if ( ad->in_bg != SoundEngine ) {
				load_sound(SoundEngine);
				ad->in_bg = SoundEngine;
				ad->in_bg_offset = 0;
			}

		}
	}
}

#endif


void
audio_free_all()
{
	AudioData * ad;
	int i;

	while( freeList != NULL ){
		ad = freeList;
		freeList = ad->next;
		FREE(ad);
	}

	if( paths != NULL ){
		for( i=0; i<NUM_SOUNDS; i++ )
			if( paths[i] != NULL ){
				FREE(paths[i]);
				paths[i] = NULL;
			}
	}
}
