#include <string.h>
#include "air.h"
#include "damage.h"
#include "dis_if.h"
#include "draw.h"
#include "effects.h"
#include "init.h"
#include "mem_manager.h"
#include "planes.h"
#include "players.h"
#include "pm.h"
#include "terrain.h"
#include "weapon.h"

#include "ccip.h"

static int select_mk82(craft *);
static int display_mk82(craft *, craftType *, viewer *, int, int);
static int update_mk82(craft *);
static int press_mk82(craft *);
static int release_mk82(craft *);

static WeaponDesc mk82Desc =
{
	select_mk82,				/* select */
	update_mk82,				/* update */
	display_mk82,				/* display procedure */
	press_mk82,				    /* fire */
	release_mk82,				/* fire button release */
};


static void
computeImpactPoint(craft *c, WorldCoordinates *ip)
{
	double t_sec, ground_speed_fps, loft_distance_feet;
	double T1_sec, T2_sec;
	double A, B, C, S4AC;
	double sin_course, cos_course;

	/*
	 * Local terrain skin location
	 */

	*ip = c->w;
	ip->z = terrain_altitude(c);

	A = - 0.5 * earth_g;
	B = - c->Cg.z;
	C = METERStoFEET ( c->w.z - ip->z );

	S4AC = sqrt ( B * B - 4.0 * A * C );

	/*
	 *  Compute time to impact using the solution to the quadratic formula:
	 *
	 *     x = 0.5 * ( - earth_g ) * t^2  +  v0 * t  +  x0
	 *
	 *   v0 = vertical velocity (up positive, feet-per-second)
	 *   x0 = height above surface (feet)
	 */

	if (S4AC >= 0.0) {

		T1_sec = ( - B + S4AC ) / ( 2.0 * A );
		T2_sec = ( - B - S4AC ) / ( 2.0 * A );

	}
	else {
		/* complex root(s) */
		T1_sec = T2_sec = 10.0; /* FIXME: ??? */
	}

	t_sec = (T1_sec > T2_sec) ? T1_sec : T2_sec;

	ground_speed_fps = sqrt ( c->Cg.x * c->Cg.x + c->Cg.y * c->Cg.y );

	loft_distance_feet = ground_speed_fps * t_sec;

	/*
	 *  Based on instantaneous velocity, not aircraft orientation
	 */

	cos_course = c->Cg.x / ground_speed_fps;
	sin_course = c->Cg.y / ground_speed_fps;

	DISUpdateWorldCoordinates(ip, cos_course, sin_course, FEETtoMETERS ( loft_distance_feet ) );

}


static int count[MAXPLAYERS];
static int hasFired[MAXPLAYERS];

static int
select_mk82( craft * c )
{
	hasFired[c->pIndex] = 0;
	count[c->pIndex] = countOrdinance(c, WEAPON_MK82);
	return 1;
}

WeaponDesc *
ccip_new(void)
{

	craftType *c;
	FILE     *f;
	dis_entity_type em1 = {2, 9, 225, 2, 73, 0, 0};
	dis_entity_type em2 = {0, 0, 0, 0, 0, 0, 0};

	c = newCraft();
	c->name = STRDUP( weapon_id_to_name(WEAPON_MK82) );

	mk82Desc.w = c;

	c->entityType = em1;
	c->altEntityType = em2;

	/*
	 *  Unverified parameters calculated using DATCOM recommendations
	 */

	c->CDOrigin = 0.081;
	c->CDFactor = 0.0;
	c->CDBOrigin = 0.0;
	c->CDBFactor = 0.0;
	c->emptyWeight = 510.0;
	c->wingS = 1.0; /* FIXME */
	c->maxFuel = 0.0;
	c->maxThrust = 0.0;
	c->spFuelConsump = 0.0;

	/*
	 * dCL/da = 4.09
	 * dCm,cg/da = -21.4
	 */

	f = acm_fopen("sa10.obv", "r");  /* FIXME */
	c->object = VReadObject(f);
	fclose(f);

	return &mk82Desc;
}


static int
dropOrdinance ( craft *c, int ind )
{
	craft *m;
	int i;
	VPoint    s, s1;
	VPoint    cY, mX, mY, mZ;
	double    v;

#ifdef HAVE_DIS
	double    disLocation[3];
	VPoint    velocity;
	double    disVelocity[3];
	double    disZeroVec[3];
	double    disOrientation[3];

#endif

	/*
	 *  Find an empty projectile entry
	 */

	for ((i = 0, m = &mtbl[0]); i < MAXPROJECTILES; (++i, ++m)) {
		if (m->type == CT_FREE) {
			m->type = CT_BOMB;
			break;
		}
	}

	if (i == MAXPROJECTILES)
		return -1;

	m->cinfo = lookupCraft("MK82");
	m->fuel = 0.0;
	m->curThrust = 0.0;
	m->owner = c->pIndex;
	m->payload = 0.0;
	m->armTimer = 0.0;
	m->prevSg = c->Sg;
	m->gvs_instance = (GVS_OBI) NULL;

/*
 *  Build trihedral based on the launching aircraft's current velocity vector
 *  rather than simply it's current direction vector.
 *
 *      (1)  build a unit velocity vector.
 *      (2)  calculate bomb local Z axis from
 *              plane's-y-axis CROSS bomb's-unit-velocity-vector
 *      (3)  calculate bomb's Y axis.
 */

	if ((v = Vmagnitude(&c->Cg)) < 1.0) {
		m->trihedral = c->trihedral;
		m->curRoll = c->curRoll;
		m->curPitch = c->curPitch;
		m->curHeading = c->curHeading;
	}
	else {
		mX = c->Cg;
		mX.x /= v;
		mX.y /= v;
		mX.z /= v;
		cY.x = c->trihedral.m[0][1];
		cY.y = c->trihedral.m[1][1];
		cY.z = c->trihedral.m[2][1];

		VCrossProd(&mX, &cY, &mZ);
		VCrossProd(&mZ, &mX, &mY);

		m->trihedral.m[0][0] = mX.x;
		m->trihedral.m[1][0] = mX.y;
		m->trihedral.m[2][0] = mX.z;
		m->trihedral.m[0][1] = mY.x;
		m->trihedral.m[1][1] = mY.y;
		m->trihedral.m[2][1] = mY.z;
		m->trihedral.m[0][2] = mZ.x;
		m->trihedral.m[1][2] = mZ.y;
		m->trihedral.m[2][2] = mZ.z;

		euler(m);
	}

	m->Cg = c->Cg;
	VTransform(&(c->cinfo->wStation[ind]), &(c->trihedral), &s1);
	VReverseTransform_(&s1, &c->XYZtoNED, &s);
	m->Sg.x = c->prevSg.x + FEETtoMETERS(s.x);
	m->Sg.y = c->prevSg.y + FEETtoMETERS(s.y);
	m->Sg.z = c->prevSg.z + FEETtoMETERS(s.z);
	DISGeocentricToWorldCoordinates((void *) & m->Sg, &m->w);
	GenerateWorldToLocalMatrix(&m->w, &m->XYZtoNED);
	m->armTimer = m->cinfo->armDelay;
	m->flags = 0;
	m->createTime = curTime;

	m->curRadarTarget = -1;

#ifdef HAVE_DIS

/*
 *  ACM bombs are DIS "tracked munitions", so we are
 *  responsible for sending entity state PDU's for them
 */

	if (disInUse) {
		VPoint    tmp;

		disLocation[0] = m->Sg.x;
		disLocation[1] = m->Sg.y;
		disLocation[2] = m->Sg.z;
		tmp.x = FEETtoMETERS(m->Cg.x);
		tmp.y = FEETtoMETERS(m->Cg.y);
		tmp.z = FEETtoMETERS(m->Cg.z);
		VReverseTransform_(&tmp, &m->XYZtoNED, &velocity);
		disVelocity[0] = velocity.x;
		disVelocity[1] = velocity.y;
		disVelocity[2] = velocity.z;
		disZeroVec[0] = 0.0;
		disZeroVec[1] = 0.0;
		disZeroVec[2] = 0.0;
		disOrientation[0] = m->curHeading;
		disOrientation[1] = m->curPitch;
		disOrientation[2] = m->curRoll;
		dis_entityEnter(c->team, m,
						&m->cinfo->entityType,
						&m->cinfo->altEntityType,
						disLocation, disVelocity,
						disZeroVec, disOrientation,
						disZeroVec, &m->disId);
#ifdef DIS_DEBUG
		fprintf(stderr, "%8.2f: dropOrdnance() mtbl[%d] disId=%d\n",
			curTime, i, m->disId);
#endif
	}
#endif
	return 0;
}


static int
update_mk82(craft * c)
{
	int i;

	if( ! hasFired[c->pIndex] )
		return 1;
	
	hasFired[c->pIndex] = 0;  /* reset fire request */

	/*
	 *  Bombs won't drop if we have "Weight on wheels"
	 *  or if we run out of ammunition.
	 */

	if( (c->flags & FL_ALL_GND_CONTACT) != 0
	|| count[c->pIndex] <= 0 )
		return 1;

	/* Get station from which to drop the bomb: */
	i = readyStation(c, WEAPON_MK82);
	if (i < 0){
		fprintf(stderr, "update_mk82() Oops. Can't find an MK82\n");
		return 1;
	}

	/*
	 *  Decrement bombs counter.
	 *  In arcade mode, we never run out of ammunition
	 */

	if (arcadeMode == 0) {
		c->station[i].id = -1;
		count[c->pIndex]--;
	}

	/* Drop bomb from station i: */
	dropOrdinance (c, i);
	/* playSound(c, SoundBombDrop); */

	return 1;
}

static int
display_mk82(craft *c, craftType *ct, viewer *u, int fpm_x, int fpm_y)
{
	char      s[16];
	WorldCoordinates ip_wc;  /* impact point (world) */
	VPoint    ip_gc;         /* impact point (geocentric XYZ coords) */
	VPoint    ip_ned;        /* impact point (NED coords) */
	VPoint    ip_eye;        /* impact point (eye space) */
	double    x0, y0;        /* cross center (viewport coords, pixels) */
	double    xscale, yscale;  /* world-to-viewport scale factors */

/*
 *  Update HUD display strings.
 */

	sprintf(s, "%d %s", count[c->pIndex], weapon_id_to_name(WEAPON_MK82));
	strcpy(c->leftHUD[3], s);
	strcpy(c->leftHUD[2], "");
	strcpy(c->leftHUD[4], "");

	/* Compute impact point ip_wc (world coords): */
 	computeImpactPoint(c, &ip_wc);

	/* Compute impact point ip_gc (geocentric coords XYZ): */
	DISWorldCoordinatesToGeocentric(&ip_wc, (void *) &ip_gc);

	/* Compute impact point ip_eye (pilot's eye coords): */
	VTransform(&ip_gc, &c->XYZtoNED, &ip_ned);
	VReverseTransform(&ip_ned, &c->AXYZtoNED, &ip_eye);
	/* Pilot's head position correction: */
	ip_eye.x -= FEETtoMETERS(c->cinfo->viewPoint.x);
	ip_eye.y -= FEETtoMETERS(c->cinfo->viewPoint.y);
	ip_eye.z -= FEETtoMETERS(c->cinfo->viewPoint.z);

	/*
		Display CCIP circle only if the impact point is ahead of us and
		closer than 45 DEG from HUD center:
	*/

	if( ip_eye.x > 0.0
	&& ip_eye.x * ip_eye.x > ip_eye.y * ip_eye.y + ip_eye.z * ip_eye.z ){

		xscale = u->v->xres * u->v->dist / ip_eye.x;
		yscale = u->v->yres * u->v->dist / ip_eye.x;

		/* Center of the CCIP circle: */
		x0 = u->v->focus.x + xscale * ip_eye.y;
		y0 = u->v->focus.y + yscale * ip_eye.z;

		draw_data *dd = draw_new();

		/* Draw CCIP circle, 5 mm radius: */
		double r = 0.005 /* m */ * u->v->xres * u->zoom/16.0;
		draw_circle(dd, x0, y0, r);

		/* Draw bomb fall line: */
		double nx = x0 - fpm_x;
		double ny = y0 - fpm_y;
		double n = sqrt(nx * nx + ny * ny);
		double fpm_r = 8.0 * u->xscaleFactor; /* FPM circle radius */
		/* Draw BFL only if FPM and CCIP are far enough: */
		if( n > fpm_r + r ){
			nx /= n;
			ny /= n;
			draw_segment(dd,
				fpm_x + nx * fpm_r,   fpm_y + ny * fpm_r,
				fpm_x + nx * (n - r), fpm_y + ny * (n - r));
		}

		draw_stroke(dd, u->v, HUDColor->cIndex);
		draw_free(dd);
	}

/*
 *  Return TRUE if we are recommending a bomb drop.
 */

	return 0;
}

static int
press_mk82(craft *c)
/* Fire! */
{
	hasFired[c->pIndex] = 1;
	return 1;
}

static int
release_mk82(craft *c)
{
	return 0;
}


int ccip_bombCalculations(craft * c)
{
	double    q, rho;
	double    FLift, FDrag, FWeight;
	double    dNorth, dEast, dmag;
	VPoint    F, Fg;

/*
 *  Check for ground impact.  We do this at the beginning to permit us to
 *  kill ground targets.
 */

	if (c->w.z < terrain_altitude(c)) {
		/* Bomb hit the ground. */
		return 1;
	}

/*
 *  Compute the resultant force vector on the bomb.
 */

	c->prevSg = c->Sg;
	rho = air_calc_rho(METERStoFEET(c->w.z), &q);
	c->VT = Vmagnitude(&c->Cg);
	q = rho * c->cinfo->wingS * c->VT * c->VT * 0.5;
	FLift = 0.0;
	FDrag = c->cinfo->CDOrigin * q;

	F.x = c->curThrust - FDrag;
	F.y = 0.0;
	F.z = 0.0;

/*
 *  Now calculate changes in position (Sg) and velocity (Cg).
 *  The bomb's trihedral matrix is managed by 
 *  trackTarget().
 */

	VTransform(&F, &c->AXYZtoNED, &Fg);
	FWeight = c->cinfo->emptyWeight;
	Fg.z += FWeight;
	calcGForces(c, &Fg, FWeight);

/*
 *  Update the bomb's position and velocity.
 */

	dNorth = FEETtoMETERS(c->Cg.x * deltaT + Fg.x / FWeight
						  * earth_g * halfDeltaTSquared);
	dEast = FEETtoMETERS(c->Cg.y * deltaT + Fg.y / FWeight
						 * earth_g * halfDeltaTSquared);
	c->w.z -= FEETtoMETERS(c->Cg.z * deltaT + Fg.z / FWeight
						   * earth_g * halfDeltaTSquared);

	dmag = sqrt(dNorth * dNorth + dEast * dEast);

	DISUpdateWorldCoordinates(&c->w, dNorth / dmag, dEast / dmag, dmag);
	DISWorldCoordinatesToGeocentric(&c->w, (void *) & c->Sg);
	GenerateWorldToLocalMatrix(&c->w, &c->XYZtoNED);

	c->Cg.x += Fg.x / FWeight * earth_g * deltaT;
	c->Cg.y += Fg.y / FWeight * earth_g * deltaT;
	c->Cg.z += Fg.z / FWeight * earth_g * deltaT;

	return 0;
}


static double distance(VPoint * a, VPoint * b, double range)
/*
	Calculate the distance between the two points.
	Return a negative value if the distance is greater then the range.
*/
{
	double dx = a->x - b->x;
	if( fabs(dx) > range )  return -1.0;
	
	double dy = a->y - b->y;
	if( fabs(dy) > range )  return -1.0;
	
	double dz = a->z - b->z;
	if( fabs(dz) > range )  return -1.0;
	
	double r = sqrt( dx * dx + dy * dy + dz * dz );

	if( r > range )
		return -1.0;

	return r;
}


int ccip_killBomb(craft * c)
{
#ifdef HAVE_DIS
	double    worldLocation[3], entityLocation[3];
	int       target_eid;
	VPoint    worldVel, tmp;
	craft *   target = NULL;

	if (c->type == CT_BOMB) {

		if (target == NULL) {
			target_eid = DIS_ID_NONE;
		}
		else {
			target_eid = target->disId;
		}

		worldLocation[0] = c->Sg.x;
		worldLocation[1] = c->Sg.y;
		worldLocation[2] = c->Sg.z;

/*
 *  killBomb's calling sequence needs to be updated to allow for the
 *  entity detonation location to be passed.
 */

		entityLocation[0] = 0.0;
		entityLocation[1] = 0.0;
		entityLocation[2] = 0.0;

		tmp.x = FEETtoMETERS(c->Cg.x);
		tmp.y = FEETtoMETERS(c->Cg.y);
		tmp.z = FEETtoMETERS(c->Cg.z);
		VReverseTransform_(&tmp, &c->XYZtoNED, &worldVel);

		if (disInUse) {
			dis_detonation(&c->cinfo->entityType,
						   ptbl[c->owner].disId,
						   target_eid,
						   DIS_ID_NONE,
						   worldLocation,
						   entityLocation,
						   (double *) &worldVel);
#ifdef HAVE_DIS
			dis_entityExit(c->disId);
#endif
		}
	}
#endif

	effects_new_explosion(&c->Sg, &(VPoint){0.0, 0.0, 0.0}, 50.0, 15.0, 1.0);
	pm_hud_strings_free(c);
	c->type = CT_FREE;

	int i;
	craft *p;
	double r, explosion_diameter_meters;

	for( i = 0; i < MAXPLAYERS; i++ ){
		p = &ptbl[i];

		/* Ignores free entries in players table: */
		if( p->type == CT_FREE )
			continue;

		/* Distance of the blast from this craft (m): */
		r = distance(&p->Sg, &c->Sg, 900.0 /* m */);
		if( r < 0.0 )
			continue;

		/* Absorb damage: */
		if( damage_absorbDISDamage(p, 
			&c->cinfo->entityType, 0, 0, 
			r,
			0.0 /* dummy */,
			&explosion_diameter_meters) == 0
		){
			killPlayer(p, "Mark 82 bomb");
		}
	}

	return 0;
}

