/*
 *  acm : an aerial combat simulator for X
 *  Copyright (C) 1991-1996  Riley Rainey
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 dated June, 1991.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program;  if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
 */

 /*
  *  The original "Smart Drone" code was created by
  *  Jason Nyberg (nyberg@ctron.com).  Enhancements added by Riley Rainey.
  *
  *  2007-07-23 Umberto Salsi <salsi@icosaedro.it>
  *  - made a proper "module"
  *  - drone now has a data structure by its own (see (craft)->drone field)
  *  - the 'aggressiveness' now sets the % of the max vertical load
  */

#include <string.h>

#include "pm.h"
#include "alarm.h"
#include "browse.h"
#include "dis_if.h"
#include "damage.h"
#include "effects.h"
#include "gear.h"
#include "planes.h"
#include "players.h"
#include "prompt.h"
#include "weapon.h"

#include "drone.h"

static DroneMode mode = DRONE_DOG_FIGHT_MODE;
static double aggressiveness = 0.5;

/* Drone status: */
#define DM_ATTACK       0     /* normal drone attack mode */
#define DM_RETURN       1     /* return to drone activation point 
								   (interceptStartPoint) */
#define DM_RETURN_CAPTURED 2  /* in return mode and moving towards return pt */

#define MIN_VEL       350     /* min speed (kt) */
#define MED_VEL       400     /* medium speed (kt) */
#define MAX_VEL       500     /* max speed (kt) */

#define MIN_ALTITUDE 15000    /* min altitude (ft) */

/*
 * "Opaque" data structure pointed by (craft)->drone.
 */
typedef struct _drone_data {
	struct _drone_data *next;

	int       curOpponent;		/* who this drone is trying to kill; -1 = none */
	int       curDroneMode;     /* drone operating mode (DM_* constants ) */
	int       holdCount;		/* non-zero when drones holding fire */
	VPoint    interceptStartPoint; /* geocentric coordinates location
									  where an end-game intercept was
									  initiated */
} drone_data;


static drone_data *free_list = NULL;


void drone_set_mode(DroneMode drone_mode)
{
	mode = drone_mode;
}


void drone_set_aggressiveness(double a)
{
	if( a < 0.05 )
		a = 0.05;
	else if( a > 1.0 )
		a = 1.0;
	aggressiveness = a;
}


static craft *
create_new_drone_craft(craft * p, int droneTeam, char *type)
{
	int       i;
	craft    *c;
	VPoint    s, tmp;
	double    v;


#ifdef HAVE_DIS
	double    disLocation[3];
	double    disZeroVec[3];
	double    disOrientation[3];
#endif

	/*
		Allocate new craft data structure:
	*/
	for (i = 0; i < MAXPLAYERS; ++i) {
		if (ptbl[i].type == CT_FREE) {
			break;
		}
	}

	if (i == MAXPLAYERS)
		return NULL;

	c = &ptbl[i];
	memset(c, '\0', sizeof(*c));

	/*
		Set craft state:
	*/
	c->pIndex = i;
	c->type = CT_DRONE;
	c->team = droneTeam;
	c->createTime = curTime;
	c->vl = NULL;
	c->trihedral = p->trihedral;
	c->Cg = p->Cg;
	c->VT = p->VT;
	c->Sg = p->Sg;
	c->w = p->w;
	c->terrain_altitude_timeout = curTime - 10.0; /* ensure will be updated */
	/* c->terrain_altitude = UNKNOWN; */
	c->XYZtoNED = p->XYZtoNED;
	c->rho = p->rho;
	c->mach1 = p->mach1;
	c->mach = p->mach;
	c->G = p->G;
	c->linAcc = p->linAcc;
	c->prevSg = p->prevSg;
	c->p = p->p;
	c->q = p->q;
	c->r = p->r;
	c->pitchComm = 0.0;
	c->rollComm = 0.0;
	c->steerComm = 0.0;
	c->rudderComm = 0.0;
	c->throttleComm = 32768;
	c->throttle = 32768;
	c->Se = p->Se;
	c->Sa = p->Sa;
	c->Sr = p->Sr;
	c->SeTrim = 0.0;
	c->SaTrim = 0.0;
	c->curHeading = p->curHeading;
	c->curPitch = p->curPitch;
	c->curRoll = p->curRoll;
	c->curThrust = 0.0;
	c->curFlap = 0.0;
	c->flapSetting = 0.0;
	c->curSpeedBrake = 0.0;
	c->speedBrakeSetting = 0.0;
	gear_up(c);
	c->throttle = 32768;
	c->rpm = 1.0;
	c->alpha = p->alpha;
	c->beta = p->beta;
	c->fuel = 0.0;
	c->payload = 0.0;
	c->flags = 0;
	c->damageBits = 0;
	c->structurePts = 0;
	c->leakRate = 0.0;
	c->damageCL = 0.0;
	c->damageCM = 0.0;
	c->radarMode = RM_ACM;
	c->nextRadarTime = 0.0;
	c->curNWDef = 0.0;
	c->cinfo = lookupCraft(type);
	c->order = 0;
	c->next = NULL;
	sprintf(c->name, "Drone%d", i);
	strcpy(c->display, "NO_DISPLAY");
	c->curRadarTarget = -1;
	c->targetDistance = 0.0;
	c->targetClosure = 0.0;
	/* skip fields: relPos, rval */
	pm_hud_strings_alloc(c);
	/* skip fields: curWeapon, station */
	/* skip fields: rinfo, rtop */
	c->showMag = FALSE;
	c->aps = NULL;
	c->curThrust = (c->cinfo->thrust) (c);


	if( mode == DRONE_DOG_FIGHT_MODE ){

		/*
			Position the drone about 1500 meters ahead of the player's craft.
		*/

		s.x = 1500.0 + 500.0 * damage_frandom2();
		s.y = 200.0 * damage_frandom2();
		s.z = 200.0 * damage_frandom2();
	
		c->curHeading = pm_normalize_yaw( p->curHeading
			+ DEGtoRAD(30.0) * damage_frandom2() );
		c->curRoll = 0.0;
		c->curPitch = DEGtoRAD(2.0);

		v = p->VT + 50.0 * damage_frandom2();

		VSetPoint(&c->Cg, v*cos(c->curHeading), v*sin(c->curHeading), 0.0);
	
	} else {  /* mode == DRONE_HUNTING_MODE */

		/*
			Position the drone up to 50 NM away, random altitude,
			random direction, random heading.
		*/

		s.x = 50000.0 * damage_frandom2();
		s.y = 50000.0 * damage_frandom2();
		s.z = 5000.0 - p->w.z + (11000.0 - 5000.0)/2.0 * damage_frandom2();

		c->curHeading = pm_normalize_yaw( 2*M_PI * damage_frandom2() );
		c->curRoll = 0.0;
		c->curPitch = DEGtoRAD(2.0);

		v = 350.0 + 50.0 * damage_frandom2();

		VSetPoint(&c->Cg, v*cos(c->curHeading), v*sin(c->curHeading), 0.0);
	
	}

	c->r = c->q = c->p = 0.0;

	eulerToMatrix(c->curRoll, c->curPitch, c->curHeading,
		&(c->trihedral));

	VTransform_(&s, &(p->trihedral), &tmp);

	/*  convert NED (meters) to Geocentric (meters)  */

	VReverseTransform_(&tmp, &(c->XYZtoNED), &c->Sg);
	c->Sg.x += p->Sg.x;
	c->Sg.y += p->Sg.y;
	c->Sg.z += p->Sg.z;
	c->prevSg = c->Sg;

	DISGeocentricToWorldCoordinates((void *) & c->Sg, &c->w);
	GenerateWorldToLocalMatrix(&c->w, &c->XYZtoNED);

/*
 *  Rearm and fuel the aircraft.
 */

	(*c->cinfo->resupply) (c);

	selectNamedWeapon(c, WEAPON_AIM9M);  /* will set properly next */

#ifdef HAVE_DIS
	if (disInUse) {
		disLocation[0] = c->Sg.x;
		disLocation[1] = c->Sg.y;
		disLocation[2] = c->Sg.z;
		disZeroVec[0] = 0.0;
		disZeroVec[1] = 0.0;
		disZeroVec[2] = 0.0;
		disOrientation[0] = c->curHeading;
		disOrientation[1] = c->curPitch;
		disOrientation[2] = c->curRoll;
		dis_entityEnter(droneTeam, c,
						&c->cinfo->entityType,
						&c->cinfo->altEntityType,
						disLocation, disZeroVec,
						disZeroVec, disOrientation,
						disZeroVec, &c->disId);
		dis_setRadarMode(c, 1, 1);
#ifdef DIS_DEBUG
		fprintf(stderr, "%8.2f: create_new_drone_craft() %s ptbl[%d] disId=%d\n",
			curTime, c->cinfo->name, i, c->disId);
#endif
	}
#endif

	return c;
}


void drone_take_commands(craft *c)
{
	drone_data *dd;

	c->type = CT_DRONE;
	/* ignore = */ selectNamedWeapon(c, WEAPON_AIM9M);

	if( free_list == NULL ){
		dd = MALLOC(sizeof(drone_data));
	} else {
		dd = free_list;
		free_list = free_list->next;
	}

	dd->next = NULL;
	dd->curOpponent = -1;
	dd->curDroneMode = DM_ATTACK;
	dd->holdCount = 0;
	/* dd->inceptStartPoint = ignored in DM_ATTACK mode; */

	c->drone = dd;
}


void drone_new(craft * p, char *plane_type)
{
	int droneTeam;
	char *type;
	craft *c;
	char s[100];

	/*
		Set drone team number:
	*/
	droneTeam = (p->team == 1)? 2 : 1;

	/*
		Set drone plane type:

		FIXME: we should look at the inventory for aircrafts actually
		available.
	*/
	if (plane_type != NULL) {
		type = plane_type;
	}
	else if (droneTeam == 1) {
		type = "F-16";
	}
	else {
		type = "MiG-29";
	}

	c = create_new_drone_craft(p, droneTeam, type);
	if( c == NULL ){
		prompt_craft_print(p, "Memory exausted -- can't create new drone");
		return;
	}

	drone_take_commands(c);

	sprintf(s, "%s drone generated for %s practice [%d]", c->name,
		(mode == DRONE_DOG_FIGHT_MODE)? "dog fight" : "hunting", mode);
	
	prompt_craft_print(p, s);
}


void drone_release_commands(craft *c)
{
	drone_data *dd;

	if( c->drone == NULL )
		return;
	
	dd = (drone_data *) c->drone;
	dd->next = free_list;
	free_list = dd;
	c->drone = NULL;

	c->type = CT_PLANE;
}


void drone_free_all()
{
	drone_data *dd;

	while( free_list != NULL ){
		dd = free_list;
		free_list = free_list->next;
		FREE(dd);
	}
}


void drone_reset_opponent(craft *c)
{
	drone_data *dd;

	if( c->drone == NULL )
		return;
	
	dd = (drone_data *) c->drone;
	dd->curOpponent = -1;
	dd->holdCount = 0;
}


/*
 *  convert target plane coords from world to drone's
 */

static void
myCoordSys(craft * c, craft * p, VPoint * pos, VPoint * vel)
{
	VPoint    tpos;

	VTransform(&p->prevSg, &c->XYZtoNED, &tpos);
	VReverseTransform_(&tpos, &c->trihedral, pos);
	VTransform_(&p->Cg, &c->XYZtoNED, &tpos);
	VReverseTransform_(&tpos, &c->trihedral, vel);
}

static void
unholdFireAlarm(void *arg1, void *arg2)
{
	craft    *c = (craft *) arg1;
	drone_data *dd;

	if( c->drone == NULL )
		return;
	
	dd = (drone_data *) c->drone;

	if (dd->holdCount > 0) {
		dd->holdCount--;
	}
}

/* 
 *  Dumbly choose the closest hostile plane to be target
 */

static int
pickTarget(craft * c)
{
	int       i, target = -1;
	craft    *p;
	double    d, min;
	VPoint    pos, vel;

	min = 100000000.0;
	for (i = 0, p = ptbl; i < MAXPLAYERS; ++i, ++p) {

		if ( p != c && p->team != c->team
		&& (p->type == CT_PLANE || 
			p->type == CT_DIS_PLANE ||
			p->type == CT_DRONE)
		) {

			myCoordSys(c, p, &pos, &vel);

			d = Vmagnitude( &pos );

			if (d < min) {
				min = d;
				target = p->pIndex;
			}
		}
	}

	return target;
}

/*
 *  droneFlyTo
 *
 *  Generate sitck/rudder controls to move the plane to the specified
 *  geocentric point.
 */

static void
droneFlyTo ( craft *c, VPoint *pos )
{
	double d, phi_rad, w, f, pitch, roll;
	craftType *p;
	VPoint q;

	/*****  Example: climb toward north
	q.x = 100.0;
	q.y = 0.0;
	q.z = -20.0;

	VReverseTransform(&q, &c->trihedral, pos);
	*****/

	/*
		Avoid too low altitudes
	*/

	if( c->w.z < FEETtoMETERS(MIN_ALTITUDE) ){
		VTransform(pos, &c->trihedral, &q);
		if( q.z >= 0.0 ){
			q.z = -0.5*sqrt(q.x*q.x + q.y*q.y);
		}
		VReverseTransform(&q, &c->trihedral, pos);
	}

	d = sqrt(pos->x * pos->x + pos->y * pos->y + pos->z * pos->z);

	/*
	 *  A lot of conventional 1V1 air combat involves keeping your lift
	 *  vector on the target aircraft.  Phi is the computed angle between our
	 *  target and the lift vector (simplified to be just the negative Z-axis).
	 */

	phi_rad = atan2 ( pos->y, -pos->z );

	/*
	 * Compute total lift force 'f':
	 */

	p = c->cinfo;
	w = p->emptyWeight + c->fuel + c->payload;
	f = c->G.z * w; /* current vertical load (pounds) */

	/*
	 *  If the target is behind our 3/9-line, we are defensive.  Pull maximum
	 *  G's into the target (after rolling into him).
	 *
	 *  If we are behind of the target, perform pure pursuit (until the code 
	 *  gets a bit smarter).
	 */

	pitch = c->pitchComm; /* target pitch command */

	if (f < -aggressiveness * p->maxLoadZPositive) {
		/*
		 * Close to the positive structural load limit.
		 */
		if (c->pitchComm < 0.0) {
			pitch = 0.95*pitch;
		}
		
	} else if (f > aggressiveness * p->maxLoadZNegative) {
		/*
		 * Close to the negative structural load limit.
		 */
		if (c->pitchComm > 0.0) {
			pitch = 0.95*pitch;
		}

	} else if (pos->x < 0.0) {
		/*
		 *  Target behind me.
		 *  Wait for lift vector to be close to where we want it before
		 *  pulling G's.
		 */
		if ((fabs(phi_rad) > DEGtoRAD(130.0))
		|| fabs(phi_rad) < DEGtoRAD(50.0)) {
			pitch = -1.0;
		}

	} else {
		pitch = pos->z / d * 3.0;

	}

	/*
	 * Smootly apply the target pitch
	 */
	if( pitch > c->pitchComm )
		c->pitchComm += 0.2*deltaT;
	else
		c->pitchComm -= 0.04*deltaT;

	/*
	 *  Put the lift vector on the target.
	 */

	roll = c->rollComm;

	if (fabs (phi_rad) > DEGtoRAD(150.0)) {
		roll = 0.2 * phi_rad;
	}
	else if (fabs (phi_rad) > DEGtoRAD(20.0)) {
		roll = -phi_rad;
	}
	else {
		roll = -0.2 * phi_rad;
	}

	/*
	 * Smootly apply the target roll
	 */
	if( roll > c->rollComm )
		c->rollComm += 2.0*deltaT;
	else
		c->rollComm -= 2.0*deltaT;

	/*
	 *  Don't use the rudder, for now.
	 */

	c->rudderComm = 0.0;

	/*
	 *  Constrain control surface positions to valid values.
	 */

	if (c->pitchComm > 1.0)
		c->pitchComm = 1.0;
	else if (c->pitchComm < -1.0)
		c->pitchComm = -1.0;

	if (c->rollComm > 1.0)
		c->rollComm = 1.0;
	else if (c->rollComm < -1.0)
		c->rollComm = -1.0;

	/* FIXME: no rudder for now
	if (c->rudderComm > 1.0)
		c->rudderComm = 1.0;
	else if (c->rudderComm < -1.0)
		c->rudderComm = -1.0;
	*/
	c->rudderComm = 0.0;

	/*
	 * Adjust engine power
	 */

	if( c->VT < KTtoFPS(MIN_VEL) ){
		c->throttleComm = 32768;  /* max power */
		pm_after_burner_on(c);
	} else if( c->VT < KTtoFPS(MED_VEL) ){
		c->throttleComm = 32768;  /* max power */
		pm_after_burner_off(c);
	} else if( c->VT < KTtoFPS(MAX_VEL) ){
		c->throttleComm = 32768/2;  /* medium power */
		pm_after_burner_off(c);
	} else {
		c->throttleComm = 32768/4;  /* idle */
		pm_after_burner_off(c);
	}

}

/*
 *  Drone flight management in Attack Mode (this is most common)
 */

static int
droneCalculationsAttackMode ( craft * c )
{
	drone_data *dd;
	double    htime;
	VPoint    pos, vel;
	_BOOL     doFire;

	dd = (drone_data *) c->drone;

	/*
	 * Our opponent has exited?  Return to engagement initiation point.
	 */

	if ( (c->flags & FL_END_GAME_DRONE ) ) {
		 if ( dd->curOpponent != -1 &&
			 ptbl[dd->curOpponent].type == CT_FREE ) {
			 dd->curDroneMode = DM_RETURN;
		 }
	}

	/*
	 *  No opponent, or opponent isn't there anymore
	 */

	else if (dd->curOpponent == -1 ||
		ptbl[dd->curOpponent].type == CT_FREE) {
		dd->curOpponent = pickTarget(c);
		dd->holdCount = 0;
	}

	if (dd->curOpponent != -1) {

		myCoordSys(c, &(ptbl[dd->curOpponent]), &pos, &vel);

		droneFlyTo ( c, &pos );

		/*
		 *  Fire at the target, if appropriate.
		 *  Tryes first with AIM-120, then AIM-9M.
		 *
		 *  We'll have to figure out a way to do lead pursuit in order to 
		 *  fire the cannon; we do pure pursuit now, which is the 
		 *  (somewhat) right thing to fire a missile.
		 */

		if (dd->holdCount == 0 ) {

			/*
				Select weapon. Try AIM-120 first, then AIM-9M.
			*/

			if( readyStation(c, WEAPON_AIM120) >= 0
			&& selectNamedWeapon(c, WEAPON_AIM120) )
				doFire = TRUE; /* Uses AIM-120 */

			else if( readyStation(c, WEAPON_AIM9M) >= 0
			&& selectNamedWeapon(c, WEAPON_AIM9M) )
				doFire = TRUE; /* Uses AIM-9M */

			else
				doFire = FALSE; /* No suitable weapon available. */

			if( doFire && doWeaponDisplay(c, (viewer *) NULL, 0, 0) == 1) {
				fireWeapon(c);
				htime = 10.0 + 10.0 * damage_frandom2();
				alarm_add(htime, unholdFireAlarm, c, NULL);
				dd->holdCount++;
			}
		}


	}

	return 0;
}

/*
 *  Drone flight management in Return Modes
 */

static int
droneCalculationsReturnMode ( craft * c )
{
	drone_data *dd;
	VPoint tpos, pos, vel;
	double dist_meters, closure_meters_per_sec;
	int result = 0;

	dd = (drone_data *) c->drone;

	/*
	 *  Generate body relative position of return point
	 */

	VTransform(&dd->interceptStartPoint, &c->XYZtoNED, &tpos);
	VReverseTransform_(&tpos, &c->trihedral, &pos);

	/*
	 *  Convert NED velocity to body-relative velocity
	 */

	VReverseTransform_(&c->Cg, &c->trihedral, &vel);

	dist_meters = Vmagnitude( &pos );

	closure_meters_per_sec = 
		( vel.x * pos.x + vel.y * pos.y + vel.z * pos.z ) / dist_meters;

	droneFlyTo ( c, &pos );

	/*
	 * If we are in return mode and turned within 30 degrees towards 
	 * the return point, enter the "return-captured" mode.
	 */

	if ( dd->curDroneMode == DM_RETURN ) {

		if ( closure_meters_per_sec > 0.866 * FEETtoMETERS(c->VT) ) {
			dd->curDroneMode = DM_RETURN_CAPTURED;
		}
	}

	/*
	 * If we are in return-capture mode and start to move away from the
	 * intercept point, then we're done; destroy the aircraft.
	 */

	else {
		if ( closure_meters_per_sec < 0.0 ) {
			result = 1;
		}
	}

	return result;
}

_BOOL
drone_calculations( craft * c )
{
	drone_data *dd;

	dd = (drone_data *) c->drone;

	switch (dd->curDroneMode) {

	case DM_ATTACK:
		return droneCalculationsAttackMode ( c );

	case DM_RETURN:
		return droneCalculationsReturnMode ( c );

	case DM_RETURN_CAPTURED:
		return droneCalculationsReturnMode ( c );

	default:
		abort();
	}
}

#ifdef USE_REQUEST_CONTROL

void
drone_endGameDistanceCheck(void * p1, void *p2)
{
	double range_meters;
	VPoint del;
	craft *p;
	craft *c =  (craft * ) p1;
	viewer *u = (viewer *) p2;
	int i;
	int done = 0;
	double threshold_meters;

	if ( c->type == CT_DIS_PLANE && (c->flags & FL_END_GAME_DRONE) ) {

		/*
		 * Determine the appropriate range threshold; if one was specified
		 * on the command line, use that. Otherwise use the lock range from
		 * the aicraft definition.
		 */

		if ( end_game_threshold_meters <= 0.0 ) {
			threshold_meters = FEETtoMETERS( c->cinfo->radarTRange * NM );
		}
		else {
			threshold_meters = end_game_threshold_meters;
		}

		for ((i = 0, p = ptbl); (i < MAXPLAYERS) && ( ! done ); (++i, ++p)) {

			/*
			 * Skip this entry if:
			 *
			 *   1) It's the entry for our own aircraft.
			 *   2) The craft isn't a hostile.
			 */

			if (p->pIndex == c->pIndex || c->team == p->team ) {
				continue;
			}

			if ( p->type == CT_PLANE || 
				 p->type == CT_DIS_PLANE ||
				 p->type == CT_DRONE) {

				del.x = p->Sg.x - c->Sg.x;
				del.y = p->Sg.y - c->Sg.y;
				del.z = p->Sg.z - c->Sg.z;

				range_meters = Vmagnitude( &del );

				/*
				 * If the distance is within our threshold, then
				 * initiate a control request.
				 */

				if ( range_meters <= threshold_meters ) {
					Entity_t * e = dis_getEntityTable();

					/*
					 *  Record start point of engagement; we will return
					 *  to this point after a kill.
					 */
					if( c->drone != NULL )
						((drone_data *)(c->drone))->interceptStartPoint = c->Sg;

					dis_requestControl ( &e[c->disId],
										 controlRequestCallback, u );

					done = 1;
				}

			}
		}
	}

	/*
	 *  This craft is no longer a DIS plane.  No need to continue proximity
	 *  testing.
	 */

	else {
		done = 1;
	}

	/*
	 *  If nothing was within our threshold range, look again after one
	 *  second has elapsed.
	 */

	if ( ! done ) {
#ifdef DEBUG
		printf ("not done: adding alarm\n");
#endif
		alarm_add(1.0, drone_endGameDistanceCheck, p1, p2);
	}

#ifdef DEBUG
	printf ("distance check exit\n");
#endif
			
}

#endif
