/*
 *  acm : an aerial combat simulator for X
 *  Copyright (C) 1991-1998  Riley Rainey
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 dated June, 1991.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program;  if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
 */

/*
 *  2014-12-29 dlc - Refer to CHANGES text file for details of modifications
 */

#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "adf.h"
#include "aps.h"
#include "box.h"
#include "browse.h"
#include "commands.h"
#include "dis_if.h"
#include "drone.h"
#include "flaps.h"
#include "gear.h"
#include "hud.h"
#include "hsi.h"
#include "instruments.h"
#include "joystick.h"
#include "mouse.h"
#include "navaid.h"
#include "planes.h"
#include "players.h"
#include "pm.h"
#include "prompt.h"
#include "radar.h"
#include "render.h"
#include "terminal.h"
#include "update.h"
#include "weapon.h"
#include "windows.h"

#define MAX_MAPPED_STRING_LEN	20
#define MAX_POPUP_STRING_LEN	40

#define FLIGHT_STATE_FILE "./acm-flight-state"

/*
	To prevent dangerous typos from causing unfortunate consequences,
	several events require double typing with a couple seconds to be
	activated: Quit (XK_Q), change stick (XK_m), and calibrate joystick
	(XK_K).  These variables hold the time of the previous key event.
*/
static double pressed_XK_Q_time = -5.0;
static double pressed_XK_m_time = -5.0;
static double pressed_XK_K_time = -5.0;

struct		keymasks {
char		b1;
char		b2;
char		b3;
char		b4;
char		b5;
char		sm;
char		lm;
char		cm;
char		m1;
char		m2;
char		m3;
char		m4;
char		m5;
char		n;
} static	kmsks = {	'.', '.', '.', '.', '.',
				'.', '.', '.',
				'.', '.', '.', '.', '.', '\0'};

KeySym		keysym;
XComposeStatus	compose;
char		buffer[MAX_MAPPED_STRING_LEN];
int		buflen = MAX_MAPPED_STRING_LEN;
KeySym		keysym;
char *		symstr;
char	  	viewermsg[121];
const char	evtfmt[] = "%s %c%c %16s\n";
const char	actfmt[] = "%8.2f         %16s %s\n";

char    	ev_buf[EV_TRACE_SIZE + 1];

static _BOOL
getEventInfo(craft * c, viewer * u, XEvent * ev) 
{

	char	dtype, etype;

	dtype = ev->type == KeyPress || ev->type == KeyRelease  ? 'K' : 'M';
	etype = ev->type == KeyPress || ev->type == ButtonPress ? 'V' : '^';
		
	if (dtype == 'K') {
		if ( ! u->hasFocus )  {
			sprintf(ev_buf, "%8.2f %s", curTime,
				"Keyboard event ignored (!focus)");
			add_ev_trace(u, ev_buf);
			fprintf(stderr, "%s\n", ev_buf);
			return FALSE;
		}

        	kmsks.b1 = ev->xkey.state & Button1Mask ? '1' : '.';
        	kmsks.b2 = ev->xkey.state & Button1Mask ? '2' : '.';
        	kmsks.b3 = ev->xkey.state & Button1Mask ? '3' : '.';
        	kmsks.b4 = ev->xkey.state & Button1Mask ? '4' : '.';
        	kmsks.b5 = ev->xkey.state & Button1Mask ? '5' : '.';
        	kmsks.sm = ev->xkey.state & ShiftMask   ? 'S' : '.';
        	kmsks.lm = ev->xkey.state & LockMask    ? 'L' : '.';
        	kmsks.cm = ev->xkey.state & ControlMask ? 'C' : '.';
        	kmsks.m1 = ev->xkey.state & Mod1Mask    ? '1' : '.';
        	kmsks.m2 = ev->xkey.state & Mod1Mask    ? '2' : '.';
        	kmsks.m3 = ev->xkey.state & Mod1Mask    ? '3' : '.';
        	kmsks.m4 = ev->xkey.state & Mod1Mask    ? '4' : '.';
        	kmsks.m5 = ev->xkey.state & Mod1Mask    ? '5' : '.';

		(void) XLookupString((XKeyEvent *) ev,
					buffer, buflen, &keysym, &compose);

 		if ((symstr = (char *) XKeysymToString(keysym)) == NoSymbol)
			symstr = &"[NoSymbol]";
	} else {
		if ( ! u->hasFocus | ! u->hasComm ) {
			sprintf(ev_buf, "%8.2f %s", curTime,
				"Keyboard event ignored (!focus||!cmds)");
			add_ev_trace(u, ev_buf);
			fprintf(stderr, "%s\n", ev_buf);
			return FALSE;
		}

	        kmsks.b1 = ev->xbutton.state & Button1Mask ? '1' : '.';
        	kmsks.b2 = ev->xbutton.state & Button1Mask ? '2' : '.';
        	kmsks.b3 = ev->xbutton.state & Button1Mask ? '3' : '.';
        	kmsks.b4 = ev->xbutton.state & Button1Mask ? '4' : '.';
        	kmsks.b5 = ev->xbutton.state & Button1Mask ? '5' : '.';
        	kmsks.sm = ev->xbutton.state & ShiftMask   ? 'S' : '.';
        	kmsks.lm = ev->xbutton.state & LockMask    ? 'L' : '.';
        	kmsks.cm = ev->xbutton.state & ControlMask ? 'C' : '.';
        	kmsks.m1 = ev->xbutton.state & Mod1Mask    ? '1' : '.';
        	kmsks.m2 = ev->xbutton.state & Mod1Mask    ? '2' : '.';
        	kmsks.m3 = ev->xbutton.state & Mod1Mask    ? '3' : '.';
        	kmsks.m4 = ev->xbutton.state & Mod1Mask    ? '4' : '.';
        	kmsks.m5 = ev->xbutton.state & Mod1Mask    ? '5' : '.';

		switch (ev->xbutton.button) {
		case Button1: symstr = &"1" ; break;
		case Button2: symstr = &"2" ; break;
		case Button3: symstr = &"3" ; break;
		case Button4: symstr = &"4" ; break;
		case Button5: symstr = &"5" ; break;
		}
	}

	sprintf(ev_buf, "%8.2f %s %c%c %s", curTime,
		(char *) &kmsks, dtype, etype, symstr);
	add_ev_trace(u, ev_buf);
	fprintf(stderr, evtfmt, (char *) &kmsks, dtype, etype, symstr);

	return TRUE;
}


static void
updateHUDColor(viewer * u)
{
	HUDColor = HUDColors[u->hColor * 4 + u->hBright];
	HUDPixel = HUDColor->cIndex;
	u->hdgScale.pixel = u->v->w->pixel[HUDPixel];
	u->altScale.pixel = u->v->w->pixel[HUDPixel];
	u->velScale.pixel = u->v->w->pixel[HUDPixel];
}


static _BOOL
doButtonEvent(craft * c, viewer * u, XEvent * ev)
{
	if (! getEventInfo(c, u, ev)) {
		sprintf(ev_buf, "%8.2f %s", curTime,
			"Button press event ignored (!getEventInfo)");
		add_ev_trace(u, ev_buf);
		fprintf(stderr, actfmt, curTime, symstr,
			"getEventInfo returned false--button event ignored");
		return FALSE;
	}

	if (ev->xbutton.button == Button1) {
		add_ev_trace(u, "Button1 calling selectCockpitItem");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    selectCockpitItem");
		selectCockpitItem( c, u, 
						   ev->xbutton.x, 
						   ev->xbutton.y, 
						   ev->xbutton.time );
	}
	else if (ev->xbutton.button == Button2) {
		add_ev_trace(u, "Button2 calling fireWeapon");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    fireWeapon");
		fireWeapon(c);
	}
	else if (ev->xbutton.button == Button3) {
		add_ev_trace(u, "Button3 calling selectWeapon");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    selectWeapon");
		selectWeapon(c);
	}
	return TRUE;
}


static _BOOL
doButtonReleaseEvent(craft * c, viewer * u, XEvent * ev, int player)
{

	if (! getEventInfo(c, u, ev))  {
		sprintf(ev_buf, "%8.2f %s", curTime,
			"Button release event ignored (!getEventInfo)");
		add_ev_trace(u, ev_buf);
		fprintf(stderr, actfmt, curTime, symstr,
			"getEventInfo returned false--button release event ignored");
		return FALSE;
	}

	if (ev->xbutton.button == Button2) {
		add_ev_trace(u, "Button2 calling ceaseFireWeapon");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    ceaseFireWeapon");
		ceaseFireWeapon(c);
	}
	return TRUE;
}


#define SWITCH_TO_NEXT_PLAYER_ENABLED

#ifdef SWITCH_TO_NEXT_PLAYER_ENABLED

static void
switchToNextPlayer(viewer * v)
{
	int    i;
	craft *p;

	for ((i = 0, p = ptbl); i < MAXPLAYERS; (++i, ++p)) {
		if ( ((p->type == CT_DRONE) || (p->type == CT_PLANE))
		&& p->vl == NULL ) {
			break;
		}
	}

	if ( i == MAXPLAYERS ) {
		prompt_viewer_print(v, "No other plane available");
		return;
	}

	v->c->vl = NULL; /* FIXME: vl might contain several viewers */
	v->c = p;
	p->vl = v;
	/****
	if ( p->type == CT_DRONE )
		v->hasComm = FALSE;
	else
		v->hasComm = TRUE;
	****/
}

#endif


static _BOOL
doKeyRelease(craft * c, viewer * u, XKeyEvent * ev)
/*
	Handle key release event. Return TRUE is player gets killed or player
	asked to terminate the program.
*/
{

	if (! getEventInfo(c, u, (XEvent *) ev)) {
		sprintf(ev_buf, "%8.2f %s", curTime,
			"Key release event ignored (!getEventInfo)");
		add_ev_trace(u, ev_buf);
		fprintf(stderr, actfmt, curTime, symstr,
			"getEventInfo returned false--key release event ignored");
		return FALSE;
	}
	if (keysym == XK_v && ! terminal_enabled(u) ) {
		add_ev_trace(u, "v^ calling ceaseFireWeapon");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    ceaseFireWeapon");
		ceaseFireWeapon(c);
	}
	return FALSE;
}


static _BOOL
doKeyEvent(craft * c, viewer * u, XKeyEvent * ev)
/*
	Handle key press event. Return TRUE is player gets killed or player
	asked to terminate the program.
*/
{

#ifdef SPECIAL_KEYS
	FILE     *fp;
	craft     pentry;
#endif

	if (! getEventInfo(c, u, (XEvent *) ev)) {
		sprintf(ev_buf, "%8.2f %s", curTime,
			"Key press event ignored (!getEventInfo)");
		add_ev_trace(u, ev_buf);
		fprintf(stderr, actfmt, curTime, symstr,
			"getEventInfo returned false--key event ignored");
		return FALSE;
	}

	if ( ! u->hasComm ){
		if ( keysym == XK_d ) {
			/* [d] must be unaffected by testKeys mode when u=>hasComm is FALSE */
			add_ev_trace(u, "[d]V attaching ACM back to X inputs");
			fprintf(stderr, actfmt, curTime, symstr,
				"attaching  ACM back to X inputs");
			u->hasComm = TRUE;
		} else {
			add_ev_trace(u, "Non-[d] key press ignored (! u->hasComm)");
			fprintf(stderr, "%s\n",
				"Non-[d] key press ignored (commands are disabled)");
		}
		return FALSE;
	}

	if ( terminal_enabled(u) && (keysym != XK_F1) ){
		/* Cannot be in terminal mode and testKeys--testKeys must be FALSE to get here */
		add_ev_trace(u, " V calling terminal_input_char");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    terminal_input_char");
		terminal_input_char(u, buffer[0]);
		{
			char line[100];
			if( terminal_read_string(u, line, 100) ) {
				add_ev_trace(u, " V calling commands_execute");
				fprintf(stderr, actfmt, curTime, symstr,
					"calling    commands_execute");
				commands_execute(u, line);
			}
		}
		return FALSE;
	}

	switch (keysym) {

	case XK_F1:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [F1] 'Terminal Facility, Toggle'");
			break;
		}
		add_ev_trace(u, "[F1]V calling terminal_toggle");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    terminal_toggle");
		terminal_toggle(u);
		break;

	case XK_F2:
		/* Obviously [F2] is not a normal pilot action key for testKeys mode */
		add_ev_trace(u, "[F2]V toggling testKeys mode");
		fprintf(stderr, actfmt, curTime, symstr,
			"toggling   testKeys mode");
		u->testKeys ^= 1;
		if (u->testKeys )
			prompt_viewer_print(u, "NOW IN TESTKEYS MODE: [F2] 'Test Keys, Toggle'");
		else	prompt_viewer_print(u, "Left testKeys mode--all pilot commands are executable again");
		break;

#ifdef SWITCH_TO_NEXT_PLAYER_ENABLED
	case XK_Tab:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [Tab] 'Stealth, Switch To Next Player'");
			break;
		}
		add_ev_trace(u, "[Tab]V calling switchToNextPlayer");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    switchToNextPlayer");
		switchToNextPlayer(u);
		break;
#endif

	case XK_d:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [d] 'X-windows Inputs, Toggle'");
			break;
		}
		add_ev_trace(u, "[d]V detaching ACM from X inputs");
		fprintf(stderr, actfmt, curTime, symstr,
			"detaching  ACM from X inputs");
		u->hasComm = FALSE;
		break;

	case XK_K:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [K] 'Joystick, Calibrate'");
			break;
		}
		if ( pressed_XK_K_time + 2.0 >= curTime ) {
			add_ev_trace(u, "[K]V calling CalibrateJoystick");
			fprintf(stderr, actfmt, curTime, symstr,
				"calling    CalibrateJoystick");
			CalibrateJoystick();
			return TRUE;
		}
		else {
			add_ev_trace(u, "[K]V arming CalibrateJoystick");
			fprintf(stderr, actfmt, curTime, symstr,
				"arming     CalibrateJoystick");
			pressed_XK_K_time = curTime;
			prompt_viewer_print(u, "Press SHIFT-K again within 2 sec. to calibrate the joystick");
		}
		break;

	case XK_m:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [m] 'Control Stick Mode, Toggle'");
			break;
		}
		if ( pressed_XK_m_time + 2.0 >= curTime ){
		    if (u->noMorJ) {
			add_ev_trace(u, "[m]V switching to Mouse/Joystick");
			fprintf(stderr, actfmt, curTime, symstr,
				"toggling   stick to mouse");
			u->noMorJ = FALSE;
			prompt_viewer_print(u, "Mouse pitch and roll now ACTIVE");
		    } else {
			add_ev_trace(u, "[m]V switching to Keystick");
			fprintf(stderr, actfmt, curTime, symstr,
				"toggling   stick to keyboard");
			u->noMorJ = TRUE;
			prompt_viewer_print(u, "Mouse pitch and roll now INACTIVE");
		    }
			return TRUE;
		}
		else {
			add_ev_trace(u, "[m]V arming stick toggle");
			fprintf(stderr, actfmt, curTime, symstr,
				"arming     stick toggle");
			pressed_XK_m_time = curTime;
			prompt_viewer_print(u, "Press [m] again within 2 sec. to switch the stick control");
		}
		break;

	case XK_U:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [u] 'Screen, View Forward'");
			break;
		}
		add_ev_trace(u, "[U]V toggling keystick cursor display");
		fprintf(stderr, actfmt, curTime, symstr,
			"toggling   keyboard cursor display");
		u->showMark ^= 1;
		break;
	case XK_v:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [v] 'Weapons, Open⇓|Cease⇑ Fire'");
			break;
		}
		add_ev_trace(u, "[v]V calling fireWeapon");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    fireWeapon");
		fireWeapon(c);
		break;

	case XK_V:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [V] 'Weapons, Cycle'");
			break;
		}
		add_ev_trace(u, "[V]V calling selectWeapon");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    selectWeapon");
		selectWeapon(c);
		break;

	case XK_G:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [G] 'Ground Color, Cycle'");
			break;
		}
		add_ev_trace(u, "[G]V cycling ground color");
		fprintf(stderr, actfmt, curTime, symstr,
			"cycling    ground color");
		if ( gtblCount < 2 )
			prompt_viewer_print(u, "Only one GROUNDC defined");
		else {
    			if ( gtblIndex == gtblCount - 1 )
				gtblIndex = 0;
			else    gtblIndex++;
    			groundColor = gtbl[gtblIndex].vcptr;
			sprintf(viewermsg, "groundColor is now %s", gtbl[gtblIndex].name);
			prompt_viewer_print(u, viewermsg);
		}
		break;

	case XK_C:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [C] 'HUD Color, Cycle'");
			break;
		}
		add_ev_trace(u, "[C]V cycling HUD color");
		fprintf(stderr, actfmt, curTime, symstr,
			"cycling    HUD color");
		u->hColor = u->hColor == 3 ? 0 : u->hColor + 1;
		updateHUDColor(u);
		break;

	case XK_B:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [B] 'HUD Brightness, Toggle'");
			break;
		}
		add_ev_trace(u, "[B]V cycling HUD brightness");
		fprintf(stderr, actfmt, curTime, symstr,
			"cycling    HUD brightness");
		u->hBright = u->hBright == 3 ? 0 : u->hBright + 1;
		updateHUDColor(u);
		break;

	case XK_H:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [H] 'HUD/Instrument Panel Display, Toggle'");
			break;
		}
		add_ev_trace(u, "[H]V toggling HUD/panel display");
		fprintf(stderr, actfmt, curTime, symstr,
			"toggling   HUD display");
		windows_set_layout(c, u, u->w->width, u->w->height, !u->hud_mode);
		break;

	case XK_A:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [A] '[AP] Hold Rate of Climb (AutoPilot), Toggle'");
			break;
		}
		add_ev_trace(u, "[A]V calling aps_ap_toggle");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    aps_ap_toggle");
		aps_ap_toggle(c);
		break;

	case XK_5:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [5] '[AP] Decrease Rate of Climb ~1 FPM'");
			break;
		}
		add_ev_trace(u, "[5]V calling aps_ap_adj_vs(-1.0)");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    aps_ap_adj_vs(-1.0)");
		aps_ap_adj_vs(c, -1.0);
		break;

	case XK_6:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [6] '[AP] Increase Rate of Climb ~1 FPM'");
			break;
		}
		add_ev_trace(u, "[6]V calling aps_ap_adj_vs(+1.0)");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    aps_ap_adj_vs(+1.0)");
		aps_ap_adj_vs(c, +1.0);
		break;

	case XK_percent:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [%] '[AP] Decrease Rate of Climb ~100 FPM'");
			break;
		}
		add_ev_trace(u, "[%]V calling aps_ap_adj_vs(-100.0)");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    aps_ap_adj_vs(-100.0)");
		aps_ap_adj_vs(c, -100.0);
		break;

	case XK_asciicircum:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [^] '[AP] Increase Rate of Climb ~100 FPM'");
			break;
		}
		add_ev_trace(u, "[^]V calling aps_ap_adj_vs(+100.0)");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    aps_ap_adj_vs(+100.0)");
		aps_ap_adj_vs(c, +100.0);
		break;

	case XK_Z:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [Z] '[AP] Hold Altitude (AutoPilot), Toggle'");
			break;
		}
		add_ev_trace(u, "[Z]V calling aps_ap_toggle");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    aps_ap_toggle");
		aps_ap_toggle(c);
		if( aps_ap_enabled(c) )
			aps_ap_set_vs(c, 0.0);
		break;

	case XK_X:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [X] '[AC] Coordinate Roll/Yaw (AutoCoordinate), Toggle'");
			break;
		}
		add_ev_trace(u, "[X]V calling aps_ac_toggle");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    aps_ac_toggle");
		aps_ac_toggle(c);
		break;

	case XK_T:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [T] '[AT] Hold True Airspeed (AutoThrottle), Toggle'");
			break;
		}
		add_ev_trace(u, "[T]V calling aps_at_toggle");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    aps_at_toggle");
		aps_at_toggle(c);
		break;

	case XK_N:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [N] '[AN] Navigate To VOR (AutoNavigate), Toggle'");
			break;
		}
		add_ev_trace(u, "[N]V calling aps_an_toggle");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    aps_an_toggle");
		aps_an_toggle(c);
		break;

	case XK_L:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [L] '[AL] ILS Approach and Landing (AutoLanding), Toggle'");
			break;
		}
		add_ev_trace(u, "[L]V calling aps_al_toggle");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    aps_al_toggle");
		aps_al_toggle(c);
		break;

	case XK_parenleft:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [(] '[AW] Maximum Bank, Decrease'");
			break;
		}
		add_ev_trace(u, "[(]V calling aps_bank_max_dec");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    aps_bank_max_dec");
		aps_bank_max_dec(c);
		break;

	case XK_parenright:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [)] '[AW] Maximum Bank, Increase'");
			break;
		}
		add_ev_trace(u, "[)]V calling aps_bank_max_inc");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    aps_bank_max_inc");
		aps_bank_max_inc(c);
		break;

	case XK_M:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [M] 'Magnetic/True Mode, Toggle'");
			break;
		}
		add_ev_trace(u, "[M]V toggling magnetic/true");
		fprintf(stderr, actfmt, curTime, symstr,
			"toggling   magnetic/true");
		/* Toggle MH<->TH display */
		c->showMag = ! c->showMag;
		break;

	case XK_less:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [<] '[AW] Hold Left Turn 3.0°/Second'");
			break;
		}
		add_ev_trace(u, "[<]V setting AW (disable AN and AL) for 3.0 left");
		fprintf(stderr, actfmt, curTime, symstr,
			"setting    AW (disable AN and AL) for 3.0 left");
		aps_an_disable(c);
		aps_al_disable(c);
		aps_aw_set(c, DEGtoRAD(-3.0));
		break;

	case XK_greater:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [>] '[AW] Hold Right Turn 3.0°/Second'");
			break;
		}
		add_ev_trace(u, "[>]V setting AW (disable AN and AL) for 3.0 right");
		fprintf(stderr, actfmt, curTime, symstr,
			"setting    AW (disable AN and AL) for 3.0 right");
		aps_an_disable(c);
		aps_al_disable(c);
		aps_aw_set(c, DEGtoRAD(3.0));
		break;

	case XK_comma:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [,] '[AW] Hold Left Turn 1.5°/Second'");
			break;
		}
		add_ev_trace(u, "[,]V setting AW (disable AN and AL) for 1.5 left");
		fprintf(stderr, actfmt, curTime, symstr,
			"setting    AW (disable AN and AL) for 1.5 left");
		aps_an_disable(c);
		aps_al_disable(c);
		aps_aw_set(c, DEGtoRAD(-1.5));
		break;

	case XK_period:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [.] '[AW] Hold Right Turn 1.5°/Second'");
			break;
		}
		add_ev_trace(u, "[.]V setting AW (disable AN and AL) for 1.5 right");
		fprintf(stderr, actfmt, curTime, symstr,
			"setting    AW (disable AN and AL) for 1.5 right");
		aps_an_disable(c);
		aps_al_disable(c);
		aps_aw_set(c, DEGtoRAD(1.5));
		break;

	case XK_bar:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [|] '[AW] Hold Wings Level {AutoWings)'");
			break;
		}
		add_ev_trace(u, "[|]V setting AW (disable AN and AL) for wings level");
		fprintf(stderr, actfmt, curTime, symstr,
			"setting    AW (disable AN and AL) for wings level");
		aps_an_disable(c);
		aps_al_disable(c);
		aps_aw_set(c, DEGtoRAD(0.0));
		break;

	case XK_slash:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [/] '[AW] [AN] [AL], Disable'");
			break;
		}
		add_ev_trace(u, "[/]V disabling AW and AN and AL");
		fprintf(stderr, actfmt, curTime, symstr,
			"disabing   AW and AN and AL");
		aps_an_disable(c);
		aps_al_disable(c);
		aps_aw_disable(c);
		break;

	case XK_Up:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [↑] 'Keystick, Forward Per Mode'");
			break;
		}
		add_ev_trace(u, "[Up]V calling getStick for ^");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    getStick for ^");
		getStick(u->c, u, 1);
		break;

	case XK_Down:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [↓] 'Keystick, Backward Per Mode'");
			break;
		}
		add_ev_trace(u, "[Down]V calling getStick for V");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    getStick for V");
		getStick(u->c, u, 2);
		break;

	case XK_Left:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [←] 'Keystick, Leftward Per Mode'");
			break;
		}
		add_ev_trace(u, "[Left]V calling getStick for <");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    getStick for <");
		getStick(u->c, u, 3);
		break;

	case XK_Right:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [→] 'Keystick, Rightward Per Mode'");
			break;
		}
		add_ev_trace(u, "[Right]V calling getStick for <");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    getStick for >");
		getStick(u->c, u, 4);
		break;

#ifdef sun
	case XK_R2:
#else
	case XK_Page_Up:
#endif
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [PageUp][R2] 'Keystick Sensitivity, Increase'");
			break;
		}
		add_ev_trace(u, "[Page_Up]V increasing keystick sensitivity");
		fprintf(stderr, actfmt, curTime, symstr,
			"increasing keystick sensitivity");
		if (u->stickM == KEYSTICK_LINEAR)
			u->stickS++;
		else	u->stickS = u->stickS << 1;
		if ( u->stickS > 128 )
			u->stickS = 128;
		break;

#ifdef sun
	case XK_R5:
#else
	case XK_Page_Down:
#endif
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [PageDown][R5] 'Keystick Sensitivity, Decrease'");
			break;
		}
		add_ev_trace(u, "[Page_Down]V decreasing keystick sensitivity");
		fprintf(stderr, actfmt, curTime, symstr,
			"decreasing keystick sensitivity");
		if (u->stickM == KEYSTICK_LINEAR)
			u->stickS--;
		else	u->stickS = u->stickS >> 1;
		if ( u->stickS < 1 )
			u->stickS = 1;
		break;

	case XK_End:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [End] 'Keystick Sensitivity Mode, Cycle'");
			break;
		}
		add_ev_trace(u, "[End]V cycling keystick mode");
		fprintf(stderr, actfmt, curTime, symstr,
			"cycling    keystick mode");
		if (u->stickM == KEYSTICK_LINEAR)
			u->stickM = KEYSTICK_DOUBLE;
		else	u->stickM = KEYSTICK_LINEAR;
		break;

	case XK_z:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [z] 'Rudder, (Node) Leftward'");
			break;
		}
		add_ev_trace(u, "[z]V increasing rudderComm (more left)");
		fprintf(stderr, actfmt, curTime, symstr,
			"increasing rudderComm (more left)");
		c->rudderComm = c->rudderComm + 0.01;
		if ( c->rudderComm > 1.0 )  c->rudderComm = 1.0;
		break;

	case XK_c:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [c] 'Rudder, (Node) Rightward'");
			break;
		}
		add_ev_trace(u, "[c]V decreasing rudderComm (more right)");
		fprintf(stderr, actfmt, curTime, symstr,
			"decreasing rudderComm (more right)");
		c->rudderComm = c->rudderComm - 0.01;
		if ( c->rudderComm < -1.0 )  c->rudderComm = -1.0;
		break;

	case XK_x:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [x] 'Rudder, Center'");
			break;
		}
		add_ev_trace(u, "[x]V zeroing rudderComm (center)");
		fprintf(stderr, actfmt, curTime, symstr,
			"zeroing    rudderComm (center)");
		c->rudderComm = 0.0;
		break;

#ifdef sun
	case XK_R7:
#else
	case XK_Home:
#endif
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [Home][R7] '[AS] and Trim, Reset'");
			break;
		}
		add_ev_trace(u, "[Home]V canceling all [AS] and trim");
		fprintf(stderr, actfmt, curTime, symstr,
			"cancling   all APS and trim settings");
		aps_off(c);
		c->SeTrim = 0.0;
		c->SaTrim = 0.0;
		break;

	case XK_j:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [j] 'Elevator, Trim To Current Pitch'");
			break;
		}
		add_ev_trace(u, "[j]V setting elevator trim to current stick");
		fprintf(stderr, actfmt, curTime, symstr,
			"setting    elevator trim to current stick");
		/* Adjust Elevator Trim */
		c->SeTrim -= c->pitchComm;
		break;

	case XK_y:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [y] 'Flaps, Retract One Notch'");
			break;
		}
		add_ev_trace(u, "[y]V calling flaps_up");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    flaps_up");
		flaps_up(c);
		break;

	case XK_h:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [h] 'Flaps, Extend One Notch'");
			break;
		}
		add_ev_trace(u, "[h]V calling flaps_down");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    flaps_down");
		flaps_down(c);
		break;

	case XK_w:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [w] 'Speed Brake, Retract One Notch'");
			break;
		}
		add_ev_trace(u, "[w]V calling flaps_speed_brakes_retract");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    flaps_speed_brakes_retract");
		flaps_speed_brakes_retract(c);
		break;

	case XK_s:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [s] 'Speed Brake, Extend One Notch'");
			break;
		}
		add_ev_trace(u, "[s]V calling flaps_speed_brakes_extend");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    flaps_speed_brakes_extend");
		flaps_speed_brakes_extend(c);
		break;

	case XK_t:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [t] 'Timer, Cycle {Start|Stop|Reset)'");
			break;
		}
		add_ev_trace(u, "[t]V cycling timer");
		fprintf(stderr, actfmt, curTime, symstr,
			"cycling    timer");
		if( u->hud_mode )
			hud_timer_toggle(u);
		else
			inst_timer_toggle(u);
		break;

	case XK_1:
/*	case XK_Next:	*/

		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [1] 'Throttle, Idle and Disable All [AS] Except [AC]'");
			break;
		}
		add_ev_trace(u, "[1]V setting idle throttle (disable all [AS] except [AC]");
		fprintf(stderr, actfmt, curTime, symstr,
			"setting    throttle to idle (disable all APS except AC)");
		aps_al_disable(c);
		aps_an_disable(c);
		aps_ap_disable(c);
		aps_at_disable(c);
		aps_aw_disable(c);

		c->throttleComm = 8192;
		break;

	case XK_2:
	case XK_KP_Subtract:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [2][KP-] 'Throttle or [AT] Airspeed, Decrease'");
			break;
		}
		add_ev_trace(u, "[2]V decrease throttle or [AT] target");
		fprintf(stderr, actfmt, curTime, symstr,
			"decreasing throttle or AT velocity");
		if ( aps_at_enabled(c) ) {
			aps_at_dec_velocity(c);
		}
		else {
			c->throttleComm -= 512;
			if ( c->throttleComm < 6553 )
				c->throttleComm = 6553;
		}
		break;

	case XK_3:
	case XK_KP_Add:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [3][KP+] 'Throttle or [AT] Airspeed, Increase'");
			break;
		}
		add_ev_trace(u, "[3]V increase throttle or [AT] target");
		fprintf(stderr, actfmt, curTime, symstr,
			"increasing throttle or AT velocity");
		if ( aps_at_enabled(c) ) {
			aps_at_inc_velocity(c);
		}
		else {
			c->throttleComm += 512;
			if ( c->throttleComm >= 32768 )
				c->throttleComm = 32768;
		}
		break;

	case XK_4:
/*	case XK_Prior:	*/

		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [4] 'Throttle, Full and Disable All [AS] Except [AC]'");
			break;
		}
		add_ev_trace(u, "[4]V setting full throttle (disable all [AS] except [AC]");
		fprintf(stderr, actfmt, curTime, symstr,
			"setting    throttle to full (disable all APS except AC)");
		aps_al_disable(c);
		aps_an_disable(c);
		aps_at_disable(c);
		aps_ap_disable(c);
		aps_aw_disable(c);

		c->throttleComm = 32768;
		break;
	
	case XK_exclam:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [!] 'Reverse Thrusters, Toggle'");
			break;
		}
		add_ev_trace(u, "[!]V calling pm_thrust_reverse_toggle");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    pm_thrust_reverse_toggle");
		pm_thrust_reverse_toggle(c);
		break;

	case XK_a:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [a] 'After-Burners, Toggle'");
			break;
		}
		add_ev_trace(u, "[a]V calling pm_after_burner_toggle");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    pm_after_burner_toggle");
		pm_after_burner_toggle(c);
		break;

	case XK_space:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [Space] 'NAVAIDs, Cycle'");
			break;
		}
		/* Switch NAVn/RNAVn receivers: */
		add_ev_trace(u, "[Space]V cycling NAVs (disables [AN] and [AL])");
		fprintf(stderr, actfmt, curTime, symstr,
			"cycling    NAVs (disables AN and AL)");
		aps_al_disable(c);
		aps_an_disable(c);
		hsi_switch_mode(u);
		break;

	case XK_7:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [7] 'OBS/ADF Heading, Decrease'");
			break;
		}
		add_ev_trace(u, "[7]V decreasing OBS/ADF heading (disables [AN])");
		fprintf(stderr, actfmt, curTime, symstr,
			"decreasing OBS/ADF heading (disables AN)");
		aps_an_disable(c);
		hsi_obs_inc(u, -1);
		adf_hdg_inc(u, -1);
		break;

	case XK_8:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [8] 'OBS/ADF Heading, Increase'");
			break;
		}
		add_ev_trace(u, "[8]V increasing OBS/ADF heading (disables [AN])");
		fprintf(stderr, actfmt, curTime, symstr,
			"increasing OBS/ADF heading (disables AN)");
		aps_an_disable(c);
		hsi_obs_inc(u, +1);
		adf_hdg_inc(u, +1);
		break;

	case XK_9:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [9] 'HSI/ADF Frequency, Decrease'");
			break;
		}
		add_ev_trace(u, "[9]V decreasing HSI/ADF freq (disables [AN] and [AL])");
		fprintf(stderr, actfmt, curTime, symstr,
			"decreasing HSI/ADF frequency (disables AN and AL)");
		aps_al_disable(c);
		aps_an_disable(c);
		hsi_frq_inc(u, -1);
		adf_frq_inc(u, -1);
		break;

	case XK_0:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [0] 'HSI/ADF Frequency, Increase'");
			break;
		}
		add_ev_trace(u, "[0]V increasing HSI/ADF freq (disables [AN] and [AL])");
		fprintf(stderr, actfmt, curTime, symstr,
			"increasing HSI/ADF frequency (disables AN and AL)");
		aps_al_disable(c);
		aps_an_disable(c);
		hsi_frq_inc(u, +1);
		adf_frq_inc(u, +1);
		break;

	case XK_F3:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [F3] 'RNAV Radial, Decrease'");
			break;
		}
		add_ev_trace(u, "[F3]V decreasing RNAV radial");
		fprintf(stderr, actfmt, curTime, symstr,
			"decreasing RNAV radial");
		hsi_theta_inc(u, -1);
		break;

	case XK_F4:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [F4] 'RNAV Radial, Increase'");
			break;
		}
		add_ev_trace(u, "[F4]V increasing RNAV radial");
		fprintf(stderr, actfmt, curTime, symstr,
			"increasing RNAV radial");
		hsi_theta_inc(u, +1);
		break;

	case XK_F5:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [F5] 'RNAV Distance, Decrease'");
			break;
		}
		add_ev_trace(u, "[F5]V decreasing RNAV distance");
		fprintf(stderr, actfmt, curTime, symstr,
			"decreasing RNAV distance");
		hsi_rho_inc(u, -1);
		break;

	case XK_F6:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [F6] 'RNAV Distance, Increase'");
			break;
		}
		add_ev_trace(u, "[F6]V increasing RNAV distance");
		fprintf(stderr, actfmt, curTime, symstr,
			"increasing RNAV distance");
		hsi_rho_inc(u, +1);
		break;

	case XK_b:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [b] 'Brakes, Toggle'");
			break;
		}
		add_ev_trace(u, "[b]V toggling brakes");
		fprintf(stderr, actfmt, curTime, symstr,
			"toggling   brakes");
		c->flags ^= FL_BRAKES;
		break;

	case XK_g:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [g] 'Landing Gear, Toggle'");
			break;
		}
		add_ev_trace(u, "[g]V calling gear_handle_toggle");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    gear_handle_toggle");
		gear_handle_toggle(c);
		break;

	case XK_l:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [l] 'Drone, Launch'");
			break;
		}
		add_ev_trace(u, "[l]V calling drone_new");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    drone_new");
		drone_new(c, NULL);
		break;
		
/* Display next hint now with array wrapping */

	case XK_KP_1:
	case XK_F:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [F][KP1] 'Hint, Display Next'");
			break;
		}
		add_ev_trace(u, "[F]V calling prompt_viewer_print_hint for next");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    prompt_viewer_print_hint for next");
		prompt_viewer_print_hint(u, 1);
		break;

/* Display previous hint now with array wrapping */

	case XK_KP_7:
	case XK_J:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [J][KP7] 'Hint, Display Previous'");
			break;
		}
		add_ev_trace(u, "[J]V calling prompt_viewer_print_hint for prev");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    prompt_viewer_print_hint for previous");
		prompt_viewer_print_hint(u, 2);
		break;

/* Display first hint now */

	case XK_KP_9:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [O][KP9] 'Hint, Display First'");
			break;
		}
	case XK_O:
		add_ev_trace(u, "[O]V calling prompt_viewer_print_hint for 1st");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    prompt_viewer_print_hint for first");
		prompt_viewer_print_hint(u, 3);
		break;

/* Display current hint now */

	case XK_KP_3:
	case XK_E:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [E][KP3] 'Hint, Display Current'");
			break;
		}
		add_ev_trace(u, "[E]V calling prompt_viewer_print_hint for reshow");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    prompt_viewer_print_hint for redisplay");
		prompt_viewer_print_hint(u, 0);
		break;

/* look forward */

	case XK_KP_8:
	case XK_u:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [u][KP8] 'Screen, View Forward'");
			break;
		}
		add_ev_trace(u, "[u]V calling setOutsideView for forward");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    setOutsideView for forward");
		setOutsideView(c, u, VIEW_FORWARD);
		prompt_viewer_print(u, "Front view");
		break;

/* look right */

	case XK_KP_6:
	case XK_p:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [p][KP6] 'Screen, View Rightward'");
			break;
		}
		add_ev_trace(u, "[p]V calling setOutsideView for rightward");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    setOutsideView for right");
		setOutsideView(c, u, VIEW_RIGHT);
		prompt_viewer_print(u, "Right view");
		break;

/* look left */

	case XK_KP_4:
	case XK_o:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [o][KP4] 'Screen, View Leftward'");
			break;
		}
		add_ev_trace(u, "[o]V calling setOutsideView for leftward");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    setOutsideView for left");
		setOutsideView(c, u, VIEW_LEFT);
		prompt_viewer_print(u, "Left view");
		break;

/* look back */

	case XK_KP_2:
	case XK_k:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [k][KP2] 'Screen, View Backward'");
			break;
		}
		add_ev_trace(u, "[k]V calling setOutsideView for backward");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    setOutsideView for aft");
		setOutsideView(c, u, VIEW_AFT);
		prompt_viewer_print(u, "Rear view");
		break;

/* look up */

#ifdef sun
	case XK_R11:
#else
	case XK_KP_5:
#endif
	case XK_e:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [e][KP5][R11] 'Screen, View Upward'");
			break;
		}
		add_ev_trace(u, "[e]V calling setOutsideView for upward");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    setOutsideView for up");
		setOutsideView(c, u, VIEW_UP);
		prompt_viewer_print(u, "Up view");
		break;

/* look down */

	case XK_KP_0:
	case XK_f:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [f][KP0] 'Screen, View Downward'");
			break;
		}
		add_ev_trace(u, "[f]V calling setOutsideView for downward");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    setOutsideView for down");
		setOutsideView(c, u, VIEW_DOWN);
		prompt_viewer_print(u, "Down view");
		break;

	case XK_n:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [n] 'Screen, View Chase/Forward, Toggle'");
			break;
		}
		add_ev_trace(u, "[n]V toggling forward/chase view");
		fprintf(stderr, actfmt, curTime, symstr,
			"toggling   chase/forward view");
		if( c->flags & FL_CHASE_VIEW ){
			setOutsideView(c, u, VIEW_FORWARD);
			prompt_viewer_print(u, "Front view");

		} else if( c->cinfo->object == NULL ){
			prompt_viewer_print(u, "Sorry, no object model available for this plane.");

		} else {
			setOutsideView(c, u, VIEW_CHASE);
			prompt_viewer_print(u, "Chase view");

		}
		break;

	case XK_q:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [q] 'Radar Targets, Cycle'");
			break;
		}
		add_ev_trace(u, "[q]V calling newRadarTarget");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    newRadarTarget");
		c->curRadarTarget = newRadarTarget(c);
#ifdef HAVE_DIS
		dis_radarTargetChanged(c);
#endif
		break;

	case XK_D:
		/* Debug */
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [D] 'Debug (GPS) Display, Toggle'");
			break;
		}
		add_ev_trace(u, "[D]V calling FSPageToggle (debug display)");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    FSPageToggle");
		FSPageToggle();
		break;

	case XK_r:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [r] 'Radar Without DIS Beam Change, Cycle'");
			break;
		}
		add_ev_trace(u, "[r]V cycling radar w/o beam (disables [AN] and [AL]");
		fprintf(stderr, actfmt, curTime, symstr,
			"cycling    radar without beam change (disables AN and AL)");
		aps_al_disable(c);
		aps_an_disable(c);

		switch (c->radarMode) {

		case RM_OFF:
		case RM_STANDBY:
			if( u->c->cinfo->radarOutput > 0.0 ){
				/* we really have a radar -- enable it */
				setRadarMode(c, RM_NORMAL);
				enableRadar(c, RM_NORMAL);
			} else {
				/* no radar -- HSI mode */
				disableRadar(c);
				setRadarMode(c, RM_HSI);
				hsi_enable(u);
			}
			break;

		case RM_NORMAL:
		case RM_ACM:
		case RM_STT:
			disableRadar(c);
			setRadarMode(c, RM_HSI);
			hsi_enable(u);
			break;

		case RM_HSI:
			hsi_disable(u);
			setRadarMode(c, RM_ADF);
			adf_enable(u);
			break;

		case RM_ADF:
			adf_disable(u);
			setRadarMode(c, RM_OFF);
			break;

		default:
			adf_disable(u);
			disableRadar(c);
			setRadarMode(c, RM_OFF);
			hsi_enable(u);
			break;
		}

		break;

#ifdef HAVE_DIS
/* this is a hack to allow change of beam parameter index
in the outgoing electromagnetic emission pdu's */
	case XK_R:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [R] 'Radar With DIS Beam Change, Cycle'");
			break;
		}
		add_ev_trace(u, "[R]V cycling radar with beam (disables [AN] and [AL]");
		fprintf(stderr, actfmt, curTime, symstr,
			"cycling    radar with DIS beam change (disables AN and AL)");
		aps_al_disable(c);
		aps_an_disable(c);
		switch (c->radarMode) {
		case RM_OFF:
		case RM_STANDBY:
			setRadarMode(c, RM_NORMAL);
			enableRadar(c, RM_NORMAL);
			break;
		case RM_NORMAL:
			disableRadar(c);
			setRadarMode(c, RM_ACM);
			enableRadar(c, RM_ACM);
			break;
		case RM_ACM:
			disableRadar(c);
			setRadarMode(c, RM_STT);
			enableRadar(c, RM_STT);
			break;
		case RM_STT:
			disableRadar(c);
			setRadarMode(c, RM_HSI);
			hsi_enable(u);
			break;
		case RM_HSI:
			disableRadar(c);
			hsi_disable(u);
			setRadarMode(c, RM_OFF);
			break;
		default:
			abort();
		}
		break;
#endif

#ifdef FIXME_INCOMPLETE_OR_BUGGED

	case XK_i:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [i] 'Flight Data File, Save'");
			break;
		}
		add_ev_trace(u, "[i]V saving FLIGHT_STATE_FILE");
		fprintf(stderr, actfmt, curTime, symstr,
			"saving     FLIGHT_STATE_FILE");
		fp = fopen(FLIGHT_STATE_FILE, "w");
		if ( fp == NULL ) {
			char s[1000];

			sprintf("Error opening the file " FLIGHT_STATE_FILE
				" for writing: %s", strerror(errno));
			prompt_craft_print(c, s);
		}
		else {
			fwrite((char *) &ptbl[c->pIndex],
				sizeof(craft), 1, fp);
			fclose(fp);
			prompt_craft_print(c, "Flight state saved in file "
				FLIGHT_STATE_FILE);
		}
		break;

	case XK_I:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [I] 'Flight Data File, Restore'");
			break;
		}
		add_ev_trace(u, "[I]V restoring FLIGHT_STATE_FILE");
		fprintf(stderr, actfmt, curTime, symstr,
			"restoring  FLIGHT_STATE_FILE");
		fp = fopen(FLIGHT_STATE_FILE, "r");
		if ( fp == NULL ) {
			char s[1000];

			sprintf("Error opening the file " FLIGHT_STATE_FILE
				" for reading: %s", strerror(errno));
			prompt_craft_print(c, s);
		}
		else {
			fread((char *) &pentry, sizeof(craft), 1, fp);
			fclose(fp);

			c->flags = pentry.flags;
			c->Sg = pentry.Sg;
			c->prevSg = pentry.prevSg;
			c->Cg = pentry.Cg;
			c->trihedral = pentry.trihedral;
			c->curHeading = pentry.curHeading;
			c->curPitch = pentry.curPitch;
			c->curRoll = pentry.curRoll;
			c->p = pentry.p;
			c->q = pentry.q;
			c->r = pentry.r;
			c->damageBits = pentry.damageBits;
			c->structurePts = pentry.structurePts;
			c->damageCL = pentry.damageCL;
			c->damageCM = pentry.damageCM;
			c->curThrust = pentry.curThrust;
			c->throttle = pentry.throttle;

			/* FIXME: there are missing fields */

			prompt_craft_print(c, "Flight state restored from file "
				FLIGHT_STATE_FILE);
		}
		break;

	case XK_semicolon:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [;] 'Debug Mode, Toggle'");
			break;
		}
		add_ev_trace(u, "[;]V toggling debug");
		fprintf(stderr, actfmt, curTime, symstr,
			"toggling   debug");
		debug ^= 1;
		break;
#endif /* FIXME_INCOMPLETE_OR_BUGGED */

	//case XK_Escape:
	case XK_Q:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [Q] 'ACM, Exit (Self-Destruct)'");
			break;
		}
		if ( pressed_XK_Q_time + 2.0 >= curTime ){
			add_ev_trace(u, "[Q]V calling killPlayer(NULL)");
			fprintf(stderr, actfmt, curTime, symstr,
				"executing  self-destruction");
			killPlayer(c, NULL);
			return TRUE;
		}
		else {
			add_ev_trace(u, "[Q]V arming killPlayer(NULL)");
			fprintf(stderr, actfmt, curTime, symstr,
				"arming     self-destruction");
			pressed_XK_Q_time = curTime;
			prompt_viewer_print(u, "Press SHIFT-Q again within 2 sec. to exit the program");
		}
		break;

	case XK_P:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [P] 'Screen, Dump Using Xwd'");
			break;
		}
		add_ev_trace(u, "[P]V calling system(xwd) for window dump");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    system(xwd) for screen dump");
		system("xwd -name " ACM " -out /tmp/acm-dump-`date +%s`.xwd");
		prompt_viewer_print(u, "Screen dumped into /tmp/acm-dump-*.xwd");
		break;

#ifdef FIXME_INCOMPLETE_OR_BUGGED

	case XK_braceleft:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [{] 'Black Box, Start Recording'");
			break;
		}
		add_ev_trace(u, "[{]V calling startBlackBoxRecording");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    startBlackBoxRecording");
		startBlackBoxRecording();
		prompt_viewer_print(u, "Black box recording started");
		break;

	case XK_braceright:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [}] 'Black Box, Stop Recording'");
			break;
		}
		add_ev_trace(u, "[}]V calling endBlackBoxRecording");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    endBlackBoxRecording");
		endBlackBoxRecording();
		prompt_viewer_print(u, "Black box recording stopped");
		break;

	case XK_bracketleft:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [[] 'Black Box, Replay Recording'");
			break;
		}
		add_ev_trace(u, "[[]V calling startBlackBoxPlayback");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    startBlackBoxPlayback");
		startBlackBoxPlayback();
		prompt_viewer_print(u, "Black box playback started");
		break;

#endif

	case XK_F7:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [F7] 'Altimeter Correction, Decrease'");
			break;
		}
		add_ev_trace(u, "[F7]V calling inst_altimeter_correction(-1)");
		fprintf(stderr, actfmt, curTime, symstr,
			"decreasing altimeter correction");
		inst_altimeter_correction(u, -1);
		break;

	case XK_F8:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [F8] 'Altimeter Correction, Increase'");
			break;
		}
		add_ev_trace(u, "[F8]V calling inst_altimeter_correction(+1)");
		fprintf(stderr, actfmt, curTime, symstr,
			"increasing altimeter correction");
		inst_altimeter_correction(u, +1);
		break;

	case XK_F9:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [F9] 'Artificial Horizon, Cage (Reset)'");
			break;
		}
		add_ev_trace(u, "[F9]V calling inst_altitude_reset; i.e., cage");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    inst_attitude_reset");
		inst_attitude_reset(u);
		break;

	case XK_F11:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [F11] 'Artificial Horizon, Decrease Pitch'");
			break;
		}
		add_ev_trace(u, "[F11]V calling inst_altitude_adjust_pitch for down");
		fprintf(stderr, actfmt, curTime, symstr,
			"decreasing attitude pitch");
		inst_attitude_adjust_pitch(u, -DEGtoRAD(0.5));
		break;

	case XK_F12:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [F12] 'Artificial Horizon, Increase Pitch'");
			break;
		}
		add_ev_trace(u, "[F12]V calling inst_altitude_adjust_pitch for up");
		fprintf(stderr, actfmt, curTime, symstr,
			"increasing attitude pitch");
		inst_attitude_adjust_pitch(u, DEGtoRAD(0.5));
		break;

	case XK_plus:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [+] 'Screen, Zoom In One Notch'");
			break;
		}
		add_ev_trace(u, "[+]V calling windows_zoom_in");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    windows_zoom_in");
		windows_zoom_in(u);
		break;

	case XK_minus:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [-] 'Screen, Zoom Out One Notch'");
			break;
		}
		add_ev_trace(u, "[-]V calling windows_zoom_out");
		fprintf(stderr, actfmt, curTime, symstr,
			"calling    windows_zoom_out");
		windows_zoom_out(u);
		break;

	case XK_dollar:
		if( u->testKeys ) {
			prompt_viewer_print(u, "testKeys: [$] 'Drone Control Mode, Toggle'");
			break;
		}
		add_ev_trace(u, "[$]V calling drone_take_commands");
		fprintf(stderr, actfmt, curTime, symstr,
			"toggling   drone commands mode");
		if( c->type == CT_PLANE )
			drone_take_commands(c);
		else
			drone_release_commands(c);
		break;

/* Handle without doing anything */

	case XK_Caps_Lock:
	case XK_Shift_Lock:
	case XK_Num_Lock:
	case XK_Shift_L:
	case XK_Shift_R:
	case XK_Control_L:
	case XK_Control_R:
	case XK_Alt_L:
	case XK_Alt_R:
	case XK_Meta_L:
	case XK_Meta_R:
	case XK_Super_L:
	case XK_Super_R:
	case XK_Hyper_L:
	case XK_Hyper_R:
		add_ev_trace(u, " V ignoring the keypress silently");
		fprintf(stderr, actfmt, curTime, symstr,
			"ignoring   silently");
		break;

/* Currently unassigned in events.c */

/*	case XK_5:	*/
/*	case XK_6:	*/
/*	case XK_E:	*/
/*	case XK_F:	*/
/*	case XK_J:	*/
/*	case XK_O:	*/
	case XK_S:
	case XK_W:
	case XK_Y:
/*	case XK_e:	*/
/*	case XK_f:	*/
/*	case XK_k:	*/
/*	case XK_o:	*/
/*	case XK_p:	*/
/*	case XK_u:	*/
	case XK_KP_Begin:
	case XK_KP_Decimal:
	case XK_KP_Delete:
	case XK_KP_Divide:
	case XK_KP_Down:
	case XK_KP_End:
	case XK_KP_Enter:
	case XK_KP_Home:
	case XK_KP_Insert:
	case XK_KP_Left:
	case XK_KP_Multiply:
	case XK_KP_Next:		/* used by browse */
	case XK_KP_Prior:		/* used by browse */
	case XK_KP_Right:
	case XK_KP_Up:
/*	case XK_F2:	*/
	case XK_F10:
	case XK_BackSpace:
	case XK_Delete:
	case XK_Escape:
	case XK_Insert:
	case XK_Menu:
	case XK_Pause:
	case XK_Print:
	case XK_Return:
	case XK_Scroll_Lock:
	case XK_ampersand:
	case XK_apostrophe:
/*	case XK_asciicircum:	*/
	case XK_asciitilde:
	case XK_asterisk:
	case XK_at:
	case XK_backslash:
	case XK_bracketright:
	case XK_colon:			/* used by browse */
	case XK_equal:
	case XK_grave:
	case XK_numbersign:
/*	case XK_percent:	*/
	case XK_question:
	case XK_quotedbl:		/* used by browse */
	case XK_underscore:
	default:

/* Unhandled input (probably typo) -- warn pilot */

		add_ev_trace(u, " V reporting probable typo to pilot");
		fprintf(stderr, actfmt, curTime, symstr,
			"warning    pilot about probable typo\a");
		sprintf(viewermsg, "%16s is not a supported input!", symstr);
		prompt_viewer_print(u, viewermsg);
		break;
	}

	return FALSE;
}


static _BOOL
doWindowEvent(viewer * u)
/*
	Return TRUE on:
	
	- close request of the window: in this case the viewer u gets closed
	  and released
	
	- player gets killed or terminated: in this case the viewer u and all
	  the other viewers attached to the craft c gets closed and the craft c
	  gets released
*/
{
	XEvent    ev;
	craft    *c;

	updateSystemClock();
	fprintf(stderr, "%s\n", "=> doWindowEvent");
	c = u->c;

/*	updateSystemClock();	*/
/*	fprintf(stderr, "%s\n", "into  XCheckWindowEvent");	*/
	if (XCheckWindowEvent(u->dpy, u->win, -1L, &ev)) {
/*		updateSystemClock();	*/
/*		fprintf(stderr, "%s\n", "outof XCheckWindowEvent--true");	*/

		switch (ev.type) {

		case MappingNotify:
			XRefreshKeyboardMapping(&ev.xmapping);
			break;

		case KeyPress:
			if (u->viewer_state == ViewerStateBrowsing ||
				u->viewer_state == ViewerStatePiggyback) {
				if (doBrowseKeyEvent( c, u, &ev, 1)) {
					return FALSE;
				}
			}
			else {
				if ( doKeyEvent(c, u, (XKeyEvent *) &ev) ){
					return TRUE;
				}
			}
			break;

		case KeyRelease:
			if ( doKeyRelease(c, u, (XKeyEvent *) &ev) ){
				return TRUE;
			}
			break;

		case ButtonPress:
			if (doButtonEvent(c, u, &ev))
				return TRUE;
			break;

		case ButtonRelease:
			if (doButtonReleaseEvent(c, u, &ev, 1))
				return TRUE;
			break;

		case ConfigureNotify:
			windows_set_layout(c, u,
				ev.xconfigure.width,
				ev.xconfigure.height, u->hud_mode);
			break;

		case FocusIn:
			u->hasFocus = TRUE;
			break;

		case FocusOut:
			u->hasFocus = FALSE;
			u->hasComm = FALSE;
			break;

		case Expose:
			ForceWindowRedraw(u->w);
			break;

		default:
			break;
		}

	} else {
/*		updateSystemClock();	*/
/*		fprintf(stderr, "%s\n", "outof XCheckWindowEvent--false");	*/
/*		fprintf(stderr, "%s\n", "into  XCheckTypedEvent--ClientMessage");	*/
		if (XCheckTypedEvent(u->dpy, ClientMessage, &ev)) {
			updateSystemClock();
			fprintf(stderr, "%s\n", "outof XCheckTypedEvent--ClientMessage--true");

			if (ev.xclient.message_type == u->protocolsAtom
			&& (ev.xclient.data.l[0] == u->deleteWindowAtom
			|| ev.xclient.data.l[0] == u->closedownAtom)) {
/*				updateSystemClock();	*/
/*				fprintf(stderr, "%s\n", "into  killPlayer");	*/
				killPlayer(c, NULL);
/*				updateSystemClock();	*/
/*				fprintf(stderr, "%s\n", "outof killPlayer");	*/
				return TRUE;
			}
		} else {
/*			updateSystemClock();	*/
/*			fprintf(stderr, "%s\n", "outof XCheckTypedEvent--ClientMessage--false");	*/
/*			fprintf(stderr, "%s\n", "into  XCheckTypedEvent--MappingNotify");	*/
			if (XCheckTypedEvent(u->dpy, MappingNotify, &ev)) {
				updateSystemClock();
				fprintf(stderr, "%s\n", "outof XCheckTypedEvent--MappingNotify--true");
/*				fprintf(stderr, "%s\n", "into  XRefreshKeyboardMapping");	*/
				XRefreshKeyboardMapping(&ev.xmapping);
				updateSystemClock();
				fprintf(stderr, "%s\n", "outof XRefreshKeyboardMapping");
				return TRUE;
			} else {
/*				updateSystemClock();	*/
/*				fprintf(stderr, "%s\n", "outof XCheckTypedEvent--MappingNotify--false");	*/
/*				fprintf(stderr, "%s\n", "into  XCheckTypedEvent--SelectionClear");	*/
				if (XCheckTypedEvent(u->dpy, SelectionClear, &ev)) {
					updateSystemClock();
					fprintf(stderr, "%s\n", "outof XCheckTypedEvent--SelectionClear--true");
				} else {
/*					updateSystemClock();	*/
/*					fprintf(stderr, "%s\n", "outof XCheckTypedEvent--SelectionClear--false");	*/
/*					fprintf(stderr, "%s\n", "into  XCheckTypedEvent--SelectionNotify");	*/
					if (XCheckTypedEvent(u->dpy, SelectionNotify, &ev)) {
						updateSystemClock();
						fprintf(stderr, "%s\n", "outof XCheckTypedEvent--SelectionNotify--true");
					} else {
/*						updateSystemClock();	*/
/*						fprintf(stderr, "%s\n", "outof XCheckTypedEvent--SelectionNotify--false");	*/
/*						fprintf(stderr, "%s\n", "into  XCheckTypedEvent--SelectionRequest");	*/
						if (XCheckTypedEvent(u->dpy, SelectionRequest, &ev)) {
							updateSystemClock();
							fprintf(stderr, "%s\n", "outof XCheckTypedEvent--SelectionRequest--true");
						} else {
							updateSystemClock();
							fprintf(stderr, "%s\n", "outof XCheckTypedEvent--SelectionRequest--false");
						}
					}
				}
			}
		}
	}

/*	updateSystemClock();	*/
/*	fprintf(stderr, "%s\n", "<= doWindowEvent returning false (no events in queue)");	*/
	return FALSE;  /* done with all pending events for this viewer */

}


void
events_window_keyb_buttons(void)
/*
	Process events for each viewer
*/
{
	viewer   *u;
	_BOOL     got_evt;

#ifdef FIXME
	int n;

	n = 0;
	u = vl_head;
	while( u != NULL ){
		n++;
		u = u->vl_next;
	}
	printf("FIXME: trovati %d viewers\n", n);
#endif

	do {

		/*
			Process one event for each viewer
		*/

		got_evt = FALSE;

		u = vl_head;

		while( u != NULL ){

			if (real_delta_t || ((cur % redraw_interval) == 0)) {
				getStick(u->c, u, 0);
			}

			if( XEventsQueued(u->dpy, QueuedAlready) ){

				got_evt = TRUE;

				if( doWindowEvent(u) ){

					/*
					Viewer or craft terminated. We can't
					continue scanning the linked list of
					viewers because we don't know how many
					viewers were attached to this craft, so
					we have to restart from the first
					viewer:
					*/

					u = vl_head;

				} else {

					u = u->vl_next;
				}

			} else {

				u = u->vl_next;
			}
		}
	
	} while( got_evt );
}


static int last_switches = 0;


int
events_joystick(craft * c, viewer * u, double throttle, int switches)
{

	int       switch_xor = switches ^ last_switches;

/*
 *  Change in state of any buttons ?
 */

	if (switch_xor != 0) {

		if (switch_xor & 1) {

/*
 *  Button 1 press
 */
			if (switches & 1) {
				fireWeapon(c);
			}

/*
 *  Button 1 release
 */

			else {
				ceaseFireWeapon(c);
			}
		}

/*
 * Button 2 press
 */

		if (switch_xor & 2) {
			if (switches & 2) {
				selectWeapon(c);
			}
		}

	}

	last_switches = switches;

	return 0;
}

