/*
 *  acm : an aerial combat simulator for X
 *  Copyright (C) 1991-1998  Riley Rainey
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 dated June, 1991.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program;  if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
 */

/*
 *  2014-12-29 dlc - Refer to CHANGES text file for details of modifications
 */
 
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "aps.h"
#include "draw.h"
#define HUD_IMPORT
#include "hud.h"
#include "damage.h"
#include "m61a1.h"
#include "navaid.h"
#include "terrain.h"
#include "vpath.h"
#include "weapon.h"
#include "mouse.h"

#define DESIGNATOR_SIZE     20

#define HUD(u)     ((hud_data *)u->hud)

static int blink = 0;
static double blink_toggle_time = 0.0;


typedef struct _hud_data {
	struct _hud_data *next;

	_BOOL  enabled;

	int timer_op;        /* timer operation: 0=do not display, 1=run, 2=stop */
	int timer_freezed;   /* store the time when timer_op is 2 */
	double timer_offset; /* store the offset time when timer_op is 1 or 2 */
} hud_data;


static hud_data *free_list = NULL;

static VPath
	* hud_vpath = NULL,
	* fpm_vpath = NULL;


void hud_enable(viewer *u)
{
	hud_data *hud;

	if( u->hud == NULL ){
		if( free_list == NULL ){
			hud = MALLOC( sizeof(hud_data) );
		} else {
			hud = free_list;
			free_list = hud->next;
			hud->next = NULL;
		}
		hud->next = NULL;
		hud->timer_op = 0;
		u->hud = hud;
	}
	HUD(u)->enabled = TRUE;
}


void hud_disable(viewer *u)
{
	if( u->hud == NULL )
		return;
	HUD(u)->enabled = FALSE;
}



void
hud_timer_toggle(viewer *u)
/*
	Actually, it cycles between the states timer_op 0, 1 and 2.
*/
{
	if( u->hud == NULL || !HUD(u)->enabled )
		return;

	HUD(u)->timer_op = (HUD(u)->timer_op + 1) % 3;

	if ( HUD(u)->timer_op == 1 ) {
		HUD(u)->timer_offset = curTime;

	} else if ( HUD(u)->timer_op == 2 ) {
		HUD(u)->timer_freezed = (int) (curTime - HUD(u)->timer_offset + 0.5);

	}
}


static void
doBankIndicator(craft * c, viewer * u)
{
	double xscale, yscale, r, r2, r3, r4, cosa, sina, bank,
		x1, y1, x2, y2, x3, y3;
	int x0, y0, a, blink_bank;
	draw_data *dd;

	xscale = u->xscaleFactor;
	yscale = u->yscaleFactor;
	r = 200.0 * xscale; /* tick mark start radius */
	r2 = 1.05*r; /* tick mark end radious */
	r3 = 0.99*r;
	r4 = 0.94*r;
	x0 = u->v->focus.x;
	y0 = u->v->focus.y + (int) (-150.0*yscale + r);
	dd = draw_new();
	VSetClipRect(u->v, &u->v->rect);

	for( a = -30; a <= 30; a += 10 ){
		cosa = cos( DEGtoRAD(a) );
		sina = sin( DEGtoRAD(a) );
		draw_segment(dd, x0 + sina*r, y0 - cosa*r, x0 + sina*r2, y0 - cosa*r2);
	}

	bank = c->curRoll;
	blink_bank = 0;
	if( bank < DEGtoRAD(-40) ){
		bank = DEGtoRAD(-40);
		blink_bank = blink;
	} else if( bank > DEGtoRAD(+40) ){
		bank = DEGtoRAD(+40);
		blink_bank = blink;
	}
	if( ! blink_bank ){
		x1 = x0 + sin(bank)*r3;
		y1 = y0 - cos(bank)*r3;
		x2 = x0 + sin(bank + DEGtoRAD(2))*r4;
		y2 = y0 - cos(bank + DEGtoRAD(2))*r4;
		x3 = x0 + sin(bank - DEGtoRAD(2))*r4;
		y3 = y0 - cos(bank - DEGtoRAD(2))*r4;
		draw_segment(dd, x1, y1, x2, y2);
		draw_segment(dd, x1, y1, x3, y3);
		draw_segment(dd, x2, y2, x3, y3);
	}

	draw_stroke(dd, u->v, u->v->w->pixel[HUDPixel]);
	draw_free(dd);
}


#define TurnAndSlipIndicatorWidth 100


static void
doTurnAndSlipIndicator(craft * c, viewer * u)
/*
	Displays the rate of turn (DEG/s), the lateral acceleration (g)
	and a timer (s).
*/
{
	double xscale, yscale, w;
	int tx, ty, t, blink_turn, blink_w;
	int w_tick;
	int w_t;
	double a;
	draw_data *dd;
	int h, m, s;
	char buffer[20];
	Color color;

	xscale = u->xscaleFactor;
	yscale = u->yscaleFactor;
	w_t = (int) (xscale * TurnAndSlipIndicatorWidth / 2.0);
	w_tick = (int) (xscale * TurnAndSlipIndicatorWidth / 4.0);
 	tx = u->v->focus.x - (int) (xscale * 160);
 	ty = u->v->focus.y + (int) (yscale * 200);
	dd = draw_new();
	VSetClipRect(u->v, &u->v->rect);
	color = u->v->w->pixel[HUDColor->cIndex];

 	/*
 		Draw scale:
 	*/
	w = 6 * yscale;
	draw_segment(dd, tx-w_t,    ty,  tx+w_t, ty);
	draw_segment(dd, tx-w_t,    ty,  tx-w_t, ty-w);     /* -3.0 DEG/s */
	draw_segment(dd, tx-w_tick, ty,  tx-w_tick, ty-w);  /* -1.5 DEG/s */
	draw_segment(dd, tx,        ty,  tx,     ty-w);     /*  0   DEG/s */
	draw_segment(dd, tx+w_tick, ty,  tx+w_tick, ty-w);  /* +1.5 DEG/s */
	draw_segment(dd, tx+w_t,    ty,  tx+w_t, ty-w);     /* +3.0 DEG/s */

	/*
		Draw turn rate needle:
	*/
	a = RADtoDEG( (c->q*sin(c->curRoll) + c->r*cos(c->curRoll))
		/ cos(c->curPitch) );
 	blink_turn = 0;
 	if( a > 3.5 ){
 		a = 3.5;
 		blink_turn = blink;
 	} else if( a < -3.5 ){
 		a = -3.5;
 		blink_turn = blink;
 	}
 	if( ! blink_turn ){
		int x0 = tx + a / 3.0 *  w_t;
		int y0 = ty - 4 * yscale;
		w = 4 * xscale;
		draw_segment(dd, x0 - w, y0, x0 - w - w, y0 - w);
		draw_segment(dd, x0 - w - w, y0 - w, x0 + w + w, y0 - w);
		draw_segment(dd, x0 + w + w, y0 - w, x0 + w, y0);
	}

	/*
		Draw lateral accel. needle:
	*/
	a = c->G.y;
	blink_turn = 0;
	if( a < -0.5 ){
		a = -0.5;
		blink_turn = blink;
	} else if( a > 0.5 ){
		a = 0.5;
		blink_turn = blink;
	}
	if( ! blink_turn ){
		t = tx - (int) rint(2.0*a*w_t);
		draw_segment(dd, t, ty+2*yscale, t, ty+9*yscale);
	}
	if( aps_ac_enabled(c)
	&& ( ! aps_ac_warn(c) || ! blink ) ){
		VDrawStrokeString(u->v,
			tx - w_t, ty + (int) (yscale * 12),
			"AC", 2,
			xscale * 8.0, color);
	}

	draw_stroke(dd, u->v, u->v->w->pixel[HUDPixel]);
	draw_free(dd);

	/*
		Draw AutoTurn status:
	*/
	if ( aps_aw_enabled(c)
	&& ( ! aps_aw_warn(c) || ! blink ) ){

		w = aps_aw_get(c);

		blink_w = FALSE;

		if( w < DEGtoRAD(-3.1) ){
			w = DEGtoRAD(-3.0);
			blink_w = blink;
		} else if( w > DEGtoRAD(3.1) ){
			w = DEGtoRAD(3.0);
			blink_w = blink;
		}

		if ( ! blink_w )
			VDrawStrokeString(u->v,
				tx + (int) (w * w_t / DEGtoRAD(3.0) - xscale * 6.0 + 0.5),
				ty + (int) (yscale * 4.0 + 0.5),
				"o", 1, xscale * 12.0, color);
	}

	/*
		Draw timer:
	*/

	if ( HUD(u)->timer_op == 0 ) {
		return;

	} else if ( HUD(u)->timer_op == 1 ) {
		t = (int) (curTime - HUD(u)->timer_offset);

	} else {
		t = HUD(u)->timer_freezed;
	}

	if (t >= 3600) {
		h = t / 3600;
		t = t % 3600;
		m = t / 60;
		s = t % 60;
		sprintf(buffer, "%d:%02d:%02d", h, m, s);
	} else if (t >= 60) {
		m = t / 60;
		s = t % 60;
		sprintf(buffer, "%d:%02d", m, s);
	} else {
		sprintf(buffer, "%d", t);
	}

	VDrawStrokeString(u->v, tx - w_t, ty + (int) (yscale * 30.0),
		buffer, strlen(buffer),
		yscale * 12.0, color);
}


#define RADAR_ALTIMETER_LEN 220
#define RADAR_ALTIMETER_ES 2500


static double map_radar_alt_to_scale(double altitude)
/*
	Rescale the range [0,RADAR_ALTIMETER_ES] to [0,1].
*/
{
	static double k = 0.0;

	if( k == 0.0 )
		k = 1.0 / (sqrt(RADAR_ALTIMETER_ES + 9.0) - 3.0);

	if( altitude < 0.0 )
		altitude = 0.0;
	else if( altitude > RADAR_ALTIMETER_ES )
		altitude = RADAR_ALTIMETER_ES;

	return k * (sqrt(altitude + 9.0) - 3.0);
}


static void
doRadarAltimeter(craft * c, viewer * u)
{
	int v, x0, y0, h, y, fontH, t;
	double xscale, yscale;
	draw_data *dd;
	char buffer[20];
	Color color;

	color = u->v->w->pixel[HUDColor->cIndex];

	v = (int) (METERStoFEET(c->w.z - terrain_altitude(c))
		- c->cinfo->rm.z - c->cinfo->Gm - c->cinfo->cmMax);

	if ( v > RADAR_ALTIMETER_ES )  return;
	if ( v < 0 )  v = 0;

	xscale = u->xscaleFactor;
	yscale = u->yscaleFactor;
 	x0 = u->v->focus.x + (int) (xscale * 300);
 	y0 = u->v->focus.y + (int) (yscale * 120);
 	h = (int) (xscale * RADAR_ALTIMETER_LEN);
 	fontH = (int) (yscale * 8);
	dd = draw_new();
	VSetClipRect(u->v, &u->v->rect);

 	/*
 		Draw scale:
 	*/
 	draw_segment(dd, x0, y0, x0, y0-h);
 	draw_segment(dd, x0-4, y0, x0+4, y0);
 	for( t = 0; t <= RADAR_ALTIMETER_ES; t += 100 ){
 		y = y0 - (int) (h * map_radar_alt_to_scale(t) - 0.5);
 		if ( t <= 500 || t % 500 == 0 ) {
			draw_segment(dd, x0, y, x0+4, y);
			sprintf(buffer, "%-2d", t/100);
			VDrawStrokeString(u->v, x0+8, y, buffer, strlen(buffer),
				fontH, color);
		}
		else {
			draw_segment(dd, x0, y, x0+2, y);
		}
 	}

 	/*
 		Draw local altitude:
 	*/
 	y = y0 - (int) (h * map_radar_alt_to_scale(v) - 0.5);
 	draw_segment(dd, x0 - 2, y0, x0 - 2, y);
 	draw_segment(dd, x0 + 2, y0, x0 + 2, y);
 	draw_segment(dd, x0 - 4, y,  x0 + 4, y);
 	draw_segment(dd, x0 - 4, y-1,x0 + 4, y-1);
	draw_stroke(dd, u->v, u->v->w->pixel[HUDPixel]);
	draw_free(dd);
}


static void draw_dashed_line(VPath *p, VPoint *a, VPoint *b, int n)
{
	int i;
	VPoint c, d;

	n = 2*n - 1;
	c = *a;
	for( i = 1; i <= n; i++ ){
		VSetPoint(&d,
			a->x + i*(b->x - a->x)/n,
			a->y + i*(b->y - a->y)/n,
			a->z + i*(b->z - a->z)/n);
		if( (i & 1) == 1 ){
			vpath_moveTo(p, &c);
			vpath_lineTo(p, &d);
		} else {
			c = d;
		}
	}
}


static VPath * build_hud_vpath()
{
	VPath * p;
	int angle, s_len;
	VPoint a, b, c;
	VMatrix m, n;
	char s[10];
	double fw, fh;

	p = vpath_new();

	/**** Bare cross:
	vpath_moveTo(p, &(VPoint){-0.1, 0.0, 1.0});
	vpath_lineTo(p, &(VPoint){+0.1, 0.0, 1.0});

	vpath_moveTo(p, &(VPoint){0.0, 0.1, 1.0});
	vpath_lineTo(p, &(VPoint){0.0, -0.1, 1.0});

	return p;
	****/

/*
	Since the HUD pitch ladder is drawn over a sphere of radius 1.0,
	it is convenient representing all the distances as angles under
	which every item can be seen from the center of the sphere.


	                 .
	                 |
	+----------      .   v...--------+  15
	|                |   D   .       |   .
	|                |   ............|   .
	                 .   ^   .       .   .
	                 |       .       .   .
	                 .<--A-->.       .   .
	                 |               .   .
	                 .<------B------>.   .
	                 |                   .
	                 .<--------D-------->.
	                 |
*/

#define A DEGtoRAD(1.0)   /* ladder inner point */
#define B DEGtoRAD(2.0)   /* ladder outer point */
#define C DEGtoRAD(0.5)   /* ladder "comma" */
#define D DEGtoRAD(2.5)   /* label yaw angle */

	/* Draw horizon line, right side: */
	VSetPoint(&a, cos(A), sin(A), 0.0);
	VSetPoint(&b, cos(2*B), sin(2*B), 0.0);
	vpath_moveTo(p, &a);
	vpath_lineTo(p, &b);

	/* Draw horizon line, left side: */
	a.y = -a.y;
	b.y = -b.y;
	vpath_moveTo(p, &a);
	vpath_lineTo(p, &b);

	/* Draw ladder pitch marks, step 5 DEG: */

	fh = sin(DEGtoRAD(0.7));
	fw = sin(DEGtoRAD(0.4));

	for( angle = -85; angle <= 85; angle += 5 ){

		if( angle == 0 )
			continue;

		VIdentMatrix(&m);
		VRotate(&m, YRotation, DEGtoRAD(angle));

		/* Draw the right side of the ladder peg: */
		VSetPoint(&a, cos(A), sin(A), 0.0);
		VSetPoint(&b, cos(B), sin(B), 0.0);
		VSetPoint(&c, cos(B), sin(B), sin(C));
		if( angle < 0 )
			c.z = -c.z;

		VTransform(&a, &m, &a);
		VTransform(&b, &m, &b);
		VTransform(&c, &m, &c);

		if( angle > 0 ){
			vpath_moveTo(p, &a);
			vpath_lineTo(p, &b);
			vpath_lineTo(p, &c);
		} else {
			draw_dashed_line(p, &a, &b, 3);
			vpath_moveTo(p, &b);
			vpath_lineTo(p, &c);
		}

		/* Reverse y, then draw the left side of the peg: */
		a.y = -a.y;
		b.y = -b.y;
		c.y = -c.y;

		if( angle > 0 ){
			vpath_moveTo(p, &a);
			vpath_lineTo(p, &b);
			vpath_lineTo(p, &c);
		} else {
			draw_dashed_line(p, &a, &b, 3);
			vpath_moveTo(p, &b);
			vpath_lineTo(p, &c);
		}

		/* Now the label, right side: */
		sprintf(s, "%d", abs(angle));
		s_len = strlen(s);
		VIdentMatrix(&n);
		VScaleMatrix(&n, fw, fh, 1.0);
		VRotate(&n, XRotation, DEGtoRAD(90));
		VRotate(&n, ZRotation, DEGtoRAD(90));
		VTranslate(&n, 1.0, sin(D) - 0.5*s_len*fw, 0.5*fh);
		VMatrixMult(&n, &m, &n);
		vpath_draw_string(p, s, s_len, &n);

		/* Label, left side: */
		VTranslate(&n, 0.0, -2.0*sin(D), 0.0);
		vpath_draw_string(p, s, s_len, &n);

	}

	/* -1, -2 and -3 pitch marks: */
	for( angle = -1; angle >= -3; angle-- ){

		VIdentMatrix(&m);
		VRotate(&m, YRotation, DEGtoRAD(angle));

		/* Draw the right side of the ladder peg: */
		VSetPoint(&a, cos(A), sin(A), 0.0);
		VSetPoint(&b, cos(B), sin(B), 0.0);
		VTransform(&a, &m, &a);
		VTransform(&b, &m, &b);
		draw_dashed_line(p, &a, &b, 3);

		/* Reverse y, then draw the left side of the peg: */
		a.y = -a.y;
		b.y = -b.y;
		draw_dashed_line(p, &a, &b, 3);
	}

	return p;
}


static VPath * build_fpm_vpath()
{
	#define FPM_A_STEP 45  /* DEG */
	VPath *p;
	int a;

	p = vpath_new();

	vpath_moveTo(p, &(VPoint){0.0, 0.0, 1.0});

	for( a = FPM_A_STEP; a <= 360; a += FPM_A_STEP ){
		vpath_lineTo(p, &(VPoint){0.0, sin(DEGtoRAD(a)), cos(DEGtoRAD(a))});
	}

	vpath_moveTo(p, &(VPoint){0.0, 0.0, -1.0});
	vpath_lineTo(p, &(VPoint){0.0, 0.0, -2.0});

	vpath_moveTo(p, &(VPoint){0.0, 1.0, 0.0});
	vpath_lineTo(p, &(VPoint){0.0, 2.0, 0.0});

	vpath_moveTo(p, &(VPoint){0.0, -1.0, 0.0});
	vpath_lineTo(p, &(VPoint){0.0, -2.0, 0.0});

	return p;
}


#define LADDER_WIDTH 325.0
#define LADDER_HEIGHT 430.0


static void
doLadder(craft * c, viewer * u, double vel)
{

	_BOOL     plotFPM;
	int       windX, windY, w, tx, ty, xscale, yscale, x, y;
	VPoint    tmp, tmp1, t1;
	VMatrix   m;
	double    v, d;
	Rect      rect;
	Color     color;

	xscale = (int) (12.0 * u->xscaleFactor + 0.5);
	yscale = (int) (12.0 * u->yscaleFactor + 0.5);

	tx = (int) (LADDER_WIDTH * u->xscaleFactor * 0.5 + 0.5);
	ty = (int) (LADDER_HEIGHT * u->yscaleFactor + 0.5);
	SetRect(&rect, u->v->focus.x - tx, u->hud_yCenter - 0.37 * ty,
		u->v->focus.x + tx, u->hud_yCenter + 0.63 * ty);
	VSetClipRect(u->v, &rect);
	if( IsEmptyRect(&u->v->rect) )
		return;

	color = u->w->pixel[HUDPixel];
		      
/*
 *  Build a transformation matrix to be used to display the flight
 *  path ladder (artificial horizon).
 *
 *  One thing we do is to keep the ladder centered on the flight path
 *  marker.  The correction angle "v" is calculated to make that happen.
 */

	if (fabs(c->Cg.x) < 5.0 && fabs(c->Cg.y) < 5.0) {
		v = 0.0;
	}
	else {
		t1.x = cos(c->curHeading) * c->Cg.x +
			sin(c->curHeading) * c->Cg.y;
		t1.y = -sin(c->curHeading) * c->Cg.x +
			cos(c->curHeading) * c->Cg.y;
		v = atan2(t1.y, t1.x);
	}

	if( hud_vpath == NULL ){
		hud_vpath = build_hud_vpath();
		fpm_vpath = build_fpm_vpath();
	}
	
	VIdentMatrix(&m);
	VRotate(&m, ZRotation, v);
	VRotate(&m, YRotation, -c->curPitch);
	VRotate(&m, XRotation, -c->curRoll);
	VRotate(&m, ZRotation, DEGtoRAD(-90));
	VRotate(&m, XRotation, DEGtoRAD(-90));
	vpath_perspective_stroke(hud_vpath, &m, u->v, color);

/*
 *  Determine the location of the flight path marker
 */

	VReverseTransform_(&c->Cg, &c->trihedral, &tmp);


	if (vel < 50.0 || tmp.x == 0.0) {
		plotFPM = ! blink;
		windX = u->v->focus.x;
		windY = u->v->focus.y;
	}
	else if (tmp.x > 0.0) {
		plotFPM = TRUE;
		windX = u->v->focus.x + ((int) (tmp.y * u->v->Scale.x / tmp.x) / 4);
		windY = u->v->focus.y + ((int) (tmp.z * u->v->Scale.y / tmp.x) / 4);
	}
	else {
		plotFPM = FALSE;
		windX = 0.0;
		windY = 0.0;
	}

	if (plotFPM) {

		d = 8.0 * u->xscaleFactor;
		VIdentMatrix(&m);
		VScaleMatrix(&m, d, d, d);
		VRotate(&m, ZRotation, DEGtoRAD(-90));
		VRotate(&m, XRotation, DEGtoRAD(-90));
		VTranslate(&m, windX, windY, -1.0);
		vpath_stroke(fpm_vpath, &m, u->w, color);

	}

/*
 *  Gather weapon display info (and maybe draw a reticle).
 *
 *  WARNING. This function must be called before drawing the target
 *  designator, as this will also set the m61a1_lcos_last_post global
 *  variable that will be used next.
 */

	doWeaponDisplay(c, u, (int) floor(windX), (int) floor(windY));

/*
 *  Draw a target designator around the current primary radar target.
 */

	if (c->curRadarTarget >= 0) {

		w = (int) (DESIGNATOR_SIZE * u->xscaleFactor);

		VTransform(&ptbl[c->curRadarTarget].Sg, &c->XYZtoNED, &tmp1);
		VReverseTransform_(&tmp1, &c->trihedral, &tmp);

/* radar target is assumed to be ahead of us (tmp.z > 0.0) */

		tx = (u->v->Middl.x + (((int) (tmp.y * u->v->Scale.x / tmp.x)))) >> 2;
		ty = (u->v->Middl.y + (((int) (tmp.z * u->v->Scale.y / tmp.x)))) >> 2;


/*
 *  If the LCOS reticule was drawn by doWeaponDisplay(), then
 *  m61a1_lcos_last_pos contains its position on the screen.  We avoid to
 *  draw the target designator if this latter is too close to the LCOS.
 *
 *  If no LCOS reticle was plotted or if the distance to the LCOS reticle is
 *  sufficient, then plot a radar target designator box.
 */

		if( c->curWeapon == WEAPON_M61A1 ){
			x = m61a1_lcos_last_pos.x;
			y = m61a1_lcos_last_pos.y;
			d = sqrt((double) ((x - tx) * (x - tx) + (y - ty) * (y - ty)));
		}

		if( c->curWeapon != WEAPON_M61A1 || d > w * 2.5 ){
			draw_data *dd;

			dd = draw_new();
			draw_segment(dd, tx - w, ty - w, tx + w, ty - w);
			draw_segment(dd, tx + w, ty - w, tx + w, ty + w);
			draw_segment(dd, tx + w, ty + w, tx - w, ty + w);
			draw_segment(dd, tx - w, ty + w, tx - w, ty - w);
			draw_stroke(dd, u->v, color);
			draw_free(dd);
		}
	}

	/* Bare nose pointer: */
	AlibDrawArc(u->v->w, u->v->focus.x - 2, u->v->focus.y - 2,
		4, 4, 0, 360 * 64, color);

}


#ifdef FIXME_TESTING_CODE_FOR_VLIB
#define DASH_LEN 7

static void draw_dashed_line2(Viewport *v, int x1, int y1, int x2, int y2)
{
	int len, n, i, d, x, y, nx, ny;
	double dx, dy;
	draw_data *dd;

	len =sqrt((x1-x2)*(x1-x2) + (y1-y2)*(y1-y2));
	n = (int) (len / DASH_LEN + 0.5);
	if( (n & 1) == 0 )
		n++;
	
	dx = (x2 - x1)/(double)n;
	dy = (y2 - y1)/(double)n;

	x = x1;  y = y1;

	dd = draw_new();
	d = 1;
	for( i = 1; i <= n; i++ ){
		nx = x1 + (int)(i*dx + 0.5);  ny = y1 + (int)(i*dy + 0.5);
		if( d ){
			draw_segment(dd, x, y, nx, ny);
			//AlibDrawLine(w, x, y, nx, ny,
			//	w->pixel[HSIMagentaColor->cIndex]);
		}
		x = nx;  y = ny;
		d = !d;
	}
	draw_stroke(dd, v, v->w->pixel[HSIMagentaColor->cIndex]);
	draw_free(dd);
}


static void FIXME_draw_dashed_rect(Viewport *v, Rect *rr)
{
	AWindow *w;
	Rect r2, *r;


	w = v->w;
	AlibSetClipRect(w, rr);

	r2 = *rr;
	ExpandRect(&r2, -1, -1);
	r = &r2;
	//printf("FIXME: r=%d %d %d %d\n", r2.a.x, r2.a.y, r2.b.x, r2.b.y);

	draw_dashed_line2(v, r->a.x, r->a.y, r->b.x-1, r->a.y);
	draw_dashed_line2(v, r->a.x, r->a.y, r->a.x, r->b.y-1);
	draw_dashed_line2(v, r->a.x, r->b.y-1, r->b.x-1, r->b.y-1);
	draw_dashed_line2(v, r->b.x-1, r->a.y, r->b.x-1, r->b.y-1);

	draw_dashed_line2(v, v->focus.x, v->focus.y - 20, v->focus.x, v->focus.y + 20);
	draw_dashed_line2(v, v->focus.x - 20, v->focus.y, v->focus.x + 20, v->focus.y);
}
#endif


void
hud_draw(viewer * u)
{
	craft    *c;
	int       i, x, x1, y, fth, xscale1, yscale1;
	double    xscale, yscale, fontH, vel;
	char      buffer[80];
	Color     color;

	//FIXME_draw_dashed_rect(u->v, &u->v->rect);

	if( u->hud == NULL || !HUD(u)->enabled )
		return;
	
	c = u->c;

	if( ! damage_isFunctioning(c, SYS_HUD) )
		return;

	xscale = u->xscaleFactor;
	yscale = u->yscaleFactor;
	color = u->v->w->pixel[HUDColor->cIndex];

	/* Stall warning */
	if( ! blink && c->damageBits & FLAG_STALL_WARN ){
		fontH = (int) (24 * yscale);
		int fontW = fontH;
		x = u->v->focus.x - 5 * fontW / 2;
		y = u->v->focus.y + 4 * fontH;
		strcpy(buffer, "STALL");
		VDrawStrokeString(u->v, x, y, buffer, strlen(buffer), fontH, color);
	}

	xscale1 = (int) (xscale * 2048.0);
	yscale1 = (int) (yscale * 2048.0);
	fontH = (int) (12.0 * yscale + 0.5);
	fth = (int) (18.0 * yscale + 0.5);

	if( curTime >= blink_toggle_time ){
		blink = ! blink;
		if( blink )
			blink_toggle_time = curTime + 0.05;
		else
			blink_toggle_time = curTime + 0.5;
	}

	if (u->viewDirection.x < 0.90)
		return;

	/* Angle of Attack */
	if ((vel = Vmagnitude(&c->Cg)) < 50.0)
		sprintf(buffer, "a=0.00");
	else
		sprintf(buffer, "a=%.2f", RADtoDEG(c->alpha));
	x = u->v->focus.x + 150 * xscale1 / 2048;
	y = u->v->focus.y - u->velScale.length / 2 - (int) (yscale * 50);
	VDrawStrokeString(u->v, x, y, buffer, strlen(buffer), fontH, color);

	/* Accelerometer */
	x1 = u->v->focus.x - 220 * xscale1 / 2048;
	sprintf(buffer, "%4.1f", - c->G.z);
	VDrawStrokeString(u->v, x1, y, buffer, strlen(buffer), fontH, color);

	/*
		APS lights:
	*/

	x = u->v->focus.x + 120.0*xscale;
	y = u->v->focus.y - 90.0*yscale;
	i = fontH * 2;  /* interline */

	/* Auto Pilot */
	if( aps_ap_enabled(c) && ( ! aps_ap_warn(c) || ! blink ) ){
		VDrawStrokeString(u->v, x, y, "AP", 2, fontH, color);
	}

	/* Auto Landing */
	y += i;
	if( aps_al_enabled(c) && ( ! aps_al_warn(c) || ! blink ) ){
		VDrawStrokeString(u->v, x, y, "AL", 2, fontH, color);
	}

	/* Auto Nav */
	y += i;
	if ( aps_an_enabled(c) && ( ! aps_an_warn(c) || ! blink ) ){
		VDrawStrokeString(u->v, x, y, "AN", 2, fontH, color);
	}

	/* Auto Throttle */
	y += i;
	if ( aps_at_enabled(c) && ( ! aps_at_warn(c) || ! blink ) ){
		VDrawStrokeString(u->v, x, y, "AT", 2, fontH, color);
	}

	/* Auto Turn */
	y += i;
	if ( aps_aw_enabled(c) && ( ! aps_aw_warn(c) || ! blink ) ){
		VDrawStrokeString(u->v, x, y, "AW", 2, fontH, color);
	}

	/* Auto Coord */
	y += i;
	if ( aps_ac_enabled(c) && ( ! aps_ac_warn(c) || ! blink ) ){
		VDrawStrokeString(u->v, x, y, "AC", 2, fontH, color);
	}

	/* Vertical velocity */
	x = u->v->focus.x + 130 * xscale1 / 2048;
	y = u->velScale.yorg + (int) (2 * yscale * fontH);
	sprintf(buffer, "%+6d", (int) (-c->Cg.z * 60.0));
	VDrawStrokeString(u->v, x, y, buffer, strlen(buffer), fontH, color);

	/*  Compass, altititude and airspeed cards */

	/*scale.c*/ doCompassScale(u->v, &(u->hdgScale),
		RADtoDEG(c->showMag? navaid_mag_heading(c) : c->curHeading) * 100.0);

	doScale(u->v, &(u->altScale), METERStoFEET(c->w.z));

	/* TAS: */
	//vel = c->VT;
	/* Airspeed along the x axis (the Pitot tube indicates this): */
	vel = c->VT * cos(c->alpha) * cos(c->beta);
	if( vel < 0.0 )  vel = 0.0;
	doScale(u->v, &(u->velScale), FPStoKT(vel));

	/*
		AutoThrottle
	*/
	if ( aps_at_enabled(c)
	&& ( ! aps_at_warn(c) || ! blink ) ){
		x = u->v->focus.x - 150 * xscale1 / 2048;
		y = u->hud_yCenter - 10 * yscale1 / 2048;
		sprintf(buffer, "AT%3d", (int) (FPStoKT( aps_at_get_velocity(c) ) + 0.5));
		VDrawStrokeString(u->v, x, y,
			buffer, strlen(buffer), fontH, color);
	}

	/* Mach number */
	if (c->mach >= 0.20) {
		sprintf(c->leftHUD[1], "   %4.2f", c->mach);
	}
	else {
		strcpy(c->leftHUD[1], "");
	}

	x = u->v->focus.x - 220 * xscale1 / 2048;
	y = u->hud_yCenter - 200 * yscale1 / 2048;

	for (i = 0; i < 5; ++i) {
		VDrawStrokeString(u->v, x, y, c->leftHUD[i],
			strlen(c->leftHUD[i]), fontH, color);
		y += fth;
	}

	x = u->v->focus.x + 160 * xscale1 / 2048;
	y = u->hud_yCenter - 200 * yscale1 / 2048;

	for (i = 0; i < 5; ++i) {
		VDrawStrokeString(u->v, x, y, c->rightHUD[i],
			strlen(c->rightHUD[i]), fontH, color);
		y += fth;
	}

	doLadder(c, u, vel);

	doBankIndicator(c, u);

	doTurnAndSlipIndicator(c, u);

	doRadarAltimeter(c, u);

}


static int fspage_show = 0; /* FIXME: put in data struct tied to the viewer */

/* on entry to add_ev_trace, this is the index of the oldest line, but on exit,   */
/* it is newest and ev_trace_oldest has been incremented, wrapping after 6.       */
/* For printing, it is the first to print.  The first time ev_trace_add is        */
/* called, it is negative to force initialization of the array and is reset to 0. */
int   ev_trace_oldest = -1;
char  ev_trace_string[EV_TRACE_NUM][EV_TRACE_SIZE + 1];

void add_ev_trace(viewer *u, char *s)
{
	unsigned int len;
	if( ev_trace_oldest < 0 ) {
		unsigned int ii, jj;
		for ( ii = 0 ; ii < EV_TRACE_NUM ; ii++ )
			for ( jj = 0 ; jj < EV_TRACE_SIZE + 1 ; jj++ )
				ev_trace_string[ii][jj] = '\0';
		ev_trace_oldest = 0;
	}
	if( ( len = strlen(s) ) == 0 )
		return; /* ignore null input lines */
	if( len > EV_TRACE_SIZE )
		len = EV_TRACE_SIZE; /* truncate excessive input lines */
	strncpy(ev_trace_string[ev_trace_oldest], s, len);
	ev_trace_string[ev_trace_oldest][len] = '\0';
	if( ++ev_trace_oldest == EV_TRACE_NUM )
		ev_trace_oldest = 0;
}

static void TermPrintLn(viewer * u, char *s)
{
	static int x = 0, y = 0, h = 0, dy = 0;

	if( s == NULL ){
		if( u->hud_mode )
			h = RectHeight(u->v->rect) / 50;
		else
			h = RectHeight(u->v->rect) / 30;
		dy = (int) (h * 1.5 + 0.5);
		x = dy;
		y = dy;
		return;
	}

	VDrawStrokeString(u->v, x, y, s, strlen(s), h,
		u->w->pixel[HUDColor->cIndex]);
	y += dy;
}


void hud_free(viewer *u)
{
	if( u->hud == NULL )
		return;
	
	HUD(u)->next = free_list;
	free_list = u->hud;
	u->hud = NULL;
}


void hud_free_all()
{
	hud_data *hud;

	while( free_list != NULL ){
		hud = free_list;
		free_list = free_list->next;
		FREE(hud);
	}

	vpath_free(hud_vpath);
	vpath_free(fpm_vpath);
}


void   FSPageToggle()
{
	fspage_show = ! fspage_show;
}


static void
doFSPage(craft * c, viewer * u)
{

	double VAR, vel, altft;
	char buf[256], buf1[256];
	char  stickType; /* 'M'=mouse 'K'=keyboard 'J'=joystick */
	char  stickMode; /* keyboard: 'L'=linear 'D'=double           */
			 /* mouse:    'F'=fast 'N'=normal 'P'=precise */
			 /* joystick: '?'=?                           */

	if (u->noMorJ) {		 	/* is keystick? Y,       */
		stickType = 'K';
		if (u->stickM == KEYSTICK_LINEAR)
			stickMode = 'L';
		else	stickMode = 'D';
	} else {				/* is mouse/joystick, so */
		stickType = 'M';
		if (mouse_stick_mode == MOUSE_FAST)
			stickMode = 'F';
		else if (mouse_stick_mode == MOUSE_NORMAL)
			stickMode = 'N';
		else	stickMode = 'P';
	}

	VSetClipRect(u->v, &u->v->rect);

	TermPrintLn(u, NULL);

/*	TermPrintLn(u, c->cinfo->name);	*/
	TermPrintLn(u, c->cinfo->description);

	sprintf (buf, "%8.2f secs  G: %+.2f %+.2f %+.2f",
		 curTime, c->G.x, c->G.y, c->G.z);
	TermPrintLn(u, buf);

	sprintf (buf, "Ail/Elev/Rdr: %+.3f %+.3f %+.3f", c->Sa, c->Se, c->Sr);
	TermPrintLn(u, buf);

	sprintf (buf, "Euler: %+.1f %+.1f %+.1f",
		RADtoDEG(c->curRoll),
		RADtoDEG(c->curPitch),
		RADtoDEG(c->curHeading));
	TermPrintLn(u, buf);
	vel = c->VT * cos(c->alpha) * cos(c->beta);
/*	if (vel < 0.0) vel = 0.0;	*/
	altft = METERStoFEET(c->w.z);
	fprintf(stderr, "%-38.38s TAS=%+9.2f altft=%+7.1f\n", buf, vel, altft);

	sprintf (buf, "Euler rates: %+.2f %+.2f %+.2f",
		RADtoDEG(c->p + c->q*tan(c->curPitch)*sin(c->curRoll) + c->r*tan(c->curPitch)*cos(c->curRoll)),
		RADtoDEG(c->q*cos(c->curRoll) - c->r*sin(c->curRoll)),
		RADtoDEG(c->q*sin(c->curRoll)/cos(c->curPitch) + c->r*cos(c->curRoll)/cos(c->curPitch)));
	TermPrintLn(u, buf);
	fprintf(stderr, "%-38.38s THr=%+.6f\n", buf, c->curHeading);

	sprintf (buf, "Omega: %+.2f %+.2f %+.2f",
		RADtoDEG(c->p), RADtoDEG(c->q), RADtoDEG(c->r));
	TermPrintLn(u, buf);

#ifdef AFDS
	sprintf (buf, "psiRef/Lambda: %+.1f, %+.2f",
		fd->psiRef_deg, fd->lambda_deg);
	TermPrintLn(u, buf);

	sprintf (buf, "psiCmd/Coupler:",
		fd->psiCommand_deg, fd->x[AFDS_COUPLER_X1]);
	TermPrintLn(u, buf);

	sprintf (buf, "   %+.1f, %+.2f",
		fd->psiCommand_deg, fd->x[AFDS_COUPLER_X1]);
	TermPrintLn(u, buf);

	AFDSGetStatusString ( c, buf, sizeof(buf));
	TermPrintLn(u, buf);
#endif /*AFDS*/

	DISLatitudeToString(buf,  c->w.latitude, LLM_DMS);
	strcat (buf, " (");
	DISLatitudeToString(buf1, c->w.latitude, LLM_SIGNED_D);
	strcat (buf, buf1);
	strcat (buf, ")");
	TermPrintLn(u, buf);
	fprintf(stderr, "%s\n", buf);

	DISLongitudeToString(buf,  c->w.longitude, LLM_DMS);
	strcat (buf, " (");
	DISLongitudeToString(buf1, c->w.longitude, LLM_SIGNED_D);
	strcat (buf, buf1);
	strcat (buf, ")  ");
	TermPrintLn(u, buf);
	fprintf(stderr, "%s\n", buf);

	sprintf (buf, "%5d m (%5d ft)",
		 (int)             (c->w.z+0.5),
		 (int)(METERStoFEET(c->w.z+0.5))
		);
	if (c->showMag) {
		strcat (buf, "   ");
		VAR = RADtoDEG( c->indicatedLocalVAR );
		sprintf(buf1, "VAR %.1f%c", fabs(VAR), VAR < 0 ? 'E' : 'W');
		strcat (buf, buf1);
	}
	TermPrintLn(u, buf);
	sprintf (buf, "%c%c: S %3i   X %+4i   Y %+4i",
			 stickType, stickMode, u->stickS,
			 u->stickX - u->w->width  / 2,
			(u->stickY - u->w->height / 2) * -1);
	TermPrintLn(u, buf);


	if (u->testKeys ) /* Only if we're if testKeys mode via F2 key */
	/* Display the input events trace table */
	{ int ev_trace_newest, ii, jj;

	  if ( ev_trace_oldest == 0 )
		  ev_trace_newest = EV_TRACE_NUM - 1;
	  else	  ev_trace_newest = ev_trace_oldest - 1;
	  ii = ev_trace_oldest;
	  for( jj = 0; jj < EV_TRACE_NUM; jj++ ) {
		  TermPrintLn(u, ev_trace_string[ii++]);
		  if ( ii == EV_TRACE_NUM )
			  ii = 0;
	  }
	}

}


void
doFlightStatusPage(craft * c, viewer * u)
{

	if ( ! fspage_show )
		return;

	doFSPage(c, u);
	return;

}
