/*
 *  acm : an aerial combat simulator for X
 *  Copyright (C) 1991-1998  Riley Rainey
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 dated June, 1991.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program;  if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
 */

/*
 *  2014-12-29 dlc - Refer to CHANGES text file for details of modifications
 */
 
#ifndef WIN32
#include <X11/Xos.h>
#endif
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <Vlib.h>
#include "adf.h"
#include "alarm.h"
#include "audio.h"
#include "aps.h"
#include "damage.h"
#include "dis_if.h"
#include "draw.h"
#include "drone.h"
#include "effects.h"
#include "hsi.h"
#include "hud.h"
#include "instruments.h"
#include "inventory.h"
#include "magnetic_compass.h"
#include "mem_manager.h"
#include "missile.h"
#include "navaid.h"
#include "panel.h"
#include "planes.h"
#include "players.h"
#include "pm.h"
#include "prompt.h"
#include "render.h"
#include "resupply.h"
#include "runway.h"
#include "terminal.h"
#include "terrain.h"
#include "vpath_gallery.h"
#include "weapon.h"

#define INIT_INTERNAL_GLOBAL_VAR
#include "init.h"

/*
 *  CLOUD_BASE and CLOUD_TOP sets the altitude of base and top of
 *  the cloud layer in feet. If base == top, no clouds are used.
 */
#ifndef CLOUD_BASE
#define CLOUD_BASE      0
#endif
#ifndef CLOUD_TOP
#define CLOUD_TOP       0
#endif


static char ** objects_dirs = NULL;

static char * default_object_dirs[] = {
	".",
	"objects",
	"../objects",
	NULL
};


/*
 *  undersampleObject
 *
 *  Remove some of the detail in an object to reduce the CPU overhead
 *  of transforming and clipping it.
 */

static void
undersampleObject (VObject * obj, int rate)
{
    int i, j, k;

    k = 0;

    for (i = 1; i < obj->numPolys; ++i) {
	for (j = 0; j < obj->polygon[i]->numVtces; j += rate) {
	    obj->polygon[i]->vertex[k++] =
		obj->polygon[i]->vertex[j];
	}
	obj->polygon[i]->numVtces = k;
    }
}

static void
acm_file_not_found(char * name)
{
	char **p;

	fprintf(stderr, "acm_file_not_found() file `%s' not found\n", name);
	if ( name[0] == '/' )  return;
	fprintf(stderr, "acm_file_not_found() hint: use the option -objects path1:path2:...\n");
	fprintf(stderr, "acm_file_not_found() directories scanned:");
	p = objects_dirs;
	while ( *p != NULL ) {
		fprintf(stderr, "acm_file_not_found()  %s", *p);
		p++;
	}
	fprintf(stderr, "acm_file_not_found() \n");
}


/*
 *  acm_fopen :  Open an important acm data file (or die trying).
 */

FILE     *
acm_fopen(char *name, char *access)
{

	FILE     *f;
	char      s[256];
	char    **p;

	if ((f = fopen(name, access)) != NULL)
		return f;

	if ( name[0] == '/' ) {
		acm_file_not_found(name);
		exit(1);
	}

	p = objects_dirs;
	while ( *p != NULL ) {
		sprintf(s, "%s/%s", *p, name);
		if ((f = fopen(s, access)) != NULL)
			return f;
		p++;
	}

	acm_file_not_found(name);
	exit(1);
}

/*
 *  acm_find_file :  Find an important acm data file (or die trying).
 */

char     *
acm_find_file(char *name)
{
	static char pname[1024];
	struct stat statbuf;
	char ** p;

	strcpy(pname, name);
	if (stat(name, &statbuf) == 0)
		return pname;

	if ( name[0] == '/' ) {
		acm_file_not_found(name);
		exit(1);
	}

	p = objects_dirs;
	while ( *p != NULL ) {
		sprintf(pname, "%s/%s", *p, name);
		if (stat(pname, &statbuf) == 0)
			return pname;
		p++;
	}

	acm_file_not_found(name);
	exit(1);
}


int split(char *s, int *argc, char *argv[], int max_argc)
{
	int n;

	n = 0;

	do {

		/* skip separators: */
		while( *s == ' ' || *s == '\t' || *s == '\r' || *s == '\n' ){
			*s = '\0';
			s++;
		}
		
		/* end of the string? */
		if( *s == '\0' ){
			*argc = n;
			return TRUE;
		}

		/* found another field: */
		if( n >= max_argc ){
			*argc = n;
			return FALSE;
		}
		argv[n] = s;
		n++;

		/* scan field: */
		while( *s != ' ' && *s != '\t' && *s != '\r' && *s != '\0' && *s != '\n' )
			s++;

	} while( TRUE );
}


typedef struct {
	char * file_name;
	FILE * file;
	int    line_no;
} FileIn;


static FileIn * file_open_read(char *file_name)
{
	FILE * file;
	FileIn * in;

	file = acm_fopen(file_name, "r");
	if( file == NULL )
		return NULL;
	in = MALLOC( sizeof(FileIn) );
	in->file_name = STRDUP(file_name);
	in->file = file;
	in->line_no = 0;
	return in;
}


static _BOOL file_read_line(FileIn *in, char *line, int line_size)
/*
	Return next line from file with leading spaces, trailing spaces and
	new-line removed. Increment internal line counter so we can generate
	meaningful error messages.

	Return TRUE on success or FALSE on end EOF or other I/O error.
*/
{
	int l;

	if( fgets(line, line_size, in->file) == NULL )
		return FALSE;

	l = strlen(line);

	/* Remove new-line: */
	while( l > 0 && (line[l-1] == '\n' || line[l-1] == '\r') ){
		line[l-1] = '\0';
		l--;
	}

	/* Remove trailing spaces: */
	while( l > 0 && isspace(line[l-1]) ){
		line[l-1] = '\0';
		l--;
	}

	/* Remove leading spaces: */
	while( isspace(line[0]) ){
		memmove(&line, &line[1], l);
		l--;
	}

	in->line_no++;
	return TRUE;
}


static void file_close(FileIn *in)
{
	fclose(in->file);
	FREE(in->file_name);
	FREE(in);
}


#define parse_error(/* Filein * */ in, ...) \
	{ \
		fprintf(stderr, "%s:%d: ", in->file_name, in->line_no); \
		fprintf(stderr, __VA_ARGS__); \
		fprintf(stderr, "\n"); \
	}


static void parse_runway_record(FileIn *in, char **argv, int argc, int *i)
/*
	Parses RWY and RWY2 records and add to the stbl[] array.

	argv, argc: fields of the scene record
	*i: next available slot in stbl[] array
*/
{
    char *airport, *runways_no, lat_lon[64];
    double heading = 0.0, mag, length=0.0, width=0.0;
    int j;
    VPoint fwd, down, local_fwd, local_right, local_down;
    WorldCoordinates w, w1, w2;
    dis_world_coordinates xyz, xyz1;
    VObject *object;
    VPolygon *p;
	VPolySet *ps;
    VMatrix RWYtoXYZ, XYZtoNED;

	if (strcmp (argv[0], "RWY") == 0) {
		
		if( argc != 10 ){
			parse_error(in, "invalid number of fields in runway record -- ignoring.");
			return;
		}

		/*
		 *  Parse Latitude/Longitude of both runway ends
		 */
		length = FEETtoMETERS (atoi (argv[4]));
		width = FEETtoMETERS (atoi (argv[5]));

		strcpy (lat_lon, argv[6]);
		strcat (lat_lon, argv[7]);
		DISStringToLatLong (lat_lon, &w);
		w.z = FEETtoMETERS (atoi (argv[3]));

		addAltitudeEntry (&w);

		strcpy (lat_lon, argv[8]);
		strcat (lat_lon, argv[9]);
		DISStringToLatLong (lat_lon, &w1);
		w1.z = w.z;

		DISWorldCoordinatesToGeocentric (&w,
					   (dis_world_coordinates *) & xyz);

		DISWorldCoordinatesToGeocentric (&w1,
					  (dis_world_coordinates *) & xyz1);

		/*
		 *  Calculate runway forward direction:
		 */

		local_fwd.x = xyz1.x - xyz.x;
		local_fwd.y = xyz1.y - xyz.y;
		local_fwd.z = xyz1.z - xyz.z;

		mag = sqrt (
			   local_fwd.x * local_fwd.x +
			   local_fwd.y * local_fwd.y +
			   local_fwd.z * local_fwd.z
		);
		local_fwd.x /= mag;
		local_fwd.y /= mag;
		local_fwd.z /= mag;

		/*
		 *  Average those two points to generate a midpoint that will be the
		 *  origin of a runway coordinate system.
		 */

		xyz.x = (xyz.x + xyz1.x) / 2.0;
		xyz.y = (xyz.y + xyz1.y) / 2.0;
		xyz.z = (xyz.z + xyz1.z) / 2.0;

		DISGeocentricToWorldCoordinates (&xyz, &w2);
		GenerateWorldToLocalMatrix (&w2, &XYZtoNED);
		down.z = 1.0;
		down.x = down.y = 0.0;
		VReverseTransform_ (&down, &XYZtoNED, &local_down);

		/*
		 * Calculate runway heading, normalized to [0,pi[ and fix the
		 * local_fwd direction accordingly:
		 */

		VPoint q;
		VTransform_(&local_fwd, &XYZtoNED, &q);
		heading = atan2(q.y, q.x);
		if( heading < 0.0 ){
			heading += pi;
			local_fwd.x = -local_fwd.x;
			local_fwd.y = -local_fwd.y;
			local_fwd.z = -local_fwd.z;
		}

	}

	else if (strcmp (argv[0], "RWY2") == 0) {

		if( argc != 9 ){
			parse_error(in, "invalid number of fields in runway record -- ignoring.");
			return;
		}

		length = FEETtoMETERS (atoi (argv[4]));
		width = FEETtoMETERS (atoi (argv[5]));
		heading = DEGtoRAD( atof(argv[8]) );

		/*
			In order to correctly place the runway numbers, the
			local_fwd vector must be oriented in the range [0,pi[:
		*/

		if( heading >= pi )
			heading -= pi;

		strcpy (lat_lon, argv[6]);
		strcat (lat_lon, argv[7]);
		DISStringToLatLong (lat_lon, &w);
		w.z = FEETtoMETERS (atoi (argv[3]));

		addAltitudeEntry (&w);

		DISWorldCoordinatesToGeocentric (&w,
					   (dis_world_coordinates *) & xyz);

		GenerateWorldToLocalMatrix (&w, &XYZtoNED);

		fwd.x = cos(heading);
		fwd.y = sin(heading);
		fwd.z = 0.0;
		VReverseTransform_ (&fwd, &XYZtoNED, &local_fwd);

		down.z = 1.0;
		down.x = down.y = 0.0;
		VReverseTransform_ (&down, &XYZtoNED, &local_down);

	} else {

		parse_error(in, "unknown runway type `%s' -- ignoring.", argv[0]);
		return;

	}


	/*
	 *  Calculate the local right vector in geocentric XYX.
	 *  A basic property of Cross Products: k x i = j
	 */

	VCrossProd (&local_down, &local_fwd, &local_right);

	/*
	 *  Generate a transformation matrix to get from "runway" coordinates to
	 *  Geocentric. The x axis of the runway coord. (local_fwd) points in the
	 *  range [0,180[ DEG of the geographical coords.
	 */
	VIdentMatrix (&RWYtoXYZ);

	RWYtoXYZ.m[0][0] = local_fwd.x;
	RWYtoXYZ.m[1][0] = local_fwd.y;
	RWYtoXYZ.m[2][0] = local_fwd.z;

	RWYtoXYZ.m[0][1] = local_right.x;
	RWYtoXYZ.m[1][1] = local_right.y;
	RWYtoXYZ.m[2][1] = local_right.z;

	RWYtoXYZ.m[0][2] = local_down.x;
	RWYtoXYZ.m[1][2] = local_down.y;
	RWYtoXYZ.m[2][2] = local_down.z;

	RWYtoXYZ.m[0][3] = xyz.x;
	RWYtoXYZ.m[1][3] = xyz.y;
	RWYtoXYZ.m[2][3] = xyz.z;

	airport = argv[1];
	runways_no = argv[2];

	ps = runway_build(runways_no, &RWYtoXYZ, length, width, heading);

	/*
		Optimize a bit here and put all the runways for a given airport
		in the same object.

		NOTE. With this optimization, a runway that cross another one
		will hide part of this latter. Otherwise without this
		optimization the two overlapping parts are rendered
		alternatively one and the other producing a bad flickering
		effect, I don't know why.
	*/
	
	if( *i == 0 )
		object = NULL;
	else
		object = stbl[*i - 1].cinfo->object;
	
	if ((object != NULL) && (strcmp(object->name, airport) == 0)) {

		/*
		 *  New runway at same airport as last .. add a polygon to the last object.
		 */

		object->polygon = REALLOC( object->polygon,
			sizeof (VPolygon *) * (object->numPolys + VPolySet_Count(ps)));
		j = 0;
		p = VPolySet_First(ps);
		while( p != NULL ){
			object->polygon[object->numPolys + j] = p;
			p = VPolySet_Next(ps);
			j++;
		}
		object->numPolys += VPolySet_Count(ps);
		VPolySet_Free(ps, FALSE);
		VComputeObjectExtent (object);
	}
	else {

		/*
		 *    New airport.  New object and craftInfo.
		 */
		object = MALLOC(sizeof(VObject));
		memset(object, 0, sizeof(VObject));
		object->name = STRDUP(airport);
		object->numPolys = VPolySet_Count(ps);
		object->polygon = MALLOC( sizeof(VPolygon *) * object->numPolys );
		p = VPolySet_First(ps);
		j = 0;
		while( p != NULL ){
			object->polygon[j++] = p;
			p = VPolySet_Next(ps);
		}
		VPolySet_Free(ps, FALSE);
		object->order = NULL;

		VComputeObjectExtent (object);

		stbl[*i].type = CT_SURFACE;
		stbl[*i].flags = FL_FIXED_OBJECT;
		stbl[*i].cinfo = newCraft();
		if (! stbl[*i].cinfo) {
			fprintf(stderr, "parse_runway_record() ERROR: Out of craft types while loading scene.\n");
			exit(1);
		}
		stbl[*i].cinfo->name = STRDUP(airport);
		stbl[*i].cinfo->object = object;
		stbl[*i].cinfo->placeProc = NULL;
		stbl[*i].curHeading = stbl[*i].curPitch = stbl[*i].curRoll = 0.0;
		stbl[*i].Sg.x = stbl[*i].Sg.y = stbl[*i].Sg.z = 0.0;
		if (++(*i) == MAXSURFACE) {
			fprintf (stderr,
			"parse_runway_record() ERROR: Only the first %d surface objects will be displayed.\n",
				 MAXSURFACE);
			*i = MAXSURFACE - 1;
		}
	}

}


/*
 *  readScene :  Process the scene description file, returns zero if successful
 *               or -1 otherwise.
 */

static int
readScene (char *ground, int len)
{
    FileIn *f;
	FILE *f1;
    char file[256], *name, *filep;
    char *argv[32], line[1000];
    char lat[64], lon[64], lat_lon[64];
    double z, heading, freq;
    int i = 0, j, k, undersample, depthcue, argc;
    craftType *c;
    VPoint scale;
    WorldCoordinates w, gsw;
    double loc_width, loc_bearing, gs_angle;

	/* initialize list of active viewers -- null list */

	vl_head = NULL;

    name = (sceneFile) ? sceneFile : "default.acmscene";

    /*
     *  The units used in the objects we've created are expressed in FEET.
     *  Internally, graphics objects should be METERS.
     */

    scale.x = scale.y = scale.z = FEETtoMETERS (1.0);
    VSetReadObjectScale (&scale);

    /*
     * In case there are no GROUNDC records, set [0] to default black
     */

    strcpy(gtbl[0].name, "_______________");
    strcpy(gtbl[0].cstring, "#000000");
    gtbl[0].vcptr = VAllocDepthCueuedColor("#000000", 1);

    /*
     * In case there are no TEAMLOC records, set [0] to default everything zero
     */

    strcpy(ttbl[0].sname, "_____");
    strcpy(ttbl[0].lname, "___________________");
    ttbl[0].loc.latitude  = 0.0;
    ttbl[0].loc.longitude = 0.0;
    ttbl[0].loc.z         = 0.0; /* altitude */
    ttbl[0].heading       = 0.0;

    f = file_open_read(name);
    fprintf(stderr, "Processing acmscene file %s\n", name);

    while( file_read_line(f, line, 1000) ){

		if( line[0] == '\0' ){
			/* Ignore empty line. */
    			fprintf(stderr, "Processing init: Ignoring empty line\n");
			continue;
		}

		if( line[0] == '#' ){
			/* Ignore comment line. */
    			fprintf(stderr, "Processing init: Ignoring comment line\n");
			continue;
		}

		if( ! split(line, &argc, argv, 32) ){
    			fprintf(stderr, "Processing init: Ignoring excessive fields\n");
			parse_error(f, "too many fields -- ignoring line");
		}

		else if ( strcmp (argv[0], "TEAMLOC") == 0 ) {
    			fprintf(stderr, "%s\n", "Processing init: Handling TEAMLOC");
			if( argc != 7 ){
				parse_error(f, "TEAMLOC: invalid no. of fields");
				return -1;
			}
			if ( MAXTEAMLOCS == ttblCount ) {
    				fprintf(stderr, "Encountered more TEAMLOC records than %u\n",
					MAXTEAMLOCS);
				return -1;
			}
			if ( strlen(argv[1]) > sizeof ttbl->sname ) {
    				fprintf(stderr, "TEAMLOC %s sname is too long\n",
					argv[1]);
				return -1;
			}
			for ( j = 0 ; j < ttblCount ; j++ ) {
				if ( strcmp(ttbl[j].sname, argv[1]) == 0 ) {
    					fprintf(stderr, "Encountered a second TEAMLOC record for sname %s\n",
						argv[1]);
					return -1;
				}
			}
			if ( strlen(argv[2]) > sizeof ttbl->lname ) {
    				fprintf(stderr, "TEAMLOC %s lname is too long\n",
					argv[2]);
				return -1;
			}
			for ( j = 0 ; j < ttblCount ; j++ ) {
				if ( strcmp(ttbl[j].lname, argv[2]) == 0 ) {
    					fprintf(stderr, "Encountered a second TEAMLOC record for lname %s\n",
						argv[2]);
					return -1;
				}
			}
			/* Use current ttblCount as ttbl index, increment it at end */
			strcpy(ttbl[ttblCount].sname, argv[1]);
			strcpy(ttbl[ttblCount].lname, argv[2]);
			strcpy(lat, argv[3]);
			strcpy(lon, argv[4]);
			z = atof(argv[5]);
			heading = atof(argv[6]);
			strcat (lat, lon);
			DISStringToLatLong (lat, &ttbl[ttblCount].loc);
			ttbl[ttblCount].loc.z = FEETtoMETERS (z);
			ttbl[ttblCount].heading = DEGtoRAD (heading);
			ttblCount++;

		} else if ( strcmp (argv[0], "TEAM_1") == 0 ) {
    			fprintf(stderr, "%s\n", "Processing init: Handling TEAM_1");
			if( argc != 2 ){
				parse_error(f, "TEAM_1: invalid no. of fields");
				return -1;
			}
			j = -1;
			for ( k = 0 ; k < ttblCount && j != 0 ; k++ ) {
				if ( strcmp(ttbl[k].sname, argv[1]) == 0 ) {
					teamLatLon[1].latitude  = ttbl[k].loc.latitude;
					teamLatLon[1].longitude = ttbl[k].loc.longitude;
					teamLatLon[1].z         = ttbl[k].loc.z;
					teamHeading[1]          = ttbl[k].heading;
					j = 0;
				}
			}
			if ( j == -1 ) {
    				fprintf(stderr, "%s\n", "The TEAMLOC record for %s was not found",
					argv[1]);
				return -1;
			}

		} else if ( strcmp (argv[0], "TEAM_2") == 0 ) {
    			fprintf(stderr, "%s\n", "Processing init: Handling TEAM_2");
			if( argc != 2 ){
				parse_error(f, "TEAM_2: invalid no. of fields");
				return -1;
			}
			j = -1;
			for ( k = 0 ; k < ttblCount && j != 0 ; k++ ) {
				if ( strcmp(ttbl[k].sname, argv[1]) == 0 ) {
					teamLatLon[2].latitude  = ttbl[k].loc.latitude;
					teamLatLon[2].longitude = ttbl[k].loc.longitude;
					teamLatLon[2].z         = ttbl[k].loc.z;
					teamHeading[2]          = ttbl[k].heading;
					j = 0;
				}
			}
			if ( j == -1 ) {
    				fprintf(stderr, "The TEAMLOC record for %s was not found\n",
					argv[1]);
				return -1;
			}

		} else if( strcmp (argv[0], "GROUNDC") == 0 ){
    			fprintf(stderr, "Processing init: Handling GROUNDC\n");

			if( argc != 3 ){
				parse_error(f, "GROUNDC: invalid no. of fields");
				return -1;
			}
			if ( strlen(argv[1]) > sizeof gtbl->name ) {
    				fprintf(stderr, "GROUNDC %s name is too long\n",
					argv[1]);
				return -1;
			}
			if ( MAXGROUNDCS == gtblCount ) {
    				fprintf(stderr, "Encountered more GROUNDC records than %u\n",
					MAXGROUNDCS);
				return -1;
			}
			for ( j = 0 ; j < gtblCount ; j++ ) {
				if ( strcmp(gtbl[j].name, argv[1]) == 0 ) {
    					fprintf(stderr, "Encountered a second GROUNDC record for %s\n",
						argv[1]);
					return -1;
				}
			}
			/* Use current gtblCount as gtbl index, increment it at end */
			strcpy(gtbl[gtblCount].name, argv[1]);
			if ( strlen(argv[2]) >= sizeof gtbl->cstring ) {
    				fprintf(stderr, "GROUNDC %s color is too long\n",
					argv[2]);
				return -1;
			}
			if ( argv[2][0] != '#' ) {
    				fprintf(stderr, "GROUNDC %s color does not begin with #\n",
					argv[2]);
				return -1;
			}
			strcpy(gtbl[gtblCount].cstring, argv[2]);
			gtbl[gtblCount].vcptr = VAllocDepthCueuedColor(argv[2], 1);
			gtblCount++;
		}

		/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
		 *
		 *  RWY : Runway record (old format)
		 *
		 *-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/

		else if (strcmp (argv[0], "RWY") == 0) {
    			fprintf(stderr, "Processing init: Handling RWY\n");

			parse_runway_record(f, argv, argc, &i);

		}

		/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
		 *
		 *  RWY2 : Runway record (new format)
		 *
		 *-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/

		else if (strcmp (argv[0], "RWY2") == 0) {
    			fprintf(stderr, "Processing init: Handling RWY2\n");

			parse_runway_record(f, argv, argc, &i);

		}

		/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
		 *
		 *  NAV: NAVAID record
		 *
		 *-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/

		else if (strcmp (argv[0], "NAV") == 0) {
    			fprintf(stderr, "Processing init: Handling NAV\n");

			/*
			 *  Parse Latitude/Longitude
			 */

			strcpy (lat_lon, argv[3]);
			strcat (lat_lon, argv[4]);
			DISStringToLatLong (lat_lon, &w);
			w.z = FEETtoMETERS (atoi (argv[5]));
			freq = strtod (argv[7], (char **) NULL);

			navaid_add_vor_dme_ndb(argv[1], argv[2], &w, argv[6], freq);
		}

		/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
		 *
		 *  ILS :  Instrument Landing System record
		 *
		 *-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/

		else if (strcmp (argv[0], "ILS") == 0) {
    			fprintf(stderr, "Processing init: Handling ILS\n");

			/*
			 *  Parse Latitude/Longitude of localizer transmitter
			 */

			strcpy (lat_lon, argv[5]);
			strcat (lat_lon, argv[6]);
			DISStringToLatLong (lat_lon, &w);
			w.z = FEETtoMETERS (atoi (argv[9]));

			/*
			 *  Is Glide Slope transmitter information present?
			 */

			if (strcmp (argv[7], "-") != 0) {
				strcpy (lat_lon, argv[7]);
				strcat (lat_lon, argv[8]);
				DISStringToLatLong (lat_lon, &gsw);
				gsw.z = w.z;
			}
			freq = strtod (argv[4], (char **) NULL);
			loc_width = strtod (argv[11], (char **) NULL);
			loc_bearing = strtod (argv[12], (char **) NULL);
			gs_angle = strtod (argv[13], (char **) NULL);

			navaid_add_ils(argv[3], argv[2], &w, &gsw, argv[10], freq,
				loc_width, loc_bearing, gs_angle);
		}

		/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
		 *
		 *  Now the simple terrain/man-made features
		 *
		 *-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/

		else if (strcmp (argv[0], "FEATURE") == 0) {
    			fprintf(stderr, "Processing init: Handling FEATURE\n");

			strcpy(file, argv[1]);
			strcpy(lat, argv[2]);
			strcpy(lon, argv[3]);
			z = atof(argv[4]);
			heading = atof(argv[5]);

			depthcue = 1;
			undersample = 0;

			if (file[0] == '@') {
				filep = &file[1];
				undersample = 1;
			}
			else if (file[0] == '+') {
				filep = &file[1];
				depthcue = 0;
			}
			else {
				filep = file;
			}
			stbl[i].type = CT_SURFACE;
			if ((c = lookupCraft (filep))) {
				stbl[i].cinfo = c;
			}
			else {
				char *p = strrchr (filep, '.');

				f1 = acm_fopen(filep, "r");
				stbl[i].cinfo = newCraft ();
				stbl[i].cinfo->name = STRDUP(filep);
				stbl[i].cinfo->object = NULL;
				if (p != NULL && (strcmp (p, ".dxf") == 0 ||
									strcmp (p, ".DXF") == 0)) {
					stbl[i].cinfo->object = VReadDepthCueuedDXFObject (f1,
								depthcue);
				}
				else {
					stbl[i].cinfo->object = VReadDepthCueuedObject (f1,
								depthcue);
				}

				if (!stbl[i].cinfo->object) {
					parse_error(f, "error reading object file");
					exit (1);
				}
				if (undersample) {
					undersampleObject (stbl[i].cinfo->object, 3);
				}
				fclose(f1);
			}
#ifdef FLAT_WORLD
			stbl[i].Sg.x = x * NM;
			stbl[i].Sg.y = y * NM;
			stbl[i].Sg.z = z;
#else
			strcat (lat, lon);
			DISStringToLatLong (lat, &stbl[i].w);
			stbl[i].w.z = FEETtoMETERS (z);
			DISWorldCoordinatesToGeocentric (&stbl[i].w,
							(void *) & stbl[i].Sg);
#endif
			stbl[i].curHeading = pm_normalize_yaw( DEGtoRAD(heading) );
			stbl[i].curPitch = stbl[i].curRoll = 0.0;
			if (++i == MAXSURFACE) {
				fprintf(stderr,
				   "readScene() ERROR: Only the first %d surface objects will be displayed.\n",
					 MAXSURFACE);
				file_close(f);
				return 0;
			}
		}

		/*
		 * Unknown record
		 */
		else {
    			fprintf(stderr, "Processing init: ignoring unknown record\n");
			parse_error(f, "readScene() unexpected record `%s' -- ignoring.", argv[0]);
		}

    }

    file_close(f);
    fprintf(stderr, "Finished   acmscene file %s\n", name);
    return 0;
}


int
init(char *objects)
{
    int i, size;
    char *s;
    craft *p;
    alarm_id_t id;
    char ground[256];
	VColor *haze;

    ttblCount = 0;

    if ( objects == NULL ) {
    	objects_dirs = MALLOC( sizeof(default_object_dirs) );
    	i = 0;
    	do {
    		if( default_object_dirs[i] == NULL ){
				objects_dirs[i] = NULL;
    			break;
    		}
			objects_dirs[i] = STRDUP( default_object_dirs[i] );
    		i++;
    	} while(1);

    } else {
    	size = 0;
		i = 0;
		objects_dirs = NULL;
		s = strtok(objects, ":");
		do {
			if ( i >= size ) {
				size = 2 * size + 10;
				objects_dirs = REALLOC(objects_dirs, size * sizeof( char * ));
			}
			if ( s == NULL ){
				/* REMEMBER: last elements of objects_dirs[] must be NULL: */
				objects_dirs[i++] = NULL;
				break;
			} else {
				objects_dirs[i++] = STRDUP(s);
			}
			s = strtok(NULL, ":");
		} while ( 1 );
    }

    cbase = -CLOUD_BASE;
    ctop = -CLOUD_TOP;

	if( render_lighting == RENDER_DAYLIGHT )
		haze = VAllocColor("#A6BBCD");
	else
		haze = VAllocColor("black");
		
	AlibSetVisibility(visibility, haze, depth_cue_steps);

    for ((i = 0, p = stbl); i < MAXSURFACE; (++i, ++p))
		p->type = CT_FREE;

    for ((i = 0, p = ptbl); i < MAXPLAYERS; (++i, ++p)) {
		p->pIndex = i;
		p->type = CT_FREE;
    }

    for ((i = 0, p = mtbl); i < MAXPROJECTILES; (++i, ++p)) {
		p->pIndex = i;
		p->type = CT_FREE;
    }

    damage_init();

        HUDColors[ 0] = VAllocColor ("#00ff00"); /* Brightest green    */
        HUDColors[ 1] = VAllocColor ("#00dd00"); /* Bright    green    */
        HUDColors[ 2] = VAllocColor ("#00aa00"); /*           green    */
        HUDColors[ 3] = VAllocColor ("#006600"); /* Dark      green    */
        HUDColors[ 4] = VAllocColor ("#ff0000"); /* Brightest red      */
        HUDColors[ 5] = VAllocColor ("#dd0000"); /* Bright    red      */
        HUDColors[ 6] = VAllocColor ("#aa0000"); /*           red      */
        HUDColors[ 7] = VAllocColor ("#660000"); /* Dark      red      */
        HUDColors[ 8] = VAllocColor ("#0000ff"); /* Brightest blue     */
        HUDColors[ 9] = VAllocColor ("#0000dd"); /* Bright    blue     */
        HUDColors[10] = VAllocColor ("#0000aa"); /*           blue     */
        HUDColors[11] = VAllocColor ("#000066"); /* Dark      blue     */
        HUDColors[12] = VAllocColor ("#ffffff"); /* Brightest gray     */
        HUDColors[13] = VAllocColor ("#bbbbbb"); /* Bright    gray     */
        HUDColors[14] = VAllocColor ("#777777"); /*           gray     */
        HUDColors[15] = VAllocColor ("#000000"); /* Dark      gray     */
	HUDColor = HUDColors[0];
	whiteColor = VAllocColor ("white");
	blackColor = VAllocColor ("black");
	yellowColor = VAllocColor ("#ee0");
	redColor = VAllocColor("#f44");
	radarColor = VAllocColor ("#0c0");
	HSIMagentaColor = VAllocColor ("magenta");
	radarBackgroundColor = VAllocColor ("#130");
	attitudeEarthColor = VAllocColor("#70462d");
	attitudeSkyColor = VAllocColor("#32b6e7");
	cloudColor = VAllocDepthCueuedColor ("#ccc", 1);

    effects_init();

    /*
     *  Set up the scene
     */

    if (readScene (ground, sizeof (ground)) != 0) {
		fprintf (stderr, "init() Fatal error reading scene file.\n");
		exit (1);
    }

/*  groundColor = VAllocDepthCueuedColor(ground, 1);	*/
    gtblIndex = gtblCount < 2 ? 0 : gtblCount - 1;
    groundColor = gtbl[gtblIndex].vcptr;

    /*
     *  Compile the aircraft inventory, DIS entity/munition maps
     */

    if (compileAircraftInventory () != 0) {
		fprintf (stderr, "init() Fatal error compiling aircraft inventory.\n");
		exit (1);
    }

    if (compileEntityMap("object-map.txt", &eo_map_count, &eo_map) != 0) {
		fprintf (stderr, "init() Fatal error\n");
		exit(1);
    }

    if (compileMunitionMap("munition-map.txt",
						   &mun_map_count, &mun_map) != 0) {
		fprintf (stderr, "init() Fatal error\n");
		exit(1);
    }

    /*
     *  Initialize weapons module
     */
	
	weapon_init();

    /*
     *  Add the periodic resupply check procedure
	 *  FIXME: resupply currently disabled
     */

    id = alarm_add(RESUPPLY_INTERVAL, resupply_doResupply, NULL, NULL);

    return 0;
}


void
freeAll(void)
{
	int i;

	for( i = 0; i < MAXPLAYERS; i++ ){
		if( ptbl[i].type != CT_FREE ){
			killPlayer(&ptbl[i], "program terminated");
		}
	}

	effects_free_all();  /* this also remove mtbl[].type == CT_EXPLOSION */

	for( i = 0; i < MAXPROJECTILES; i++ ){
		if( mtbl[i].type == CT_MISSILE ){
			killMissile(&mtbl[i], NULL);
		}
	}

	freeRendering();
	freeCraftTypes();
	navaid_free_all();
	dis_close();
	Vlib_FreeAll();
	runway_free_all();
	alarm_free_all();
	audio_free_all();
	aps_free_all();
	adf_free_all();
	draw_free_all();
	drone_free_all();
	hsi_free_all();
	hud_free_all();
	inst_free_all();
	magnetic_compass_free_all();
	prompt_free_all();
	terminal_free_all();
	panel_free_all();
	vpath_gallery_free_all();

	for( i=0; ; i++){
		if( objects_dirs[i] != NULL )
			FREE(objects_dirs[i]);
		else
			break;
	}
	FREE(objects_dirs);

	if( sceneFile != NULL )
		FREE(sceneFile);

	if( mun_map_count > 0 )
		FREE(mun_map);

	if( eo_map_count > 0 )
		FREE(eo_map);

	MEM_MANAGER_REPORT();
}
