/*
 *  acm : an aerial combat simulator for X
 *  Classic instruments module
 *  Copyright (C) 2007  Umberto Salsi
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 dated June, 1991.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program;  if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <math.h>
#include "aps.h"
#include "pm.h"
#include "Vlibmath.h"
#include "draw.h"
#include "vpath.h"


typedef struct _inst_data {
	struct _inst_data *next;

	_BOOL enabled;

	Color white, black;

	/* Anemometer: */
	double cos_alpha_stall;

	/* Used by inst_attitude_draw(): */
	double attitude_last_upd;
	double attitude_pitch_offset;
	VPoint attitude_gyro; /* gyro axis pointer, components in NED */

	/* Altimeter: */
	int altimeter_p0; /* altitude correction 2800-3100 inHg/100 */

	/* Vertical speed indicator (VSI): */
	double vsi_last_upd;
	double vsi_last_vs;

	/* Timer: */
	int timer_op;        /* timer operation: 0=off, 1=run, 2=stop */
	int timer_freezed;   /* store the time when timer_op is 2 */
	double timer_offset; /* store the offset time when timer_op is 1 or 2 */

} inst_data;

static inst_data *free_list = NULL;


#define INST(u) ((inst_data *)(u->inst))

static _BOOL blink = FALSE;
static double blink_toggle_time = 0.0;

/* These proptotypes apparently missing from math.h, I don't know why: */
double fmin(double x, double y);
double fmax(double x, double y);

/*
	TURN AND SLIP INDICATOR
	=======================
*/

static void inst_turnslip_draw(viewer *u, double xo, double yo, double width)
{
	AWindow *w;
	double radius, x1, y1, r, co, si, a, r1, r2, l, m;
	draw_data *dd;
	int j;

	w = u->v->w;
	radius = 0.48 * width;

	dd = draw_new();

	draw_circle(dd, xo, yo, radius);

	/*
		Yaw rate:
	*/
	x1 = xo;
	y1 = yo + 0.25 * radius;

	r = 0.90 * radius;
	for( j=-30; j<= 30; j+=15 ){
		co = cos( DEGtoRAD(j) );
		si = sin( DEGtoRAD(j) );
		draw_segment(dd, x1 + r*si, y1 - r*co,  x1 + radius*si, y1 - radius*co);
	}

	/*
		This instrument measures the rotational speed around the craft
		z axis:
	*/

	a = u->c->r;
	a = fmin(a, DEGtoRAD(5.0));
	a = fmax(a, -DEGtoRAD(5.0));
	draw_pointer(dd, x1, y1, -M_PI/2.0 + a*30.0/3.0, 0.87*radius);

	/*
		Slip indicator:
	*/

	x1 = xo;
	y1 = yo - 1.5*radius;

	/* curved glass: */
	r1 = 2.0*radius;
	r2 = 2.25*radius;
	l = DEGtoRAD(15.0);
	draw_arc(dd, x1, y1, r1, DEGtoRAD(90.0) - l, DEGtoRAD(90.0) + l);
	draw_arc(dd, x1, y1, r2, DEGtoRAD(90.0) - l, DEGtoRAD(90.0) + l);
	co = cos(l);
	si = sin(l);
	draw_segment(dd, x1 - r1*si, y1 + r1*co,  x1 - r2*si, y1 + r2*co);
	draw_segment(dd, x1 + r1*si, y1 + r1*co,  x1 + r2*si, y1 + r2*co);

	/* center marks of the curved glass: */
	co = cos(DEGtoRAD(4.0));
	si = sin(DEGtoRAD(4.0));
	draw_segment(dd, x1 - r1*si, y1 + r1*co,  x1 - r2*si, y1 + r2*co);
	draw_segment(dd, x1 + r1*si, y1 + r1*co,  x1 + r2*si, y1 + r2*co);

	/* ball: */
	a = atan2(-u->c->G.y, -u->c->G.z);
	/* max angular displacement of the ball (approximated for little angles): */
	m = l - (r2-r1)/(r1+r2);
	a = fmin(a, m);
	a = fmax(a, -m);
	co = cos(a);
	si = sin(a);
	draw_circle(dd, x1 + (r1+r2)/2.0*si, y1 + (r1+r2)/2.0*co, 0.95*(r2-r1)/2.0);

	draw_stroke(dd, u->v, w->pixel[whiteColor->cIndex]);
	draw_free(dd);
}


/*
	ANEMOMETER
	==========
*/


#define ANEMOMETER_MAX_VEL 530.0
#define ANEMOMETER_MAX_ANGLE DEGtoRAD(320.0)
#define ANEMOMETER_LINEARITY 0.8
#define ANEMOMETER_OFFSET 30.0


static void inst_anemometer_init(viewer * u)
{
	inst_data *d = INST(u);
	craftType *p = u->c->cinfo;

	d->cos_alpha_stall = cos(p->alpha_stall);
}


static double map_vel_to_angle(double vel)
{
	double a;
	static double k = 0.0;

	if( k == 0.0 )
		k = 1.0 / (sqrt(ANEMOMETER_MAX_VEL - ANEMOMETER_OFFSET + ANEMOMETER_LINEARITY*ANEMOMETER_LINEARITY) - ANEMOMETER_LINEARITY);
	vel = fmax(vel - ANEMOMETER_OFFSET, 0.0);
	a = k*(sqrt(vel + ANEMOMETER_LINEARITY*ANEMOMETER_LINEARITY) - ANEMOMETER_LINEARITY);
	a = fmin(a, 1.0);
	return a * ANEMOMETER_MAX_ANGLE;
}


static void inst_draw_bold_arc(draw_data *dd, double xo, double yo, double r,
	double a1, double a2)
{
	draw_arc(dd, xo, yo, r, a1, a2);
	draw_arc(dd, xo, yo, r - 1.0, a1, a2);
	draw_arc(dd, xo, yo, r + 1.0, a1, a2);
}


static VPath *anemometer_vpath = NULL;


static VPath * build_anemometer_vpath()
{
	int i, j, n, l;
	double radius, r, r1, r2, r3, fh, fw, si, co, vel, a;
	VPath *p;
	double v[200];
	char buf[10];
	VMatrix m;

	p = vpath_new();

	radius = 1.0;

	/*
		Build array v[] of the velocity to display
	*/
	n = 0;
	/* add speeds from 40 up to 150 kt step 10: */
	for( i = 40; i <= 250; i+= 10 )
		v[n++] = (double) i;
	/* add speeds from 200 up to ANEMOMETER_MAX_VEL kt step 50: */
	for( i = 200; i <= ANEMOMETER_MAX_VEL; i += 10 )
		v[n++] = (double) i;
	v[n++] = -1.0;

	i = 0;
	j = 0;
	r = 0.83 * radius;  /* labels center radius */
	r1 = 0.60 * radius; /* nock - inner radius */
	r2 = 0.65 * radius; /* short nock - outer radius */
	r3 = 0.72 * radius; /* long nock - outer radius */
	while( v[j] >= 0.0 ){
		vel = v[j];
		a = map_vel_to_angle(vel);
		co = cos(a);
		si = sin(a);
		
		if(
			(vel <= 120.0 && (int) vel % 20 == 0)
		||  ((int) vel % 50 == 0)
		){
			sprintf(buf, "%d", (int) vel);
			l = strlen(buf);
			VIdentMatrix(&m);
			fh = 0.10*radius;
			fw = 2.0 * fh / l;  /* same width (fw*l) for any length (l) */
			VScaleMatrix(&m, fw, fh, 1.0);
			VTranslate(&m, r*si - 0.5*fw*l, - r*co + 0.5*fh, 0.0);
			vpath_draw_string(p, buf, l, &m);

			vpath_moveTo(p, &(VPoint){r1*si, - r1*co, 0.0});
			vpath_lineTo(p, &(VPoint){r3*si, - r3*co, 0.0});
		} else {
			vpath_moveTo(p, &(VPoint){r1*si, - r1*co, 0.0});
			vpath_lineTo(p, &(VPoint){r2*si, - r2*co, 0.0});
		}

		j++;
	}


	/*
		"KT":
	*/

	strcpy(buf, "KT");
	l = strlen(buf);
	fh = 0.15 * radius;
	fw = fh;
	VIdentMatrix(&m);
	VScaleMatrix(&m, fw, fh, 1.0);
	VTranslate(&m, -0.5*fw*l, 0.3*radius + 0.5*fh, 0.0);
	vpath_draw_string(p, buf, l, &m);

	return p;
}


static void inst_anemometer_draw(viewer *u, double xo, double yo, double width)
{
	AWindow *w;
	double radius, a, b, r1, r2, vel;
	draw_data *dd;
	VMatrix m;

	w = u->v->w;
	radius = 0.48 * width;

	dd = draw_new();
	draw_circle(dd, xo, yo, radius);
	double cos_alpha_stall = INST(u)->cos_alpha_stall;

	/*
		Speed limit arcs
	*/

	craftType *p = u->c->cinfo;

	r1 = 0.55 * radius;
	r2 = 0.50 * radius;

	// White arc:
	if( p->Vs0 > 0.0 && p->Vfe > 0.0 ){
		inst_draw_bold_arc(dd, xo, yo, r1,
			map_vel_to_angle(p->Vs0 * cos_alpha_stall) - 0.5*pi,
			map_vel_to_angle(p->Vfe) - 0.5*pi);
	}
	if( p->Vs0 > 0.0 && p->Vs1 > 0.0 ){
		inst_draw_bold_arc(dd, xo, yo, r2,
			map_vel_to_angle(p->Vs0 * cos_alpha_stall) - 0.5*pi,
			map_vel_to_angle(p->Vs1 * cos_alpha_stall) - 0.5*pi);
	}
	
	// Green arc:
	if( p->Vs1 > 0.0 && p->Vfe > 0.0 && p->Vno > 0.0 ){
		draw_data *green = draw_new();
		b = map_vel_to_angle(p->Vno) - 0.5*pi;
		inst_draw_bold_arc(green, xo, yo, r1,
			map_vel_to_angle(p->Vfe) - 0.5*pi, b);
		inst_draw_bold_arc(green, xo, yo, r2,
			map_vel_to_angle(p->Vs1 * cos_alpha_stall) - 0.5*pi, b);
		draw_stroke(green, u->v, u->w->pixel[radarColor->cIndex]);
		draw_free(green);
	}

	// Yellow arc:
	if( p->Vno > 0.0 && p->Vne > 0.0 ){
		draw_data *yellow = draw_new();
		a = map_vel_to_angle(p->Vno) - 0.5*pi;
		b = map_vel_to_angle(p->Vne) - 0.5*pi;
		inst_draw_bold_arc(yellow, xo, yo, r1, a, b);
		inst_draw_bold_arc(yellow, xo, yo, r2, a, b);
		draw_stroke(yellow, u->v, u->w->pixel[yellowColor->cIndex]);
		draw_free(yellow);
	}

	// Red line (actually here we draw a short arc):
	if( p->Vne > 0.0 ){
		a = map_vel_to_angle(p->Vne);
		draw_data *red = draw_new();
		inst_draw_bold_arc(red, xo, yo, r1,
			a - 0.5*pi,
			a + DEGtoRAD(3) - 0.5*pi);
		inst_draw_bold_arc(red, xo, yo, r2,
			a - 0.5*pi,
			a + DEGtoRAD(3) - 0.5*pi);
		draw_stroke(red, u->v, u->w->pixel[redColor->cIndex]);
		draw_free(red);
	}

	if( anemometer_vpath == NULL )
		anemometer_vpath = build_anemometer_vpath();
	
	VIdentMatrix(&m);
	VScaleMatrix(&m, radius, radius, radius);
	VTranslate(&m, xo, yo, -1.0);
	vpath_stroke(anemometer_vpath, &m, u->w, INST(u)->white);

	/*
		Draw pointer:
	*/

	vel = FPStoKT( u->c->VT * cos(u->c->alpha) * cos(u->c->beta) );
	if( vel < ANEMOMETER_OFFSET )
		vel = ANEMOMETER_OFFSET;
	a = map_vel_to_angle( vel );
	draw_pointer(dd, xo, yo, a - M_PI/2.0, 0.60 * radius);
	draw_stroke(dd, u->v, INST(u)->white);
	draw_free(dd);
}


/*
	ATTITUDE INDICATOR
	==================
*/

#define ATTITUDE_MAX_PITCH_OFFSET DEGtoRAD(10.0)
#define ATTITUDE_MAX_PITCH DEGtoRAD(70.0)
#define ATTITUDE_MAX_BANK  DEGtoRAD(70.0)


static VPath * wings_vpath = NULL;


static void inst_attitude_init(viewer *u)
{
	INST(u)->attitude_pitch_offset = 0.0;

	/* Init gyro randomly oriented: */
	INST(u)->attitude_gyro.x = 0.2;
	INST(u)->attitude_gyro.y = 0.2;
	INST(u)->attitude_gyro.z = -0.959;

	INST(u)->attitude_last_upd = curTime;
}


void inst_attitude_adjust_pitch(viewer * u, double delta)
{
	if( u->inst == NULL )
		return;

	INST(u)->attitude_pitch_offset += delta;
	INST(u)->attitude_pitch_offset = fmin(INST(u)->attitude_pitch_offset,
		ATTITUDE_MAX_PITCH_OFFSET);
	INST(u)->attitude_pitch_offset = fmax(INST(u)->attitude_pitch_offset,
		-ATTITUDE_MAX_PITCH_OFFSET);
}


void inst_attitude_reset(viewer * u)
{
	VPoint g;

	if( u->inst == NULL )
		return;

	g.x = 0.0;  g.y = 0.0;  g.z = -1.0;
	VTransform_(&g, &u->c->trihedral, &INST(u)->attitude_gyro);
}


static VPath * ball_ladder_vpath = NULL;


static void build_ball_draw_arc(
	double pitch1, double yaw1,
	double pitch2, double yaw2)
/*
	Draw arc on attitude ball. The arc starts from (pitch1,yaw1) and ends on
	(pitch2,yaw2). Longer arcs are splitted in two or more segments.
*/
{
	int n, i;
	VPoint a, b, c;
	double m;

	pitch1 = DEGtoRAD(pitch1);
	yaw1  = DEGtoRAD(yaw1);
	pitch2 = DEGtoRAD(pitch2);
	yaw2  = DEGtoRAD(yaw2);

	n = (int) (fmax( fabs(pitch2-pitch1), fabs(yaw2-yaw1) )
		/ DEGtoRAD(5) + 0.5);

	if( n < 1 )
		n = 1;
	
	VSetPoint(&a, cos(pitch1)*sin(yaw1), -sin(pitch1), -cos(pitch1)*cos(yaw1));
	VSetPoint(&b, cos(pitch2)*sin(yaw2), -sin(pitch2), -cos(pitch2)*cos(yaw2));

	vpath_moveTo(ball_ladder_vpath, &a);
	
	for( i=1; i<=n; i++ ){

		VSetPoint(&c, a.x + (double)i/n*(b.x-a.x),
			a.y - (double)i/n*(b.y-a.y),
			a.z + (double)i/n*(b.z-a.z));

		m = Vmagnitude(&c);

		VSetPoint(&c, c.x/m, c.y/m, c.z/m);

		vpath_lineTo(ball_ladder_vpath, &c);
	}
}


static void build_ball_draw_str(int pitch, int yaw)
{
	VMatrix m;
	double fh, fw;
	char s[10];
	int s_len;

	sprintf(s, "%d", abs(pitch));
	s_len = strlen(s);

	fh = 0.05;
	fw = 0.05;

	VIdentMatrix(&m);
	VScaleMatrix(&m, fw, fh, 1.0);
	VTranslate(&m, -s_len*fw/2, fh/2, -1.0);
	VRotate(&m, YRotation, DEGtoRAD(yaw));
	VRotate(&m, XRotation, DEGtoRAD(-pitch));
	vpath_draw_string(ball_ladder_vpath, s, s_len, &m);

	VIdentMatrix(&m);
	VScaleMatrix(&m, fw, fh, 1.0);
	VTranslate(&m, -s_len*fw/2, fh/2, -1.0);
	VRotate(&m, YRotation, DEGtoRAD(-yaw));
	VRotate(&m, XRotation, DEGtoRAD(-pitch));
	vpath_draw_string(ball_ladder_vpath, s, s_len, &m);
}


static void build_ball_ladder()
{
	ball_ladder_vpath = vpath_new();

	/* Pitch ladder: */
	build_ball_draw_arc( 30,-15,  30,+15);  build_ball_draw_str( 30, 20 );
	build_ball_draw_arc( 25, -3,  25, +3);
	build_ball_draw_arc( 20,-10,  20,+10);  build_ball_draw_str( 20, 15 );
	build_ball_draw_arc( 15, -3,  15, +3);
	build_ball_draw_arc( 10, -5,  10, +5);  build_ball_draw_str( 10, 10 );
	build_ball_draw_arc(  5, -3,   5, +3);
	build_ball_draw_arc(  0,-45,   0,+45);
	build_ball_draw_arc( -5, -3,  -5, +3);
	build_ball_draw_arc(-10, -5, -10, +5);  build_ball_draw_str(-10, 10 );
	build_ball_draw_arc(-15, -3, -15, +3);
	build_ball_draw_arc(-20,-10, -20,+10);  build_ball_draw_str(-20, 15 );
	build_ball_draw_arc(-25, -3, -25, +3);
	build_ball_draw_arc(-30,-15, -30,+15);  build_ball_draw_str(-30, 20 );
}


static VPath * build_wings()
{
	VPath * wings = vpath_new();
	vpath_moveTo(wings, &(VPoint){-0.5, 0.0, 0.0});
	vpath_lineTo(wings, &(VPoint){-0.2, 0.0, 0.0});
	vpath_lineTo(wings, &(VPoint){-0.1, 0.1, 0.0});
	vpath_lineTo(wings, &(VPoint){0.0, 0.0, 0.0});
	vpath_lineTo(wings, &(VPoint){0.1, 0.1, 0.0});
	vpath_lineTo(wings, &(VPoint){0.2, 0.0, 0.0});
	vpath_lineTo(wings, &(VPoint){0.5, 0.0, 0.0});
	return wings;
}


static void inst_attitude_update(viewer *u)
{
	VPoint *gyro, g;
	double m, k;

	if( curTime < INST(u)->attitude_last_upd + 0.2 )
		return;

	/*
		Erection system: the gyro axis (smoothly) follows the local
		vertical:
	*/
	gyro = &INST(u)->attitude_gyro;
	VTransform_(&u->c->G, &u->c->trihedral, &g);
	m = Vmagnitude(&g);
	if( m > 0.98  &&  m < 1.02 /* erection cut-out */){
		k = 0.003 / m;
		g.x = gyro->x + k * g.x;
		g.y = gyro->y + k * g.y;
		g.z = gyro->z + k * g.z;
		m = Vmagnitude(&g);
		if( m > 1e-6 ){
			k = 1.0 / m;
			gyro->x = k * g.x;
			gyro->y = k * g.y;
			gyro->z = k * g.z;
		}
	}
	INST(u)->attitude_last_upd = curTime;
}


static void inst_attitude_sky_earth_colors(viewer * u, VMatrix *R)
{
	AWindow *w;
	VPoint p;
	Point sky[20], earth[20];
	double a;
	int i;

	w = u->v->w;
	i = 0;

	/*
		Add points of the horizon line
	*/
	for( a=-M_PI/4.0; a <= M_PI/4.0 + 1e-5; a += M_PI/18 ){
		p = (VPoint) {sin(a), 0.0, -cos(a)};
		VTransform(&p, R, &p);
		sky[i].x = p.x + 0.5;  sky[i].y = p.y + 0.5;
		earth[i].x = p.x + 0.5;  earth[i].y = p.y + 0.5;
		i++;
	}

	/*
		Close polygons at arbitrary upper/lower points
	*/
	p = (VPoint) {0.707, -0.707, -0.2};
	VTransform(&p, R, &p);
	sky[i].x = p.x;  sky[i].y = p.y;

	p = (VPoint) {0.707, 0.707, -0.2};
	VTransform(&p, R, &p);
	earth[i].x = p.x;  earth[i].y = p.y;
	i++;

	p = (VPoint) {-0.707, -0.707, -0.2};
	VTransform(&p, R, &p);
	sky[i].x = p.x;  sky[i].y = p.y;

	p = (VPoint) {-0.707, 0.707, -0.2};
	VTransform(&p, R, &p);
	earth[i].x = p.x;  earth[i].y = p.y;
	i++;

	if( i > 20 )
		abort();

	AlibFillPolygon(w, sky, i, attitudeSkyColor->cIndex);

	AlibFillPolygon(w, earth, i, attitudeEarthColor->cIndex);

}


static void inst_attitude_draw(viewer * u, double xo, double yo, double width)
{
	AWindow *w;
	VPoint g;
	double pitch, bank, l;
	_BOOL hit_limit;
	int j;
	draw_data *dd;
	double radius, a, r, co, si;
	VMatrix R;
	Rect rect;

	w = u->v->w;

	/*
		Compute gyro angles of bank and pitch:
	*/

	VReverseTransform_(&INST(u)->attitude_gyro, &u->c->trihedral, &g);
	bank = atan2(-g.y, -g.z); /* positive turning right */
	pitch = atan2(g.x, sqrt(g.y*g.y + g.z*g.z)); /* positive pulling up */

	hit_limit = FALSE;

	if( bank > ATTITUDE_MAX_BANK ){
		hit_limit = TRUE;
		bank = ATTITUDE_MAX_BANK;
	} else if( bank < -ATTITUDE_MAX_BANK ){
		hit_limit = TRUE;
		bank = -ATTITUDE_MAX_BANK;
	}

	if( pitch > ATTITUDE_MAX_PITCH ){
		hit_limit = TRUE;
		pitch = ATTITUDE_MAX_PITCH;
	} else if( pitch < -ATTITUDE_MAX_PITCH ){
		hit_limit = TRUE;
		pitch = -ATTITUDE_MAX_PITCH;
	}

	if( hit_limit ){
		g.x = sin(pitch) * cos(bank);
		g.y = -sin(bank);
		g.z = -cos(pitch) * cos(bank);
		VTransform_(&g, &u->c->trihedral, &INST(u)->attitude_gyro);
	}

	radius = 0.48 * width;

	/*
		Draw horizon and pitch scale. First, build a rotational matrix
		acting on vectors of the screen frame: x axis pointing right,
		y axis down, z forward:
	*/

	SetRect(&rect, (int)(xo - 0.3*width), (int)(yo - 0.25*width),
		(int)(xo + 0.3*width), (int)(yo + 0.25*width));
	AlibSetClipRect(w, &rect);

	r = 0.60 * width;  /* radius of the rotating ball */

	VIdentMatrix(&R);
	VRotate(&R, XRotation, pitch);
	VRotate(&R, ZRotation, -bank);
	VScaleMatrix(&R, r, r, r);
	VTranslate(&R, xo, yo, 0.0);

	inst_attitude_sky_earth_colors(u, &R);

	if( ball_ladder_vpath == NULL )
		build_ball_ladder();
	
	vpath_stroke(ball_ladder_vpath, &R, w, whitePixel);

	dd = draw_new();
	draw_free(dd);

	/*
		Aircraft frame:
	*/

	if( wings_vpath == NULL )
		wings_vpath = build_wings();

	VIdentMatrix(&R);
	VScaleMatrix(&R, radius, radius, 1.0);
	VTranslate(&R, xo, yo - radius * INST(u)->attitude_pitch_offset, -1.0);
	vpath_stroke(wings_vpath, &R, u->w, w->pixel[blackColor->cIndex]);
	VTranslate(&R, 0.0, 1.0, 0.0);
	vpath_stroke(wings_vpath, &R, u->w, w->pixel[blackColor->cIndex]);

	/*
		Draw bank scale:
	*/

	AlibSetClipRect(w, &u->attitude);
	dd = draw_new();

	draw_circle(dd, xo, yo, radius);

	r = 0.85 * radius;
	draw_segment(dd,  xo, yo - radius,  xo, yo - r );
	for( j = 10; j <= 90; j+=10 ){
		a = DEGtoRAD(j);
		co = cos(a);
		si = sin(a);
		if( j <= 30 || j % 30 == 0 ){
			if( j % 30 == 0 )
				r = 0.85 * radius;
			else
				r = 0.92 * radius;
			draw_segment(dd,  xo + si*radius, yo - co*radius, xo + si*r, yo - co*r );
			draw_segment(dd,  xo - si*radius, yo - co*radius, xo - si*r, yo - co*r );
		}
	}

	draw_stroke(dd, u->v, (unsigned short) (w->pixel[whiteColor->cIndex]));
	draw_free(dd);

	/*
		Draw bank pointer:
	*/

	r = 0.80 * radius;
	l = 0.68 * radius;
	dd = draw_new();
	draw_segment(dd,  xo + r*sin(bank), yo - r*cos(bank),
		xo + l*sin(bank - DEGtoRAD(5.0)), yo - l*cos(bank - DEGtoRAD(5.0)) );
	draw_segment(dd,  xo + r*sin(bank), yo - r*cos(bank),
		xo + l*sin(bank + DEGtoRAD(5.0)), yo - l*cos(bank + DEGtoRAD(5.0)) );
	draw_segment(dd,  xo + l*sin(bank - DEGtoRAD(5.0)), yo - l*cos(bank - DEGtoRAD(5.0)),
		xo + l*sin(bank + DEGtoRAD(5.0)), yo - l*cos(bank + DEGtoRAD(5.0)) );

	draw_stroke(dd, u->v, w->pixel[whiteColor->cIndex]);
	draw_free(dd);
}


/*
	ALTIMETER
	=========
*/


static void inst_altimeter_init(viewer *u)
{
	INST(u)->altimeter_p0 = 2992;
}


void inst_altimeter_correction(viewer *u, int delta)
{
	int p0;

	if( u->inst == NULL )
		return;

	p0 = INST(u)->altimeter_p0;

	p0 += delta;
	if( p0 < 2800 )
		p0 = 2800;
	else if( p0 > 3100 )
		p0 = 3100;
	
	INST(u)->altimeter_p0 = p0;
}


static void inst_altimeter_draw(viewer * u, double xo, double yo, double width)
{
	AWindow *w;
	double alt, radius, r, r1, r2, a, fw, fh, co, si, x;
	int j, hi, lo, p0;
	draw_data *dd;
	char buf[20];

	w = u->v->w;
	AlibSetClipRect(w, &u->altimeter);

	radius = 0.48 * width;

	dd = draw_new();

	draw_circle(dd, xo, yo, radius);

	j = 0;
	r = 0.85 * radius;  /* tick mark with number */
	r1 = 0.92 * radius; /* tick mark without number */
	r2 = 0.70 * radius; /* center circonf. for numbers */
	fh = 0.15 * radius;
	for( a = 0.0; a < DEGtoRAD(359.9); a += DEGtoRAD(360.0/20.0) ){
		co = cos(a);
		si = sin(a);
		if( j % 10 == 0 ){
			sprintf(buf, "%d", j/10);
			draw_string_centered(u->v, xo + r2*si, yo - r2*co, fh, buf, INST(u)->white);
			draw_segment(dd, xo + r*si, yo - r*co,   xo + radius*si, yo - radius*co);
		} else {
			draw_segment(dd, xo + r1*si, yo - r1*co,   xo + radius*si, yo - radius*co);
		}
		j = j+5;
	}

	/* FIXME: what if alt<0? */

	alt = METERStoFEET(u->c->w.z);
	/* apply altitude correction: */
	p0 = INST(u)->altimeter_p0;
	alt = alt + 949.9 * (p0*0.01 - 29.92);
	if( alt < -0.5 )
		alt = 1e5+alt;

	j = (int) (alt + 0.5);
	j = (j + 5) / 10 * 10;
	lo = j % 100;
	hi = j / 100;

	a = fmod(alt, 1000.0) * M_PI / 500.0;

	draw_pointer(dd, xo, yo, a - DEGtoRAD(90.0), 0.6*radius);

	fh = 0.18*radius;
	fw = VFontWidthPixels(u->v, (int) (fh+0.5));
	sprintf(buf, "%03d", hi);
	x = fw * 1.0;
	VDrawStrokeString(u->v,
		(int) (xo - fw * (strlen(buf)-1) + 0.5),
		(int) (yo - 0.2*radius + 0.5),
		buf, strlen(buf), (int)(fh+0.5), INST(u)->white);

	fh = 0.12*radius;
	fw = VFontWidthPixels(u->v, (int)(fh+0.5));
	sprintf(buf, "%02d", lo);
	VDrawStrokeString(u->v,
		(int) (xo + x + 0.5),
		(int) (yo - 0.2*radius + 0.5),
		buf, strlen(buf), (int)(fh+0.5), INST(u)->white);

	/*
		Draw altitude correction:
	*/
	draw_string_centered(u->v, xo, yo + 0.1*radius, 0.10*radius, "inHg   hPa", INST(u)->white);
	sprintf(buf, "%2.2f %04d", 0.01*p0, (int)(1013.25/29.92 * 0.01 * p0 + 0.5));
	draw_string_centered(u->v, xo, yo+0.30*radius, 0.10*radius, buf, INST(u)->white);

	draw_string_centered(u->v, xo, yo-0.5*radius, 0.10*radius, "FEET", INST(u)->white);

	draw_stroke(dd, u->v, w->pixel[whiteColor->cIndex]);
	draw_free(dd);
}


/*
	Vertical speed indicator (VSI)
	==============================
*/

static double vario[] = {0.0, 100.0, 200.0, 300.0, 400.0, 500.0, 1000.0,
	2000.0, 3000.0, 4000.0, -1.0};


static double map_vario_to_angle(double v)
{
	static double k = 0.0;
	double a;

	if( k == 0.0 ){
		/* limit of the scale at 4000 ft/min: */
		k = M_PI / (sqrt(4500.0 + 15.0*15.0) - 15.0);
	}

	v = fmin(v, 4000.0);
	v = fmax(v, -4000.0);

	a = k * (sqrt(fabs(v) + 15.0*15.0) - 15.0);

	if( v < 0.0 )
		a = -a;
	
	return a;
}


static void inst_vsi_init(viewer *u)
{
	INST(u)->vsi_last_upd = curTime;
	INST(u)->vsi_last_vs = - u->c->Cg.z * 60.0;   /* vertical speed, ft/min */
}


#define VSI_K (1.0/6.0)
/*
	If the vertical speed changes abruptly from vs(0) to some vs_actual,
	then in our model of the instrument lag the indicated vertical speed
	will be:

		vs(t) = (vs(0) - vs_actual) * exp(-VSI_K*t) + vs_actual
	
	where 1/VSI_K is the "time constant" (i.e. an extimation of the lag time).
*/


static void inst_vsi_update(viewer *u)
{
	double vs, dt;

	vs = - u->c->Cg.z * 60.0;    /* vertical speed, ft/min */
	dt = curTime - INST(u)->vsi_last_upd;
	INST(u)->vsi_last_vs += VSI_K * (vs - INST(u)->vsi_last_vs) * dt;
	INST(u)->vsi_last_upd = curTime;
}


static void inst_vsi_draw(viewer * u, double xo, double yo, double width)
{
	AWindow *w;
	double climb, radius, r, r1, r2, r3, a, fw, fh, co, si;
	int j;
	draw_data *dd;
	char buf[20];

	w = u->v->w;
	AlibSetClipRect(w, &u->vsi);

	radius = 0.48 * width;

	dd = draw_new();

	draw_circle(dd, xo, yo, radius);

	r1 = 0.87 * radius; /* tick mark with number */
	r2 = 0.92 * radius; /* tick mark without number */
	r3 = 0.70 * radius; /* center of the number */
	fh = 0.12 * radius;
	fw = VFontWidthPixels(u->v, (int)(fh+0.5));
	j = 0;
	while( vario[j] >= 0 ){
		climb = vario[j];
		a = map_vario_to_angle(climb);
		co = cos(a);
		si = sin(a);
		sprintf(buf, "%.0f", climb/100);
		if( (int) (climb + 0.5) % 500 == 0 ){
			draw_string_centered(u->v, xo - r3*co, yo - r3*si, fh, buf, INST(u)->white);
			draw_string_centered(u->v, xo - r3*co, yo + r3*si, fh, buf, INST(u)->white);
			r = r1;
		} else {
			r = r2;
		}
		draw_segment(dd, xo - r*co, yo - r*si,   xo - radius*co, yo - radius*si);
		/* FIXME: because of a bug in Vlib, drawing a segment over another
		segment results in no segment drawn at all! This if() fixes that: */
		if( j != 0 )
			draw_segment(dd, xo - r*co, yo + r*si,   xo - radius*co, yo + radius*si);
		j++;
	}

	a = map_vario_to_angle(INST(u)->vsi_last_vs);
	draw_pointer(dd, xo, yo, a - M_PI, 0.80*radius);

#ifdef DEBUG
	/* actual v.s.: */
	climb = - u->c->Cg.z * 60.0;    /* climb ratio, ft/min */
	a = map_vario_to_angle(climb);
	draw_pointer(dd, xo, yo, a - M_PI, 0.40*radius);
#endif

	draw_string_centered(u->v, xo, yo-0.3*radius, 0.15*radius, "FPM", INST(u)->white);
	draw_string_centered(u->v, xo, yo+0.3*radius, 0.12*radius, "x100", INST(u)->white);

	draw_stroke(dd, u->v, w->pixel[whiteColor->cIndex]);
	draw_free(dd);
}


/*
	AutoPilot System lights panel
	=============================
*/

static void inst_ap_panel(viewer *u, double x1, double y1, double x2, double y2)
{
	AWindow *w;
	double margin, width, h, fh;

	void draw_light(char *label, int order, _BOOL on)
	{
		double xa, ya, xb, yb;
		Rect r;

		xa = x1 + order*width + margin;
		ya = y1 + margin;
		xb = x1 + order*width + width - margin;
		yb = y1 + h - margin;
		if( on ){
			SetRect(&r, xa, ya, xb, yb);
			AlibFillRectangle(w, &r, HSIMagentaColor->cIndex);
		}
		draw_string_centered(u->v, (xa+xb)/2.0, (ya+yb)/2.0,
			fh, label, INST(u)->black);
	}


	w = u->w;
	AlibSetClipRect(w, &u->stripe);

	/* size of every light: */
	width = (x2-x1)/6.0;
	h = y2-y1;

	/* margin around each light: */
	margin = 0.1*h;

	/* font height: */
	fh = 0.5 * (h - 2.0*margin);

	draw_light("Pilot", 0, aps_ap_enabled(u->c)
		&& ( ! aps_ap_warn(u->c) || ! blink ));

	draw_light("Nav", 1, aps_an_enabled(u->c)
		&& ( ! aps_an_warn(u->c) || ! blink ));

	draw_light("Land", 2, aps_al_enabled(u->c)
		&& ( ! aps_al_warn(u->c) || ! blink ));

	draw_light("Turn", 3, aps_aw_enabled(u->c)
		&& ( ! aps_aw_warn(u->c) || ! blink ));

	draw_light("Thr", 4, aps_at_enabled(u->c)
		&& ( ! aps_at_warn(u->c) || ! blink ));

	draw_light("Coord", 5, aps_ac_enabled(u->c)
	    && ( ! aps_ac_warn(u->c) || ! blink ));
}


/*
	TIMER
	=====
*/


static void inst_timer_init(viewer *u)
{
	INST(u)->timer_op = 0; /* do not display */
}


void inst_timer_toggle(viewer *u)
{
	if( u->inst == NULL )
		return;

	INST(u)->timer_op = (INST(u)->timer_op + 1) % 3;

	if ( INST(u)->timer_op == 1 ) {
		INST(u)->timer_offset = curTime;

	} else if ( INST(u)->timer_op == 2 ) {
		INST(u)->timer_freezed = (int) (curTime - INST(u)->timer_offset + 0.5);

	}
}


static void inst_timer_draw(viewer *u, double x1, double y1, double x2, double y2)
{
	AWindow *w;
	int t, l;
	double h, fh, fw;
	char buffer[20];
	Rect rect;

	if ( INST(u)->timer_op == 0 ) {
		return;

	} else if ( INST(u)->timer_op == 1 ) {
		t = (int) (curTime - INST(u)->timer_offset);

	} else {
		t = INST(u)->timer_freezed;
	}

	if (t >= 3600) {
		int h, m, s;
		h = t / 3600;
		t = t % 3600;
		m = t / 60;
		s = t % 60;
		sprintf(buffer, "%d:%02d:%02d", h, m, s);
	} else if (t >= 60) {
		int m, s;
		m = t / 60;
		s = t % 60;
		sprintf(buffer, "%d:%02d", m, s);
	} else {
		sprintf(buffer, "%d", t);
	}

	/*
		Background:
	*/
	w = u->v->w;
	h = y2 - y1;
	SetRect(&rect, x1, y1, x2, y2);
	AlibSetClipRect(w, &rect);
	
	fh = h/2.0;
	fw = VFontWidthPixels(u->v, (int)(fh+0.5));
	l = strlen(buffer);

	VDrawStrokeString(u->v,
		(int) (x2 - (l+2)*fw + 0.5),
		(int) (y1 + h/2.0 + fh/2.0 + 0.5),
		buffer, l, (int)(fh+0.5), INST(u)->white);
}


void inst_draw(viewer * u)
{
	AWindow *w;
	double width;
	Rect rect;

	if( u->inst == NULL || ! INST(u)->enabled )
		return;
	
	w = u->v->w;

	/* We have only the flux gate, not an inertial platform: */
	u->c->showMag = TRUE;

	inst_attitude_update(u);
	inst_vsi_update(u);
	
	/* Update blink flag: */
	if( curTime >= blink_toggle_time ){
		blink = ! blink;
		if( blink )
			blink_toggle_time = curTime + 0.1;
		else
			blink_toggle_time = curTime + 0.3;
	}

	/* Clear area of the classic instruments: */
	/* FIXME -- every instrument must take care of its own area */
	SetRect(&rect, u->stripe.a.x, u->stripe.a.y, u->engine.b.x, u->engine.b.y);
	AlibSetClipRect(w, &rect);
	AlibFillRectangle(w, &rect, radarBackgroundColor->cIndex);

	/* FIXME: should we show the instruments while looking elsewere?
	if( u->viewDirection.x < 0.90 )
		return;
	*/

	width = RectWidth(u->vsi); /* FIXME: see FIXME above */

	/* Instruments that occupy a standard slot: */
	inst_turnslip_draw  (u, u->turn.a.x+width/2.0, u->turn.a.y+width/2.0, width);
	inst_anemometer_draw(u, u->anemometer.a.x+width/2.0, u->anemometer.a.y+width/2.0, width);
	inst_attitude_draw  (u, u->attitude.a.x+width/2.0, u->attitude.a.y+width/2.0, width);
	inst_altimeter_draw (u, u->altimeter.a.x+width/2.0, u->altimeter.a.y+width/2.0, width);
	inst_vsi_draw       (u, u->vsi.a.x+width/2, u->vsi.a.y+width/2, width);

	/* Timer and auto-pilot system lights goes in the stripe: */
	inst_timer_draw     (u, u->stripe.a.x, u->stripe.a.y, width, u->stripe.b.y);
	inst_ap_panel       (u, width, u->stripe.a.y, u->stripe.b.x, u->stripe.b.y);

#ifdef DEBUG
	{ /* FIXME -- Alib bugs tests */
	Rect r;
	Point p[4];

	AlibSetClipRect(u->w, &u->w->rect);

	SetRect(&r, 10, 10, 20, 20);
	AlibDrawRect(u->w, &r, whitePixel);
	ExpandRect(&r, -1, -1);
	AlibSetClipRect(u->w, &r);
	/****
	p[0] = (Point){11, 11};
	p[1] = (Point){18, 11};
	p[2] = (Point){18, 18};
	p[3] = (Point){11, 18};
	****/
	p[0] = (Point){-11, -11};
	p[1] = (Point){180, -11};
	p[2] = (Point){180, 180};
	p[3] = (Point){-11, 180};
	AlibFillPolygon(u->w, p, 4, blackPixel);
	AlibSetClipRect(u->w, &u->w->rect);

	SetRect(&r, 30, 10, 40, 20);
	AlibDrawRect(u->w, &r, whitePixel);
	AlibDrawLine(u->w, 30, 10, 39, 19, blackPixel);
	AlibDrawLine(u->w, 39, 10, 30, 19, blackPixel);
	AlibDrawLine(u->w, 41, 10, 41, 19, blackPixel);
	AlibDrawLine(u->w, 30, 21, 39, 21, blackPixel);

	SetRect(&r, 10, 30, 20, 40);
	AlibDrawRect(u->w, &r, whitePixel);
	p[0] = (Point){15, 30};
	p[1] = (Point){19, 35};
	p[2] = (Point){15, 39};
	p[3] = (Point){10, 35};
	AlibFillPolygon(u->w, p, 4, blackPixel);

	SetRect(&r, 30, 30, 40, 40);
	AlibDrawRect(u->w, &r, whitePixel);
	ExpandRect(&r, -1, -1);
	AlibFillRectangle(u->w, &r, blackPixel);

	}
#endif

}


void inst_enable(viewer * u)
{
	AWindow *w;

	if( u->inst == NULL ){
		if( free_list == NULL ){
			u->inst = MALLOC( sizeof(inst_data) );
		} else {
			u->inst = free_list;
			free_list = free_list->next;
			INST(u)->next = NULL;
		}
		w = u->v->w;
		INST(u)->white = whiteColor->cIndex;
		INST(u)->black = blackColor->cIndex;
		inst_anemometer_init(u);
		inst_attitude_init(u);
		inst_altimeter_init(u);
		inst_vsi_init(u);
		inst_timer_init(u);
	}
	INST(u)->enabled = TRUE;
}


void inst_disable(viewer * u)
{
	if( u->inst == NULL )
		return;
	INST(u)->enabled = FALSE;
}


_BOOL inst_enabled(viewer * u)
{
	return (u->inst != NULL) && INST(u)->enabled;
}


void inst_free(viewer *u)
{
	if( u->inst == NULL )
		return;
	
	INST(u)->next = free_list;
	free_list = u->inst;
	u->inst = NULL;
}


void inst_free_all()
{
	inst_data *p;

	while( free_list != NULL ){
		p = free_list;
		free_list = free_list->next;
		FREE(p);
	}

	vpath_free(ball_ladder_vpath);
	ball_ladder_vpath = NULL;
	vpath_free(wings_vpath);
	wings_vpath = NULL;

	vpath_free(anemometer_vpath);
	anemometer_vpath = NULL;
}

/* End of the instruments module. */
