/*
 *  acm : an aerial combat simulator for X
 *  Magnetic compass module
 *  Copyright (C) 2007  Umberto Salsi
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 dated June, 1991.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program;  if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
 */

/*
	This module simulates the magnetic compass for didactic purposes.

	The compass rotates under the effect of the torque produced by the Earth
	magnetic field and it is braked by its inertia and by the viscosity of
	the fluid.
	
	A simple model for the Earth magnetic field is also implemented, with
	the magnetic deviation guessed from the nearest NAVAID and the magnetic
	dip guessed from the latitude. Since the resulting magnetic vector has a
	vertical component, the angle of bank causes the rotation of the compass,
	that exhibits all those interesting behaviors pilots must account for.

	References
	----------
	http://www.aero.polimi.it/Antasme/Deliverable71.pdf
		Design description of the library for Space Flight Dynamics
	http://gsc.nrcan.gc.ca/geomag/field/magref_e.php
	http://gravmag.ou.edu/mag_earth/mag_earth.html
		These sites describe the Earth magnetic field and provide a
		mathematical model for it. Actually, this module implements a
		simplified model, faster and precise enough for our purposes.
	http://www.pilotfriend.com/training/flight_training/fxd_wing/comp.htm
		This site describes the magnetic compass and its behavior from the
		point of view of the pilot.
*/

#include <string.h>
#include <math.h>
#include "pm.h"
#include "draw.h"
#include "navaid.h"
#include "vpath.h"


typedef struct _magnetic_compass_data {
	struct _magnetic_compass_data *next;

	_BOOL enabled;

	double last_time;  /* last update */
	double hdg;        /* current indicated value (RAD) */
	double hdg_dot;    /* current rotation speed (RAD/s) */
	double x_rot, y_rot;  /* compass roll and pitch angles in aircraft frame */
} magnetic_compass_data;


static magnetic_compass_data *free_list = NULL;


void magnetic_compass_enable(viewer *u)
{
	magnetic_compass_data *mc;

	if( u->magnetic_compass == NULL ){

		/* Allocate memory for magnetic compass status: */
		if( free_list == NULL ){
			mc = MALLOC( sizeof(magnetic_compass_data) );
		} else {
			mc = free_list;
			free_list = mc->next;
		}

		/* Initialize: */
		mc->last_time = curTime;
		mc->hdg = navaid_mag_heading(u->c);
		mc->hdg_dot = 0.0;
		mc->x_rot = 0.0;
		mc->y_rot = 0.0;

		u->magnetic_compass = mc;
	} else {
		mc = u->magnetic_compass;
	}

	mc->enabled = TRUE;
}


void magnetic_compass_disable(viewer *u)
{
	magnetic_compass_data *mc;

	mc = u->magnetic_compass;
	if( mc == NULL )
		return;
	
	mc->enabled = FALSE;
}


#define TILT_ANGLE DEGtoRAD(18)

static void update_compass_bearing(viewer *u)
{
	magnetic_compass_data *mc;
	double x_rot, y_rot, delta, theta, MH, dt;
	_BOOL tilted;
	VPoint H, aH, cH;
	VMatrix R;

	mc = u->magnetic_compass;
	if( mc == NULL )
		return;

	dt = curTime - mc->last_time;
	mc->last_time = curTime;

	/*
		Update compass roll and pitch angles based on the local vertical
		u->c->G:
	*/
	tilted = FALSE;

	x_rot =  atan2(u->c->G.y, -u->c->G.z);
	y_rot = -atan2(u->c->G.x, -u->c->G.z);

	mc->x_rot += (x_rot - mc->x_rot) * 0.002 / dt;
	if( mc->x_rot < -TILT_ANGLE ){
		mc->x_rot = -TILT_ANGLE;
		tilted = TRUE;
	}
	if( mc->x_rot > TILT_ANGLE ){
		mc->x_rot = TILT_ANGLE;
		tilted = TRUE;
	}

	mc->y_rot += (y_rot - mc->y_rot) * 0.002 / dt;
	if( mc->y_rot < -TILT_ANGLE ){
		mc->y_rot = -TILT_ANGLE;
		tilted = TRUE;
	}
	if( y_rot > TILT_ANGLE ){
		y_rot = TILT_ANGLE;
		tilted = TRUE;
	}

	if( tilted )
		mc->hdg_dot = 0.0;

	/*
		Local magnetic deviation (positive east):
	*/
	delta = u->c->curHeading - navaid_mag_heading(u->c);

	/*
		Local magnetic dip. This very crude extimation gives:

		latitude    theta
		(DEG)       (DEG)
		--------    -----
		0           0
		30          47
		60          75
		90          0

		FIXME: improve using the IGRF model
	*/
	theta = 0.60504 * u->c->w.latitude * (3.1067 - u->c->w.latitude);
	if( theta > M_PI/2 )
		theta = M_PI/2;
	else if( theta < -M_PI/2 )
		theta = M_PI/2;
	
	/* Local magnetic vector H in NED frame, unit length: */
	H.x = cos(theta) * cos(delta);
	H.y = cos(theta) * sin(delta);
	H.z = sin(theta);

	//printf("H=%f, %f, %f\n", H.x, H.y, H.z);

	/*
		Local magnetic vector aH in aircraft frame, unit length:
	*/
	VReverseTransform_(&H, &u->c->trihedral, &aH);

	/*
		Local magnetic vector cH in compass frame, unit length:
	*/
	VIdentMatrix(&R);
	VRotate(&R, YRotation, mc->y_rot);
	VRotate(&R, XRotation, mc->x_rot);
	VReverseTransform_(&aH, &R, &cH);

	/*
		MH is the magnetic heading in the compass frame, i.e. projection
		if cH on the local horizontal plane x-y of the compass frame:
	*/
	if( fabs(cH.x) + fabs(cH.y) < 1e-5 )
		return;
	MH = atan2(-cH.y, cH.x);
	if( MH < 0.0 )
		MH += 2*M_PI;

	mc->hdg_dot +=

		/* apply torque due to the magnetic field: */
		0.1 * dt * sin( MH - mc->hdg ) * sqrt(cH.x*cH.x + cH.y*cH.y)

		/* apply viscosity: */
		- 0.007 * mc->hdg_dot;

	mc->hdg += mc->hdg_dot * dt;

	while( mc->hdg < 0.0 )
		mc->hdg += 2*M_PI;
	while( mc->hdg >= 2*M_PI )
		mc->hdg -= 2*M_PI;
}


static VPath * get_compass_scale_path()
/*
	Return the scale of the compass, drawn on a cone of radius 1.0
	and height 1.0. The axis of the cone is y and the pivot point is
	the origin.
*/
{
	VPath *path;
	int a, s_len;
	double r, notch_short, notch_long, notch_len;
	double fh, fw, slant, margin;
	VPoint p, q;
	VMatrix M, L;
	char *s;
	static char *labels[] = {"N", "3", "6", "E", "12", "15", "S", "21",
		"24", "W", "30", "33"};

	path = vpath_new();

	r = 1.0;  /* radius and cone height */
	notch_short = 0.12 * r;
	notch_long = 0.25 * r;
	slant = DEGtoRAD(45);
	fh = 0.20 * r;  /* font height */
	fw = 0.12 * r;  /* font width */
	margin = 0.03 * r;  /* long notch to char margin */

	for( a = 0; a < 360; a += 5 ){

		/* Transformation for this notch + label */
		VIdentMatrix(&M);
		VRotate(&M, XRotation, -slant);
		VTranslate(&M, 0.0, r, -r);
		VRotate(&M, YRotation, -DEGtoRAD(a));

		/* Draw notch */
		if( a % 10 == 0 )
			notch_len = notch_long;
		else
			notch_len = notch_short;
		
		VSetPoint(&p, 0.0, 0.0, 0.0);
		VSetPoint(&q, 0.0, -notch_len, 0.0);
		VTransform(&p, &M, &p);
		VTransform(&q, &M, &q);
		vpath_moveTo(path, &p);
		vpath_lineTo(path, &q);

		/* Draw label */
		if( a % 30 == 0 ){

			s = labels[a/30];
			s_len = strlen(s);

			VIdentMatrix(&L);
			VScaleMatrix(&L, fw, fh, 1.0);
			VTranslate(&L, -0.5 * (fw * s_len), -notch_len - margin, 0.0);
			VMatrixMult(&L, &M, &L);

			vpath_draw_string(path, s, s_len, &L);
		}

	}

	return path;
}


static VPath * compass_scale_path = NULL;


void magnetic_compass_draw(viewer * u)
{
	magnetic_compass_data *mc;
	AWindow  *w;
	double    x, y, width, h, r, xmargin, ymargin;
	VMatrix   M;
	Rect      rect;

	mc = u->magnetic_compass;
	if( mc == NULL || ! mc->enabled )
		return;
	
	update_compass_bearing(u);

	if( u->viewDirection.x < 0.90 )
		return;
	
	w = u->v->w;
	x = 0.85 * w->width;
	y = 0.0;
	width = w->width - x;
	h = 0.50 * width;
	r = 0.45 * width; /* actual radius of the scale cone */

	SetRect(&rect, x, 0, w->width, h);
	AlibSetClipRect(w, &rect);
	AlibFillRectangle(w, &rect, w->pixel[radarBackgroundColor->cIndex]);
	
	VIdentMatrix(&M);

	/* Set scale size */
	VScaleMatrix(&M, r, r, r);

	/* Rotate based on the current magnetic compass status */
	VRotate(&M, YRotation, mc->hdg);
	VRotate(&M, XRotation, mc->y_rot);
	VRotate(&M, ZRotation, mc->x_rot);

	/* Move to the final screen location */
	VTranslate(&M, x + width/2, y - r + h*2/3, 0.0);

	/* Draw compass scale */
	xmargin = 0.20 * width;
	ymargin = 0.10 * h;
	SetRect(&rect, x + xmargin, y + ymargin, x + width - xmargin, y + h - ymargin);
	AlibSetClipRect(w, &rect);
	if( compass_scale_path == NULL )
		compass_scale_path = get_compass_scale_path();
	vpath_stroke(compass_scale_path, &M, w, w->pixel[whiteColor->cIndex]);

	/* Draw reference line */
	AlibDrawLine(w, x + 0.5*width, y + 0.10*h, x + 0.5*width, y + 0.90*h,
		w->pixel[HSIMagentaColor->cIndex]);
}


void magnetic_compass_free(viewer *u)
{
	magnetic_compass_data *mc;

	mc = u->magnetic_compass;
	if( mc == NULL )
		return;
	
	mc->next = free_list;
	free_list = u->magnetic_compass;
	u->magnetic_compass = NULL;
}


void magnetic_compass_free_all()
{
	magnetic_compass_data *p;

	while( free_list != NULL ){
		p = free_list;
		free_list = free_list->next;
		FREE(p);
	}

	vpath_free(compass_scale_path);
	compass_scale_path = NULL;
}

/* End of the "magnetic_compass" module */
