/*
 *  acm : an aerial combat simulator for X
 *  Copyright (C) 1991-1998  Riley Rainey
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 dated June, 1991.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program;  if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
 */

/*
 *  2014-12-29 dlc - Refer to CHANGES text file for details of modifications
 */

#if defined(SVR4)
#include <sys/filio.h>
#endif

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include "air.h"
#include "audio.h"
#include "damage.h"
#include "dis_if.h"
#include "drone.h"
#include "effects.h"
#include "init.h"
#include "joystick.h"
#include "list.h"
#include "manifest.h"
#include "mem_manager.h"
#include "missile.h"
#include "mouse.h"
#include "patchlevel.h"
#include "planes.h"
#include "players.h"
#include "pm.h"
#include "redraw.h"
#include "render.h"
#include "transfer.h"
#include "viewer.h"
#include "update.h"

#undef a

#ifdef SVR4
/*
 * This forces the inclusion of filio.h (and other things) in ioctl.h 
 * on NCR SVR4.
 */
#define BSD_COMP 1
#endif							/* SVR4 */

#include <sys/ioctl.h>
#include <sys/time.h>
#ifdef _AIX
#include <sys/select.h>
#endif
#include <netinet/in.h>
#include <netdb.h>
#include <setjmp.h>

static char *objects = NULL; /* value of the -objects PATHs option */
static _BOOL doUpdate = FALSE;
static char * display = NULL;
static char * name = NULL;

static sigset_t sigset_mask, sigset_omask;

/* FIXME: NOT USED
static int   listen_socket;
*/

#ifdef HAVE_DIS

#include "dis_if.h"

int       dis_site = DIS_SITE;
int       dis_application = DIS_APPLICATION;
int       dis_exercise = DIS_EXERCISE;

static void
disEntityEnterCb(int eid, dis_entity_type * etype, int force, craft ** cptr)
{
	int       i, top, mtype;
	craft    *c;
	int       team;
	craft    *tbl;

	*cptr = NULL;

#ifdef DIS_DEBUG
	char     *kindstr;

	if      (etype->kind == DISKindOther)         kindstr = "Other";
	else if (etype->kind == DISKindPlatform)      kindstr = "Platform";
	else if (etype->kind == DISKindMunition)      kindstr = "Munition";
	else if (etype->kind == DISKindLifeForm)      kindstr = "LifeForm";
	else if (etype->kind == DISKindEnvironmental) kindstr = "Environmental";
	else if (etype->kind == DISKindCultural)      kindstr = "Cultural";
	else if (etype->kind == DISKindSupply)        kindstr = "Supply";
	else if (etype->kind == DISKindRadio)         kindstr = "Radio";
	else                                          kindstr = "???";
	fprintf(stderr, "%8.2f: disEntityEnterCb() eid=%d kind=%s\n",
		curTime, eid, kindstr );
#endif

	if (etype->kind == DISKindPlatform) {
		tbl = ptbl;
		top = MAXPLAYERS;
		mtype = CT_DIS_PLANE;
	}
	else if (etype->kind == DISKindMunition) {
		tbl = mtbl;
		top = MAXPROJECTILES;
		mtype = CT_DIS_MUNITION;
	}
	else {
		return;
	}

/*
 *  most of this is a copy of the newDrone code in newPlane.c
 */

	for (i = 0; i < top; ++i) {

		if (tbl[i].type == CT_FREE) {

			if (force == DISForceFriendly) {
				team = 1;
			}
			else {
				team = 2;
			}

			c = &tbl[i];
			*cptr = c;

			c->createTime = curTime;
			c->vl = NULL;
			c->disId = eid;
			c->team = team;
			c->vl = NULL;
			c->type = mtype;
			c->cinfo = lookupCraftByEntityType(etype);
			drone_reset_opponent(c);

/*
 *  If we don't know about a given entity type, then we will not track it.
 */

			if (!c->cinfo) {
				*cptr = NULL;
				c->type = CT_FREE;
				return;
			}
			strncpy(c->name, "DIS", sizeof(c->name));

			c->curNWDef = 0.0;
			c->flags = 0;
			c->radarMode = RM_OFF;
			c->curRadarTarget = -1;

			pm_hud_strings_alloc(c);

			break;
		}
	}

	if (i == top)
		fprintf(stderr, "disEntityEnterCb() Out of players (increase MAXPLAYERS)\n");
}

static void
disDetonationCb(int ftype, 
				int firingEid, 
				int targetEid, 
				double time, 
				double *worldLocation, 
				double *entityLocation, 
				craft * m, 
				dis_detonation_pdu *dpdu)
{
	craft    *c;
	int       j;
	VPoint    Sg, rloc, rvel, tmp;
	double   exp_diameter, dist_meters, vel_meters_per_sec;

	if (m) {
		killMissile(m, (craft *) NULL);
	}

#ifdef DIS_DEBUG
	fprintf(stderr, "%8.2f: disDetonationCb() target %d\n", curTime, targetEid);
#endif

	Sg.x = worldLocation[0];
	Sg.y = worldLocation[1];
	Sg.z = worldLocation[2];

	for (c = ptbl, j = 0; j < MAXPLAYERS; ++j, ++c)

		if (c->type != CT_FREE && c->disId == targetEid) {

			/* found the target */

			if (c->type != CT_DIS_PLANE) {

				/* impact distance from C.G. */

				rloc.x = c->Sg.x - Sg.x;
				rloc.y = c->Sg.y - Sg.y;
				rloc.z = c->Sg.z - Sg.z;
				dist_meters = Vmagnitude(&rloc);

				/* impact velocity */

				tmp.x = FEETtoMETERS(c->Cg.x);
				tmp.y = FEETtoMETERS(c->Cg.y);
				tmp.z = FEETtoMETERS(c->Cg.z);
				VReverseTransform_(&tmp, &c->XYZtoNED, &rvel);
				rvel.x = dpdu->vel.x - rvel.x;
				rvel.y = dpdu->vel.y - rvel.y;
				rvel.z = dpdu->vel.z - rvel.z;
				vel_meters_per_sec = Vmagnitude(&rvel);

				/* the target is a local player, damage him */
				if (damage_absorbDISDamage(c, 
									&dpdu->burst.munition,
									dpdu->burst.warhead, 
									dpdu->burst.fuze,
									dist_meters,
									vel_meters_per_sec,
									&exp_diameter) == 0) {
					killPlayer(c, ftype == DIS_FIRE_M61A1 ?
						"cannon fire" :
						"something (presumably a missile)");
				}
				if (exp_diameter > 0.0) {
					VPoint vel = { 0, 0, 0 };
					effects_new_explosion(&Sg, &vel, exp_diameter * 0.3, exp_diameter, 3.0);
				}

			}
			break;
		}
}

static int
disInit(void)
{
	int       err;

	err = dis_init(dis_exercise,
				   dis_site, dis_application,
				   disEntityEnterCb, disDetonationCb);
	dis_setDRThresholds(DIS_LOCATION_THRESHOLD, DIS_ORIENTATION_THRESHOLD);

	return err;
}

#endif


static acm_sig_t
killed(int i)
{
	printf("\ninterrupt\n");
	/* FIXME: NOT USED
	shutdown(listen_socket, 2);
	close(listen_socket);
	*/
#ifdef HAVE_DIS
	dis_close();
#endif
	exit(0);
}


#ifdef HAVE_STRUCT_SIGACTION
static struct sigaction alrm;
#else
static struct sigvec    alrm;
#endif


static acm_sig_t
myalarm(int s)
{
	doUpdate = TRUE;
#ifdef HAVE_STRUCT_SIGACTION
	sigaction(SIGALRM, &alrm, (struct sigaction *) 0);
#else
	sigvec   (SIGALRM, &alrm, (struct sigvec *)    0);
#endif
}


static void
main_loop(void)
{

	fd_set    fdset, ofdset;
	int       news = -1, pno;
	struct itimerval update;
	struct timeval zero_timeout, update_timeout;

	signal(SIGINT, killed);
	signal(SIGQUIT, killed);

	zero_timeout.tv_sec = 0;
	zero_timeout.tv_usec = 0;
	update_timeout.tv_sec = 0;
	update_timeout.tv_usec = (long) (update_interval_millisec * 1000);

	/*
	 *  If we aren't supposed to be running full-bore, then setup the
	 *  real time interval clock.
	 */

	if (real_delta_t == 0) {

#ifdef HAVE_STRUCT_SIGACTION
		alrm.sa_handler = myalarm;
		sigemptyset( &alrm.sa_mask );
		alrm.sa_flags = 0;
		sigaction( SIGALRM, &alrm, (struct sigaction *) 0 );
#else
		alrm.sv_handler = myalarm;
		sigemptyset( &alrm.sv_mask );
#ifdef __hpux
		alrm.sv_flags = SV_BSDSIG;
#else
		alrm.sv_flags = SV_INTERRUPT;
#endif
		sigvec(SIGALRM, &alrm, (struct sigvec *) 0);
#endif

		/*
		 *  Set real time clock to interrupt us at the appropriate interval
		 */

		update.it_interval.tv_sec = 0;
		update.it_interval.tv_usec = (long) (update_interval_millisec * 1000);
		update.it_value.tv_sec = 0;
		update.it_value.tv_usec = (long) (update_interval_millisec * 1000);

		setitimer(ITIMER_REAL, &update, 0);

	}

	FD_ZERO(&ofdset);
	FD_ZERO(&fdset);
	/* FIXME: NOT USED
	FD_SET(listen_socket, &ofdset);
	*/

	/*
		Loop until there is at least a viewer:
	*/

	while ( vl_head != NULL ) {

		sigprocmask(SIG_SETMASK, &sigset_omask, NULL);

		fdset = ofdset;

		if (real_delta_t == 0) {
			pno = select(32, &fdset, (fd_set *) NULL, (fd_set *) NULL,
						 (struct timeval *) NULL);
		}
		else {
			if (ptblCount == 0) {
				pno = select(32, &fdset, (fd_set *) NULL, (fd_set *) NULL,
							 &update_timeout);
			}
			else {
			pno = select(32, &fdset, (fd_set *) NULL, (fd_set *) NULL,
						 &zero_timeout);
			}
			doUpdate = TRUE;
		}

		(void) sigemptyset(&sigset_mask);
		(void) sigaddset(&sigset_mask, SIGALRM);
		(void) sigprocmask(SIG_BLOCK, &sigset_mask, &sigset_omask);

		if (pno < 0) {
			/* FIXME: NOT USED
			FD_CLR(listen_socket, &fdset);
			*/
			if (news > 0)
				FD_CLR(news, &fdset);
		}
		if (doUpdate) {
			doUpdate = FALSE;
			update_redraw();
			audio_update();
		}
	}
}


static void
help()
{
	printf("ACM %s\n", REVISION_STRING);
	printf(
"Copyright (C) 1991-1998  Riley Rainey (rrainey@ix.netcom.com)\n"
"Updated and modified by Umberto Salsi (salsi@icosaedro.it)\n"
"Updated and modified by David L. Craig (dlc.usa@gmail.com)\n"
"ACM comes with ABSOLUTELY NO WARRANTY.\n"
"This is free software, and you are welcome to distribute it under the\n"
"conditions described in the COPYING file.\n");

#ifdef HAVE_DIS
	if (disInUse)
		printf("* DIS protocol active.\n");
#endif

#if defined(NETAUDIO)
	printf("* This ACM program was built with sound support (netaudio).\n");
#else
#if defined(_HPUX_SOURCE)
	printf("* This ACM program was built with sound support (HP AAPI).\n");
#endif
#endif
}


static list *
read_switches_from_file(const char *fname)
{
	FILE *f;
	char commands[4096];
	char *argv[100];
	int  argc;
	int  n, err, i;
	list *l;

	argv[0] = NULL;
	argc = 1;

	f = fopen (fname, "r");
	if( f == NULL ){
		fprintf(stderr, "read_switches_from_file() ERROR: failed to read file `%s'\n", fname);
		return NULL;
	}
	n = fread(commands, 1, 4096 - 1, f);
	err = ferror(f);
	fclose(f);
	if( err != 0 ){
		fprintf(stderr, "read_switches_from_file() ERROR: failed to read file `%s'\n", fname);
		return NULL;
	}
	if( n == 4096 - 1 ){
		fprintf(stderr, "read_switches_from_file() ERROR: file `%s' too long\n", fname);
		return NULL;
	}

	if( n < 1 )
		return NULL;

	commands[n] = '\0';
	if( ! split(commands, &argc, argv, 100) ){
		fprintf(stderr, "read_switches_from_file() ERROR: too many arguments in file `%s'\n", fname);
		return NULL;
	}
	if( argc == 0 )
		return NULL;

	l = list_new();
	for( i = 0; i < argc; i++ )
		list_add_elem(l, argv[i]);
	return l;
}


static list *
processCommandSwitches(list *args)
{
	int i;
	char *a;
	list *switches;

	switches = list_new();

	for (i = 0; i < args->n; ++i) {

		a = args->arr[i];
#ifdef TRACE_ARGS
		fprintf(stderr, "processCommandSwitches() is processing switch `%s'\n", a);
#endif

		if (strcmp(a, "-help") == 0 || strcmp(a, "-version") == 0
		|| strcmp(a, "-copyright") == 0) {
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() is invoking help()\n");
#endif
			help();
		}
		else if (strcmp(a, "-objects") == 0 && ++i < args->n) {
			objects = STRDUP(args->arr[i]);
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() set objects to '%s'\n", objects);
#endif
		}
		else if (strcmp(a, "-display") == 0 && ++i < args->n) {
			if( display != NULL )
				FREE(display);
			display = STRDUP(args->arr[i]);
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() set display to '%s'\n", display);
#endif
		}
#ifdef HAVE_DIS
		else if (strcmp(a, "-no-dis") == 0) {
			dis_site = 0;
			dis_application = 0;
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() set dis_site and dis_application to 0\n");
#endif
		}
		else if (strcmp(a, "-dis-site") == 0 && ++i < args->n) {
			dis_site = strtol(args->arr[i], (char **) NULL, 0);
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() set dis_site to %d\n", dis_site);
#endif
		}
		else if (strcmp(a, "-dis-appl") == 0 && ++i < args->n) {
			dis_application = strtol(args->arr[i], (char **) NULL, 0);
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() set dis_application to %d\n", dis_application);
#endif
		}
		else if (strcmp(a, "-dis-exercise") == 0 && ++i < args->n) {
			dis_exercise = strtol(args->arr[i], (char **) NULL, 0);
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() set dis_exercise to %d\n", dis_exercise);
#endif
		}
		else if (strcmp(a, "-dis-absolute-time") == 0) {
			disAbsoluteTime = 1;
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() set disAbsoluteTime to 1\n");
#endif
		}
		else if (strncmp(a, "-dis", 4) == 0) {
			fprintf(stderr, "Acm DIS arguments:\n"
					" -dis-exercise <exercise number> (default %d)\n"
					" -dis-site <site number> (default %d)\n"
					" -dis-appl <application number> (default %d)\n"
					" -dis-absolute-time\n",
					DIS_EXERCISE, DIS_SITE, DIS_APPLICATION);
		}
#endif
		else if (strcmp(a, "-init") == 0 && ++i < args->n) {
	                list *file_switches, *init_switches ;

	                file_switches = list_new();
		        file_switches = read_switches_from_file(args->arr[i]);
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() is processing init file contents\n");
#endif
                        init_switches = processCommandSwitches(file_switches);
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() has finished processing init file contents\n");
#endif
			list_free(file_switches);
			list_add_list(switches, init_switches);
			list_free(init_switches);
		}
		else if (strcmp(a, "-arcade") == 0) {
			arcadeMode = 1;
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() set arcadeMode to 1\n");
#endif
		}
		else if (strcmp(a, "-cmap") == 0) {
			list_add_elem(switches, "-cmap");
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '-cmap' to switches list\n");
#endif
		}
		else if (strcmp(a, "-depth_steps") == 0 && ++i < args->n) {
			depth_cue_steps = atoi(args->arr[i]);
			if (depth_cue_steps < 1) {
				depth_cue_steps = 1;
			}
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() set depth_cue_steps to %i\n", depth_cue_steps);
#endif
		}
		else if (strcmp(a, "-da") == 0 && ++i < args->n) {
			drone_set_aggressiveness( atof(args->arr[i]) );
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() invoked drone_set_aggresiveness(%f)\n", atof(args->arr[i]));
#endif
		}
		else if (strcmp(a, "-drone-mode") == 0 && ++i < args->n) {
			if( strcmp(args->arr[i], "DOG_FIGHT") == 0 ) {
				drone_set_mode(DRONE_DOG_FIGHT_MODE);
#ifdef TRACE_ARGS
				fprintf(stderr, "processCommandSwitches() invoked drone_set_mode(%u)\n", DRONE_DOG_FIGHT_MODE);
#endif
			}
			else if( strcmp(args->arr[i], "HUNTING") == 0 ) {
				drone_set_mode(DRONE_HUNTING_MODE);
#ifdef TRACE_ARGS
				fprintf(stderr, "processCommandSwitches() invoked drone_set_mode(%u)\n", DRONE_HUNTING_MODE);
#endif
			}
			else {
				fprintf(stderr, "processCommandSwitches() ERROR: invalid argument for -drone-mode option, must be DOG_FIGHT or HUNTING\n");
				exit(1);
			}
		}
		else if (strcmp(a, "-geometry") == 0 && ++i < args->n) {
			list_add_elem(switches, "-geometry");
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '-geometry' to switches list\n");
#endif
			list_add_elem(switches, args->arr[i]);
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '%s' to switches list\n", args->arr[i]);
#endif
		}
		else if (strcmp(a, "-eye_to_screen_cm") == 0 && ++i < args->n) {
			eye_to_screen_cm = atof(args->arr[i]);
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() set eye_to_screen_cm to %f\n", eye_to_screen_cm);
#endif
		}
		else if (strcmp(a, "-downward_view_angle_deg") == 0 && ++i < args->n) {
			downward_view_angle_rad = DEGtoRAD( atof(args->arr[i]) );
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() set downward_view_angle_rad to %f\n", downward_view_angle_rad);
#endif
		}
		else if (strcmp(a, "-js") == 0 && i+1 < args->n) {
			if (args->arr[i + 1][0] != '-') {
				SetJoystickPort(args->arr[++i]);
#ifdef TRACE_ARGS
				fprintf(stderr, "processCommandSwitches() invoked SetJoystickPort('%s')\n", args->arr[i]);
#endif
			}
		}
		else if (strcmp(a, "-groundc") == 0 && ++i < args->n) {
			list_add_elem(switches, "-groundc");
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '-groundc' to switches list\n");
#endif
			list_add_elem(switches, args->arr[i]);
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '%s' to switches list\n", args->arr[i]);
#endif
		}
		else if (strcmp(a, "-name") == 0 && ++i < args->n) {
			name = STRDUP(args->arr[i]);
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() set name to '%s'\n", name);
#endif
		}
		else if (strcmp(a, "-plane") == 0 && ++i < args->n) {
			list_add_elem(switches, "-plane");
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '-plane' to switches list\n");
#endif
			list_add_elem(switches, args->arr[i]);
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '%s' to switches list\n", args->arr[i]);
#endif
		}
		else if (strcmp(a, "-fuel") == 0 && ++i < args->n) {
			list_add_elem(switches, "-fuel");
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '-fuel' to switches list\n");
#endif
			list_add_elem(switches, args->arr[i]);
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '%s' to switches list\n", args->arr[i]);
#endif
		}
		else if (strcmp(a, "-payload") == 0 && ++i < args->n) {
			list_add_elem(switches, "-payload");
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '-payload' to switches list\n");
#endif
			list_add_elem(switches, args->arr[i]);
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '%s' to switches list\n", args->arr[i]);
#endif
		}
		else if (strcmp(a, "-scene") == 0 && ++i < args->n) {
			if( sceneFile != NULL )
				FREE(sceneFile);
			sceneFile = STRDUP(args->arr[i]);
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() set sceneFile to '%s'\n", sceneFile);
#endif
		}
		else if (strcmp(a, "-stealth") == 0) {
			list_add_elem(switches, "-stealth");
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '-stealth' to switches list\n");
#endif
		}
		else if (strcmp(a, "-end-game") == 0) {
			list_add_elem(switches, "-end-game");
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '-end-game' to switches list\n");
#endif
		}
		else if (strcmp(a, "-threshold-range") == 0 && ++i < args->n) {
			double end_game_threshold_nm;

			end_game_threshold_nm = atof( args->arr[i] );
			if (end_game_threshold_nm < 1.0) {
				end_game_threshold_nm = 1.0;
			}
			end_game_threshold_meters = FEETtoMETERS(end_game_threshold_nm*NM);
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() set end_game_threshold_meters to %f\n", end_game_threshold_meters);
#endif
		}
		else if (strcmp(a, "-team") == 0 && ++i < args->n) {
			list_add_elem(switches, "-team");
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '-team' to switches list\n");
#endif
			list_add_elem(switches, args->arr[i]);
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '%s' to switches list\n", args->arr[i]);
#endif
		}
		else if (strcmp(a, "-team_1") == 0 && ++i < args->n) {
			list_add_elem(switches, "-team_1");
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '-team_1' to switches list\n");
#endif
			list_add_elem(switches, args->arr[i]);
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '%s' to switches list\n", args->arr[i]);
#endif
/*			int ii, jj;

			jj = -1;
                        for ( ii = 0 ; ii < ttblCount && jj != 0 ; ii++ ) {
                                if ( strcmp(ttbl[ii].sname, args->arr[i]) == 0 ) {
                                        teamLatLon[1].latitude  = ttbl[ii].loc.latitude;
                                        teamLatLon[1].longitude = ttbl[ii].loc.longitude;
                                        teamLatLon[1].z         = ttbl[ii].loc.z;
                                        teamHeading[1]          = ttbl[ii].heading;
                                        jj = 0;
                                }
                        }
                        if ( jj == -1 ) {
				fprintf(stderr, "processCommandSwitches() ERROR: The TEAMLOC record for %s was not found\n",
                                        args->arr[i]);
				exit(1);
                        }
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() set team_1 location values to those of %s\n",
                        	args->arr[i]);
#endif
 */
		}
		else if (strcmp(a, "-team_2") == 0 && ++i < args->n) {
			list_add_elem(switches, "-team_2");
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '-team_2' to switches list\n");
#endif
			list_add_elem(switches, args->arr[i]);
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '%s' to switches list\n", args->arr[i]);
#endif
/*			int ii, jj;

			jj = -1;
                        for ( ii = 0 ; ii < ttblCount && jj != 0 ; ii++ ) {
                                if ( strcmp(ttbl[ii].sname, args->arr[i]) == 0 ) {
                                        teamLatLon[2].latitude  = ttbl[ii].loc.latitude;
                                        teamLatLon[2].longitude = ttbl[ii].loc.longitude;
                                        teamLatLon[2].z         = ttbl[ii].loc.z;
                                        teamHeading[2]          = ttbl[ii].heading;
                                        jj = 0;
                                }
                        }
                        if ( jj == -1 ) {
				fprintf(stderr, "processCommandSwitches() ERROR: The TEAMLOC record for %s was not found\n",
                                        args->arr[i]);
				exit(1);
                        }
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() set team_2 location values to those of %s\n",
                        	args->arr[i]);
#endif
 */
		}
		else if (strcmp(a, "-latitude") == 0 && ++i < args->n) {
			list_add_elem(switches, "-latitude");
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '-latitude' to switches list\n");
#endif
			list_add_elem(switches, args->arr[i]);
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '%s' to switches list\n", args->arr[i]);
#endif
		}
		else if (strcmp(a, "-longitude") == 0 && ++i < args->n) {
			list_add_elem(switches, "-longitude");
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '-longitude' to switches list\n");
#endif
			list_add_elem(switches, args->arr[i]);
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '%s' to switches list\n", args->arr[i]);
#endif
		}
		else if (strcmp(a, "-altitude") == 0 && ++i < args->n) {
			list_add_elem(switches, "-altitude");
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '-altitude' to switches list\n");
#endif
			list_add_elem(switches, args->arr[i]);
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '%s' to switches list\n", args->arr[i]);
#endif
		}
		else if (strcmp(a, "-heading") == 0 && ++i < args->n) {
			list_add_elem(switches, "-heading");
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '-heading' to switches list\n");
#endif
			list_add_elem(switches, args->arr[i]);
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '%s' to switches list\n", args->arr[i]);
#endif
		}
		else if (strcmp(a, "-airspeed-kt") == 0 && ++i < args->n) {
			list_add_elem(switches, "-airspeed-kt");
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '-airspeed-kt' to switches list\n");
#endif
			list_add_elem(switches, args->arr[i]);
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '%s' to switches list\n", args->arr[i]);
#endif
		}
		else if (strcmp(a, "-lighting") == 0 && ++i < args->n) {
			a = args->arr[i];
			if( strcmp(a, "daylight") == 0 || strcmp(a, "'daylight'") == 0 ) {
				render_lighting = RENDER_DAYLIGHT;
#ifdef TRACE_ARGS
				fprintf(stderr, "processCommandSwitches() set render_lighting to %u\n", render_lighting);
#endif
			}
			else if( strcmp(a, "night") == 0 || strcmp(a, "'night'") == 0 ) {
				render_lighting = RENDER_NIGHT;
#ifdef TRACE_ARGS
				fprintf(stderr, "processCommandSwitches() set render_lighting to %u\n", render_lighting);
#endif
			}
			else {
				fprintf(stderr, "processCommandSwitches() ERROR: Invalid lighting \"%s\", must be daylight|night.\n", a);
				exit (1);
			}
		}
		else if (strcmp(a, "-visibility") == 0 && ++i < args->n) {
			visibility = atof(args->arr[i]);
			if (visibility < 0.1) {
				visibility = 0.1;
			}
			else if (visibility > 200.0) {
				visibility = 200.0;
			}
			visibility = NMtoMETERS(visibility);
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() set visibility to %f (meters)\n", visibility);
#endif
		}
		else if (strcmp(a, "-wind") == 0 && ++i < args->n) {
			double wd, wv;
			if( sscanf(args->arr[i], "%lf/%lf", &wd, &wv) != 2 ){
				fprintf(stderr, "processCommandSwitches() ERROR: invalid parameter for -wind. Expected DIRECTION/VELOCITY.\n");
				exit(1);
			}
			air_set_wind(wd, wv);
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() invoked air_set_wind(%f, %f)\n", wd, wv);
#endif
		}
		else if (strcmp(a, "-gust") == 0 && ++i < args->n) {
			air_set_gust( atof(args->arr[i]) );
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() invoked air_set_gust(%f)\n", atof(args->arr[i]));
#endif
		}
		else if (strcmp(a, "-no-sound") == 0 ) {
			list_add_elem(switches, "-no-sound");
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '-no-sound' to switches list\n");
#endif
		}
		else if (strcmp(a, "-frame-rate") == 0 && ++i < args->n) {
			double val = atof(args->arr[i]);
			frame_interval_millisec = 1000.0 / val;
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() set frame_interval_millisec to %f\n", frame_interval_millisec);
#endif
			real_delta_t = 0;
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() set read_delta_t to 0\n");
#endif
	    	}
		else if (strcmp(a, "-update-rate") == 0 && ++i < args->n) {
            		double val = atof(args->arr[i]);
			update_interval_millisec = 1000.0 / val;
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() set update_interval_millisec to %f\n", update_interval_millisec);
#endif
			real_delta_t = 0;
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() set read_delta_t to 0\n");
#endif
	    	}
		else if (strcmp(a, "-watch-frame-rate") == 0) {
			watch_frame_rate = 1;
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() set watch_frame_rate to 1\n");
#endif
	    	}
		else if (strcmp(a, "-transfer-entity-mode") == 0 && ++i < args->n) {
			transferEntityIdBits = strtol ( args->arr[i], NULL, 0 );
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() set transferEntityIdBits to %u\n", transferEntityIdBits);
#endif
		}
		else if (strcmp(a, "-subject-entity-id") == 0 && ++i < args->n) {
			dis_entity_id id;
			if (DISParseEntityID ( &id, args->arr[i], 
								   strlen(args->arr[i])+1, ":/." ) == 0) {
				subjectEntityID = id;
				subjectEntitySpecified = 1;
#ifdef TRACE_ARGS
				fprintf(stderr, "processCommandSwitches() set subjectEntityID.sim_id.site_id to %u\n",
					subjectEntityID.sim_id.site_id);
				fprintf(stderr, "processCommandSwitches() set subjectEntityID.sim_id.application_id to %u\n",
					subjectEntityID.sim_id.application_id);
				fprintf(stderr, "processCommandSwitches() set subjectEntityID.entity_id to %u\n",
					subjectEntityID.entity_id);
				fprintf(stderr, "processCommandSwitches() set subjectEntitySpecified to 1\n");
#endif
			}
			else {
				fprintf(stderr, "processCommandSwitches() Invalid subject-entity-id \"%s\"\n", args->arr[i]);
				exit (1);
			}
		}
		else if (strcmp(a, "-mouse-mode") == 0 && ++i < args->n) {
			a = args->arr[i];
			if( strcmp(a, "fast") == 0 ) {
				mouse_stick_mode = MOUSE_FAST;
#ifdef TRACE_ARGS
				fprintf(stderr, "processCommandSwitches() set mouse_stick_mode to %u\n", mouse_stick_mode);
#endif
			}
			else if( strcmp(a, "normal") == 0 ) {
				mouse_stick_mode = MOUSE_NORMAL;
#ifdef TRACE_ARGS
				fprintf(stderr, "processCommandSwitches() set mouse_stick_mode to %u\n", mouse_stick_mode);
#endif
			}
			else if( strcmp(a, "precise") == 0 ) {
				mouse_stick_mode = MOUSE_PRECISE;
#ifdef TRACE_ARGS
				fprintf(stderr, "processCommandSwitches() set mouse_stick_mode to %u\n", mouse_stick_mode);
#endif
			}
			else {
				fprintf(stderr, "iprocessCommandSwitches() ERROR: Invalid mouse mode \"%s\", must be fast|normal|precise.\n", a);
				exit (1);
			}
		}
		else if (strcmp(a, "-hud-mode") == 0 ) {
			list_add_elem(switches, "-hud-mode");
#ifdef TRACE_ARGS
			fprintf(stderr, "processCommandSwitches() added '-hud-mode' to switches list\n");
#endif
		}
		else {
			fprintf(stderr, "processCommandSwitches() ERROR: invalid switch `%s'\n", args->arr[i]);
			exit(1);
		}
	}

	/*
		FIXME: the "flight by night" currently is mostly an hack
		implemented simply omitting to draw ground and clouds. If
		the visibility is set too low, runways becomes nearly white,
		that wouls be unrealistic. Workaround: by night, force
		visibility to 50 NM so that runways are rendered in dark gray:
	*/
	if( render_lighting == RENDER_NIGHT )
		visibility = NMtoMETERS(50.0);
	
	return switches;
}


int
main(int argc, char **argv)
{
	list      *clopts, *newPlayerSwitches;
	int       i;
	craft     *p;

	updateSystemClock();

	curTime = 0.0;

	real_delta_t = 1;

	depth_cue_steps = 8;

	mouse_stick_mode = MOUSE_NORMAL;

	render_lighting = RENDER_DAYLIGHT;

	eye_to_screen_cm = 50.0;
	downward_view_angle_rad = DEGtoRAD(15.0);

	/*
	 * An endGameThreshold of -1.0 means "use the radar lock range for
	 * the current aircraft.
	 */

	end_game_threshold_meters = -1.0;
	end_game_mode = 0;

	/*
	 *  When accepting control of an entity, the default is to use our
	 *  site ID and keep everything else the same.
	 */

	transferEntityIdBits = 0x4;

	/*
	 * these only apply to non-real time updating
	 */

	update_interval_millisec = 1000.0 / 50.0;  /* 50.0 hertz */
	frame_interval_millisec = 1000.0 / 15.0;   /* 15.0 hertz */

	watch_frame_rate = 0;

#ifdef HAVE_DIS
	disAbsoluteTime = 0;
#endif

	ptblCount = ctblCount = 0;
	visibility = NMtoMETERS(50.0);

	/*
	 *  Define handler for DIS transfer control requests
	 */

	dis_setTransferControlRequestCallback ( transferControlRequestHandler );

/*
 *  parse arguments
 */

	clopts = list_new();
	for ( i = 1; i < argc; i++ )
		list_add_elem(clopts, argv[i]);
	newPlayerSwitches = processCommandSwitches(clopts);
	list_free(clopts);

	if (real_delta_t == 0) {

		redraw_interval = (int) 
			(frame_interval_millisec / update_interval_millisec + 0.5);
		if ( redraw_interval < 1 ) {
			fprintf(stderr, "main() The frame rate must be <= 2*update rate.\n");
			exit(1);
		}

		/*
		 *  Set time intervals.
		 */

		deltaT = update_interval_millisec / 1000.0;

		halfDeltaTSquared = 0.5 * deltaT * deltaT;

	}

#ifdef HAVE_DIS
	{

/*
 *  If the user has not specified a DIS application ID and has not referred
 *  us to a SIM/x server, automatically generate an application id based
 *  on the low-order 16-bits of this host's IP address.  This is a bit of a
 *  hack, but it will assure that we get a unique application id so long as we
 *  are not using any sort of DIS bridge software.
 */

		char      name[256], *p;
		struct hostent *h;

		gethostname(name, sizeof(name));

		if (dis_application == DIS_APPLICATION) {
			if ((h = gethostbyname(name)) != (struct hostent *) NULL) {
				p = h->h_addr;
				dis_application = (p[2] << 8) | p[3];
			}
		}
	}

	if (disInit() == 0)
		disInUse = 1;
	else
		disInUse = 0;
#endif

	if ( display == NULL ) {
		if (getenv("DISPLAY")) {
			display = STRDUP( getenv("DISPLAY") );
		}
		else {
			fprintf(stderr, "main() Excuse me, but you have no DISPLAY.\n"
					   "How do you use X, anyway?\n");
			exit(1);
		}
	}

	if( name == NULL )
		name = STRDUP("Anonymous");

	init( objects );

	for ((i = 0, p = stbl); i < MAXSURFACE; (++i, ++p))
		if ( p->type != CT_FREE ) {
			fprintf(stderr, "stbl[%u] name<%-20.20s> type<%s> flags<",
				i, p->cinfo->name,
				p->type > CT_MAX ? "[invalid!]" : ctypes[p->type]);
			printFlags(p->flags);
		}
	for ((i = 0, p = mtbl); i < MAXPROJECTILES; (++i, ++p))
		if ( p->type != CT_FREE ) {
			fprintf(stderr, "mtbl[%u] name<%-20.20s> type<%s> flags<",
				i, p->cinfo->name,
				p->type > CT_MAX ? "[invalid!]" : ctypes[p->type]);
			printFlags(p->flags);
		}
	for (i = 0; i < MAXPLAYERS; ++i)
        	if ( ptbl[i].type != CT_FREE ) {
               		fprintf(stderr, "ptbl[%u] name<%s> type<%s> flags<",
				i, ptbl[i].cinfo->name,
				ptbl[i].type > CT_MAX ? "[invalid!]" : ctypes[ptbl[i].type]);
			printFlags(ptbl[i].flags);
        	}

        fprintf(stderr, "%s\n", "Invoking newPlayer to create the user's aircraft");
	if( newPlayer(display, name, newPlayerSwitches) == 0 ){
		for (i = 0; i < MAXPLAYERS; ++i)
        		if ( ptbl[i].type != CT_FREE ) {
                		fprintf(stderr, "ptbl[%u] name<%s> type<%s> flags<",
					i, ptbl[i].cinfo->name,
					ptbl[i].type > CT_MAX ? "[invalid!]" : ctypes[ptbl[i].type]);
				printFlags(ptbl[i].flags);
        		}
        	fprintf(stderr, "%s\n", "Entering main_loop");
		main_loop();

	} else {

		/*  failed to create player  */
        	fprintf(stderr, "%s\n", "newPlayer failed to create user's aircraft--terminating");

	}

	list_free(newPlayerSwitches);
	FREE(display);
	FREE(name);
	FREE(objects);
	freeAll();

	return 0;
}


