#define MEM_MANAGER_DEBUG


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef MEM_MANAGER_DEBUG

#define MEM_MAGIC 0x1f2f3f4f

typedef struct _myblock {
	int magic;
	int s;       /* size of the requested block (B) */
	char * f;    /* source file name */
	int l;       /* source line number of the MALLOC() or REALLOC() */
	struct _myblock * next, * prev;  /* double chained list */
	char m[0];   /* the client data */
} myblock;


static myblock *mem_start = NULL;


void * mem_manager_malloc(int s, char * f, int l)
{
	myblock *b;

	if( s < 0 ){
		fprintf(stderr, "mem_manager_malloc() %s:%d: allocating %d bytes\n", f, l, s);
		abort();
	}
	b = malloc(sizeof(myblock) + s);
	if( b == NULL ){
		fprintf(stderr, "mem_manager_malloc() %s:%d: out of memory allocating %d bytes\n", f, l, s);
		exit(1);
	}
	b->magic = MEM_MAGIC;
	b->s = s;
	b->f = f;
	b->l = l;
	if( mem_start != NULL )
		mem_start->prev = b;
	b->next = mem_start;
	b->prev = NULL;
	mem_start = b;
	//printf(__FILE__ " FIXME: allocating from %s:%d\n", f, l);
	return &b->m[0];
}


void * mem_manager_realloc(void * p, int s, char * f, int l)
{
	myblock *b, *b1;

	if( s < 0 ){
		fprintf(stderr, "mem_manager_realloc() %s:%d: reallocating %d bytes\n", f, l, s);
		abort();
	}

	if( p == NULL )
		return mem_manager_malloc(s, f, l);

	b = (myblock *) ((char *)p - sizeof(myblock));
	if( b->magic != MEM_MAGIC ){
		fprintf(stderr, "mem_manager_realloc() %s:%d: reallocating invalid block\n", f, l);
		abort();
	}
	b1 = realloc(b, sizeof(myblock) + s);
	if( b1 == NULL ){
		fprintf(stderr, "mem_manager_realloc() %s:%d: out of memory reallocating %d bytes\n", f, l, s);
		exit(1);
	}

	if( b1 != b ){
		if( b1->next != NULL ) b1->next->prev = b1;
		if( b1->prev != NULL ) b1->prev->next = b1;
		if( mem_start == b )  mem_start = b1;
		b = b1;
	}

	b->s = s;
	b->f = f;
	b->l = l;
	//printf(__FILE__ " FIXME: re-allocating from %s:%d\n", f, l);
	return &b->m[0];
}


void mem_manager_free(void *p, char * f, int l)
{
	myblock *b;

	if( p == NULL )
		return;

	b = (myblock *) ((char *)p - sizeof(myblock));
	if( b->magic != MEM_MAGIC ){
		fprintf(stderr, "mem_manager_free() %s:%d: deallocating invalid block\n", f, l);
		abort();
	}
	b->magic = 0;
	if( b == mem_start )
		mem_start = b->next;
	if( b->prev != NULL )
		b->prev->next = b->next;
	if( b->next != NULL )
		b->next->prev = b->prev;
	free(b);
}


char * mem_manager_strdup(char * s, char * f, int l)
{
	myblock *b;
	int size;

	if( s == NULL )
		return NULL;

	size = strlen(s) + 1;

	b = (myblock *) ((char *) mem_manager_malloc(size, f, l) - sizeof(myblock));
	if( b == NULL )
		return NULL;

	return (char *) memcpy(&b->m[0], s, size);
}


void mem_manager_report()
{
	myblock *b;
	unsigned int tot;

	if( mem_start == NULL )
		return;
	fprintf(stderr, "mem_manager_report() Memory report -- unallocated blocks:\n");
	b = mem_start;
	tot = 0;
	while( b != NULL ){
		fprintf(stderr, "mem_manager_report() \t%s:%d: %d bytes\n", b->f, b->l, b->s);
		fflush(stderr);
		if( b->magic != MEM_MAGIC ){
			fprintf(stderr, "mem_manager_report()     INVALID MAGIC NUMBER\n");
			fflush(stderr);
		}
		tot += b->s;
		b = b->next;
	}
	fprintf(stderr, "mem_manager_report() TOTAL: %u B\n", tot);
}


/****************************/

#else


void *
mem_manager_malloc(int s)
{
	void * p;

	p = malloc(s);
	if( p == NULL ){
		fprintf(stderr, "mem_manager_malloc() Out of memory trying to allocate %d B.\n", s);
		exit(1);
	}
	return p;
}


void *
mem_manager_realloc(void * p, int s)
{
	void * q;

	q = realloc(p, s);
	if( q == NULL ){
		fprintf(stderr, "mem_manager_realloc() Out of memory trying to reallocate %d B.\n", s);
		exit(1);
	}
	return q;
}


char *
mem_manager_strdup(char *s)
{
	char * d;

	d = strdup(s);
	if( d == NULL ){
		fprintf(stderr, "mem_manager_strdup() Out of memory trying to duplicate a string.\n");
		exit(1);
	}
	return d;
}

void mem_manager_free(void *p)
{
	free(p);
}

#endif /* MEM_MANAGER_DEBUG */
