/*
 *  acm : an aerial combat simulator for X
 *  Copyright (C) 1991,1992,1997  Riley Rainey
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 dated June, 1991.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program;  if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
 */

#include <string.h>
#include <math.h>
#include "dis_if.h"
#include "mem_manager.h"
#include "navaid.h"
#include "pm.h"

static navaid_t *test_navaid = NULL;


/*
 * localVAR(loc) returns the local magnetic variation in radiants.
 * loc must be in geocentric coord. The returned value is positive
 * for West deviation. For example:
 *
 *    TH = MH - localVAR(loc)
 *
 * Implementation: search for the nearest NAVAID and returns its VAR.
 */

static double
localVAR(VPoint * loc)
{
	navaid_t *nearest, *n;
	double    nearest_dist, dist;

	nearest = NULL;
	nearest_dist = 0.0;
	for (n = test_navaid; n != NULL; n = n->next) {
		dist = fabs(loc->x - n->Sg.x) + fabs(loc->y - n->Sg.y)
			+ fabs(loc->z - n->Sg.z);
		//printf("%s %.0f\t", n->id, dist);
		if (nearest == NULL) {
			nearest = n;
			nearest_dist = dist;
		}
		else if (dist < nearest_dist) {
			nearest = n;
			nearest_dist = dist;
		}
	}

	if (nearest == NULL) {
		return 0.0;
	}
	else {
		//printf("\nfound navaid %s\n", nearest->id);
		return nearest->magvar;
	}
}


/*
 * magHeading(c) returns the MH of the aircraft c (RAD).
 */

double
navaid_mag_heading(craft * c)
{
	if (curTime >= c->updActualLocalVAR) {
		c->actualLocalVAR = localVAR(&c->Sg);
		c->updActualLocalVAR = curTime + 2.0;
	}

	if (curTime >= c->updIndicatedLocalVAR) {
		c->indicatedLocalVAR
			+= 0.01 * (c->actualLocalVAR - c->indicatedLocalVAR);
		c->updIndicatedLocalVAR = curTime + 0.1;
	}

	return pm_normalize_yaw( c->curHeading + c->indicatedLocalVAR );
}


navaid_t * navaid_reception_check(craft * c, freq_t f)
{
	navaid_t *n;
	VPoint    p;
	double    range, dist, d;

	// nearest VOR station found:
	navaid_t *last_found = NULL;
	double    last_dist = 0.0;

	for (n = test_navaid; n; n = n->next) {
		if (f == n->frequency) {
			if (n->flags & NAVAID_LOC) {
				VTransform(&c->Sg, &n->lt, &p);
				if (
					p.x > 0.0 && fabs(p.y / p.x) < 1.192 /* = 50 DEG */
					&& Vmagnitude(&p) < 33e3 /* LOC, typical max range 18 NM */
				) {
					// LOC in range found: return immediately
					return n;
				}
			}
			else if (n->flags & (NAVAID_VOR|NAVAID_DME)){
				if ( n->frequency < 80 /* 112.00 MHz */ ) {
					/* terminal VOR, range 40 NM */
					range = NMtoMETERS(40);
				}
				else {
					/* en-route VOR, range 200 NM */
					range = NMtoMETERS(200);
				}

				/*
					Compute aircraft position relative to the station
					in its magnetic north/east/down frame:
				*/
				VTransform(&c->Sg, &n->lt, &p);

				/*
					Ignore station if it can't be received:
				*/
				if( p.z > 0.0 ){
					/*
						Aircraft below horizon of the radio station.
						Reduce radio range due to the Earth curvature.

						The value of the constant K below was derived from
						a pure geometric calculation based on an Earth mean
						radius of 6367 Km, giving a maximum reception range of

							d=K*(sqrt(transmitter_alt)+sqrt(aircraft_alt))

						Other sources suggest a greater value K=4126 with
						all the distances in meters, but I adopted a more
						conservative approach.
					*/
					#define K 3568.0
					d = K*( sqrt(abs(n->loc.z)) + sqrt(abs(c->w.z)) ); /* m */
					if( d < range )
						range = d;
				}
				dist = Vmagnitude(&p);
				if( dist > range ) /* too far from station */{
					continue;
					printf("%s: too far\n", n->id);
				}

				if ( last_found == NULL || last_dist > dist ) {
					// VOR found: continue looping searching
					// for possible nearest stations with same freq.
					last_found = n;
					last_dist = dist;
				}
			}
			else if(n->flags & NAVAID_NDB){
				if (n->flags & (NAVAID_OMARKER|NAVAID_MMARKER|NAVAID_IMARKER)) {
					range = NMtoMETERS(20);
				}
				else {
					range = NMtoMETERS(100);
				}
				VTransform(&c->Sg, &n->lt, &p);
				dist = Vmagnitude(&p);
				if ( dist < range
				&& (last_found == NULL || last_dist > dist) ) {
					// NDB in range found: continue looping searching
					// for possible nearest stations
					last_found = n;
					last_dist = dist;
				}
			}
			else {
				fprintf(stderr, "navaid_reception_check() Unexpected navaid type %d\n", n->flags);
				abort();
			}
		}
	}

	if( last_found == NULL ){
		// no stations found in range
		return NULL;
	}
	else {
		// nearest in range station found
		return last_found;
	}
}


void
navaid_free_all(void)
{
	navaid_t *n = test_navaid, *p;
	while (n) {
		p = n;
		n = n->next;
		FREE(p);
	}
}


static double
MagneticVariationToDEG(char * id, char * s)
{
	char      s2[10];
	int       len;

	strncpy(s2, s, 10);
	s2[9] = '\0';
	len = strlen(s2);
	if ( s2[len-1] == 'E' ) {
		s2[len-1] = '\0';
		return -strtod(s2, NULL);
	}
	else if ( s2[len-1] == 'W' ) {
		s2[len-1] = '\0';
		return strtod(s2, NULL);
	}
	else {
		fprintf(stderr, "MagneticVariationToDEG() ERROR: invalid format for the magnetic variation `%s' in %s\n", s, id);
		return 0.0;
	}
}

int
navaid_add_vor_dme_ndb(char *ident, char *type, WorldCoordinates * w,
		  char *magvar, double freq)
{
	navaid_t *n;

	n = (navaid_t *) MALLOC(sizeof(navaid_t));
	n->bearing = n->magvar = DEGtoRAD( MagneticVariationToDEG(ident, magvar) );

	strncpy(n->id, ident, 5);
	n->id[4] = '\0';
	n->loc = *w;
	GenerateWorldToLocalMatrix(w, &n->lt);
	VRotate(&n->lt, ZRotation, n->magvar);
	DISWorldCoordinatesToGeocentric(w, (void *) & n->Sg);

	if (strcmp(type, "VORTAC") == 0) {
		n->flags = NAVAID_VOR | NAVAID_DME;
	}
	else if (strcmp(type, "TACAN") == 0) {
		n->flags = NAVAID_VOR | NAVAID_DME;
	}
	else if (strcmp(type, "VOR/DME") == 0) {
		n->flags = NAVAID_VOR | NAVAID_DME;
	}
	else if (strcmp(type, "VOR") == 0) {
		n->flags = NAVAID_VOR;
	}
	else if (strcmp(type, "DME") == 0) {
		n->flags = NAVAID_DME;
	}
	else if (strcmp(type, "NDB") == 0) {
		n->flags = NAVAID_NDB;
	}
	else if (strcmp(type, "OMARKER") == 0) {
		n->flags = NAVAID_NDB | NAVAID_OMARKER;
	}
	else if (strcmp(type, "OMARKER/COMLO") == 0) {
		n->flags = NAVAID_NDB | NAVAID_OMARKER;
	}
	else if (strcmp(type, "MMARKER") == 0) {
		n->flags = NAVAID_NDB | NAVAID_MMARKER;
	}
	else if (strcmp(type, "IMARKER") == 0) {
		n->flags = NAVAID_NDB | NAVAID_IMARKER;
	}
	else if (strcmp(type, "NDB") == 0) {
		n->flags = NAVAID_NDB;
	}
	else {
		fprintf(stderr, "navaid_add_vor_dme_ndb() Unexpected NAV of the type `%s'. Ignore.\n", type);
		FREE(n);
		return -1;
	}

	if (n->flags & (NAVAID_VOR | NAVAID_DME)) {
		n->frequency = VOR_CHANNEL_MIN + (int) ((freq - 108.00) * 20.0 + 0.5);
		if( n->frequency < VOR_CHANNEL_MIN
		|| n->frequency > VOR_CHANNEL_MAX ){
			fprintf(stderr, "navaid_add_vor_dme_ndb() Frequency out of range %g for VOR/DME station. Ignore.\n", freq);
			FREE(n);
			return -1;
		}
	}
	else {
		n->frequency = (int) freq;
		if( n->frequency < NDB_CHANNEL_MIN
		|| n->frequency > NDB_CHANNEL_MAX ){
			fprintf(stderr, "navaid_add_vor_dme_ndb() Frequency out of range %g for NDB station. Ignore.\n", freq);
			FREE(n);
			return -1;
		}
	}
	n->next = test_navaid;
	test_navaid = n;
	return 0;
}

int
navaid_add_ils(char *ident, char *type, WorldCoordinates * w,
	WorldCoordinates * gsw, char *magvar, double freq, double loc_width,
	double loc_bearing, double gs_angle)
{
	navaid_t *n;

	n = (navaid_t *) MALLOC(sizeof(navaid_t));

	n->magvar = DEGtoRAD( MagneticVariationToDEG(ident, magvar) );

	strncpy(n->id, ident, 5);
	n->id[4] = '\0';
	n->bearing = DEGtoRAD(loc_bearing);

	n->loc = *w;
	GenerateWorldToLocalMatrix(w, &n->lt);
	VRotate(&n->lt, ZRotation, -n->bearing + n->magvar - M_PI);
	DISWorldCoordinatesToGeocentric(w, (void *) & n->Sg);

	n->gs_loc = *gsw;
	GenerateWorldToLocalMatrix(gsw, &n->gst);
	VRotate(&n->gst, ZRotation, -n->bearing + n->magvar - M_PI);

	n->slope = DEGtoRAD(gs_angle);
	n->beam_width = DEGtoRAD(loc_width);

	if (strcmp(type, "ILS") == 0) {
		n->flags = NAVAID_LOC | NAVAID_GS;
	}
	else if (strcmp(type, "ILS/DME") == 0) {
		n->flags = NAVAID_LOC | NAVAID_GS | NAVAID_DME;
	}
	else if (strcmp(type, "LOCALIZER") == 0) {
		n->flags = NAVAID_LOC;
	}
	else if (strcmp(type, "LOC/GS") == 0) {
		n->flags = NAVAID_LOC | NAVAID_GS;
	}
	else if (strcmp(type, "LOC/DME") == 0) {
		n->flags = NAVAID_LOC | NAVAID_DME;
	}
	else {
		FREE(n);
		return -1;
	}

	n->frequency = VOR_CHANNEL_MIN + (int) ((freq - 108.00) * 20.0 + 0.5);
	if( n->frequency < VOR_CHANNEL_MIN
	|| n->frequency > VOR_CHANNEL_MAX ){
		fprintf(stderr, "navaid_add_ils() Frequency out of range %g for ILS station. Ignore.\n", freq);
		FREE(n);
		return -1;
	}

	n->next = test_navaid;
	test_navaid = n;
	return 0;
}


/* End of the "navaid" module. */
