/*
 *  acm : an aerial combat simulator for X
 *  Copyright (C) 1991,1992  Riley Rainey
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 dated June, 1991.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program;  if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
 */

#ifndef _navaid_h
#define _navaid_h

#include "pm.h"

/*
	Every navaid_t station has a frequency encoded as int number
	whose value is here indicated as "channel".

	Channels in the range 0-199 encode frequency between 108.00 MHz
	and 117.95 MHz (VOR, DME and ILS stations).

	Channels in the range 200-499 encode frequency between 200 KHz
	and 499 KHz (NDB stations).
*/

#define VOR_CHANNEL_MIN    0
#define VOR_CHANNEL_MAX  199
#define NDB_CHANNEL_MIN  200
#define NDB_CHANNEL_MAX  499

typedef int freq_t;

typedef struct _navaid_t {
	struct _navaid_t *next;		/* links navaids on the same channel */
	char      id[5];			/* identifier (e.g. "IHNL") */
	freq_t    frequency;		/* channel number */
	double    magvar;			/* converts True to Magnetic (rad) */
	unsigned short flags;
	WorldCoordinates loc;		/* lat/long/altitude */
	VPoint    Sg;				/* geocentric location */
	WorldCoordinates gs_loc;	/* lat/long/altitude of glide-slope */
	double    bearing;			/* geodetic azimuth of loc bearing */
	double    beam_width;		/* beam width for localizers (rad) */
	/*see navaid.c on how to calculate it */
	double    slope;			/* angle of glide slope (rad) */
	VMatrix   lt;				/* transform geocentric to loc local coordinates */
	                            /* the 'x' axis is oriented as the local magnetic north */
	VMatrix   gst;				/* transform geocentric to gs local coordinates */
} navaid_t;


typedef struct {
	freq_t    frequency;		/* active frequency */
	char      dme;              /* Current DME reading (NM). Negative=invalid */
	navaid_t *station;			/* station being received */
} radio_t;


/*
	Definition of the navaid flag word.
	The allowed combination of flags are:

	NAVAID_LOC (possibly + NAVAID_LOC + NAVAID_GS)
	NAVAID_DME
	NAVAID_VOR (possibly + NAVAID_DME)
	NAVAID_NDB (possibly + NAVAID_OMARKER|NAVAID_MMARKER|NAVAID_IMARKER)
*/

#define NAVAID_LOC	0x0001
#define NAVAID_VOR	0x0002
#define NAVAID_DME	0x0004
#define NAVAID_GS	0x0008
#define NAVAID_NDB	0x0010
#define NAVAID_OMARKER	0x0020
#define NAVAID_MMARKER	0x0040
#define NAVAID_IMARKER	0x0080

int navaid_add_ils(char *ident, char *type, WorldCoordinates * w,
	WorldCoordinates * gsw, char *magvar, double freq, double loc_width,
	double loc_bearing, double gs_angle);

int navaid_add_vor_dme_ndb(char *ident, char *type, WorldCoordinates * w,
	char *magvar, double freq);

navaid_t * navaid_reception_check(craft * c, freq_t f);
/*
	Search for a NAVAID receivable at the frequency specified.
	Return the station found, or NULL if no station received.
*/

double navaid_mag_heading(craft * c);
/*
	Return current magnetic heading. The MH is calculated based on the VAR
	of the nearest NAVAID. The value returned varies smootly passing from
	a NAVAID to another.
*/

void navaid_free_all(void);
/*
	Release the memory about NAVAIDs and other internal data structures.
*/

#endif
