/*
 *  acm : an aerial combat simulator for X
 *  Copyright (C) 1991-1998  Riley Rainey
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 dated June, 1991.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program;  if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
 */

/*
 *  2014-12-29 dlc - Refer to CHANGES text file for details of modifications
 */

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <memory.h>
#include "damage.h"
#include "drone.h"
#include "effects.h"
#include "gear.h"
#include "hsi.h"
#include "init.h"
#include "mem_manager.h"
#include "navaid.h"
#include "panel.h"
#include "pm.h"
#include "terrain.h"
#include "weapon.h"

#ifdef AFDS
#include <afds.h>
#endif
#ifdef HAVE_DIS
	#include "dis_if.h"
#endif

static craftType ctype[MAXCRAFTTYPES];


void
freeCraftTypes (void)
{
	int i, j;

	for (i = 0; i < MAXCRAFTTYPES; ++i) {

		if (ctype[i].name != (char *) NULL) {

			FREE(ctype[i].name);

			if (ctype[i].object) {
				VDestroyObject (ctype[i].object);
			}

			if (ctype[i].objname) {
				FREE(ctype[i].objname);
			}

			if (ctype[i].description) {
				FREE(ctype[i].description);
			}

			if (ctype[i].modelname) {
				FREE(ctype[i].modelname);
			}

			if (ctype[i].CLift) {
				FREE(ctype[i].CLift->entry);
				FREE(ctype[i].CLift);
			}
			if (ctype[i].CDb) {
				FREE(ctype[i].CDb->entry);
				FREE(ctype[i].CDb);
			}
			if (ctype[i].CnBeta) {
				FREE(ctype[i].CnBeta->entry);
				FREE(ctype[i].CnBeta);
			}
			if (ctype[i].ClBeta) {
				FREE(ctype[i].ClBeta->entry);
				FREE(ctype[i].ClBeta);
			}
			if (ctype[i].Thrust) {
				FREE(ctype[i].Thrust->entry);
				FREE(ctype[i].Thrust);
			}
			if (ctype[i].ABThrust) {
				FREE(ctype[i].ABThrust->entry);
				FREE(ctype[i].ABThrust);
			}
			for (j=0; j<STATIONS; ++j) {
				if (ctype[i].station[j].type) {
					FREE(ctype[i].station[j].type);
				}
			}
			ctype[i].name = NULL;
		}
	}
}


#ifdef FIXME_NOT_USED
static int
getCraftCount(void)
{
	int       i;

	for (i = 0; i < MAXCRAFTTYPES; ++i) {
		if (ctype[i].name == (char *) NULL) {
			break;
		}
	}

	return i;
}
#endif


#ifdef FIXME_NOT_USED
static craftType *
lookupCraftByIndex(int craft_index)
{
	return &ctype[craft_index];
}
#endif


craftType *
lookupCraft(const char *name)
{
	int       i;

	for (i = 0; i < MAXCRAFTTYPES; ++i) {
		if (ctype[i].name != (char *) NULL) {
			if (strcmp(name, ctype[i].name) == 0)
				return &ctype[i];
		}
	}

	return NULL;
}

craftType *
newCraft(void)
{
	int       i;

	for (i = 0; i < MAXCRAFTTYPES; ++i)
		if (ctype[i].name == (char *) NULL) {
			memset (&ctype[i], 0, sizeof (craft));
			ctype[i].name = "*allocated*";
			return &ctype[i];
		}

	return (craftType *) NULL;
}

#ifdef HAVE_DIS

/*
 *  Find (or possible generate) the craftType entry associated with a
 *  given DIS entity type.
 */

craftType *
lookupCraftByEntityType( const dis_entity_type * id )
{
	int       i, depthcue = 1;
	craftType *c = NULL;
	entity_object_map *ep;

	for (i = 0; i < MAXCRAFTTYPES; ++i) {
		if (ctype[i].name != (char *) NULL) {

/*
 * This search could be done better; we could stop when the entityType
 * field being tested is zero.
 */

			if (ctype[i].entityType.kind == id->kind &&
				ctype[i].entityType.domain == id->domain &&
				ctype[i].entityType.country == id->country &&
				ctype[i].entityType.category == id->category &&
				ctype[i].entityType.subcategory == id->subcategory &&
				ctype[i].entityType.specific == id->specific &&
				ctype[i].entityType.extra == id->extra) {
				return &ctype[i];
			}
		}
	}

	/*
	 * Well, there wasn't a craft type defined that matched the desired
	 * entity type.  So, we'll look for an entry in the patterns contained
	 * in the eo_map.  If we find a match, create a new craftType entry
	 * and return it.
	 */

	ep = eo_map;

	for (i = 0; i < eo_map_count; ++ i) {

		FILE *f1;

		if (damage_entityWildcardMatch(id, &ep->entity_type, &ep->entity_mask)) {

			c = newCraft();
			if (c) {
				char *p = strrchr (ep->object_name, '.');

				f1 = acm_fopen (ep->object_name, "r");
				c->name = STRDUP(ep->object_name);
				c->objname = STRDUP(ep->object_name);
				c->entityType = *id;
				c->altEntityType = *id;
				c->object = NULL;
				if (p != NULL && (strcmp (p, ".dxf") == 0 || 
								  strcmp (p, ".DXF") == 0)) {
					c->object = VReadDepthCueuedDXFObject (f1, depthcue);
				}
				else {
					c->object = VReadDepthCueuedObject (f1, depthcue);
				}

				if (c->object) {
					ep->obj = c->object;
				}
				else {
					printf ("error reading object file: %s\n", ep->object_name);
				}
			 
				fclose (f1);
				return c;
			}
			else {
				printf ("craft table overflow in lookupCraftByEntityType()\n");
			}
		}
		++ ep;
	}

	return c;
}

#endif


void
printValidAircraft()
{
	int       i;

	for (i = 0; i < MAXCRAFTTYPES; ++i) {
		if (ctype[i].name != NULL) {
			if (ctype[i].CLift) {
				printf("%s\n", ctype[i].name);
			}
		}
	}
}


double
genericThrust( craft * c )
{
	double t, ts;

	if (c->flags & FL_AFTERBURNER) {
		t = interpolate(c->cinfo->ABThrust, c->mach) * c->cinfo->maxABThrust;
	}
	else {
		t = interpolate(c->cinfo->Thrust, c->mach) * c->cinfo->maxThrust;
	}

	if ( c->fuel > 0.0 )
		c->throttle = c->throttleComm;
	else
		c->throttle = 0;
	ts = c->throttle / 32768.0;
	c->rpm = (c->rpm - ts) * exp(deltaT / c->cinfo->engineLag) + ts;

	return t * c->rpm * c->rpm * c->rho / 0.002377;
}

double
fuelUsed( craft * c )
{
	double    spFuelConsump;

	if (c->flags & FL_AFTERBURNER) {
		spFuelConsump = c->cinfo->spABFuelConsump;
	}
	else {
		spFuelConsump = c->cinfo->spFuelConsump;
	}
	return spFuelConsump * c->curThrust * deltaT / 3600.0;
}

void
genericResupply(craft * c)
{

	int       i;

	if( c->type == CT_DRONE ){
		/* half refueling for drones: */
		c->fuel = 0.5 * c->cinfo->maxFuel;
	} else {
		/* +10% every 30 s while on ground: */
		c->fuel += c->cinfo->maxFuel / 10.0;
		if ( c->fuel > c->cinfo->maxFuel )
			c->fuel = c->cinfo->maxFuel;
	}

	for (i = 0; i < c->cinfo->sCount; i++) {
		c->station[i] = c->cinfo->station[i];
	}
	damage_reset(c);
}


int
newPlane(const char *planeType, int reservation_index)
{

	int       i;
	craft    *c;

	/*
	 *  Begin our search for an entry at the specified reservation location
	 */

	if (reservation_index == -1) {
		i = 0;
	}
	else {
		i = reservation_index;
	}

	for (; i < MAXPLAYERS; ++i) {
		if (ptbl[i].type == CT_FREE || ptbl[i].type == CT_RESERVED) {
			break;
		}
	}

	if (i == MAXPLAYERS)
		return -1;

	c = &ptbl[i];
	if ((c->cinfo = lookupCraft(planeType)) == NULL) {
		return -2;
	}

	c->type = CT_PLANE;
	c->Cg.x = 0.0;
	c->Cg.y = 0.0;
	c->Cg.z = 0.0;

	/*
		Ensure terrain_altitude() will update the local altitude at the
		first call
	*/

	c->terrain_altitude_timeout = curTime - 10.0;
	/* c->terrain_altitude = UNKNOW; */

	c->p = c->q = c->r = 0.0;
	c->pitchComm = c->rollComm = c->steerComm = c->rudderComm = 0.0;
	c->Se = c->Sr = c->Sa = 0.0;
	c->SeTrim = c->SaTrim = 0.0;

	c->throttleComm = 8192; /* 25% */
	c->throttle = 0;
	c->curThrust = (c->cinfo->thrust) (c);
	c->rpm = (double) c->throttle / 32768.0;
	c->thrust_reverse_on = FALSE;
	c->thrust_reverse_pos = 1.0;

	c->curFlap = 0.0;
	c->flapSetting = 0.0;
	c->curSpeedBrake = 0.0;
	c->speedBrakeSetting = 0.0;
	c->curHeading = c->curPitch = c->curRoll = 0.0;
	VIdentMatrix(&(c->trihedral));
	c->curNWDef = 0.0;
	c->flags = 0;
	c->curRadarTarget = -1;
	pm_hud_strings_alloc(c);

	/* AutoPilot System (automatically allocated by aps.c): */
	c->aps = NULL;

	/* Magnetic VAR handling: */
	c->showMag = 0;
	c->actualLocalVAR = 0.0;
	c->updActualLocalVAR = 0.0;
	c->indicatedLocalVAR = 0.0;
	c->updIndicatedLocalVAR = 0.0;
	c->radarMode = RM_OFF;

	gear_down(c);

/*
 *  rearm and fuel the aircraft.
 */

	(*c->cinfo->resupply) (c);

	c->w.z = 0.0; /* sea level -- caller must set properly */

	/* Set c->curWeapon: */
	selectNamedWeapon(c, WEAPON_NULL);
#ifdef DIS_DEBUG
        fprintf(stderr, "%8.2f: newPlane() %s ptbl[%d] disId=%d\n",
                curTime, c->cinfo->name, i, c->disId);
#endif

	return i;
}
