/*
 *  acm : an aerial combat simulator for X
 *  Copyright (C) 1991-1998  Riley Rainey
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 dated June, 1991.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program;  if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
 */

/*
 *  2014-12-29 dlc - Refer to CHANGES text file for details of modifications
 */
 
#include <stdio.h>
#include <X11/Xutil.h>
#ifndef __hpux
	#include <X11/Xos.h>
#endif
#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <X11/cursorfont.h>

#include "Vlib.h"
#include "adf.h"
#include "audio.h"
#include "aps.h"
#include "box.h"
#include "browse.h"
#include "drone.h"
#include "effects.h"
#include "gear.h"
#include "hsi.h"
#include "hud.h"
#include "instruments.h"
#include "list.h"
#include "magnetic_compass.h"
#include "mem_manager.h"
#include "navaid.h"
#include "planes.h"
#include "pm.h"
#include "prompt.h"
#include "redraw.h"
#include "render.h"
#include "terrain.h"
#include "terminal.h"
#include "update.h"
#include "viewer.h"
#include "windows.h"
#ifdef HAVE_DIS
	#include "dis_if.h"
#endif

#if defined(HAVE_STDLIB_H)
#ifndef bzero
#define bzero(s,n)      memset ((s), 0, (n))
#endif
#endif							/* HAVE_STDLIB_H */

#define ARG_BORDER_COLOR        "bordercolor"
#define ARG_BORDER              "borderwidth"
#define ARG_GEOMETRY            "geometry"
#define DEFAULT_BACKGROUND      "#7c99b6"
#define DEFAULT_BORDER          "black"

#define SW_BORDER	      1
#define SW_BACKGROUND	  2
#define SW_GEOM		      4
#define SW_TEAM		      5
#define SW_TEAM_1	      6
#define SW_TEAM_2	      7
#define SW_DEFAULT_VISUAL 8
#define SW_PLANE	      9
#define SW_LIST_PLAYER	 10
#define SW_MONOCHROME	 11
#define SW_HUD_MODE      12
#define SW_PASSIVE       13
#define SW_LATITUDE      14
#define SW_LONGITUDE     15
#define SW_ALTITUDE      16
#define SW_AIRSPEED_KT   17
#define SW_HEADING       18
#define SW_END_GAME      19
#define SW_NO_SOUND      20
#define SW_FUEL          21
#define SW_PAYLOAD       22
#define SW_GROUNDC       23
#define SW_LAST 23             /* last entry in list */


static struct {
	char     *sw;
	int       value;
} swt[] = {
	{"-bw",        SW_BORDER},
	{"-skycolor",  SW_BACKGROUND},
	{"-geometry",  SW_GEOM},
	{"-team",      SW_TEAM},
	{"-team_1",    SW_TEAM_1},
	{"-team_2",    SW_TEAM_2},
	{"-stealth",   SW_PASSIVE},
	{"-cmap",      SW_DEFAULT_VISUAL},
	{"-mono",      SW_MONOCHROME},
	{"-plane",     SW_PLANE},
	{"-list",      SW_LIST_PLAYER},
	{"-latitude",  SW_LATITUDE},
	{"-longitude", SW_LONGITUDE},
	{"-altitude",  SW_ALTITUDE},
	{"-airspeed-kt", SW_AIRSPEED_KT},
	{"-heading",   SW_HEADING},
	{"-end-game",  SW_END_GAME},
	{"-no-sound",  SW_NO_SOUND},
	{"-fuel",      SW_FUEL},
	{"-payload",   SW_PAYLOAD},
	{"-hud-mode",   SW_HUD_MODE},
	{"-groundc",    SW_GROUNDC},
	{NULL, 0}
}, *swp;


static Visual   *
get_pseudo_visual(Display * display, int screen, unsigned int *depth)
{

	XVisualInfo vTemplate;
	XVisualInfo *visualList;
	int       i, visualsMatched;

/*
 *  Get all Visuals on this screen
 */

	vTemplate.screen = screen;
	visualList = XGetVisualInfo(display, VisualScreenMask,
								&vTemplate, &visualsMatched);

	if (visualsMatched == 0)
		return NULL;

/*
 *  look first for a suitable PseudoColor visual, then a GreyScale visual
 *  if no PseudoColor is present
 */

	for (i = 0; i < visualsMatched; ++i) {
		if (visualList[i].visual->class == PseudoColor &&
			visualList[i].depth >= 4) {
			*depth = visualList[i].depth;
			XFree((char *) visualList);
			return visualList[i].visual;
		}
	}

	for (i = 0; i < visualsMatched; ++i) {
		if (visualList[i].visual->class == GrayScale &&
			visualList[i].depth >= 4) {
			*depth = visualList[i].depth;
			XFree((char *) visualList);
			return visualList[i].visual;
		}
	}

	XFree((char *) visualList);

	return NULL;
}


int
newPlayer(char *display, char *logname, list *switches)
{
	int       argc;
	char    **argv;
	XSizeHints xsh;				/* Size hints for window manager */
	Colormap  cmap;
	unsigned long bd;			/* Pixel values */
	unsigned long bw;			/* Border width */
	char     *tempstr;			/* Temporary string */
	XColor    color;			/* Temporary color */
	char     *geomSpec;			/* Window geometry string */
	XWMHints  xwmh;				/* Window manager hints */
	Cursor    cursor;
	char     *c;
	static char *background = NULL;
	int       borderWidth = -1;
	int       player;
	viewer   *u;
	craft    *cf;
	int       i;
	int       screen;
	int       team = 1;
	char     *plane = NULL;		/* name of plane type */
	int       width, height;	/* dimensions of main window */
	unsigned  depth, mono = 0;
	long      win_attr_mask;
	Visual   *theVisual;
	XSetWindowAttributes window_attributes;
	int       useDefaultVisual = 1;
	int       passive = 0;
	Atom      atom[2];
	_BOOL     hud_mode;
	int       overrides[SW_LAST+1];
	char     *overrideLatitude = NULL;
	char     *overrideLongitude = NULL;
	double    overrideAltitude = 0.0,
	          overrideHeading_rad = 0.0, overrideAirspeed_fps = 0.0,
	          overrideFuel = 0.0,
			  overridePayload = 0.0;
	int       end_game = 0;
	int no_sound = 0;
        int       ii, jj;

#ifdef HAVE_DIS
	double    disLocation[3];
	double    disZeroVec[3];
	double    disOrientation[3];

#endif


/*
 *  Parse command line
 */

	memset(overrides, 0, sizeof(overrides));
	geomSpec = NULL;
	hud_mode = FALSE;

 	if( switches == NULL ){
		argc = 0;
		argv = NULL;  /* make happy gcc -Wall */
	} else {
		argc = switches->n;
		argv = switches->arr;
	}

	for( i=0; i<argc; i++ ){
		c = argv[i];
		if( *c == '-' ){
			for (swp = &swt[0]; swp->value != 0; ++swp) {
				if (strcmp(swp->sw, c) == 0) {

					switch (swp->value) {

					case SW_GEOM:
						i++;
						if( i < argc )
							geomSpec = argv[i];
						break;

					case SW_END_GAME:
						end_game = 1;
						break;

					case SW_BORDER:
						i++;
						if( i < argc )
							borderWidth = atoi( argv[i] );
						break;

					case SW_BACKGROUND:
						i++;
						if( i < argc )
							background = argv[i];
						break;

					case SW_TEAM:
						i++;
						if( i < argc )
							team = atoi( argv[i] );
						break;

					case SW_TEAM_1:
						i++;
						if( i < argc ) {
                        				jj = -1;
                        				for ( ii = 0 ; ii < ttblCount && jj != 0 ; ii++ ) {
                                				if ( strcmp(ttbl[ii].sname, argv[i]) == 0 ) {
                                        				teamLatLon[1].latitude  = ttbl[ii].loc.latitude;
                                        				teamLatLon[1].longitude = ttbl[ii].loc.longitude;
                                        				teamLatLon[1].z         = ttbl[ii].loc.z;
                                        				teamHeading[1]          = ttbl[ii].heading;
                                        				jj = 0;
                                				}
                        				}
                        				if ( jj == -1 ) {
                                				fprintf(stderr, "newPlayer() ERROR: The TEAMLOC record for %s was not found\n",
                                        				argv[i]);
                                				exit(1);
                        				}
                        			}
						break;

					case SW_TEAM_2:
						i++;
						if( i < argc ) {
                        				jj = -1;
                        				for ( ii = 0 ; ii < ttblCount && jj != 0 ; ii++ ) {
                                				if ( strcmp(ttbl[ii].sname, argv[i]) == 0 ) {
                                        				teamLatLon[2].latitude  = ttbl[ii].loc.latitude;
                                        				teamLatLon[2].longitude = ttbl[ii].loc.longitude;
                                        				teamLatLon[2].z         = ttbl[ii].loc.z;
                                        				teamHeading[2]          = ttbl[ii].heading;
                                        				jj = 0;
                                				}
                        				}
                        				if ( jj == -1 ) {
                                				fprintf(stderr, "newPlayer() ERROR: The TEAMLOC record for %s was not found\n",
                                        				argv[i]);
                                				exit(1);
                        				}
                        			}
						break;

					case SW_PASSIVE:
					    passive = 1;
						break;

					case SW_PLANE:
						i++;
						if( i < argc )
							plane = argv[i];
						break;

					case SW_MONOCHROME:
						mono = 1;
						break;

					case SW_LATITUDE:
						i++;
						if( i < argc ){
							overrideLatitude = argv[i];
							overrides[SW_LATITUDE] = 1;
						}
						break;

					case SW_LONGITUDE:
						i++;
						if( i < argc ){
							overrideLongitude = argv[i];
							overrides[SW_LONGITUDE] = 1;
						}
						break;

					case SW_ALTITUDE:
						i++;
						if( i < argc ){
							overrideAltitude = FEETtoMETERS(atof( argv[i] ));
							overrides[SW_ALTITUDE] = 1;
						}
						break;

					case SW_AIRSPEED_KT:
						i++;
						if( i < argc ){
							overrideAirspeed_fps = KTtoFPS(atof( argv[i] ));
							if (overrideAirspeed_fps > KTtoFPS(2500.0)) {
								printf ("You really should slow down.\n");
								printf ("At least to less than 2500 knots.\n");
							}
							overrides[SW_AIRSPEED_KT] = 1;
						}
						break;

					case SW_HEADING:
						i++;
						if( i < argc ){
							overrideHeading_rad = DEGtoRAD(atof( argv[i] ));
							overrides[SW_HEADING] = 1;
						}
						break;

					case SW_LIST_PLAYER:
						printf("\nname\t\tnumber\n");
						printf("-------------------------\n");
						for (team = 0; team < MAXPLAYERS; team++) {
							if (ptbl[team].type == CT_PLANE) {
								printf("%-16s  %d\n", ptbl[team].name, team);
							}
						}
						return -1;

					case SW_DEFAULT_VISUAL:
						useDefaultVisual = 0;
						break;

					case SW_NO_SOUND:
						no_sound = 1;
						break;

					case SW_FUEL:
						i++;
						if( i < argc ){
							overrideFuel = atof( argv[i] );
							overrides[SW_FUEL] = 1;
						}
						break;

					case SW_PAYLOAD:
						i++;
						if( i < argc ){
							overridePayload = atof( argv[i] );
							overrides[SW_PAYLOAD] = 1;
						}
						break;

					case SW_HUD_MODE:
						hud_mode = TRUE;
						break;

					case SW_GROUNDC:
						i++;
						if( i < argc ) {
                        				jj = -1;
                        				for ( ii = 0 ; ii < gtblCount && jj != 0 ; ii++ ) {
                                				if ( strcmp(gtbl[ii].name, argv[i]) == 0 ) {
                                        				groundColor  = gtbl[ii].vcptr;
                                        				jj = 0;
                                				}
                        				}
                        				if ( jj == -1 ) {
                                				fprintf(stderr, "newPlayer() ERROR: The GROUNDC record for %s was not found\n",
                                        				argv[i]);
                                				exit(1);
                        				}
                        			}
						break;

					default:  abort();
					}
					break;
				}
			}

			if (swp->value == 0) {
#ifdef TRACE_ARGS
				fprintf(stderr, "newPlayer() %s invalid players switch %s ignored\n", ACM, argv[i]);
#endif
/*				return -1;
 */
			}
		}
	}

	if (!plane) {
		plane = (team == 1) ? "F-16" : "MiG-29";
	}

	player = newPlane(plane, -1);

	if (player < 0) {
		if (player == -1) {
			printf("Sorry, no room for any more players at this moment.\n");
		}
		else {
			printf("You have selected an unknown plane type. Choose one among these types:\n");
			printValidAircraft();
		}
		return -1;
	}

	cf = &ptbl[player];

/*
 *  assign an initial location based on the player's team.
 */

	if (team != 2) {
		team = 1;
	}

/*
	Set fuel and paylod
	===================
*/

	if (overrides[SW_FUEL]) {
		if (overrideFuel > cf->cinfo->maxFuel) {
			printf("Too much fuel for this plane. Maximum is %.0f lb.\n",
				cf->cinfo->maxFuel);
			cf->fuel = cf->cinfo->maxFuel;
		} else {
			cf->fuel = overrideFuel;
		}
	} else {
		cf->fuel = cf->cinfo->maxFuel;
	}

	if (overrides[SW_PAYLOAD]) {
		cf->payload = overridePayload;
	} else {
		cf->payload = 150.0;  /* weight of the pilot (lb) */
	}

/*
	Set position
	============
*/

	cf->w.latitude  = teamLatLon[team].latitude;
	cf->w.longitude = teamLatLon[team].longitude;
	
	if ( (overrides[SW_LATITUDE] && ! overrides[SW_LONGITUDE])
	|| (! overrides[SW_LATITUDE] && overrides[SW_LONGITUDE]) ) {
		printf("You must specify both the `-latitude' option AND the `-longitude' option, not only one. Ignoring invalid setting.\n");
	}

	if (overrides[SW_LATITUDE] && overrides[SW_LONGITUDE] ){

		char s[1000];  /* FIXME: fix possible buffer overflow */
		WorldCoordinates q;
		sprintf(s, "%s %s", overrideLatitude, overrideLongitude);
		if ( DISStringToLatLong(s, &q) ) {
			cf->w.latitude = q.latitude;
			cf->w.longitude = q.longitude;
		}
		else {
			printf("Invalid syntax in position `%s', using default team position.\n", s);
		}

	}

	/*
		Set some reasonable initial altitude:
	*/

	cf->w.z = 0.0; /* dummy value (sea level) */
	DISWorldCoordinatesToGeocentric(&cf->w, (void *) & cf->Sg);
	
	/* Force terrain altitude recalculation: */
	cf->terrain_altitude_timeout = 0.0;

	if (overrides[SW_ALTITUDE]) {
		cf->w.z = overrideAltitude;
		cf->curPitch = DEGtoRAD(3.0);  /* ensure some initial lift */
		gear_up(cf);
	} else {
		/*
			FIXME: remove teamLatLon from source, manual and
			sceneries. Or change its meaning: if near to the
			ground use gear_ground_altitude() otherwise set
			this value verbatim.

			cf->w.z = teamLatLon[team].z;
		*/
		double h;
		gear_ground_altitude_pitch(cf,  &h, &cf->curPitch);
		/* WARNING. terrain_altitude() requires cf->Sg be already defined. */
		cf->w.z = terrain_altitude(cf) + FEETtoMETERS(h);
		gear_down(cf);
	}
	
/*
	Update resulting final position:
	===============================
*/

	DISWorldCoordinatesToGeocentric(&cf->w, (void *) & cf->Sg);
	
	cf->prevSg = cf->Sg;

/*
	Set magnetic variation based on the actual position
	===================================================
*/
	cf->updActualLocalVAR = curTime - 1.0;  /* force VAR recalculation */
	/* ignore = */ navaid_mag_heading(cf); /* set cf->actualLocalVAR */
	cf->indicatedLocalVAR = cf->actualLocalVAR;
	cf->updIndicatedLocalVAR = curTime + 2.0;

/*
	Set attitude and heading
	========================
*/

	if (overrides[SW_HEADING]) {
		cf->curHeading = pm_normalize_yaw( overrideHeading_rad - cf->actualLocalVAR );
	} else {
		cf->curHeading = pm_normalize_yaw( teamHeading[team] );
	}

	cf->curRoll = 0.0;

	cf->r = cf->q = cf->p = 0.0;

/*
	Generate trihedral and world-to-aircraft matrices
	=================================================
*/

	eulerToMatrix(cf->curRoll, cf->curPitch, cf->curHeading,
		&(cf->trihedral));

	GenerateWorldToLocalMatrix(&cf->w, &cf->XYZtoNED);

/*
	Set airspeed
	============
*/

	if (overrides[SW_AIRSPEED_KT]) {
		VMatrix turn;
		VPoint v = { overrideAirspeed_fps, 0.0, 0.0 };
		VIdentMatrix(&turn);
		VRotate(&turn, ZRotation, cf->curHeading);
		VTransform_(&v, &turn, &cf->Cg );
	} else {
		VSetPoint(&cf->Cg, 0, 0, 0);
		cf->flags |= FL_BRAKES; /* Ensure brakes are if not moving at start */
	}

/*
	Open a new viewer
	=================
*/

	cf->vl = NULL;

	u = viewer_new(cf);

	u->hud_mode = hud_mode;

	/*
	 *  If we're passive (stealth mode), wipe out aircraft information;
	 *  this entry will become a placeholder for our browsing view.
	 */

	if ( passive ) {

		u->viewer_state = ViewerStateBrowsing;
		cf->type = CT_DIS_STEALTH;
		cf->cinfo = NULL;
		cf->radarMode = RM_DIS_BROWSE;

		if ( end_game ) {
			end_game_mode = 1;
		}

		/*
		 *  Stealth a specific entity?  Go ahead and set it as the subject.
		 *
		 *  This requires snooping for entities in the PDU stream prior
		 *  to looking for the entity in the table.
		 */

		if ( subjectEntitySpecified ) {
			craft *c;

			printf ("Building entity database ... ");
			fflush ( stdout );

			updateSimTimeFromSystemClock ();
			dis_snoop ( 5500 );

			printf ("done.\n");

			c = locateCraftByDISEntityID ( & subjectEntityID );
			if ( c ) {

				stealthCraft ( c, u, -1, 1);

			}
		}

	}
	else {

		if ( end_game ) {
			printf("The -end-game switch is only valid when used " );
			printf("with stealth mode.\n" );
			printf("The switch will be ignored.\n\n" );
		}

		if ( subjectEntitySpecified ) {
			printf("The -subject-entity-id switch is only valid when used " );
			printf("with stealth mode.\n" );
			printf("The switch will be ignored.\n\n" );
		}

	}

	if ((u->dpy = XOpenDisplay(display)) == (Display *) NULL) {
		cf->type = CT_FREE;
		printf("%s: can't open %s\n", ACM, display);
		FREE(u);
		return -1;
	}
	screen = DefaultScreen(u->dpy);

	setOutsideView(u->c, u, VIEW_FORWARD);

/*
 *  If the player has specified that he want the default Visual, simply
 *  give 'em that along with the default Colormap.
 */

	if (useDefaultVisual) {

		theVisual = DefaultVisual(u->dpy, screen);
		cmap = DefaultColormap(u->dpy, screen);
		depth = DisplayPlanes(u->dpy, screen);

	}

/*
 *  Look for a visual; if we can't find one,
 *  fall back to monochrome mode.
 */

	else {
		if ((theVisual = get_pseudo_visual(u->dpy, screen, &depth)) == NULL) {
			theVisual = DefaultVisual(u->dpy, screen);
			cmap = DefaultColormap(u->dpy, screen);
			depth = DefaultDepth(u->dpy, screen);
		}
		else {
			cmap = XCreateColormap(u->dpy, RootWindow(u->dpy, screen),
								   theVisual, AllocNone);
		}
	}

/*
 * Select colors for the border, the window background, and the
 * foreground.  We use the default colormap to allocate the colors in.
 */

	if (background == NULL)
		background = DEFAULT_BACKGROUND;

	if ((tempstr = XGetDefault(u->dpy, ACM, ARG_BORDER_COLOR)) == NULL)
		tempstr = DEFAULT_BORDER;
	if (XParseColor(u->dpy, cmap, tempstr, &color) == 0) {
		XCloseDisplay(u->dpy);
		FREE(u);
		cf->type = CT_FREE;
		printf("Can't get border color %s\n", tempstr);
		return -1;
	}

	if (depth == 1)
		bd = BlackPixel(u->dpy, screen);
	else {
		if (XAllocColor(u->dpy, cmap, &color) == 0) {
			XCloseDisplay(u->dpy);
			FREE(u);
			cf->type = CT_FREE;
			printf("Cannot allocate color cells\n");
			return -1;
		}
		bd = color.pixel;
	}

/*
 * Set the border width of the window, and the gap between the text
 * and the edge of the window, "pad".
 */

	if (borderWidth >= 0)
		bw = borderWidth;
	else if ((tempstr = XGetDefault(u->dpy, ACM, ARG_BORDER)) == NULL)
		bw = 1;
	else
		bw = atoi(tempstr);

/*
 * Deal with providing the window with an initial position & size.
 * Fill out the XSizeHints struct to inform the window manager.
 */

	if (geomSpec == NULL)
		geomSpec = XGetDefault(u->dpy, ACM, ARG_GEOMETRY);

/*
	If the defaults database doesn't contain a specification of the
	initial size, set the size leaving to the window manager to set
	the position:
*/

	if (geomSpec == NULL) {
		xsh.flags = /* PPosition | */ PSize;
		xsh.height = DisplayHeight(u->dpy, screen) * 3 / 4;
		xsh.width = DisplayWidth(u->dpy, screen) * 3 / 4;
		/**
		xsh.x = (DisplayWidth(u->dpy, screen) - xsh.width) / 2;
		xsh.y = (DisplayHeight(u->dpy, screen) - xsh.height) / 2;
		**/
		xsh.x = 0; /* ignored */
		xsh.y = 0; /* ignored */
	}
	else {
		int       bitmask;

		bzero((char *) &xsh, sizeof(xsh));
		bitmask = XParseGeometry(geomSpec, &xsh.x, &xsh.y,
			 (unsigned int *) &xsh.width,
			 (unsigned int *) &xsh.height);
		if (bitmask & (XValue | YValue)) {
			xsh.flags |= USPosition;
		}
		if (bitmask & (WidthValue | HeightValue)) {
			xsh.flags |= USSize;
		}
	}

	width = xsh.width;
	height = xsh.height;

/*
 * Create the Window with the information in the XSizeHints, the
 * border width,  and the border & background pixels.
 */

	win_attr_mask = CWColormap | CWBitGravity | CWBackPixel;
	window_attributes.colormap = cmap;
	window_attributes.bit_gravity = NorthWestGravity;
	window_attributes.background_pixel = BlackPixel(u->dpy, screen);

	u->win = XCreateWindow(u->dpy, RootWindow(u->dpy, screen),
						   xsh.x, xsh.y, xsh.width, xsh.height, bw, depth,
			  InputOutput, theVisual, win_attr_mask, &window_attributes);

/*
 * Set the standard properties and hints for the window managers.
 */

	XSetStandardProperties(u->dpy, u->win, ACM, ACM, None, argv, argc, &xsh);
	xwmh.flags = InputHint | StateHint;
	xwmh.input = True;
	xwmh.initial_state = NormalState;
	XSetWMHints(u->dpy, u->win, &xwmh);
	cursor = XCreateFontCursor(u->dpy, XC_tcross);
	XDefineCursor(u->dpy, u->win, cursor);

/*
 *  Tell the window manager that we'd like to participate in the
 *  WM_CLOSEDOWN and WM_DELETE_WINDOW protocols.
 */

	u->protocolsAtom = XInternAtom(u->dpy, "WM_PROTOCOLS", False);
	atom[0] =
		u->closedownAtom = XInternAtom(u->dpy, "WM_CLOSEDOWN", False);
	atom[1] =
		u->deleteWindowAtom = XInternAtom(u->dpy, "WM_DELETE_WINDOW", False);
	XSetWMProtocols(u->dpy, u->win, atom, 2);

/*
 *  Fill-in the viewer structure
 */
	XSelectInput(u->dpy, u->win, KeyPressMask | KeyReleaseMask
		| ButtonPressMask | StructureNotifyMask | ButtonReleaseMask
		| ExposureMask | FocusChangeMask);
	cf->team = team;

	strncpy(cf->name, logname, sizeof(cf->name));
	strncpy(cf->display, display, sizeof(cf->display));

/*
 *  Connect to the audio server
 */

	if ( ! no_sound && ! audio_init(u, display) ) {
		printf("Sound is not available on workstation %s\n", display);
		printf("Unfortunately, this program is really limited in sound support, sorry :->.\n");
		printf("Disable the sound at all to remove this boring message (acm -no-sound).\n");
	}

#ifdef USE_PIXMAP_ANIMATION
	if (depth != 1)
		_VDefaultWorkContext->usePixmaps = 1;
#endif

	_VDefaultWorkContext->usePixmaps = mono ? 1 : 0;

	u->w = AlibOpen(u->dpy, screen, u->win, cmap, theVisual, width, height);

/*
 *  If this is a color-rich visual, then enable depth-cueing mode
 */

	if (depth >= 6) {
		AlibSetDepthCueing(u->w, 1);
	}

	if (VBindColors(u->w, background) < 0) {
		XCloseDisplay(u->dpy);
		VCloseViewport(u->v);
		AlibClose(u->w);
		FREE(u);
		ptbl[player].type = CT_FREE;
		printf("Error in binding colors, try reducing depth cueing parameter.\n");
		return -1;
	}

	whitePixel = whiteColor->cIndex;
	blackPixel = blackColor->cIndex;
	HUDPixel = HUDColor->cIndex;

/*
 * Map the window to make it visible.
 */

	XMapWindow(u->dpy, u->win);
	if (depth != 1) {
#ifdef NOTDEF
		register int curPixel;

		VExposeBuffer(u->v, u->gc);
		curPixel = *(u->v->pixel);
		XSetForeground(u->dpy, u->gc, curPixel);
		XFillRectangle(u->dpy, u->win, u->gc, 0, 0, u->width, u->height);
#endif
	}

	if (depth == 1)
		XSetWindowBackground(u->dpy, u->win, BlackPixel(u->w->display, screen));
	else
		XSetWindowBackground(u->dpy, u->win, VConstantColor(u->w, blackPixel));

	u->z.color = u->w->pixel[HUDPixel];
	u->z.depth = 1;

	u->rz.color = u->w->pixel[radarColor->cIndex];
	u->rz.depth = 2;

	/*
		Now that all the colors are set we can enable the istruments
		so they can grab their colors:
	*/

	windows_set_layout(cf, u, width, height, hud_mode);

	if( hud_mode ){
		hud_enable(u);
	} else {
		inst_enable(u);
	}

	++ptblCount;

	/*
	 *  Transmit initial DIS entity state PDU
	 */
	
#ifdef HAVE_DIS
	if ( disInUse && cf->type != CT_DIS_STEALTH ) {

		DISWorldCoordinatesToGeocentric(&cf->w,
			  (void *) disLocation);
		disZeroVec[0] = 0.0;
		disZeroVec[1] = 0.0;
		disZeroVec[2] = 0.0;
		disOrientation[0] = cf->curHeading;
		disOrientation[1] = 0.0;
		disOrientation[2] = 0.0;

		dis_entityEnter(team, cf,
						&cf->cinfo->entityType,
						&cf->cinfo->altEntityType,
						disLocation, disZeroVec, disZeroVec,
						disOrientation, disZeroVec, &cf->disId);
	}
#endif

	cf->showMag = TRUE;

	magnetic_compass_enable(u);

	cf->radarMode = RM_HSI;
	hsi_enable(u);

        u->stickM = KEYSTICK_DOUBLE; /* Initialize the keyboard cursor mode */
        u->stickS = 1;               /* Initialize the keyboard cursor sensitivity */
	u->showMark = TRUE;          /* Initialize disabled cursor display */
	u->testKeys = FALSE;         /* Initialize pilot commands dry run mode */
	u->hColor = 0;               /* HUD: 0=green, 1=red, 2=blue, 3=gray */
	u->hBright = 0;              /* HUD: 0=brightest, 1=bright, 2=normal, 3=dark */
	u->hint = 0;                 /* Initialize hints array index */

	return 0;
}


int
killPlayer(craft *c, char *reason)
{

	viewer   *v, *vn;
	int       i;
	VPoint	 vel = { 0, 0, 0 };

	audio_playSound(c, SoundCrash, FALSE);

/*
 *  Notify to all the players the reason way this craft is dead
 */

	if( (reason != NULL) && (c->type == CT_PLANE || c->type == CT_DRONE) ){
		/* FIXME: check for buffer overflow */
		char s[1000]; /* FIXME: remove static limit */
		sprintf(s, "KILLED %s %s team %d: %s",
			c->name, c->cinfo->name, c->team, reason);
		prompt_broadcast_print(s);

		/* The owner of the window can see the message, use its terminal: */
		if( c->type == CT_PLANE )
			printf("%s\n", s);
	}

/*
 *  Decrement the player count, iff this is a real person that just got
 *  killed.
 */

	if (c->type == CT_PLANE && (c->flags & FL_BLACK_BOX) == 0) {
		--ptblCount;
	}

/*
 *  Erase our radar emissions
 */

	for (i = 0; i < MAXPLAYERS; ++i)
		ptbl[i].rval[c->pIndex] = 0.0;

/*
 *  Replace the plane with an explosion.
 */

	effects_new_explosion(&(c->Sg), &vel, 30.0, 15.0, 4.0);

/*
 *  Release services tied to the craft
 */

 	aps_free(c);

	drone_release_commands(c);

 	pm_hud_strings_free(c);  /* FIXME: HUD data should go in hud.c */

/*
 *  Close all the viewers tied to this craft
 */

	while( c->vl != NULL ){

		v = c->vl;

		/*
		 *  If this was a situation where we had grabbed control in steath
		 *  mode and then died, return to the browsing state.
		 */

		if ( v->viewer_state == ViewerStateNormal
		&& v->watchedCraft != NULL
		&& v->watchedCraft->type == CT_DIS_STEALTH
		) {
			printf ("player killed: returning to stealth browsing mode\n");
			v->viewer_state = ViewerStateBrowsing;

			/*
				FIXME: the current status of the viewe does not match
				that of the watched craft, we should release the services
				(audio, terminal, etc. as in viewer_free()) and then
				initialize them properly.
			*/
			v->c = v->watchedCraft;

			v->watchedCraft = NULL;
			v = NULL;
		}
		else {
			vn = c->vl;
		}

		viewer_free(v);
	}

	if (c->flags & FL_RECORD) {
		--recordCount;
	}

	if (c->flags & FL_BLACK_BOX)
		blackBoxKillPlayer(c->pIndex);

#ifdef HAVE_DIS
	if (c->type == CT_PLANE || c->type == CT_DRONE) {
		/* active (broadcasting) player, announce its death */
		fprintf(stderr, "%8.2f: killPlayer() %s ptbl[%d] disId=%d\n",
			curTime, c->cinfo->name, c->pIndex, c->disId);
		dis_entityExit(c->disId);
	}
#endif

	c->type = CT_FREE;
	return 0;
}
