/*
 *  acm : an aerial combat simulator for X
 *  Copyright (C) 1991-1998  Riley Rainey
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 dated June, 1991.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program;  if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
 *
 */

/*
 *  2008-03-28 Umberto Salsi, main structural changes of this module:
 *  - the landing gear now returns forces and moments that are added to the
 *    aerodynamic ones
 *  - twoOrder() modified to account for constant term (landing gear moment)
 *  - the "gear" module now completely deal with the c->flags concerning
 *    the status of the landing gear
 */

#include <stdio.h>
#include <math.h>
#include <string.h>

#define PM_IMPORT
#include "pm.h"

#include "aps.h"
#include "damage.h"
#include "air.h"
#include "dis_if.h"
#include "gear.h"
#include "planes.h"
#include "prompt.h"
#include "terrain.h"
#ifdef AFDS
	#include <afds.h>
#endif

#ifdef WIN32
#include <float.h>
#endif


double    deltaT;				/* Update interval in seconds */
double    halfDeltaTSquared;	/* 0.5 * deltaT * deltaT */
static double    CM, CN, C_Y;

char * ctypes[] = {
	"FREE",         /*  0 an unused craft entry */
	"PLANE",        /*  1 a player */
	"MISSILE",      /*  2 an air to air missile */
	"CANNON",       /*  3 a stream of cannon fire */
	"SURFACE",      /*  4 surface object (e.g. a runway) */
/*	"CHASER",           5 a non-player "watcher" chasing a plane (FIXME: not used) */
	"UNASSIGNED_5", /*  5 currently unassigned--should never exist */
	"DRONE",        /*  6 a target drone */
	"EXPLOSION",    /*  7 an explosion */
#ifdef HAVE_DIS
	"DIS_PLANE",    /*  8 external player (on different server) */
	"DIS_MUNITION", /*  9 external tracked munition */
	"DIS_CANNON",   /* 10 external untracked munition (cannon) */
	"RESERVED",     /* 11 a reserved, uninitialized entry */
	"DIS_STEALTH",  /* 12 a place-holder for stealths */
	"BOMB",         /* 13 gravity bomb */
#endif
};

char * flags[] = { /* indexed by FX_name macro value */
	"RECORD",            /* (1<< 0) x000001 activate recording function */
	"AFTERBURNER",       /* (1<< 1) x000002 afterburner state */
	"BRAKES",            /* (1<< 2) x000004 wheel brake state */
	"RUNWAY",            /* (1<< 3) x000008 this object is a runway */
	"HAS_GYRO",          /* (1<< 4) x000010 missile is gyroscope equiped */
	"GHANDLE_DN",        /* (1<< 5) x000020 landing gear handle state (1 = DOWN) */
	"GEAR0_UP",          /* (1<< 6) x000040 set when nose gear is completely up */
	"GEAR1_UP",          /* (1<< 7) x000080 set when right gear is completely up */
	"GEAR2_UP",          /* (1<< 8) x000100 set when left gear is completely up */
	"NW_GND_CONTACT",    /* (1<< 9) x000200 nose wheel gnd contact -- we can brake */
	"RW_GND_CONTACT",    /* (1<<10) x000400 right wheel in contact with ground */
	"LW_GND_CONTACT",    /* (1<<11) x000800 left wheel in contact with ground */
	"FIXED_OBJECT",      /* (1<<12) x001000 fixed (surface) object */
	"CHASE_VIEW",        /* (1<<13) x002000 chase plane view */
	"BLACK_BOX",         /* (1<<14) x004000 object is from a black box recording */
	"BALLISTIC",         /* (1<<15) x008000 munition has no target */
	"RADAR_MODE_CHANGE", /* (1<<16) x010000 set when user changes radar mode */
	"END_GAME_DRONE",    /* (1<<17) x020000 drone in CALSPAN end-game mode */
	"UNASSIGNED_18",     /* (1<<18) x040000 */
	"UNASSIGNED_19",     /* (1<<19) x080000 */
	"UNASSIGNED_20",     /* (1<<20) x100000 */
	"UNASSIGNED_21",     /* (1<<21) x200000 */
	"UNASSIGNED_22",     /* (1<<22) x400000 */
	"UNASSIGNED_23",     /* (1<<23) x800000 */
};

/*
 *  twoorder: solve linear second-order differential equation with initial
 *      conditions known.
 *
 *  y_dot_dot - d * y_dot - k * y - q = 0
 *
 *  given the initial conditions:
 *
 *        y(0) == y.
 *        y_dot(0) == v.
 *
 *  Results are *newy=y(deltaT) and *newv=y_dot(deltaT) for deltaT seconds
 *  into the future.
 *  
 */

static void
twoOrder(double k, double d, double q, double y, double v,
	double *newy, double *newv)
{

	double  qk, s, s1, s2, t, ac, c1, c2, exp_s1_x, exp_s2_x;

/*
	Prevent numerical unstabilities: since y,v are pitch/yaw angle and its
	derivative, it would be perfectly unuseful to consider factors too
	little.
*/

	if( fabs(d) < 1e-9 )
		d = 0.0;

	if( fabs(k) < 1e-9 )
		k = 0.0;
	
	if( k == 0.0 ){
		if( d == 0.0 ){
			/* The equation reduces to: y_dot_dot = q. */
			*newy = 0.5 * q * deltaT * deltaT + v * deltaT + y;
			*newv = q * deltaT + v;
		} else {
			/* The equation reduces to: y_dot_dot - d*y_dot - q = 0. */
			*newy = (v + q/d)/d*exp(d*deltaT) - q/d*deltaT + y - (v + q/d)/d;
			*newv = (v + q/d)*exp(d*deltaT) - q/d;
		}
		*newy = y;
		*newv = v;
		return;
	}
	qk = q/k;

	ac = d * d + 4.0 * k;

	if ( ac > 0.0 ) {
		/*
			The general solution for ac > 0 is:
			y = c1 * exp(s1*x) + c2 * exp(s2*x) - q/k
			where:
			s1 = ...see code...
			s2 = ...see code...
			c1 = ...see code...
			c2 = ...see code...
		*/
		t = sqrt(ac);
		s1 = (d + t) * 0.5;
		s2 = (d - t) * 0.5;
		c1 = (-s2*y - s2*qk + v) / t;
		c2 = ( s1*y + s1*qk - v) / t;
		exp_s1_x = exp(s1 * deltaT);
		exp_s2_x = exp(s2 * deltaT);
		*newy = c1 * exp_s1_x + c2 * exp_s2_x - qk;
		*newv = c1 * s1 * exp_s1_x + c2 * s2 * exp_s2_x;
	}
	else if (ac < 0.0) {
		/*
			The general solution for ac < 0 is:
			y = exp(s*x) * (c1 * cos(t*x) + c2 * sin(t*x)) - qk
			where:
			s = d/2
			t = sqrt(-ac)/2
			c1 = y0 + qk
			c2 = (v-c1*s)/t
		*/
		double cos_t_x, sin_t_x;
		s = d * 0.5;
		t = sqrt(-ac) * 0.5;
		exp_s1_x = exp(s*deltaT);
		cos_t_x = cos(t*deltaT);
		sin_t_x = sin(t*deltaT);
		c1 = y + qk;
		c2 = (v - c1*s ) / t;
		*newy = exp_s1_x * ( c1*cos_t_x + c2*sin_t_x ) - qk;
		*newv = exp_s1_x * ( (c1*s+c2*t) * cos_t_x + (c2*s-c1*t) * sin_t_x );
	}
	else {
		/*
			The general solution for ac = 0 is:
			y = (c1 + c2*x) * exp(s*x) - qk
			where:
			s = d/2
			c1 = y0 + qk
			c2 = v - c1*s
		*/
		s = 0.5*d;
		c1 = y + qk;
		c2 = v - c1*s;
		exp_s1_x = exp(s*deltaT);
		*newy = (c1 + c2*deltaT) * exp_s1_x - qk;
		*newv = (c2 + (c1 + c2*deltaT)*s) * exp_s1_x;
	}
}


/*
 *  calcCoefficients :  Calculate CLift and friends
 */

static void
calcCoefficients(craft * c, double *CLift, double *CDrag)
{

	double    CDAlpha, CDBeta;
	craftType *p = c->cinfo;
	double    h, base, k;

/*
 *  We used to interpolate these values, but now use several characteristic
 *  equations to compute these values for a given alpha value. The basic
 *  formulas are:
 *
 *
 *   C  = C        + (alpha * (C       + sin(curFlap) * cFlap ))
 *    L    LOrigin              LSlope
 *
 *
 *   C  = zero-lift-wave-and-body-drag + induced-drag +
 *    D     speed-brake-drag + flap drag + landing-gear-drag +
 *      drag-based-on-sideslip
 *
 *  There are independent equations defining drag resulting from alpha
 *  and beta values.  The hypoteneuse of those two values becomes the
 *  resultant CDrag value.
 */

	*CLift = interpolate(p->CLift, c->alpha);

	if( damage_isFunctioning(c, SYS_FLAPS) )
		*CLift += sin(c->curFlap) * p->cFlap;

	/*
		Rough estimation of the ground effect correction to the CLift:
		FIXME: also the drag coeff. should be corrected
	*/

	/* Height of the wing aerodynamic center above the ground (ft): */
	h = METERStoFEET(c->w.z - terrain_altitude(c)) + p->wingh;
	/* Limit ground effect, or we would "float" over the terrain... */
	if( h < 1.0 )
		h = 1.0;
	base = h / p->wings;
	/* Compute the expensive pow() only if k will be > 1.001: */
	if ( base < 8.5 ) {
		/* Ground effect correction coefficient to the CLift: */
		k = 1.0 + 0.025 * pow(base, -1.5);
		*CLift *= k;
	}

	CM = p->cmSlope + c->damageCM;

	CDAlpha = interpolate(p->CDb, c->mach) +
		*CLift * *CLift / (pi * p->aspectRatio);
	if( damage_isFunctioning(c, SYS_SPEEDBRAKE ) )
		CDAlpha += sin(c->curSpeedBrake) * p->cSpeedBrake;
	if( damage_isFunctioning(c, SYS_FLAPS) )
		CDAlpha += sin(c->curFlap) * p->cFlapDrag;

	if( (c->flags & (FL_GEAR0_UP | FL_GEAR1_UP | FL_GEAR2_UP)) == 0 )
		CDAlpha += (sin(c->curGear[0])
					+ sin(c->curGear[1])
					+ sin(c->curGear[2])) / 3.0 * p->cGearDrag;

	if (fabs(c->beta) > p->betaStall)
		CN = interpolate(p->CnBeta, fabs(c->alpha)) * fabs(sin(c->beta));
	else
		CN = interpolate(p->CnBeta, fabs(c->alpha));

	CDBeta = p->CDBOrigin + p->CDBFactor *
		sin(c->beta + p->CDBPhase);

	*CDrag = sqrt(CDAlpha * CDAlpha + CDBeta * CDBeta);

	C_Y = p->CYbeta * c->beta /* * fabs(cos(c->beta))*/;

}


static double ATAN2(double y, double x)
{
	double a;

	a = atan2(y, x);
	if( a >= -M_PI && a <= M_PI )
		return a;
	else
		return 0.0;
}


double
heading(VPoint * x)
{

	double    m;

	m = ATAN2(x->y, x->x);

	if (m < 0.0)
		return (pi * 2.0 + m);
	else
		return m;
}

/*
 *  Convert a transformation matrix into the equivalent
 *  heading, pitch and roll angles.
 */

#define EPSILON 1.0e-6

void
matrixToEuler(VMatrix * mt, double *heading, double *pitch, double *roll)
{
	double    sin_theta;

	sin_theta = -mt->m[2][0];

	if (fabs(sin_theta) > 1.0 - EPSILON) {
		/* we have the nose pointing very close to straight up or straight down,
		   set roll to zero and compute the resulting heading */

		*heading = ATAN2(-mt->m[0][1], mt->m[1][1]);
		if (*heading < 0.0)
			*heading += 2.0 * M_PI;

		if (sin_theta > 0.0)
			*pitch = M_PI / 2.0;
		else
			*pitch = -M_PI / 2.0;

		*roll = 0.0;
	}
	else {
		*heading = ATAN2(mt->m[1][0], mt->m[0][0]);
		if (*heading < 0.0)
			*heading += 2.0 * M_PI;

		*pitch = asin(sin_theta);

		*roll = ATAN2(mt->m[2][1], mt->m[2][2]);
	}
}

void
euler(craft * c)
{
	double lastcurHeading;
/*
	FIXME: this is simply

	matrixToEuler(&c->AXYZtoNED, &c->curHeading, &c->curPitch, &c->curRoll);
*/

	double i, j, k, m;

	i = c->AXYZtoNED.m[0][0];
	j = c->AXYZtoNED.m[1][0];
	k = c->AXYZtoNED.m[2][0];

	lastcurHeading = c->curHeading; m = -99999.99999;
	if (i == 0.0 && j == 0.0)
		c->curHeading = 0.0;
	else if ((m = ATAN2(j, i)) < 0.0)
		c->curHeading = pi * 2.0 + m;
	else
		c->curHeading = m;

	fprintf(stderr, "euler: last curHeading %.6f, new curHeading %.6f, m %12e\n",
		lastcurHeading, RADtoDEG(c->curHeading), m);

	c->curPitch = -asin(k);

	c->curRoll = ATAN2(c->AXYZtoNED.m[2][1], c->AXYZtoNED.m[2][2]);
}


void
calcGForces(craft * c, VPoint * f, double w)
{

	VPoint  t, t1;
	double	m_slugs;

	m_slugs = w / earth_g;

	t = *f;
	t.x = t.x / m_slugs;
	t.y = t.y / m_slugs;
	t.z = t.z / m_slugs;

	VReverseTransform_(&t, &c->AXYZtoNED, &c->linAcc);

	t.z -= earth_g;

	VReverseTransform_ (&t, &c->AXYZtoNED, &t1);

	c->G.x = t1.x / earth_g;
	c->G.y = t1.y / earth_g;
	c->G.z = t1.z / earth_g;

}


static void
calcAlphaBetaVT(craft * c, double *alpha, double *beta, double *VT)
{

	VPoint    C, air, *wind;

	wind = air_get_wind(c->w.z);

	air.x = c->Cg.x + wind->x;
	air.y = c->Cg.y + wind->y;
	air.z = c->Cg.z + wind->z;

	VReverseTransform_(&air, &c->AXYZtoNED, &C);

	*alpha = ATAN2(C.z, sqrt(C.y * C.y + C.x * C.x));
	*beta = ATAN2(C.y, C.x);
	*VT = Vmagnitude(&C);
}


void eulerToMatrix(double roll, double pitch, double heading, VMatrix * m)
{

	register double sinPhi, cosPhi, sinTheta, cosTheta, sinPsi, cosPsi;

	sinPhi = sin(roll);
	cosPhi = cos(roll);
	sinTheta = sin(pitch);
	cosTheta = cos(pitch);
	sinPsi = sin(heading);
	cosPsi = cos(heading);

	m->m[0][0] = cosTheta * cosPsi;
	m->m[0][1] = sinPhi * sinTheta * cosPsi - cosPhi * sinPsi;
	m->m[0][2] = cosPhi * sinTheta * cosPsi + sinPhi * sinPsi;
	m->m[1][0] = cosTheta * sinPsi;
	m->m[1][1] = sinPhi * sinTheta * sinPsi + cosPhi * cosPsi;
	m->m[1][2] = cosPhi * sinTheta * sinPsi - sinPhi * cosPsi;
	m->m[2][0] = -sinTheta;
	m->m[2][1] = sinPhi * cosTheta;
	m->m[2][2] = cosPhi * cosTheta;
	m->m[0][3] = m->m[1][3] = m->m[2][3] = 0.0;
	m->m[3][0] = m->m[3][1] = m->m[3][2] = 0.0;
	m->m[3][3] = 1.0;

}


double
elevatorSetting(craft * c, double qS, double w)
{

	register double s, n, L, an;
	register craftType *p = c->cinfo;

	s = c->pitchComm + aps_ap_get_delta_elevator(c) + c->SeTrim;
	if (s < -1.0) s = -1.0;
	if (s > +1.0) s = +1.0;
	c->Se = s;

/*
 *  Limit the load factor that will be generated.
 */

	L = p->effElevator * p->CLSlope;

	an = cos(s * p->effElevator);

	n = an * (- s * L + p->CLOrigin) * qS / w;

	if (n > 9.5)
		s = - (9.5 / (an * qS) * w - p->CLOrigin) / L;
	else if (n < -3.0)
		s = - (-3.0 / (an * qS) * w - p->CLOrigin) / L;
	
	return s;
}

static double
aileronSetting(craft * c)
{
	double    s;

	if( ! damage_isFunctioning(c, SYS_WINGS) )
		return 0.0;

	s = c->rollComm + c->SaTrim;
	if (s < -1.0) s = -1.0;
	if (s > +1.0) s = +1.0;
	c->Sa = s;

	return s;
}

static double
rudderSetting(craft * c)
{
	double    s;

	s = c->rudderComm + aps_ac_get_delta_rudder(c);
	if (s < -1.0) s = -1.0;
	if (s > +1.0) s = +1.0;
	c->Sr = s;

	return s;
}

char *
flightCalculations(craft * c)
{

	craftType *p = c->cinfo;
	double    qS, s, CLift, CDrag, Sr;
	double    ClBeta;
	double    FLift, FDrag, FWeight, FSideForce;
	double    deltaRoll, deltaPitch, deltaYaw;
	double    y, newy;
	double    xa, inertia, torque, xd, r0;
	double    dNorth, dEast, dmag, dHeading_rad;
	double	  mass_slugs;
	VPoint    F, M, Fg, aeroF, aeroM, gearF, gearM, thrF, thrM;
	VMatrix   mtx, new_AXYZtoNED;
	dis_entity_appearance appearance;
	char      *killReason;

	c->prevSg = c->Sg;
	c->rho = air_calc_rho(METERStoFEET(c->w.z), &c->mach1);
	calcAlphaBetaVT(c, &c->alpha, &c->beta, &c->VT);
	Sr = rudderSetting(c);

/*
 *  A note about thrust:  Normal thrust diminishes in proportion to the
 *  decrease in air density.
 */

	c->mach = c->VT / c->mach1;

	appearance = 0;
	if (disInUse) {
		appearance = dis_getEntityAppearance(c->disId);
		appearance &= ~(DISAppearanceAirAfterburnerOn | DISAppearancePlatformPowerplantOn);
	}

/*
	Update thrust reverse position
*/

	if( c->thrust_reverse_on ){
		/* deploy thrust reverser whitin 3 s */
		c->thrust_reverse_pos -= deltaT / 3.0;
		if( c->thrust_reverse_pos < -0.5 )
			c->thrust_reverse_pos = -0.5;
	} else {
		/* retract thrust reverser whitin 3 s */
		c->thrust_reverse_pos += deltaT / 3.0;
		if( c->thrust_reverse_pos > 1.0 )
			c->thrust_reverse_pos = 1.0;
	}

/*
	Total engine force and moment.
*/

	c->curThrust = (*p->thrust) (c);
	if (c->flags & FL_AFTERBURNER) {
		appearance |= DISAppearanceAirAfterburnerOn;
	}
	appearance |= DISAppearancePlatformPowerplantOn;
	if (disInUse) {
		dis_setEntityAppearance(c->disId, appearance);
	}

	VSetPoint(&thrF, c->thrust_reverse_pos * c->curThrust, 0.0, 0.0);
	VSetPoint(&thrM, 0.0, 0.0, 0.0);

/*
	Compute the resultant aerodynamic force vector on the aircraft. By the
	way, the variable "qS" should more properly be named "qS" -- it is the
	dynamic pressure times S, the reference wing area.
*/

	calcCoefficients(c, &CLift, &CDrag);
	ClBeta = interpolate(p->ClBeta, fabs(c->alpha));

	qS = c->rho * p->wingS * c->VT * c->VT * 0.5;
	if( damage_isFunctioning(c, SYS_WINGS) )
		FLift = CLift * qS;
	else
		FLift = 0.0;
	FDrag = CDrag * qS;
	FSideForce = C_Y * qS;
	FWeight = p->emptyWeight + c->fuel + c->payload;

/*
	Detect wings structural maximum limit. Note that FLift sign is reversed
	respect to the aircraft z axis so that FLinft is positive when the
	craft is pulled up.
*/

	if( FLift > p->maxLoadZPositive ){
		c->damageBits |= SYS_WINGS | SYS_FLAPS
			| SYS_SPEEDBRAKE | SYS_HYD1 | SYS_HYD2;
		prompt_craft_print(c, "WINGS FAILURE due to excessive positive load");
	} else if( FLift < - p->maxLoadZNegative ){
		c->damageBits |= SYS_WINGS | SYS_FLAPS
			| SYS_SPEEDBRAKE | SYS_HYD1 | SYS_HYD2;
		prompt_craft_print(c, "WINGS FAILURE due to excessive negative load");
	}

/*
	Warn the pilot when the wings load exceeds 75% of maximum load.
*/

	if( FLift > 0.75*p->maxLoadZPositive
	|| FLift < -0.75 * p->maxLoadZNegative )
		c->damageBits |= FLAG_MAX_G_LOAD;
	else
		c->damageBits &= ~FLAG_MAX_G_LOAD;

/*
 *  Set stall warning flag.
 */
	
	if( (c->flags & FL_ALL_GND_CONTACT) == 0
	&& c->alpha > c->cinfo->alpha_stall )
		c->damageBits |= FLAG_STALL_WARN;
	else
		c->damageBits &= ~FLAG_STALL_WARN;

/*
	Total aerodynamic force and moment.
	These expressions convert lift and drag forces from wind axes to the
	aircraft fixed axes.  The conversion is based on the wind to
	aircraft transformation matrix supplied in "Airplane Design" by
	Donald Crawford (page 90).
*/

	VSetPoint(&aeroF,
		FLift * sin(c->alpha) - FDrag * cos(c->alpha) * cos(c->beta),
		- FDrag * sin(c->beta) + FSideForce,
		-FLift * cos(c->alpha) - FDrag * cos(c->beta) * sin(c->alpha));
	
/*
 * Aerodynamic torque (for future improvements, for example engine torque,
 * landing gear torque, etc.):
 */

	VSetPoint(&aeroM, 0.0, 0.0, 0.0);

/*
	Detect if speed is greater than the maximum allowed speed Vne:
*/

	if( p->Vne > 0.0
	&& c->VT > KTtoFPS(p->Vne)
	&& (c->damageBits & SYS_WINGS) == 0 ){
		c->damageBits |= SYS_WINGS | SYS_FLAPS | SYS_SPEEDBRAKE;
		prompt_craft_print(c, "WINGS FAILURE due to excessive airspeed");
	}

/*
	With flaps deployed, detect if speed is greater then the maximum allowed
	speed Vfe:
*/

	if( p->Vfe > 0.0
	&& c->curFlap > 0.0
	&& c->VT > KTtoFPS(p->Vfe)
	&& (c->damageBits & SYS_WINGS) == 0 ){
		c->damageBits |= SYS_FLAPS;
		c->curFlap = 0.0;
		c->flapSetting = 0.0;
		prompt_craft_print(c, "FLAPS FAILURE due to excessive airspeed");
	}


/*
 *  Compute forces and moments due to the wheels friction on the terrain
 */

 	killReason = gear_ground_dynamics(c, &gearF, &gearM);
	if( killReason != NULL )
		return killReason;

/*
	Total force and moment.
	FIXME: weight still missing.
 */

	F.x = aeroF.x + gearF.x + thrF.x;
	F.y = aeroF.y + gearF.y + thrF.y;
	F.z = aeroF.z + gearF.z + thrF.z;

	M.x = aeroM.x + gearM.x + thrM.x;
	M.y = aeroM.y + gearM.y + thrM.y;
	M.z = aeroM.z + gearM.z + thrM.z;

/*
 *  Compute fuel consumption
 */

	c->fuel -= fuelUsed(c) + c->leakRate * deltaT;
	if (c->fuel < 0.0)
		c->fuel = 0.0;

/*
 *  Resolve roll-axis (X-axis) changes
 */

/*
	FIXME: to avoid divisions by zero, c->VT cannot be == 0, that's why
	of this 0.001 term.
*/
	xa = p->wings * p->wings * p->wingS * c->rho * (c->VT + 0.001) * p->Clp;
	inertia = p->I.m[0][0];
/*
	FIXME: the momentum M is inserted in a quite arbitrary
	point, check if this is the correct one.
*/
	torque = qS * p->wings * 2.0 *
		(p->Clda * - aileronSetting(c) * p->maxAileron
		+ ClBeta * c->beta
		+ p->Cldr * Sr * p->maxRudder)
		//+ M.x  /* FIXME: check */
		+ c->damageCL * qS;
	xd = c->p + torque / xa;
	r0 = xd * inertia / xa;
	deltaRoll = xd * inertia / xa * exp(xa / inertia * deltaT)
		- deltaT * torque / xa - r0;
	//c->p = xd * exp(xa / inertia * deltaT) - torque / xa;
	c->p = xd * exp(xa / inertia * deltaT) - torque / xa + gearM.x / inertia * deltaT;

/*
 *  Resolve pitch-axis (Y-axis) changes
 */

	y = c->alpha + elevatorSetting(c, qS, FWeight) *
		p->effElevator;
	twoOrder(CM * qS * p->c / p->I.m[1][1],
		(0.25 * p->wingS * c->rho * p->c * p->c *
		c->VT * p->Cmq) / p->I.m[1][1],
		M.y / p->I.m[1][1],
		y, c->q, &newy, &(c->q));
	deltaPitch = newy - y;

/*
 *  Resolve yaw-axis (Z-axis) changes.
 *
 *  We do some trickery here.
 *  If the absolute value of the sideslip angle is greater than 90 degrees,
 *  we trick the code into believing that the sideslip angle is the negative
 *  of its reciprocal value (e.g. -176 becomes -4 degrees).  We do this with
 *  the (somewhat inaccurate) assumption that the CN value for that angle is
 *  roughly equal to the other.
 */

	y = Sr * p->effRudder - c->beta;

	if (y > pi / 2.0) {
		y = pi - y;
	}
	else if (y < -pi / 2.0) {
		y = -pi - y;
	}

	s = p->wings;
	twoOrder(CN * qS * s / p->I.m[2][2],
		(p->wingS * c->rho * s * s *
		c->VT * p->Cnr) / p->I.m[2][2],
		M.z / p->I.m[2][2],
		y, c->r, &newy, &(c->r));
	deltaYaw = newy - y;

/*
 *  Compute new aicraft trihedral, but don't set it yet.
 */

	eulerToMatrix(deltaRoll, deltaPitch, deltaYaw, &mtx);
	VMatrixMultByRank(&mtx, &c->AXYZtoNED, &new_AXYZtoNED, 3);


/*
 *  Tranform the total force F in AXYZ into Fg in NED and add weight
 */

	VTransform_(&F, &c->AXYZtoNED, &Fg);
	Fg.z += FWeight;

	calcGForces(c, &Fg, FWeight);
		
/*
 *  Update our position (in flight mode).
 */
	mass_slugs = FWeight / earth_g;

	dNorth  = FEETtoMETERS(c->Cg.x * deltaT + Fg.x / mass_slugs * halfDeltaTSquared);
	dEast   = FEETtoMETERS(c->Cg.y * deltaT + Fg.y / mass_slugs * halfDeltaTSquared);
	c->w.z -= FEETtoMETERS(c->Cg.z * deltaT + Fg.z / mass_slugs * halfDeltaTSquared);

	dmag = sqrt(dNorth * dNorth + dEast * dEast);
	fprintf(stderr, "fC: dN=%12e dE=%12e wz=%12e dmag=%12e curHeading %.6f\n",
		dNorth, dEast, c->w.z, dmag, RADtoDEG(c->curHeading));

	dHeading_rad = 0.0;

	if( fabs(dmag) > 1e-6 /* epsilon */ ) {
		DISUpdateWorldCoordinatesEx (&c->w,
			dNorth / dmag, dEast / dmag, dmag, &dHeading_rad);
		fprintf(stderr, "fC: DUWCE returned heading %12e radians, %.6f degrees\n",
			dHeading_rad, RADtoDEG(dHeading_rad));
	}

/*
 *  Update velocity vector based on acceleration
 */

	c->Cg.x += Fg.x / mass_slugs * deltaT;
	c->Cg.y += Fg.y / mass_slugs * deltaT;
	c->Cg.z += Fg.z / mass_slugs * deltaT;

/*
 *  Now rotate the trihedral and velocity vector to reflect the change
 *  in heading at the new spheroid location.
 */

/* 20200120 dlc -- try avoiding transport wander effect at the poles this way:	*/
	if (c->w.latitude > (M_PI_2 * 0.95) || c->w.latitude < (-M_PI_2 * 0.95) ) {
		c->curHeading = dHeading_rad;
	}

	VIdentMatrix( &mtx );
	VRotate( &mtx, ZRotation, dHeading_rad );
	VMatrixMultByRank( &mtx, &new_AXYZtoNED, &c->AXYZtoNED, 3 );
	c->AXYZtoNED = new_AXYZtoNED;
	VTransform_(&c->Cg, &mtx, &c->Cg);

	euler(c);

	if( c->w.z - terrain_altitude(c) <= 0.0 ){
		if( ! damage_isFunctioning(c, SYS_WINGS) )
			return "crash due to structural failure";
		else
			return "main landing gear wasn't down and locked or it was damaged";
	}

	DISWorldCoordinatesToGeocentric(&c->w, (void *) & c->Sg);

	GenerateWorldToLocalMatrix(&c->w, &c->XYZtoNED);

	aps_update( c );

	if (c->w.latitude > (M_PI_2 * 0.97) || c->w.latitude < (-M_PI_2 * 0.97) ) {
		fprintf(stderr, "leaving fC near pole, curHeading is %.6f radians, %.6f degrees\n",
			c->curHeading, RADtoDEG(c->curHeading));
	}


	return NULL;
}


double pm_normalize_roll(double roll)
{
	while( roll < M_PI )  roll += 2*M_PI;
	while( roll > M_PI )  roll -= 2*M_PI;
	return roll;
}


double pm_normalize_pitch(double pitch)
{
	while( pitch < -M_PI/2 ) pitch += 2*M_PI;
	while( pitch > M_PI/2 ) pitch -= 2*M_PI;
	return pitch;
}


double pm_normalize_yaw(double yaw)
{
	while( yaw < 0 )  yaw += 2*M_PI;
	while( yaw >= 2*M_PI )  yaw -= 2*M_PI;
	return yaw;
}


void pm_hud_strings_alloc(craft *c)
{
	int i;

	for( i=0; i<6; i++ ){
		if( c->leftHUD[i] == NULL )
			c->leftHUD[i] = MALLOC(32);
		c->leftHUD[i][0] = '\0';

		if( c->rightHUD[i] == NULL )
			c->rightHUD[i] = MALLOC(32);
		c->rightHUD[i][0] = '\0';
	}
}


void pm_hud_strings_free(craft *c)
{
	int i;

	for( i=0; i<6; i++ ){
		FREE(c->leftHUD[i]);   c->leftHUD[i] = NULL;
		FREE(c->rightHUD[i]);  c->rightHUD[i] = NULL;
	}
}


void pm_thrust_reverse_toggle(craft *c)
{
	if( c->thrust_reverse_on ){
		c->thrust_reverse_on = FALSE;

	/* thrust setting must be idle to activate reverser: */
	} else if( c->cinfo->hasThrustReverser
	&& c->throttleComm <= 32768 * 25 / 100 ){
		c->thrust_reverse_on = TRUE;
	}
}


void pm_after_burner_toggle(craft *c)
{
	if ( c->cinfo->maxThrust != c->cinfo->maxABThrust )
		c->flags ^= FL_AFTERBURNER;
}


void pm_after_burner_on(craft *c)
{
	if ( c->cinfo->maxThrust != c->cinfo->maxABThrust )
		c->flags |= FL_AFTERBURNER;
}


void pm_after_burner_off(craft *c)
{
	c->flags &= ~FL_AFTERBURNER;
}


static void printFlagsFail()
{
	fprintf(stderr, "\n%s\n", "in printFlags, string area too short--aborting!");
	exit(2);
}

void printFlags(unsigned long int ct)
{
	unsigned int c, i, j;
	char str[256];

	c = j = 0;
	for ( i = 0; i <= FX_MAX; i++ ) {
		if (ct & (1 << i)) {
			if (c > 0)
				str[j++] = ',';
			if ((j + strlen(flags[i])) > sizeof(str))
				printFlagsFail;
			j += snprintf(&str[j], sizeof(str) - j, "%s", flags[i]);
			c++;
        	}
        }
	if (j < sizeof(str))
		str[j++] = 0;
	else	printFlagsFail;

	fprintf(stderr, "%s>\n", str);
	return;
}
