/*
 *  acm : an aerial combat simulator for X
 *  Copyright (C) 1991-1998  Riley Rainey
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 dated June, 1991.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program;  if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
 */

/*
 *  2014-12-29 dlc - Refer to CHANGES text file for details of modifications
 */

#ifndef _pm_h
#define _pm_h

#include <Vlib.h>
#include <VObjects.h>
#include "manifest.h"
#include "scale.h"
#include "interpolate.h"
/*
 * eventually we will include gv_sys.h
 */
#define HAVE_GVS 1
#define GVS_OBI void *
#define GVS_OBD void *


#ifdef WIN32
	typedef BOOL _BOOL;
#else
	typedef int _BOOL;
#endif

#ifndef TRUE
	#define TRUE    1
	#define FALSE   0
#endif


struct _craft;

#ifdef HAVE_DIS

	#include <dis/dis.h>

#else

	/*
	 *  We need this typedef to store information from the inventory file,
	 *  even if we are not using DIS.
	 */

	struct dis_entity_type {
		unsigned char kind;
		unsigned char domain;
		unsigned short country;
		unsigned char category;
		unsigned char subcategory;
		unsigned char specific;
		unsigned char extra;
	};
	typedef struct dis_entity_type dis_entity_type;

#endif                          /* HAVE_DIS */


#ifdef PM_IMPORT
	#define EXTERN
	#undef PM_IMPORT
#else
	#define EXTERN extern
#endif


#ifdef FLAT_WORLD

typedef struct _worldcoordinates {
	double    latitude;         /* [RAD, north positive] */
	double    longitude;        /* [RAD, east positive] */
	double    z;                /* above reference ellipsoid [meters] */
} WorldCoordinates;

#endif

#ifdef AFDS

	#include <fplan.h>

#endif

/*
 *  redraw.c contains some simple procedures to help minimize
 *  redrawing panel items.
 */

typedef struct _drawnItem {
	struct _drawnItem *next;
	int       redraw;
} drawnItem;

typedef struct _drawnItemController {
	drawnItem *head;
} drawnItemController;


/*
 *  Each rendering window has a _viewer structure which may be in one
 *  of several states.
 */

#define ViewerStateNormal         0  /* just a plane that we control */
#define ViewerStateBrowsing       1  /* not yet associated with an entity */
#define ViewerStatePiggyback      2  /* a stealth glued to an entity */
#define ViewerStateControlPending 3  /* requested control of an entity */
//#define ViewerStateControlling    4  /* have control of another entity (FIXME: not used) */

typedef struct _viewer {

	/*
	 *   Linked list of all active rendering viewers
	 */

	struct _viewer *vl_next;

	/*
		Next viewer attached to this craft. Since usually there is
		just only one viewer per craft, this field is usually NULL.
	*/

	struct _viewer *next;

	/*
	 *  Viewers used to be tied completely to a given aircraft; this is
	 *  no longer the case.  The state value is used to track relocation
	 *  from one entity to the next.
	 */

	int viewer_state;

	struct _craft *c;            /* back pointer to associated craft */
	struct _craft *watchedCraft; /* back pointer to associated craft, 
									when in stealth mode */

	_BOOL     hasFocus;         /* win can receive input from keyb|mouse? */
	_BOOL     hasComm;          /* pilot sitting at the commands? */
	_BOOL     noMorJ;           /* pilot disabled mouse/joystick movements */
	enum 				{KEYSTICK_LINEAR, KEYSTICK_DOUBLE}
	          stickM;           /* keystick mode (sensitivity type) */
	int       stickC, stickS;   /* keyboard stick command and sensitivity info */
	int       stickX, stickY;   /* current cursor X-coordinates */
	int       hColor, hBright;  /* HUD display color and brightness */
	_BOOL     showMark;         /* display disabled cursor location */
	_BOOL     testKeys;         /* only show "Pilot Action" name (dry run mode) */
	AWindow  *w;                /* Alib window */
	Viewport *v;                /* cockpit project. Viewport inside AWindow */
	VPoint    viewDirection;    /* view direction in AXYZ */
	VPoint    viewUp;           /* "up" view direction in AXYZ */

/* FIXME: most of this X Window related data already available in Alib: */
	Display  *dpy;              /* display for this user */
	Window    win;              /* window for this user */
	GC        gc;               /* GC for drawing */
	int       cn;               /* X connection fd */
	Atom      protocolsAtom;
	Atom      deleteWindowAtom;
	Atom      closedownAtom;
/* FIXME: X Window related data ends here. */

	int       hud_yCenter;      /* organize HUD drawing around (xCenter,hud_yCenter) */

	int       zoom;             /* zoom factor *16 */
	int       hint;             /* index of next hint to display */

	_BOOL  hud_mode;
	/*
		HUD or classic instruments panel? If TRUE, the HUD is displayed,
		otherwise the classic instruments panel is displayed.  This is a
		read-only flag, as it set by (windows)resizePlayerWindow()
	*/

	/* Slots used only in "classic instruments" mode: */
	Rect      stripe;          /* timer and APS lights */
	Rect      turn;            /* turn and slip indicator */
	Rect      anemometer;      /* airspeed indicator */
	Rect      attitude;        /* pitch and roll indicator */
	Rect      altimeter;       /* altitude indicator */
	Rect      vsi;             /* vertical speed indicator */

	/* Slots used both in HUD mode and classic instruments mode:         */
	Rect      lights;          /* status lights */
	Rect      radar_altimeter; /* radar altimeter */
	Rect      tuner;           /* HSI tuner, ADF tuner and TEWS indicator share this */
	Rect      indicator;       /* HSI, ADF and radar indicators share this */
	Rect      gear;            /* gear and flaps indicators */
	Rect      engine;          /* engine gauge */
	Rect      curmark;         /* cursor location marker in noMorJ mode */

	scaleControl altScale;     /* altitude scale control */
	scaleControl velScale;     /* airspeed scale control */
	scaleControl hdgScale;     /* heading scale control */

	double    xscaleFactor;    /* overall scaling factor for HUD */
	double    yscaleFactor;    /* overall scaling factor for HUD */
	ZInfo     z;               /* depth/color information for HUD */
	ZInfo     rz;              /* depth/color information for radar */
	Segment   radarImage[1024];  /* the last radar frame */
	int       radarImageCount;

	int       browseBase;      /* first drawn DIS entity in RM_DIS_BROWSE */
	int       browseSelectedItem;
	unsigned long browseClickTime;

	/*
		Private data structures handled by the corresponding modules:
	*/
	void     *hud;              /* see hud module */
	void     *magnetic_compass; /* see magnetic_compass module */
	void     *hsi;              /* see hsi module */
	void     *adf;              /* see adf module */
	void     *audio;            /* see audio module */
	void     *inst;             /* see instruments module */
	void     *terminal;         /* see terminal module */
	void     *prompt;           /* see prompt module */
} viewer;


typedef struct _wctl {
	int       id;               /* weapon type WEAPON_* */
	char     *type;             /* type of weapon at this station */
	int       info;             /* extra info (weapon specific) */
	int       info2;            /* "      " */
	int       info3;            /* "      " */
	double    info4;            /* time offset to next round firing (s) */
} weaponStation;

/*
 *  This structure describes a class of aircraft (e.g. an F-16C).
 */

typedef struct _craftType {
	char     *name;             /* short name of aircraft class */
	char     *description;      /* long name */

	char     *modelname;        /* compiled object model to drive this aircraft type */

	double    aspectRatio;      /* wing aspect ratio */

	double    CLOrigin, CLSlope;    /* Defines the CL characteristic eqn */
	double    betaStall;        /* Stall angle for rudder */
	double    CDOrigin, CDFactor;   /* Defines the CD Characteristic eqn */
	ITable   *CLift;            /* compute lift coefficent */
	ITable   *CDb;              /* compute body + wave drag coeff */
	ITable   *CnBeta;           /* compute yaw moment due to sideslip */
	ITable   *ClBeta;           /* compute roll moment due to sideslip */
	double    CDBOrigin, CDBFactor;     /* Defines the CD Characteristic eqn */
	double    CDBPhase;
	double    CYbeta;           /* Side-force from side-slip (dCY/dBeta) */
	double    Clda;             /* roll moment from aileron offset */
	double    Cldr;             /* roll moment from rudder offset */
	double    Clp;              /* roll damping */

	double    Cmq;              /* damping in pitch */
	double    Cnr;              /* damping in yaw */

	double    maxAileron;       /* maximum aileron offset */
	double    maxRudder;        /* maximum rudder offset */

	double    effElevator;      /* Elevator effectiveness */
	double    effRudder;        /* Rudder effectiveness */

	VMatrix   I;                /* Moments of Inertia about CG in AXYZ (lb ft^2) */
	double    cmSlope;          /* CmAlpha curve slope */

	double    maxFlap;          /* maximum flap setting (RAD) */
	double    cFlap;            /* lift coefficient of flaps */
	double    cFlapDrag;        /* drag coefficient of lowered flaps */
	double    flapRate;         /* flap movement rate (RAD/s) */

	double    cGearDrag;        /* drag coefficient of lowered gear */
	double    gearRate;         /* landging gear movement rate (RAD/s) */

	double    maxSpeedBrake;    /* maximum speed brake setting (RAD) */
	double    cSpeedBrake;      /* drag coefficient of 90 degree speed brake */
	double    speedBrakeRate;   /* rate of speed brake movement (RAD/s) */
	double    speedBrakeIncr;   /* number of radians than one keystroke moves brake */

	double    wingS;            /* wing area (ft^2) */
	double    wings;            /* wing half-span (ft) */
	double    wingh;            /* wing areodyn. center height above the CM (ft) */
	double    c;                /* mean aerodynamic chord (MAC) (ft) */
	double    emptyWeight;      /* empty mass (lb) */
	double    maxLoadZPositive; /* max climb wing load (lbf) */
	double    maxLoadZNegative; /* max sink wing load (lbf) */
	double    maxFuel;          /* maximum internal fuel (lb) */

	double    maxThrust;        /* max static thrust, military power (lbf) */
	double    maxABThrust;      /* max static thrust, afterburner on  (lbf) */
	double    (*thrust) (struct _craft *);  /* computes current thrust */
	ITable   *Thrust;           /* Change in thrust due to mach number */
	ITable   *ABThrust;         /* Change in thrust due to mach number */
	int       hasThrustReverser;/* 0 = no, otherwise yes */
	double    engineLag;        /* controls lag between throttle and RPM */
	double    spFuelConsump;    /* specific fuel consump(lb fuel/lbf T x hr) */
	double    spABFuelConsump;
	VPoint    viewPoint;        /* pilot's viewing location wrt CG (ft) */

	double    muStatic;         /* static coefficient of friction no-brakes */
	double    muKinetic;        /* moving coefficient of friction no-brakes */
	double    muBStatic;        /* static brakes-on */
	double    muBKinetic;       /* moving brakes-on */

	double    maxNWDef;         /* maximum nosewheel deflection (RAD) */
	VPoint    rm, rn;           /* location if main/nose gear attachments */
	double    Dm, Dn;           /* main/nose oleo damping factor */
	double    Km, Kn;           /* main/nose oleo spring factor */
	double    Gm, Gn;           /* main/nose strut length with tire */
	double    cmMax, cnMax;     /* main/nose maximum oleo extension distance */
	VPoint    tailExtent;       /* as we rotate, this part may drag */

	/* Speed limits at MTOW (0.0 = unknown): */
	double    MTOW;             /* max takeoff weigth (lb) */
	double    alpha_stall;      /* stall AoA (RAD) */
	double    Vs0;              /* stall speed, full flaps (kt) */
	double    Vs1;              /* stall speed, no flaps (kt) */
	double    Vfe;              /* max speed with flaps (kt) */
	double    Vno;              /* normal operation speed (kt) */
	double    Vne;              /* never exceed speed (kt) */

	double    armDelay;         /* arming delay for missiles */
	long      damageBits;       /* initial bit mask of damaged systems */
	long      structurePts;     /* maximum structural damage */

	double    radarOutput;      /* radar output (watts) */
	double    radarTRange;      /* tracking (lock) radar range (nm) */
	double    radarDRange;      /* detection radar range (nm) */
	double    TEWSThreshold;

	long      sCount;           /* number of weapon stations */
	VPoint    wStation[STATIONS];   /* weapon's stations (launch points) */
	weaponStation station[STATIONS];    /* whatcan be at each weapon station */

	int       (*placeProc) (Viewport *, struct _craft *, VMatrix *, VPolySet *ps);
	/* Object placement procedure (for special craft; see place.c) */
	char     *objname;          /* name of file containing the object */
	VObject  *object;           /* what it looks like */
	void      (*resupply) (struct _craft *);    /* the plane's rearm & refuel procedure */
	dis_entity_type entityType; /* craft type used in DIS */
	dis_entity_type altEntityType;  /* alternate craft type used in DIS */

#ifdef HAVE_GVS
	GVS_OBD gvs_object;
#endif

} craftType;


/*
 *  Radar tracking information
 */

typedef struct _radarInfo {
	int       beamID;           /* index number of beam */
	VPoint    rel;              /* location relative to radar set */
	double    d;                /* distance to target (ft) */
	int       targetID;         /* craft id of target */
	int       locked;           /* 1=we have a "lock" */
	int       altDisplay_kft;   /* altitude readout (ft x 1000) */
	int       headingDisplay_deg; /* ground track of target */
	int       x, y;
} radarInfo;



typedef struct _craft {
/*
	This structure describes a particular instance of aircraft.

	The involved reference frames are all right-handled:

	XYZ           world reference frame

	AXYZ          aircraft reference frame, center in CM, x longitudinal
				  pointing forward, y pointing rightward and z pointing
				  downward

	NED           center=aircraft CM, x toward geographic north, y east,
				  z downard

	The involved matrices are:

	c->trihedral or c->AXYZtoNED is a pure rotational matrix that transforms
				  from AXYZ to NED

	c->XYZtoNED   is a roto-translational matrix that transforms from
				  XYZ to NED

	The Euler angles, giving the angular rotations to pass from NED to AXYZ,
	are so normalized (sorted in order of application to obtain the actual
	aircraft orientation):

	c->curHeading rotation around z, [0,2*PI]
	c->curPitch   rotation around y, [-PI/2,PI/2]
	c->curRoll    rotation around x, [-PI,PI]

	The angular velocity vector omega=[p,q,r] with components in AXYZ is:

	c->p          angular speed around x
	c->q          angular speed around y
	c->r          angular speed around z
*/
	int       pIndex;           /* index number of this element in the vector */
	unsigned long flags;        /* flag word */
	int       type;             /* craft type */
	char      name[32];         /* logname of player */
	char      display[32];      /* display name of player */
	int       team;             /* team number, also used for chased plane if this is a chaser */
	double    createTime;       /* creation time of this craft */
	viewer   *vl;               /* list of viewers */
	VMatrix   trihedral;        /* transforms acftXYZ to NED [north-east-down] */
#define AXYZtoNED trihedral
								/* x = forward, y=right wing, z=down */
	VPoint    Cg;               /* Velocity vector in NED system [ft/s] */
	double    VT;               /* airspeed [ft/s] */
	VPoint    Sg;               /* Position in Geocentric system [meters] */
	WorldCoordinates w;         /* The current  (geodetic) position */
	WorldCoordinates wlast;     /* The previous (geodetic) position */
	double    terrain_altitude; /* local altitude of the terrain [m]      */
								/* WARNING. Clients MUST NOT use directly */
								/* this field, but should call            */
								/* terrain_altitude() instead.            */
	double    terrain_altitude_timeout; /* when to update terrain_altitude */
	VMatrix   XYZtoNED;         /* converts geocentric to north-east-down */
	double    rho;              /* current rho value */
	double    mach1;            /* current value of mach 1.0 (fps) */
	double    mach;             /* current mach number */
	double    prevcurH;         /* last curHeading--used near poles for correcting heading */
	double    curHeading, curPitch, curRoll;    /* Euler angles for acft (RAD) */
	double    p, q, r;          /* roll, pitch, and yaw rates (RAD/s) */
	double    alpha, beta;      /* angles of attack and sideslip (RAD) */
	VPoint    G;                /* g-force vector in acft system (units of "G") */
	VPoint    linAcc;           /* acceleration in acft system (ft/s^2) */
	VPoint    prevSg;           /* last interval's Sg value (ft) */

	double    pitchComm;        /* pilot's pitch command  (-1.0 .. 1.0) */
	double    rollComm;         /* pilot's roll command   (-1.0 .. 1.0) */
	double    steerComm;        /* pilot's steer command  (-1.0 .. 1.0) */
	double    rudderComm;       /* pilot's rudder command (-1.0 .. 1.0) */
	int       throttleComm;     /* pilot's throttle command (0 .. 32768) */

	double    Se, Sa, Sr;       /* control surface positions (-1.0 .. 1.0) */
	double    SeTrim;           /* pitch trim setting */
	double    SaTrim;           /* roll trim setting */
	double    curThrust;        /* Current thrust value (lbf) */
	_BOOL     thrust_reverse_on; /* pilot's thrust rev. command */
	double    thrust_reverse_pos; /* thr. rev. factor (-0.5 .. +1.0) */
		/* NOTE: the actual thrust that push the aircraft is a vector
		directed as the x axis and with module curThrust*thrust_reverse_pos,
		possibly yielding a negative value. */
	double    curFlap;          /* current flap setting (RAD) */
	double    flapSetting;      /* current target flap setting (RAD) */
	double    curSpeedBrake;    /* current speed brake position (RAD) */
	double    speedBrakeSetting;    /* current target speed brake setting(rad) */
	double    curGear[3];       /* current ldg gear location (0.0 = up) */
								/* [0] - nose, [1] - right, [2] - left */
	double    curNWDef;         /* Current nosewheel deflection (RAD) */
	int       throttle;         /* thrust setting 0 - 32768 */
	double    rpm;              /* actual engine RPM (0.0 .. 1.0) */
	double    fuel;             /* current fuel on board (lb) */
	double    payload;          /* payload (lb) */

	long      damageBits;       /* bit flags of damaged subsystems */
	long      structurePts;     /* damage pts that can be absorbed */
	double    leakRate;         /* fuel leakage rate (lb/s) */
	double    damageCL;         /* damage induced roll */
	double    damageCM;         /* damage induced pitch */

	int       radarMode;        /* radar mode or HSI or ADF */
	craftType *cinfo;           /* General craft information */
	double    order;            /* temporary value used to sort craft */
	struct _craft *next;        /* next craft in sorted list */

	double    nextRadarTime;    /* time of next Radar frame */
	short     curRadarTarget;   /* our primary "threat" */
	double    targetDistance;   /* distance to primary target [ft] */
	double    targetClosure;    /* closure rate on primary target [fps] */
	VPoint    relPos[MAXPLAYERS];  /* used by radar and missile */
	double    rval[MAXPLAYERS]; /* radar strength seen by us from other craft */
	radarInfo rinfo[32];        /* radar target information */
	int       rtop;             /* number of entries used in rinfo */
								/* relative positions of possible targets */
	char     *leftHUD[6];       /* strings in upper left corner of HUD */
								/* (reserved for weapons' status */
	char     *rightHUD[6];      /* strings in upper right corner of HUD */
								/* (reserved for future use) */
	int       curWeapon;        /* index of currently selected weapon, wtbl[] */
	weaponStation station[STATIONS];    /* what's at each weapon station */
#ifdef HAVE_DIS
	int       disId;            /* DIS identity */
#endif

	/*
		Magnetic heading.
		The pilot can choose to see the MH rather than the TH. In this
		case we need the local magnetic variation (VAR). Since
		calculating the VAR is quite expensive, we store here the last
		value found and the time we expect it be updated.
		See navaid.c for details.
	*/
	int        showMag;              /* show magnetic heading? (flag) */
	double     actualLocalVAR;       /* actual local VAR, positive W */
	double     updActualLocalVAR;    /* when to upd targetLocalVar */
	double     indicatedLocalVAR;    /* indicated local VAR */
	double     updIndicatedLocalVAR; /* when to upd localVAR */

	/*
		Private data structures handled by the corresponding module:
	*/
	void      *drone;                /* see drone module */
	void      *aps;                  /* see aps module */
	void      *effects;              /* see effects module */

#ifdef HAVE_GVS
	GVS_OBI   gvs_instance;     /* ptr to GVS object instance */
#endif

	/* Munition handling (see dis_if.c, ccip.c, missile.c): */
	double    offset;
	long      tracerMod;
	int       tracerVal;
	double    interval;
	short     rounds;
	int       owner;
	double    armTimer;

} craft;

/*
 *  This structure describes a particular instance of observers.
 */

typedef struct _observer {
	VPoint    Sg;               /* Position in Geoparallel system */
	VPoint    viewPoint;        /* Actual viewpoint */
	viewer   *vl;               /* list of viewers (actually 1 struct) */

} observer;

/*
 *  This structure describes a particular instance of a GROUNDC record.
 */

typedef struct _groundc {
	char             name[16];     /* 2nd field of this GROUNDC record */
	char             cstring[16];  /* #RRGGBB string; e.g., "#5EA70F" */
	VColor *         vcptr;        /* Vlib color struct for this groundc */

} groundc;

/*
 *  This structure describes a particular instance of a TEAMLOC record.
 */

typedef struct _teamloc {
	char             sname[6];  /* short name of this TEAMLOC record */
	char             lname[20]; /* long name of this TEAMLOC record */
	WorldCoordinates loc;       /* latitude, longitude, and z (altitude) */
	double           heading;   /* true heading in radians */

} teamloc;

/*
 * The entity_object_map is used to map DIS entity types to a
 * displayable object.
 */

typedef struct _entity_object_map {
	dis_entity_type entity_type;
	dis_entity_type entity_mask;
	char *          object_name;
	VObject         *obj;
} entity_object_map;

/*
 * The munition_map is used to map munition entity/warhead combinations
 * to explosions sizes and damage.
 */

typedef struct _munition_map {
	dis_entity_type entity_type;
	dis_entity_type entity_mask;
	u_short         warhead_type;
	u_short         warhead_mask;
	double          explosion_diameter_meters;
	double          damage_factor;
	int             kinetic_flag;
} munition_map;


/*
 *  Craft type definitions (craft->cinfo->type)
 */

#define CT_FREE          0 /* an unused craft entry */
#define CT_PLANE         1 /* a player */
#define CT_MISSILE       2 /* an air to air missile */
#define CT_CANNON        3 /* a stream of cannon fire */
#define CT_SURFACE       4 /* surface object (e.g. a runway) */
//#define CT_CHASER      5 /* a non-player "watcher" chasing a plane (FIXME: not used) */
#define CT_UNASSIGNED_5  5 /* currently unassigned--should never exist */
#define CT_DRONE         6 /* a target drone */
#define CT_EXPLOSION     7 /* an explosion */
#ifdef HAVE_DIS
#define CT_DIS_PLANE     8 /* external player (on different server) */
#define CT_DIS_MUNITION  9 /* external tracked munition */
#define CT_DIS_CANNON   10 /* external untracked munition (cannon) */
#define CT_RESERVED     11 /* a reserved, uninitialized entry */
#define CT_DIS_STEALTH  12 /* a place-holder for stealths */
#define CT_BOMB         13 /* gravity bomb */
#endif
#define CT_MAX          13 /* any ctype higher than this is invalid */

/*  Flag names table index values */

#define FX_RECORD             0 /* x000001 activate recording function */
#define FX_AFTERBURNER        1 /* x000002 afterburner state */
#define FX_BRAKES             2 /* x000004 wheel brake state */
#define FX_RUNWAY             3 /* x000008 this object is a runway */
#define FX_HAS_GYRO           4 /* x000010 missile is gyroscope equiped */
#define FX_GHANDLE_DN         5 /* x000020 landing gear handle state (1 = DOWN) */
#define FX_GEAR0_UP           6 /* x000040 set when nose gear is completely up */
#define FX_GEAR1_UP           7 /* x000080 set when right gear is completely up */
#define FX_GEAR2_UP           8 /* x000100 set when left gear is completely up */
#define FX_NW_GND_CONTACT     9 /* x000200 nose wheel gnd contact -- we can brake */
#define FX_RW_GND_CONTACT    10 /* x000400 right wheel in contact with ground */
#define FX_LW_GND_CONTACT    11 /* x000800 left wheel in contact with ground */
#define FX_FIXED_OBJECT      12 /* x001000 fixed (surface) object */
#define FX_CHASE_VIEW        13 /* x002000 chase plane view */
#define FX_BLACK_BOX         14 /* x004000 object is from a black box recording */
#define FX_BALLISTIC         15 /* x008000 munition has no target */
#define FX_RADAR_MODE_CHANGE 16 /* x010000 set when user changes radar mode */
#define FX_END_GAME_DRONE    17 /* x020000 drone in CALSPAN end-game mode */
#define FX_UNASSIGNED_18     18 /* x040000 */
#define FX_UNASSIGNED_11     19 /* x080000 */
#define FX_UNASSIGNED_20     20 /* x100000 */
#define FX_UNASSIGNED_21     21 /* x200000 */
#define FX_UNASSIGNED_22     22 /* x400000 */
#define FX_UNASSIGNED_23     23 /* x800000 */
#define FX_MAX               23 /* Any index greater than this is invalid */

/*  Flag word definitions */
/* 20191216 dlcusa -- add object-is-a-runway bit for proper rendering */

#define FL_RECORD            (1<< 0) /* x000001 activate recording function */
#define FL_AFTERBURNER       (1<< 1) /* x000002 afterburner state */
#define FL_BRAKES            (1<< 2) /* x000004 wheel brake state */
#define FL_RUNWAY            (1<< 3) /* x000008 this object is a runway */
#define FL_HAS_GYRO          (1<< 4) /* x000010 missile is gyroscope equiped */
#define FL_GHANDLE_DN        (1<< 5) /* x000020 landing gear handle state (1 = DOWN) */
#define FL_GEAR0_UP          (1<< 6) /* x000040 set when nose gear is completely up */
#define FL_GEAR1_UP          (1<< 7) /* x000080 set when right gear is completely up */
#define FL_GEAR2_UP          (1<< 8) /* x000100 set when left gear is completely up */
#define FL_NW_GND_CONTACT    (1<< 9) /* x000200 nose wheel gnd contact -- we can brake */
#define FL_RW_GND_CONTACT    (1<<10) /* x000400 right wheel in contact with ground */
#define FL_LW_GND_CONTACT    (1<<11) /* x000800 left wheel in contact with ground */
#define FL_FIXED_OBJECT      (1<<12) /* x001000 fixed (surface) object */
#define FL_CHASE_VIEW        (1<<13) /* x002000 chase plane view */
#define FL_BLACK_BOX         (1<<14) /* x004000 object is from a black box recording */
#define FL_BALLISTIC         (1<<15) /* x008000 munition has no target */
#define FL_RADAR_MODE_CHANGE (1<<16) /* x010000 set when user changes radar mode */
#define FL_END_GAME_DRONE    (1<<17) /* x020000 drone in CALSPAN end-game mode */
#define FL_UNASSIGNED_18     (1<<18) /* x040000 */
#define FL_UNASSIGNED_19     (1<<19) /* x080000 */
#define FL_UNASSIGNED_20     (1<<20) /* x100000 */
#define FL_UNASSIGNED_21     (1<<21) /* x200000 */
#define FL_UNASSIGNED_22     (1<<22) /* x400000 */
#define FL_UNASSIGNED_23     (1<<23) /* x800000 */

#define FL_ALL_GND_CONTACT \
	(FL_NW_GND_CONTACT | FL_RW_GND_CONTACT | FL_LW_GND_CONTACT)

/* Radar mode definitions. Since radar+TEWS, ADF and HSI+HSI panel all share
   the same slots in the panel, these constants are global rather than be
   defined inside the respective modules radar.h, adf.h and hsi.h. */

#define RM_OFF      0       /* radar is off */
#define RM_STANDBY  1       /* standby */
#define RM_NORMAL   2       /* track while scan */
#define RM_ACM      4       /* 20x30 acm */
#define RM_STT      8       /* single target track */
#define RM_HSI     16       /* HSI mode */
#define RM_ADF     32       /* ADF mode */
#define RM_DIS_BROWSE 64    /* browse entities */

EXTERN int       debug;

EXTERN viewer *vl_head;   /* linked list of all rendering windows */

EXTERN int recordCount; /* number of players recording info   */
EXTERN int ptblCount;   /* number of active entries in ptbl   */
EXTERN int otblCount;   /* number of entries in otbl          */
EXTERN int ctblCount;   /* number of chasers                  */
EXTERN int ttblCount;   /* number of entries in teamloc table */
EXTERN int gtblCount;   /* number of entries in groundc table */
EXTERN int gtblIndex;   /* index of current groundColor       */
EXTERN double curTime;  /* current time */
EXTERN time_t curTodSec;	/* tv_sec  from most recent gettimeofday */
EXTERN suseconds_t curTodUsec;	/* tv_usec from most recent gettimeofday */

EXTERN craft stbl[MAXSURFACE];  /* table of surface objects */
EXTERN craft ptbl[MAXPLAYERS];  /* table of player aircraft */
EXTERN craft mtbl[MAXPROJECTILES];  /* table of missiles and cannon streams */
EXTERN craft ctbl[MAXPLAYERS];  /* table of chasers */
EXTERN observer otbl[MAXOBSERVERS]; /* table of observers not counted as players */
EXTERN groundc gtbl[MAXGROUNDCS]; /* table of ground colors */
EXTERN teamloc ttbl[MAXTEAMLOCS]; /* table of starting team locations */

#ifdef FLAT_WORLD
EXTERN VPoint teamLoc[3];   /* the centers of each team's airport */
#else
EXTERN WorldCoordinates teamLatLon[3];
#endif
EXTERN double teamHeading[3];   /* heading of plane */

EXTERN double deltaT;   /* Update interval in seconds */
EXTERN double halfDeltaTSquared;        /* 0.5 * deltaT * deltaT */
EXTERN int HUDPixel;        /* index of HUD color in viewport(s) */
EXTERN VColor *HUDColors[16];
EXTERN VColor *HUDColor;
EXTERN int whitePixel;  /* index of white in viewport(s) */
EXTERN VColor *whiteColor;  /* white in viewport(s) */
EXTERN int blackPixel;  /* index of black in viewport(s) */
EXTERN VColor *blackColor;  /* black in viewport(s) */
EXTERN VColor *yellowColor; /* yellow in viewport(s) */
EXTERN VColor *redColor;    /* red in viewport(s) */
EXTERN VColor *groundColor; /* ground color in viewport(s) */
EXTERN VColor *radarColor;  /* radar color */
EXTERN VColor *radarBackgroundColor;        /* radar CRT surface color */
EXTERN VColor *HSIMagentaColor; /* radar color */
EXTERN VColor *attitudeEarthColor;
EXTERN VColor *attitudeSkyColor;
EXTERN VColor *cloudColor;  /* cloud color */
EXTERN int arcadeMode;  /* set by -a switch */
EXTERN double visibility;   /* visibility (meters) */
EXTERN double cbase, ctop;  /* cloud base and tops (ft) */

#ifdef HAVE_DIS
EXTERN int disInUse;        /* true if DIS protocol is used */
EXTERN int disAbsoluteTime; /* this host have true UTC time */
#endif

EXTERN entity_object_map *eo_map;
EXTERN int eo_map_count;
EXTERN munition_map *mun_map;
EXTERN int mun_map_count;

EXTERN int real_delta_t;    /* 1 when using real-time clock */
EXTERN int redraw_interval; /* render scene every n update frames */
EXTERN double frame_interval_millisec;
EXTERN double update_interval_millisec;
EXTERN int watch_frame_rate;
EXTERN int depth_cue_steps;
EXTERN double end_game_threshold_meters;
EXTERN int end_game_mode;
EXTERN int transferEntityIdBits;
EXTERN dis_entity_id subjectEntityID;
EXTERN int subjectEntitySpecified;     /* =1 if subject entity was
										specfied on the command line */
EXTERN double eye_to_screen_cm;  /* pilot eye dist. from screen (cm) */

EXTERN double downward_view_angle_rad; /* downward view angle over the nose (RAD) */

EXTERN char * ctypes[];
EXTERN char * flags[];

#define earth_g 32.17               /* acceleration due to gravity (ft/s^2) */
#define pi  3.14159265358979323846

#define NM              6076.115 /* ft/NM */
#define FPStoKT(v)      ((v) / 6076.115 * 3600.0)
#define KTtoFPS(v)      ((v) * 6076.115 / 3600.0)
#define FEETtoMETERS(v) ((v) * 0.30480060960)
#define METERStoFEET(v) ((v) * 3.2808399)
#define METERStoNM(x)   ((x) / 1853.18)
#define NMtoFEET(x)     ((x) * NM)
#define NMtoMETERS(x)   ((x) * 1853.18)


void   calcGForces(craft * c, VPoint * f, double w);

char * flightCalculations(craft * c);
/*
	Performs flight dynamics calculations. Return NULL on success,
	otherwise return a string reporting the reason why the aircraft
	was destroyed.
*/

double elevatorSetting(craft * c, double qS, double w);

void   eulerToMatrix(double roll, double pitch, double heading, VMatrix * m);
/*
	Build a transformation matrix based on the supplied euler angles.
*/

void   matrixToEuler(VMatrix * mt, double *heading, double *pitch, double *roll);
/*
	Calculate Euler's angles based on the supplied rotational matrix.
*/

void   euler(craft * c);
/*
	Calculate Euler angles c->curRoll, c->curPitch and c->curHeading
	from the matrix c->trihedral. It is a shorter version of

	matrixToEuler(&c->trihedral, &c->curHeading, &c->curPitch, &c->curRoll);
*/

double pm_normalize_roll(double roll);
double pm_normalize_pitch(double pitch);
double pm_normalize_yaw(double yaw);
/*
	These functions return the program idea of normalized Euler's
	angles, as explained above describing the "craft" data type.
*/

double heading(VPoint * x);

void pm_hud_strings_alloc(craft *c);
void pm_hud_strings_free(craft *c);

void pm_thrust_reverse_toggle(craft *c);

void pm_after_burner_toggle(craft *c);
void pm_after_burner_on(craft *c);
void pm_after_burner_off(craft *c);
void printFlags(unsigned long ct);

#undef EXTERN

#endif
