/*
 *  acm : an aerial combat simulator for X
 *  Copyright (C) 1991-1998  Riley Rainey
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 dated June, 1991.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program;  if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
 */

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <Vlib.h>
#include "audio.h"
#include "draw.h"
#include "pm.h"
#include "damage.h"
#ifdef HAVE_DIS
	#include "dis_if.h"
#endif
#include "redraw.h"

#define RADAR_WINDOW_WIDTH  200

#define scanSlope 	2.1445

#define SetSegment(s, xa, ya, xb, yb) {s.x1=u->indicator.a.x+(xa); \
	s.x2=u->indicator.a.x+(xb); s.y1=u->indicator.a.y+(ya); s.y2=u->indicator.a.y+(yb); }


#ifdef FIXME_NOT_USED

typedef enum (
	RM_OFF,         /* radar is off */
	RM_STANDBY,     /* standby */
	RM_NORMAL,      /* track while scan */
	RM_ACM,         /* 20x30 acm */
	RM_STT          /* single target track */
) radar_mode;

typedef struct _radar_data {
	struct _radar_data *next;

	_BOOL  enabled;

	radar_mode mode;

	Segment   radarImage[1024]; /* the last radar frame */
	int       radarImageCount;

} radar_data;

static radar_data *free_list = NULL;

#endif


static char *thirty = "30";
static Point tgt[4];
static _BOOL tgt_valid = FALSE;
static int rftw, rfth, slot_size;


/*
	We'll build an array that contains the amount of radar radiation that
	each craft can "see" coming from another player.  This is used to build
	each players radar threat display.

	rval[i][j] will represent the amount of radar energy that player j sees
	coming from player i.

	For neatness, the rval array has been moved to the ptbl vector.
*/


void
enableRadar(craft * c, int mode)
{
	c->curRadarTarget = -1;

	c->radarMode = mode;

	strcpy(c->rightHUD[0], "R_+__");
	c->rightHUD[1][0] = '\0';
	c->rightHUD[2][0] = '\0';

#ifdef HAVE_DIS
	dis_setRadarMode(c, 1, 1);
#endif
}


void
disableRadar(craft * c)
{
	c->curRadarTarget = -1;

	c->radarMode = RM_OFF;

	c->rightHUD[0][0] = '\0';
	c->rightHUD[1][0] = '\0';
	c->rightHUD[2][0] = '\0';

#ifdef HAVE_DIS
	dis_setRadarMode(c, 0, 1);
#endif
}


static void
GenerateWorldToRadarMatrix(craft * c, VMatrix * m1,
	double el_center, double az_center)
{
	WorldCoordinates *w = &c->w;
	VMatrix   m, ABCtoNED;

	VIdentMatrix(&m);

	VRotate(&m, ZRotation, -w->longitude);
	VRotate(&m, YRotation, w->latitude);
	VRotate(&m, YRotation, DEGtoRAD(90.0));

	VRotate(&m, YRotation, el_center);
	VRotate(&m, ZRotation, -az_center);
	VMatrixTranspose(&c->trihedral, &ABCtoNED);
	VMatrixMultByRank(&m, &ABCtoNED, m1, 3);

}


static void
plotNormalTarget(viewer * u, int x, int y)
{

	SetSegment(u->radarImage[u->radarImageCount], x - 2, y - 2, x - 2, y + 2);
	u->radarImageCount++;
	SetSegment(u->radarImage[u->radarImageCount], x - 2, y + 2, x + 2, y + 2);
	u->radarImageCount++;
	SetSegment(u->radarImage[u->radarImageCount], x + 2, y + 2, x + 2, y - 2);
	u->radarImageCount++;
	SetSegment(u->radarImage[u->radarImageCount], x + 2, y - 2, x - 2, y - 2);
	u->radarImageCount++;
}


static void
plotPrimaryTarget(craft * c, viewer * u, int i, int x, int y)
{

	int       xp, yp, interline;
	char      s[16], lr;
	VPoint    tmp, rel, deltaV;

	double    d, cl, targetAspectAngle;

	xp = u->indicator.a.x + slot_size * 6 / 8;
	yp = u->indicator.a.y + slot_size * 5 / 8;
	interline = (int) (1.5 * rfth + 0.5);

	tgt_valid = TRUE;
	tgt[0].x = u->indicator.a.x + x;
	tgt[0].y = u->indicator.a.y + y - 4;
	tgt[1].x = u->indicator.a.x + x + 4;
	tgt[1].y = u->indicator.a.y + y;
	tgt[2].x = u->indicator.a.x + x;
	tgt[2].y = u->indicator.a.y + y + 4;
	tgt[3].x = u->indicator.a.x + x - 4;
	tgt[3].y = u->indicator.a.y + y;
	FillPolygon(u->v->w, tgt, 4, &u->rz);

/*
 *  Heading of target
 */

	sprintf(s, "  %3.3d", (int) (RADtoDEG(ptbl[i].curHeading)));
	VGetStrokeString(u->v, xp, yp,
					 u->radarImage, &u->radarImageCount,
					 s, strlen(s), rfth);

/*
 *  Groundspeed of target
 */

	rel.x = ptbl[i].Cg.x;
	rel.y = ptbl[i].Cg.y;
	rel.z = 0.0;
	sprintf(s, "%3.3d K", (int) (Vmagnitude(&rel) / NM * 3600.0));
	yp += interline;
	VGetStrokeString(u->v, xp, yp,
					 u->radarImage, &u->radarImageCount,
					 s, strlen(s), rfth);

/*
 *  Relative heading to target.
 */

	tmp.x = METERStoFEET(ptbl[i].Sg.x - c->Sg.x);
	tmp.y = METERStoFEET(ptbl[i].Sg.y - c->Sg.y);
	tmp.z = METERStoFEET(ptbl[i].Sg.z - c->Sg.z);
	VTransform_(&tmp, &c->XYZtoNED, &rel);
	targetAspectAngle = ptbl[i].curHeading - heading(&rel);
	if (targetAspectAngle > 0.0) {
		lr = 'R';
	}
	else {
		lr = 'L';
		targetAspectAngle = -targetAspectAngle;
	}

	sprintf(s, "%3.3d %c", (int) (RADtoDEG(targetAspectAngle) + 0.5), lr);
	yp += interline;
	VGetStrokeString(u->v, xp, yp,
					 u->radarImage, &u->radarImageCount,
					 s, strlen(s), rfth);

/*
 *  Closure rate
 */

	deltaV.x = ptbl[i].Cg.x - c->Cg.x;
	deltaV.y = ptbl[i].Cg.y - c->Cg.y;
	deltaV.z = ptbl[i].Cg.z - c->Cg.z;
	d = Vmagnitude(&rel);
	cl = -(deltaV.x * rel.x + deltaV.y * rel.y + deltaV.z * rel.z) / d;
	c->targetDistance = d;
	c->targetClosure = cl;
	yp += interline;
	sprintf(s, "%3.3d", (int) (cl / NM * 3600.0));
	VGetStrokeString(u->v, xp, yp,
					 u->radarImage, &u->radarImageCount,
					 s, strlen(s), rfth);

/*
 *  Range to target
 */

#define XA	40
#define XB	150

	xp = XA * slot_size / RADAR_WINDOW_WIDTH;
	yp = rfth + 4;
	sprintf(s, "%d", (int) (d / NM));
	VGetStrokeString(u->v, xp + u->indicator.a.x, yp + u->indicator.a.y,
					 u->radarImage, &u->radarImageCount,
					 s, strlen(s), rfth);

/*
 *  Altitude of target
 */

	xp = XB * slot_size / RADAR_WINDOW_WIDTH;
	yp = rfth + 4;
	sprintf(s, "%d", (int) (METERStoFEET(ptbl[i].w.z) / 1000.0));
	VGetStrokeString(u->v, xp + u->indicator.a.x, yp + u->indicator.a.y,
					 u->radarImage, &u->radarImageCount,
					 s, strlen(s), rfth);

	sprintf(c->rightHUD[0], "R_+__");
	sprintf(c->rightHUD[1], "%05.1f", d / NM);
	sprintf(c->rightHUD[2], " %4.4d", (int) (cl / NM * 3600.0));
}


static double
radarFrameInterval(craft * c)
{
	double    dt = 1.0;

	switch (c->radarMode) {
	case RM_OFF:
	case RM_STANDBY:
		dt = 1.0;
		break;
	case RM_NORMAL:
		dt = 1.0;
		break;
	case RM_STT:
		dt = 1.0;
		break;
	}
	return dt;
}


/*
 *  Add information about a target that this aircraft's radar can "see".
 */

static void
clearRadarInfo(craft * c)
{
	c->rtop = 0;
}


static void
addRadarInfo(craft * c, int i, int beam, double d)
{
	if (c->rtop < 16) {
		c->rinfo[c->rtop].rel = c->relPos[i];
		c->rinfo[c->rtop].d = d;
		c->rinfo[c->rtop].targetID = i;
		c->rinfo[c->rtop].beamID = 0;
		c->rinfo[c->rtop].locked = 0;
		++c->rtop;
	}
}


/*
 *  Compute the radar energy arriving at the target.
 */

static void
computeRadarEnergy(craft * c, int i, VMatrix * m, double els, double azs,
				   double energy)
{
	VPoint    r1;
	double    xs, ys, d;

/*
 *  Calculate the coordinates of the target relative to the craft's frame
 *  and the position of the radar search volume.
 */

	r1.x = METERStoFEET(ptbl[i].Sg.x - c->Sg.x);
	r1.y = METERStoFEET(ptbl[i].Sg.y - c->Sg.y);
	r1.z = METERStoFEET(ptbl[i].Sg.z - c->Sg.z);

	VTransform(&r1, m, &c->relPos[i]);

	ptbl[i].rval[c->pIndex] = 0.0;

	if (c->relPos[i].x <= 0.0)
		return;

	xs = c->relPos[i].y / c->relPos[i].x;
	ys = c->relPos[i].z / c->relPos[i].x;

/*
 *  if the absolute values of xs and ys are both less than 1.0, then
 *  we are painting this target with radar energy.  Estimate the value of
 *  the energy that the target sees.
 */

	if ((fabs(xs) <= azs) && (fabs(ys) <= els)) {

		d = Vmagnitude(&c->relPos[i]);

		ptbl[i].rval[c->pIndex] = energy / (4.0 * M_PI * d * d);

/*
 *  Can we detect that ?
 */

/*  return_energy = ptbl[i].rval[c->pIndex] / (4.0 * M_PI * d * d); */

/*
 *  Lookdown?
 */

		if (ys > 0.0) {
			d *= 4.0 / 3.0;
		}

		/*
		 *  Reject targets that would not appear on our radar set
		 */

		/*
		 *  Not a DIS platform?
		 */

		if (ptbl[i].cinfo->entityType.kind != DISKindPlatform) {
			return;
		}

		/*
		 *  Ground-based entity?  Ignore it.
		 */

		if (ptbl[i].cinfo->entityType.kind == DISKindPlatform &&
			ptbl[i].cinfo->entityType.domain == DISDomainLand) {
			return;
		}

/*
 *  For now, compute radar visibility based on fixed ranges.  A better
 *  solution would include factoring estimated target radar cross sections,
 *  antenna size and radar set sensitivity.  Wow. Where would you get that
 *  sort of information?
 */

		if (ptbl[i].flags & FL_ALL_GND_CONTACT) {
			return;
		}

		if (d >= c->cinfo->radarDRange) {
			return;
		}
		else if (d >= c->cinfo->radarTRange) {
			addRadarInfo(c, i, 0, d);
			return;
		}
		else {
			addRadarInfo(c, i, 0, d);
			return;
		}
	}
	else {
		return;
	}
}


static void
computeRadar(craft * c)
{
	VMatrix   m1;
	register int i, j, k;
	double    el_center, az_center, el_width, az_width, e, el_slope, az_slope;
	craft    *c1;

	clearRadarInfo(c);
	for (i = 0; i < MAXPLAYERS; ++i) {
		ptbl[i].rval[c->pIndex] = 0.0;
	}

	if (c->type != CT_DIS_PLANE && 
		c->type != CT_DIS_STEALTH &&
		(c->radarMode == RM_OFF ||
				 c->radarMode == RM_STANDBY || c->radarMode == RM_HSI)) {
		return;
	}

	i = dis_getBeamCount(c);

	for (j = 0; j < i; ++j) {
		dis_getRadarParameters(c, j, &az_center, &az_width,
							   &el_center, &el_width, &e);

/*
 *  Note: the energy value returned by DIS for EM emissions is in units of
 *        dBm.  dBm = 10.0 * log10 (e_watts / 0.001).
 *
 *  Sooo, to convert to watts, we'd do this:
 *
 *  e = 1000.0 * exp10(e / 10.0);
 */

/*
 * "Transmission Test Procedures for DIS Compliance: Distributed  Emission
 *  Regeneration Protocol Family" says that the elevation value is
 *  up-postive, and azimith is left-postive; units in radians (!)
 */

		GenerateWorldToRadarMatrix(c, &m1, el_center, az_center);
		el_slope = atan(el_width / 2.0);
		az_slope = atan(az_width / 2.0);

		for (c1 = ptbl, k = 0; k < MAXPLAYERS; ++k, ++c1) {
			if (c1->type == CT_FREE ||
				c1->type == CT_DIS_STEALTH || 
				c->pIndex == k) {
				continue;
			}
			computeRadarEnergy(c, k, &m1, el_slope, az_slope, e);
		}
	}
}


void
doRadar(craft * c, viewer * u)
{
	register radarInfo *p;
	AWindow  *w;
	int       i, x, y, j = -1, state;
	int       xc, yc;
	double    xs, ys, dmax;
	char     *buf;
	XSegment *seg;
	int       primary = 0;
	_BOOL     time_up = FALSE;
	Rect      rect;

	if ( ! ( c->radarMode == RM_STANDBY
	|| c->radarMode == RM_NORMAL
	|| c->radarMode == RM_ACM
	|| c->radarMode == RM_STT ) )
		return;

	rftw = (int) (0.025 * RectWidth(u->indicator) + 0.5);
	rfth = (int) (1.3 * rftw + 0.5);
	slot_size = RectWidth(u->indicator);

	w = u->v->w;
	AlibSetClipRect(w, &u->indicator);

/*
 *  Compute radar emissions
 */

	seg = u->radarImage;

	computeRadar(c);

/*
 *  Are we ready to give the pilot the next frame ? If not, mark the
 *  radar display as an "Unchanged" area, and return.
 */

	if (curTime >= c->nextRadarTime) {
		time_up = TRUE;
	}

/*
 *  If we fall through to here, then a redraw is required.
 *  There are four possible scenarios:
 *
 *  1) The frame redraw timer has expired.
 *     An entire radar frame, including target information,
 *     must be generated and displayed.  A logical copy
 *     of the image is saved for possible future use.
 *
 *  2) The frame redraw timer has not expired, but some
 *     external event has triggered a frame redraw (e.g.
 *     the window was uncovered after being hidden).
 *     In this case, the frame is redrawn from the
 *     information used to generate the last frame.
 *
 *  3) The frame redraw timer has not expired, but
 *     the user changed radar modes.  In this case, we should
 *     redraw just the legend information (i.e. no targets),
 *     save a copy of what was drawn, and reset the frame
 *     timer.
 *
 *  4) If none of the first three cases are true, then simply
 *     mark the screen region associated with the radar display
 *     as unchanged, and return.
 */

	state = 4;

	if (time_up) {
		state = 1;
	}
	else {
		/*** FIXME
		if (isRedrawRequired(&u->radarState) == TRUE) {
			state = 2;
		}
		***/
		if (c->flags & FL_RADAR_MODE_CHANGE) {
			state = 3;
		}
	}

	/*
	 *  Case 4; nothing interesting happens.  Mark the display as unchanged
	 */

	if (state == 4) {
		/****
			FIXME This code should prevent the updating of the screen
			area, but it does not work because it erases the slot to its
			right (gear+flap) for some obscure reason, probably a bug in
			V/lib/SLDiff.c.	 Skip and continue with the regular redraw.

		ZInfo z;
		z.depth = --w->depth;
		z.color = UnchangedColor;
		AlibFillRectangle(u->w, &rect, u->w->pixel[UnchangedColor]);
		return;
		****/

		state = 1; /* full redraw */
	}

	c->flags &= ~FL_RADAR_MODE_CHANGE;

	if (state == 2) {

		AlibFillRectangle(w, &rect, w->pixel[radarBackgroundColor->cIndex]);
		
		VDrawSegments(u->v,
			u->radarImage, u->radarImageCount,
			w->pixel[radarColor->cIndex]);

		if (tgt_valid) {
			FillPolygon(w, tgt, 4, &u->rz);
		}
		return;
	}

	tgt_valid = FALSE;

/*
 *  Verify that the current target is still visible.
 */

	if (c->curRadarTarget != -1) {
		for (i = 0, p = c->rinfo; i < c->rtop; ++i, ++p) {
			if (c->curRadarTarget == p->targetID) {
				p->locked = 1;
				j = i;
				break;
			}
		}
		if (j == -1) {
			c->curRadarTarget = -1;
			dis_radarTargetChanged(c);
		}
	}

/*
 *  Go find the closest radar return ... that becomes our victim.
 */

	if (state == 1) {
		dmax = 1000.0 * NM;

		if (c->curRadarTarget == -1) {
			for (i = 0, p = c->rinfo, j = -1; i < c->rtop; ++i, ++p) {
				if (p->d < dmax) {
					dmax = p->d;
					j = i;
				}
			}
			if (j != -1) {
				c->curRadarTarget = c->rinfo[j].targetID;
				c->rinfo[j].locked = 1;
				dis_radarTargetChanged(c);
				audio_playSound(c, SoundAPGLockAcquired, FALSE);
			}
		}
	}

	xc = (slot_size + 1) / 2;
	yc = (slot_size + 1) / 2;

/*
 *  Fill background
 */

	AlibFillRectangle(w, &u->indicator, w->pixel[radarBackgroundColor->cIndex]);

/*
 *  Draw reference "thing" and the border
 */

	u->radarImageCount = 0;
	SetSegment(seg[u->radarImageCount], xc, yc, xc - 5, yc + 5);
	u->radarImageCount++;
	SetSegment(seg[u->radarImageCount], xc - 10, yc, xc - 5, yc + 5);
	u->radarImageCount++;
	SetSegment(seg[u->radarImageCount], xc - 10, yc, xc - 18, yc);
	u->radarImageCount++;
	SetSegment(seg[u->radarImageCount], xc, yc, xc + 5, yc + 5);
	u->radarImageCount++;
	SetSegment(seg[u->radarImageCount], xc + 10, yc, xc + 5, yc + 5);
	u->radarImageCount++;
	SetSegment(seg[u->radarImageCount], xc + 10, yc, xc + 18, yc);
	u->radarImageCount++;

/*
 *  Radar dead?
 */

	if (damage_isFunctioning(c, SYS_RADAR) == 0) {
		AlibDrawSegments(w, seg, u->radarImageCount,
			w->pixel[radarColor->cIndex]);
		return;
	}

	/*
	 *  Radar off?
	 */

	if (c->radarMode <= RM_STANDBY) {
		if (c->radarMode == RM_OFF)
			buf = "RADAR OFF";
		else
			buf = "RADAR STBY";
		c->curRadarTarget = -1;
		VGetStrokeString(u->v, u->indicator.a.x + rftw * 4,
						 u->indicator.a.y + rfth * 3,
					seg, &u->radarImageCount, buf, strlen(buf), rfth);
		return;
	}

	c->nextRadarTime = curTime + radarFrameInterval(c);

/*
 *  Show acm and stt HACK
 */

	if (c->radarMode == RM_ACM) {
		buf = "ACM 20x30";
		VGetStrokeString(u->v, u->indicator.a.x + rftw * 4,
			u->indicator.a.y + rfth * 3,
			seg, &u->radarImageCount, buf, strlen(buf), rfth);
	}
	else if (c->radarMode == RM_STT) {
		buf = "STT";
		VGetStrokeString(u->v, u->indicator.a.x + rftw * 4,
			u->indicator.a.y + rfth * 3,
			seg, &u->radarImageCount, buf, strlen(buf), rfth);
	}

/*
 *  Plot reference lines
 */

	i = (slot_size + 3) / 4;
	y = rfth + 3;
	SetSegment(seg[u->radarImageCount], xc, y, xc, y + 4);
	u->radarImageCount++;
	y = slot_size - 10 - rfth;
	SetSegment(seg[u->radarImageCount], xc, y, xc, y + 4);
	u->radarImageCount++;

	x = xc - i;
	y = rfth + 6;
	SetSegment(seg[u->radarImageCount], x, y, x, y + 4);
	u->radarImageCount++;
	y = slot_size - 10 - rfth;
	SetSegment(seg[u->radarImageCount], x, y, x, y + 4);
	u->radarImageCount++;
	VGetStrokeString(u->v, x - rftw + u->indicator.a.x, y + 6 + rfth + u->indicator.a.y,
		seg, &u->radarImageCount, thirty, 2, rfth);

	x = xc + i;
	y = rfth + 6;
	SetSegment(seg[u->radarImageCount], x, y, x, y + 4);
	u->radarImageCount++;
	y = slot_size - 10 - rfth;
	SetSegment(seg[u->radarImageCount], x, y, x, y + 4);
	u->radarImageCount++;
	VGetStrokeString(u->v, x - rftw + u->indicator.a.x, y + 6 + rfth + u->indicator.a.y,
		seg, &u->radarImageCount, thirty, 2, rfth);

	i = (slot_size + 3) / 4;
	x = 2;
	SetSegment(seg[u->radarImageCount], x, yc, x + 4, yc);
	u->radarImageCount++;
	x = slot_size - 6;
	SetSegment(seg[u->radarImageCount], x, yc, x + 4, yc);
	u->radarImageCount++;

	x = 5 + 2 * rftw;
	y = yc - i;
	SetSegment(seg[u->radarImageCount], x, y, x + 4, y);
	u->radarImageCount++;
	x = 3;
	VGetStrokeString(u->v, x + u->indicator.a.x, y + (rfth + 1) / 2 + u->indicator.a.y,
		seg, &u->radarImageCount, thirty, 2, rfth);
	x = slot_size - 6;
	SetSegment(seg[u->radarImageCount], x, y, x + 4, y);
	u->radarImageCount++;

	x = 5 + 2 * rftw;
	y = yc + i;
	SetSegment(seg[u->radarImageCount], x, y, x + 4, y);
	u->radarImageCount++;
	x = 3;
	VGetStrokeString(u->v, x + u->indicator.a.x, y + (rfth + 1) / 2 + u->indicator.a.y,
		seg, &u->radarImageCount, thirty, 2, rfth);
	x = slot_size - 6;
	SetSegment(seg[u->radarImageCount], x, y, x + 4, y);
	u->radarImageCount++;

/*
 *  If the user was simply changing radar modes skip looking
 *  for new targets ..
 */

	if (state != 3) {

/*
 *  Now plot all targets visible to the radar system.
 */


/*
 *  Scan the radar information table and plot targets
 */

		for (i = 0, p = c->rinfo; i < c->rtop; ++i, ++p) {
			if (p->rel.x >= 0.0) {
				xs = p->rel.y / (p->rel.x * scanSlope);
				ys = p->rel.z / (p->rel.x * scanSlope);
				if (fabs(xs) <= 1.0 && fabs(ys) <= 1.0) {
					x = (int) (xs * (double) xc + xc);
					y = (int) (ys * (double) yc + yc);
					if (p->locked) {
						plotPrimaryTarget(c, u, p->targetID, x, y);
						primary = 1;
					}
					else {
						plotNormalTarget(u, x, y);
					}
				}
			}
		}
	}

	if (primary == 0) {
		strcpy(c->rightHUD[0], "R_+__");
		strcpy(c->rightHUD[1], "");
		strcpy(c->rightHUD[2], "");
	}

	AlibDrawSegments(w, u->radarImage, u->radarImageCount,
		w->pixel[radarColor->cIndex]);

	return;
}

/*
 *  Get a new radar target.
 */

int
newRadarTarget(craft * c)
{
	int       i, j = -1;
	radarInfo *p;

	if (c->curRadarTarget == -1) {
		return -1;
	}

/*
 *  Locate the locked target in the list of visible targets.
 */

	for (i = 0, p = c->rinfo; i < c->rtop; ++i, ++p) {
		if (c->curRadarTarget == p->targetID) {
			j = i;
			p->locked = 0;
			break;
		}
	}
	if (j == -1) {
		return -1;
	}
	++j;
	i = (j == c->rtop) ? 0 : j;
	c->rinfo[i].locked = 1;
	return c->rinfo[i].targetID;
}

void
doTEWS(craft * c, viewer * u)
{
	AWindow  *w;
	int       i, x, y, hostile_found = 0;
	VPoint    rel, tmp;
	double    r, r1, m, unit, xo, yo;
	draw_data *dd;

	if ( (c->radarMode & (RM_STANDBY | RM_NORMAL | RM_ACM | RM_STT)) == 0
	&& c->radarMode != RM_OFF )
		return;
	
	slot_size = RectWidth(u->indicator);

	w = u->v->w;
	AlibSetClipRect(w, &u->tuner);
	AlibFillRectangle(w, &u->tuner, w->pixel[radarBackgroundColor->cIndex]);

	if ( c->radarMode == RM_OFF )
		return;

	r = 0.45 * slot_size;
	r1 = 0.06 * r;
	xo = u->tuner.a.x + 0.5*slot_size;
	yo = u->tuner.a.y + 0.5*slot_size;

	dd = draw_new();

	draw_circle(dd, xo, yo, r);

	if( damage_isFunctioning(c, SYS_TEWS) ){
		for (i = 0; i < MAXPLAYERS; ++i) {

			if (c->pIndex == i)
				continue;

			if (c->rval[i] > c->cinfo->TEWSThreshold) {
				VTransform(&ptbl[i].Sg, &c->XYZtoNED, &tmp);
				VReverseTransform_(&tmp, &c->trihedral, &rel);
				m = 1.0/Vmagnitude(&rel);
				rel.x *= m;
				rel.y *= m;
				rel.z *= m;
				unit = sqrt(rel.x * rel.x + rel.y * rel.y);
				if (unit == 0.0) {
					rel.x = 1.0;
					rel.y = 0.0;
				}
				x = (int) (xo + (int) (rel.y / unit * r * 0.7));
				y = (int) (yo - (int) (rel.x / unit * r * 0.7));
				if (c->team == ptbl[i].team) {

					draw_circle(dd, x, y, r1);

				}
				else {

					draw_data *ee;

					ee = draw_new();
					draw_circle(ee, x, y, r1);
					/* FIXME: draw_fill() does not actually fill! */
					draw_fill(ee, u->v, w->pixel[HUDPixel]);
					draw_free(ee);

					hostile_found = 1;
				}
			}
		}

		if (c->vl == u && hostile_found == 1) {
			audio_playSound(c, SoundLockWarning, FALSE);
		}
	}

	draw_stroke(dd, u->v, w->pixel[HUDPixel]);

	draw_free(dd);

}


void
doDroneRadar(craft * c)
{
	int       i, j = -1;
	radarInfo *p;
	double    dmax = 500.0 * NM;
	VPoint    tmp, rel, deltaV;

/*
 *  Compute radar emissions
 */

	computeRadar(c);

/*
 *  If this is not one of our drones, we're done ...
 */

	if (c->type != CT_DRONE) {
		return;
	}

/*
 *  Go find the closest radar return ... that becomes our victim.
 */

	if (c->curRadarTarget == -1) {
		for (i = 0, p = c->rinfo; i < c->rtop; ++i, ++p) {
			if (p->d < dmax) {
				dmax = p->d;
				j = i;
			}
		}
		c->curRadarTarget = j;
		dis_radarTargetChanged(c);
	}

/*
 *  Is the current radar target still visible to the radar set?
 */

	else {
		for (i = 0, p = c->rinfo; i < c->rtop; ++i, ++p) {
			if (c->curRadarTarget == p->targetID) {
				j = i;
			}
		}
		if (j == -1) {
			c->curRadarTarget = -1;
			dis_radarTargetChanged(c);
		}
	}

/*
 *  Compute tracking parameters
 */

	if ((i = c->curRadarTarget) != -1) {
		tmp.x = METERStoFEET(ptbl[i].Sg.x - c->Sg.x);
		tmp.y = METERStoFEET(ptbl[i].Sg.y - c->Sg.y);
		tmp.z = METERStoFEET(ptbl[i].Sg.z - c->Sg.z);
		VTransform_(&tmp, &c->XYZtoNED, &rel);
		deltaV.x = ptbl[i].Cg.x - c->Cg.x;
		deltaV.y = ptbl[i].Cg.y - c->Cg.y;
		deltaV.z = ptbl[i].Cg.z - c->Cg.z;
		c->targetDistance = Vmagnitude(&rel);
		c->targetClosure = -(deltaV.x * rel.x +
				deltaV.y * rel.y + deltaV.z + rel.z) / c->targetDistance;
	}
}


void
setRadarMode(craft * c, int mode)
{
	int       dis_mode;

	switch (mode) {
	case RM_OFF:
		dis_mode = 0;
		break;
	case RM_STANDBY:
		dis_mode = 0;
		break;
	case RM_NORMAL:
		dis_mode = 1;
		break;
	case RM_ACM:
		dis_mode = 2;
		break;
	case RM_STT:
		dis_mode = 3;
		break;
	case RM_HSI:
	case RM_ADF:
	case RM_DIS_BROWSE:
	default:
		dis_mode = 0;
		break;
	}

	c->radarMode = mode;
	c->nextRadarTime = curTime + radarFrameInterval(c);
	c->flags |= FL_RADAR_MODE_CHANGE;
	c->curRadarTarget = -1;
	if (disInUse) {
		dis_setRadarMode(c, dis_mode, 1);
	}
}

