/*
 *  acm : an aerial combat simulator for X
 *  Copyright (C) 1991-1998  Riley Rainey
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 dated June, 1991.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program;  if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
 */

/*
 *  2014-12-29 dlc - Refer to CHANGES text file for details of modifications
 */

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "adf.h"
#include "audio.h"
#include "browse.h"
#include "damage.h"
#include "hsi.h"
#include "hud.h"
#include "instruments.h"
#include "magnetic_compass.h"
#include "panel.h"
#include "place.h"
#include "planes.h"
#include "pm.h"
#include "prompt.h"
#include "radar.h"
#include "redraw.h"
#define RENDER_INCLUDE
#include "render.h"
#include "terminal.h"
#include "terrain.h"


#define lim	((double) 200 * NM)

static VPoint    hex[6] =
{
	{1.00000, 0.00000, 0.0},
	{0.50000, 0.86603, 0.0},
	{-0.50000, 0.86603, 0.0},
	{-1.00000, 0.00000, 0.0},
	{-0.50000, -0.86603, 0.0},
	{0.50000, -0.86603, 0.0}
};


static craft *sortList;

void
freeRendering (void)
{
	/* FIXME: none to release */
}

static void
addGround(craft * c, viewer *u, VColor * color, double z, VPolySet *ps)
{
	VPoint    hex1[6], tmp;
	double    d;
	int       i, j, k;
	VPolygon *p;

	if( render_lighting == RENDER_NIGHT )
		return;

	/*
	 *  We should really be drawing all these hexagons if any of the viewers
	 *  is in depth-cueued mode, not just the real player.
	 */

	if (u->v->flags & VPDepthCueing) {
		k = _VDefaultWorkContext->depthCueSteps - 1;
		for (i = 0; i < _VDefaultWorkContext->depthCueSteps; ++i, --k) {
			d = (i == 0) ? lim : (sqrt(_VDefaultWorkContext->visTable[k]));
			for (j = 0; j < 6; ++j) {

				/* coordinates in NED (meters) frame */

				hex1[j].x = hex[j].x * d;
				hex1[j].y = hex[j].y * d;
				hex1[j].z = z;

				/* convert NED to XYZ */

				VReverseTransform(&hex1[j], &c->XYZtoNED, &tmp);
				hex1[j] = tmp;
			}
			p = VCreatePolygon(6, hex1, color);
			p->assignedDepth = k;
			VPolySet_Add(ps, p);
		}
	}
	/*
	 *  Otherwise draw just a single "ground" hexagon.
	 */

	else {
		d = FEETtoMETERS(lim);
		for (j = 0; j < 6; ++j) {
			hex1[j].x = hex[j].x * d;
			hex1[j].y = hex[j].y * d;
			hex1[j].z = z;
			VReverseTransform(&hex1[j], &c->XYZtoNED, &tmp);
			hex1[j] = tmp;
		}
		VPolySet_Add(ps, VCreatePolygon(6, hex1, color));
	}
}


/*
 *  insertCraft :  Insert a craft into sortList in descending order.
 */

static void
insertCraft(craft * c, craft * p)
{

	double    dist, offset;
	VPoint    ds;
	craft    *q, *prev;

	/*
	 *  Here's a kludge for you:  to avoid polygon clipping, I'm going to 
	 *  cheat and hack a way to get ground objects to display properly.
	 *  if the Z coordinate of an object is zero (i.e. on ground objects),
	 *  I'll add a huge offset to their distance values to force them to be
	 *  plotted first -- and in roughly their correct drawing order.
	 *
	 *  To support automated world outline maps, stbl[0] is considered to be
	 *  the world map, and will always be plotted in the background.
	 */

	offset = (p->Sg.z == 0.0) ? 1000.0 * NM : 0.0;
	if (p == &stbl[0]) {
		offset += 500.0 * NM;
	}
	ds.x = p->Sg.x - c->Sg.x;
	ds.y = p->Sg.y - c->Sg.y;
	ds.z = p->Sg.z - c->Sg.z;
	p->order = dist = Vmagnitude(&ds) + offset;

	if (sortList != (craft *) NULL) {
		for ((q = sortList, prev = (craft *) NULL); q != (craft *) NULL;) {
			if (q->order < dist) {
				p->next = (struct _craft *) q;
				if (prev == (craft *) NULL)
					sortList = p;
				else
					prev->next = (struct _craft *) p;
				break;
			}
			prev = q;
			q = (craft *) q->next;
		}
		if (q == (craft *) NULL) {
			prev->next = (struct _craft *) p;
			p->next = (struct _craft *) NULL;
		}
	}
	else {
		sortList = p;
		p->next = (struct _craft *) NULL;
	}
}


static void
renderCockpitView ( craft *c, viewer *u )
{

	/*
	 *  Build a vector of polygons for all objects in the scene.
	 *
	 *  This vector should be ordered from "most distant" to "closest" so that
	 *  the final display will end up correct.  Rather than generalizing this
	 *  to death, we'll use a few heuristics to get very close to what we need:
	 *
	 *  (0) Build a single polygon to represent the ground.
	 *  (1) Objects on the surface (stbl) are collected first.
	 *  (2) Planes and projectiles (ptbl and mtbl) are first sorted in 
	 *      descending order by their distance from the observer and then 
	 *      polygons are collected.
	 */

	VPolySet *ps, *ps_clipped;
	VPolygon *poly, *poly_clipped;
	craft *p;
	int j;
	VPoint    tmp, vp, fwd, up;
	double    d, local_z;
	short     bgIndex;
	static _BOOL     blink = FALSE;
	static double    blink_toggle_time = 0.0;
	craft     *vc;

	VSetClipRect(u->v, &u->v->rect);

	/*
	 *  pay attention: vc will be the viewer craft information, c will be
	 *  the watched entity craft information
	 */

	vc = c;
	if ( c->type == CT_DIS_STEALTH ) {

		if ( c->vl->viewer_state == ViewerStatePiggyback ) {
			
			c = c->vl->watchedCraft;
		}

	}

	/*
	 *  Set up the eyespace transformation for this viewpoint
	 */

	if (vc->flags & FL_CHASE_VIEW) {

		vp.x = c->Sg.x - c->prevSg.x;
		vp.y = c->Sg.y - c->prevSg.y;
		vp.z = c->Sg.z - c->prevSg.z;
		d = Vmagnitude(&vp);

		/*
		 *  If our velocity is greater than 2 fps, point the view along 
		 *  the velocity vector (approximately); otherwise the view will 
		 *  be along the direction of aircraft orientation.
		 */

		if (d / deltaT > 2.0) {
			vp.x *= FEETtoMETERS(-100.0) / d;
			vp.y *= FEETtoMETERS(-100.0) / d;
			vp.z *= FEETtoMETERS(-100.0) / d;
			vp.x += c->Sg.x;
			vp.y += c->Sg.y;
			vp.z += c->Sg.z;
			fwd = c->Sg;
		}
		else {
			fwd.x = -100.0;
			fwd.y = fwd.z = 0.0;
			VTransform_(&fwd, &c->trihedral, &tmp);
			VReverseTransform_(&tmp, &c->XYZtoNED, &vp);
			vp.x = c->Sg.x + FEETtoMETERS(vp.x);
			vp.y = c->Sg.y + FEETtoMETERS(vp.y);
			vp.z = c->Sg.z + FEETtoMETERS(vp.z);
			
			if (c->cinfo) {
				VTransform_(&c->cinfo->viewPoint, &c->trihedral, &tmp);
			}
			else {
				VPoint zero = {0, 0, 0};
				tmp = zero;
			}
			VReverseTransform_(&tmp, &c->XYZtoNED, &fwd);
			fwd.x = c->Sg.x + FEETtoMETERS(fwd.x);
			fwd.y = c->Sg.y + FEETtoMETERS(fwd.y);
			fwd.z = c->Sg.z + FEETtoMETERS(fwd.z);
		}
		
		tmp.x = tmp.y = 0.0;
		tmp.z = -1.0;
		VReverseTransform_(&tmp, &c->XYZtoNED, &up);
		up.x += vp.x;
		up.y += vp.y;
		up.z += vp.z;
		
	}
	else {
		if (c->cinfo) {
			VTransform_(&c->cinfo->viewPoint, &(c->trihedral), &tmp);
		}
		else {
			tmp.x = tmp.y = tmp.z = 0.0;
		}
		tmp.x = FEETtoMETERS(tmp.x);
		tmp.y = FEETtoMETERS(tmp.y);
		tmp.z = FEETtoMETERS(tmp.z);
		VReverseTransform(&tmp, &c->XYZtoNED, &vp);
		
		VTransform_(&u->viewDirection, &(c->trihedral), &tmp);
		VReverseTransform_(&tmp, &c->XYZtoNED, &fwd);
		fwd.x += vp.x;
		fwd.y += vp.y;
		fwd.z += vp.z;
		
		VTransform_(&u->viewUp, &(c->trihedral), &tmp);
		VReverseTransform_(&tmp, &c->XYZtoNED, &up);
		up.x += vp.x;
		up.y += vp.y;
		up.z += vp.z;
	}
	
	local_z = terrain_altitude(c);
	
	VGetEyeSpace(u->v, vp, fwd, up);
	/*
		In this call to the VGetEyeSpace() function:
		vp    Position of the eye in XYZ world frame (meters)
		fwd   The point you are looking at in XYZ world frame (meters)
		up    The point above your head in XYZ world frame (meters)
	*/
	
	ps = VPolySet_New();

	sortList = (craft *) NULL;

	/*
	 *  With our simple cloud model, we are either above, in, or below the clouds
	 *
	 *  What we display in the outside-cockpit view will depend of where we are.
	 *
	 *  cbase == ctop means no clouds
	 */

	bgIndex = 0;
	if (cbase == ctop) {
		
		addGround(c, u, groundColor, -(local_z - c->w.z), ps);

		poly = VPolySet_First(ps);
		while( poly != NULL ){
			VTransformPolygon(poly, &u->v->eyeSpace);
			poly = VPolySet_Next(ps);
		}
		
		for ((j = 0, p = stbl); j < MAXSURFACE; (++j, ++p)) {
			if (p->type != CT_FREE)
				insertCraft(c, p);
		}
		
		for ((j = 0, p = ptbl); j < MAXPLAYERS; (++j, ++p)) {
			if (p->type != CT_FREE &&
				p->type != CT_DIS_STEALTH &&
				(p != c || c->flags & FL_CHASE_VIEW)) {
				insertCraft(c, p);
			}
		}
		
		for ((j = 0, p = mtbl); j < MAXPROJECTILES; (++j, ++p)) {
			if (p->type != CT_FREE) {
				insertCraft(c, p);
			}
		}
	}
	else if (c->w.z > ctop) {
		
		addGround(c, u, cloudColor, -(ctop - c->w.z), ps);

		poly = VPolySet_First(ps);
		while( poly != NULL ){
			VTransformPolygon(poly, &u->v->eyeSpace);
			poly = VPolySet_Next(ps);
		}
		
		for ((j = 0, p = ptbl); j < MAXPLAYERS; (++j, ++p)) {
			if (p->type != CT_FREE &&
				p->type != CT_DIS_STEALTH &&
				(p != c || c->flags & FL_CHASE_VIEW) &&
				p->w.z > ctop) {
				insertCraft(c, p);
			}
		}
		
		for ((j = 0, p = mtbl); j < MAXPROJECTILES; (++j, ++p)) {
			if (p->type != CT_FREE && p->w.z > ctop) {
				insertCraft(c, p);
			}
		}
	}
	else if (c->w.z > cbase) {
		bgIndex = cloudColor->cIndex;
	}
	else {
		
		addGround(c, u, groundColor, -(local_z - c->w.z), ps);
		addGround(c, u, cloudColor, -(cbase - c->w.z), ps);

		poly = VPolySet_First(ps);
		while( poly != NULL ){
			VTransformPolygon(poly, &u->v->eyeSpace);
			poly = VPolySet_Next(ps);
		}
		
		for ((j = 0, p = stbl); j < MAXSURFACE; (++j, ++p)) {
			if (p->type != CT_FREE)
				insertCraft(c, p);
		}
		
		for ((j = 0, p = ptbl); j < MAXPLAYERS; (++j, ++p)) {
			if (p->type != CT_FREE &&
				p->type != CT_DIS_STEALTH &&
				(p != c || c->flags & FL_CHASE_VIEW) &&
				p->w.z < cbase) {
				insertCraft(c, p);
			}
		}
		
		for ((j = 0, p = mtbl); j < MAXPROJECTILES; (++j, ++p)) {
			if (p->type != CT_FREE && p->w.z < cbase) {
				insertCraft(c, p);
			}
		}
	}
	
	for (p = sortList; p != (craft *) NULL; p = (craft *) p->next)
		placeCraft(u->v, c, u, p, ps);

	/*
	 *  Display this image for each viewer associated with this craft
	 */
	
	/* FIXME: to do */

	/*
     *  Clip all polygons against the view frustrum.
	 *
	 *  Attention: we can't release the ps set of polygons right
	 *  after the clipping because polygons completely visible
	 *  are shared with ps_clipped set.
	 */

	ps_clipped = VPolySet_New();
	poly = VPolySet_First(ps);
	while( poly != NULL ){

		VPolySet_Set(ps, NULL);

		poly = VClipSidedPolygon(poly, u->v->clipPoly);

		if( poly != NULL )
			VPolySet_Add(ps_clipped, poly);
	
		poly = VPolySet_Next(ps);
	}

	/*
	 *  Fill in the sky color
	 */

	if( render_lighting == RENDER_DAYLIGHT )
		VFillRectangle(u->v, &u->v->rect, u->w->pixel[bgIndex]);
	else
		VFillRectangle(u->v, &u->v->rect, u->w->pixel[blackColor->cIndex]);

	/*
	 *  Now the polygons ..
	 */

	poly_clipped = VPolySet_First(ps_clipped);
	while( poly_clipped != NULL ){

		VFillPolygon(u->v, poly_clipped);

		poly_clipped = VPolySet_Next(ps_clipped);
	}

	/*
	 *  Release polygons
	 */

	VPolySet_Free(ps, TRUE);
	ps = NULL;

	VPolySet_Free(ps_clipped, TRUE);
	ps_clipped = NULL;
		
	hud_draw(u);
	
	terminal_draw(u);

	doBrowsePage( vc, u );

	if ( u->viewer_state == ViewerStateNormal ) {

		prompt_draw(u);

		magnetic_compass_draw(u);

		inst_draw(u);

		panel_lights_draw(c, u);

		doRadar(c, u);      /* Radar */
		doTEWS(c, u);		/* TEWS -- doRadar must be called first */

		adf_panel_draw(u);  /* ADF panel */
		adf_draw(u);        /* ADF indicator */

		hsi_panel_draw(u);  /* HSI panel */
		hsi_draw(u);        /* HSI indicator */

		doFlightStatusPage(c, u);

		panel_gear_and_flaps_draw(c, u);

		panel_engine_draw(c, u);

	}

	if( curTime > blink_toggle_time ){
		blink = ! blink;
		if( blink )
			blink_toggle_time = curTime + 0.05;
		else
			blink_toggle_time = curTime + 1.2;
	}

	if (u->noMorJ && u->showMark) { /* Draw cursor marker if appropriate */
	    SetRect(&u->curmark, u->stickX - 11, u->stickY - 11,
				 u->stickX + 14, u->stickY + 12);
	    AlibSetClipRect(u->w, &u->curmark);
	    AlibFillRectangle(u->w, &u->curmark,
				u->w->pixel[redColor->cIndex]);
	    SetRect(&u->curmark, u->stickX -  7, u->stickY -  7,
				 u->stickX + 10, u->stickY +  9);
	    AlibSetClipRect(u->w, &u->curmark);
	    AlibFillRectangle(u->w, &u->curmark,
				u->w->pixel[yellowColor->cIndex]);
	}

	/* Detached window? */
	#define TAKE_COMM "[PRESS `d' TO TAKE COMMANDS]"
	if ( u->hasFocus && ! u->hasComm && ! blink ) {
		int fw, fh;
		int x, y, len;

		len = strlen(TAKE_COMM);
		fw = (int) (0.6 * RectWidth(u->v->rect) / (double) len);
		fh = fw;  /* FIXME */
		x = RectMiddleX(u->v->rect) - (fw * len / 2);
		y = RectMiddleY(u->v->rect) + (fh / 2);
		AlibSetClipRect(u->w, &u->v->rect);
		VDrawStrokeString(u->v, x+1, y, TAKE_COMM, len, fh,
			u->v->w->pixel[blackColor->cIndex]);
		VDrawStrokeString(u->v, x, y, TAKE_COMM, len, fh,
			u->v->w->pixel[whiteColor->cIndex]);
	
	/* Flying in drone mode? */
	#define DRONE_MODE "[DRONE]"
	} else if( u->hasFocus && c->type == CT_DRONE && ! blink ){
		int fw, fh;
		int x, y, len;

		len = strlen(DRONE_MODE);
		fw = (int) (0.4 * RectWidth(u->v->rect) / (double) len);
		fh = fw;  /* FIXME */
		x = RectMiddleX(u->v->rect) - (fw * len / 2);
		y = RectMiddleY(u->v->rect) + (fh / 2);
		AlibSetClipRect(u->w, &u->v->rect);
		VDrawStrokeString(u->v, x+1, y, DRONE_MODE, len, fh,
			u->v->w->pixel[blackColor->cIndex]);
		VDrawStrokeString(u->v, x, y, DRONE_MODE, len, fh,
			u->v->w->pixel[whiteColor->cIndex]);
	
	}

	/*  Expose the completed drawing  */
	VExposeBuffer(u->v);

}


static void
setAudio(craft * c)
{
	/* Set engine sound: */
	if( damage_isFunctioning(c, SYS_ENGINE1 ) )
		audio_setBackgroundSound(c,
			c->rpm,
			(c->flags & FL_AFTERBURNER) ? 1 : 0,
			0.0 /* dummy, ignored */);
	
	/* Stall warning: */
	if( c->damageBits & FLAG_STALL_WARN )
		audio_playSound(c, SoundStallWarning, FALSE);
	else
		audio_stopSound(c, SoundStallWarning);
}


void
doViews(void)
{
	viewer *u;

	for (u = vl_head; u != NULL; u=u->vl_next ) {

		renderCockpitView( u->c, u );

		setAudio(u->c);

	}

}


#ifdef WIN32
static char *flaps_id[] =
{
	MAKEINTRESOURCE(IDB_FLAPS0),
	MAKEINTRESOURCE(IDB_FLAPS1),
	MAKEINTRESOURCE(IDB_FLAPS2),
	MAKEINTRESOURCE(IDB_FLAPS3)
};

#endif

void
setOutsideView(craft * c, viewer *u, ViewDirection v)
{
	if (v == VIEW_CHASE) {
		c->flags |= FL_CHASE_VIEW;
	}
	else {
		c->flags &= ~FL_CHASE_VIEW;
	}

	switch (v) {
	case VIEW_CHASE:
	case VIEW_FORWARD:
		u->viewDirection.x = 1.0;
		u->viewDirection.y = 0.0;
		u->viewDirection.z = 0.0;
		u->viewUp.x = 0.0;
		u->viewUp.y = 0.0;
		u->viewUp.z = -1.0;
		break;
	case VIEW_UP:
		u->viewDirection.x = 0.0;
		u->viewDirection.y = 0.0;
		u->viewDirection.z = -1.0;
		u->viewUp.x = -1.0;
		u->viewUp.y = 0.0;
		u->viewUp.z = 0.0;
		break;
	case VIEW_DOWN:
		u->viewDirection.x = 0.0;
		u->viewDirection.y = 0.0;
		u->viewDirection.z = 1.0;
		u->viewUp.x = 1.0;
		u->viewUp.y = 0.0;
		u->viewUp.z = 0.0;
		break;
	case VIEW_LEFT:
		u->viewDirection.x = 0.0;
		u->viewDirection.y = -1.0;
		u->viewDirection.z = 0.0;
		u->viewUp.x = 0.0;
		u->viewUp.y = 0.0;
		u->viewUp.z = -1.0;
		break;
	case VIEW_RIGHT:
		u->viewDirection.x = 0.0;
		u->viewDirection.y = 1.0;
		u->viewDirection.z = 0.0;
		u->viewUp.x = 0.0;
		u->viewUp.y = 0.0;
		u->viewUp.z = -1.0;
		break;
	case VIEW_AFT:
		u->viewDirection.x = -1.0;
		u->viewDirection.y = 0.0;
		u->viewDirection.z = 0.0;
		u->viewUp.x = 0.0;
		u->viewUp.y = 0.0;
		u->viewUp.z = -1.0;
		break;
	}
}
