/*
 *  acm : an aerial combat simulator for X
 *  Copyright (C) 1991-1998  Riley Rainey
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 dated June, 1991.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program;  if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
 */

#include <X11/Xutil.h>
#include <sys/time.h>
#include "pm.h"
#include "alarm.h"
#include "box.h"
#include "ccip.h"
#include "effects.h"
#include "missile.h"
#include "missilec.h"
#include "drone.h"
#include "events.h"
#include "flaps.h"
#include "m61a1.h"
#include "players.h"
#include "radar.h"
#include "render.h"
#include "viewer.h"
#include "weapon.h"

static double    lastTime = 0.0, sum = 0.0;
static long      frameCount = 0;

int              cur = 0;
static int       warned = 0;

#ifdef HAVE_DIS

#include "dis_if.h"
static long      sendCount = 0;

#define M_TO_FT(x)	((x)/0.3048)

/*
 *  Update DIS entity state corresponding to te given craft
 */

static void
disBroadcast(craft * p)
{
	VPoint    velocity, linearAcceleration;
	double    location[3], disVelocity[3], disLinearAcceleration[3];
	double    orientation[3], angularVelocity[3];
	VPoint    tmp;
	VMatrix   ABCtoXYZ, NEDtoXYZ;
	int       res;
	static double base = -1;

	if (base < 0)
		base = curTime;

/*
 *  Well, this is a bit strange, but ACM's coordinate system for positions
 *  is meters in the Geocentric; for velocities are expressed as feet per
 *  second in the local NED [north-east-down] system.
 */

	location[0] = p->Sg.x;
	location[1] = p->Sg.y;
	location[2] = p->Sg.z;

	tmp.x = FEETtoMETERS(p->Cg.x);
	tmp.y = FEETtoMETERS(p->Cg.y);
	tmp.z = FEETtoMETERS(p->Cg.z);
	VReverseTransform_(&tmp, &p->XYZtoNED, &velocity);
	disVelocity[0] = velocity.x;
	disVelocity[1] = velocity.y;
	disVelocity[2] = velocity.z;

/*
 *  Derive ECI [Geocentric] heading, pitch, roll
 */
	VMatrixTranspose(&p->XYZtoNED, &NEDtoXYZ);
	/* the trihedral is an "ABCtoNED" transformation */
	VMatrixMultByRank(&p->trihedral, &NEDtoXYZ, &ABCtoXYZ, 3);
	matrixToEuler(&ABCtoXYZ, &orientation[0], &orientation[1], &orientation[2]);
	/*
	 *  Body frame angular velocities.
	 */

	angularVelocity[0] = p->p;	/* x-axis */
	angularVelocity[1] = p->q;	/* y-axis */
	angularVelocity[2] = p->r;	/* z-axis */

	/*
	 *  Transform linear acceleration vector
	 *  from body coordinates to ECI system
	 */

	tmp = p->linAcc;
	tmp.x = FEETtoMETERS(tmp.x);
	tmp.y = FEETtoMETERS(tmp.y);
	tmp.z = FEETtoMETERS(tmp.z);
	VTransform_(&tmp, &ABCtoXYZ, &linearAcceleration);
	disLinearAcceleration[0] = linearAcceleration.x;
	disLinearAcceleration[1] = linearAcceleration.y;
	disLinearAcceleration[2] = linearAcceleration.z;

#ifdef notdef
	printf("X = %7.4f %7.4f %7.4f\n",
		   XYZtoABC.m[0][0], XYZtoABC.m[0][1], XYZtoABC.m[0][2]);
	printf("Y = %7.4f %7.4f %7.4f\n",
		   XYZtoABC.m[1][0], XYZtoABC.m[1][1], XYZtoABC.m[1][2]);
	printf("Z = %7.4f %7.4f %7.4f\n",
		   XYZtoABC.m[2][0], XYZtoABC.m[2][1], XYZtoABC.m[2][2]);

	eulerToMatrix(orientation[2], orientation[1], orientation[0],
					 &XYZtoABC);
	printf("---\nX = %7.4f %7.4f %7.4f\n",
		   XYZtoABC.m[0][0], XYZtoABC.m[0][1], XYZtoABC.m[0][2]);
	printf("Y = %7.4f %7.4f %7.4f\n",
		   XYZtoABC.m[1][0], XYZtoABC.m[1][1], XYZtoABC.m[1][2]);
	printf("Z = %7.4f %7.4f %7.4f\n",
		   XYZtoABC.m[2][0], XYZtoABC.m[2][1], XYZtoABC.m[2][2]);

	printf("ECI orientation [deg] (h,p,r) = %7.2f, %7.2f, %7.2f\n",
		   RADtoDEG(orientation[0]), RADtoDEG(orientation[1]),
		   RADtoDEG(orientation[2]));

	printf("ECI angular velocities [deg/sec] (h,p,r) = %7.2f, %7.2f, %7.2f\n",
		   RADtoDEG(angularVelocity[0]), RADtoDEG(angularVelocity[1]),
		   RADtoDEG(angularVelocity[2]));
#endif

	res = dis_entityState(p->disId, location, disVelocity, disLinearAcceleration,
						  orientation, angularVelocity);
	if (res) {
		sendCount++;
#ifdef notdef
		printf("%6.2f: h=%8.2f, p=%8.2f, r=%8.2f, p=%7.2f, q=%7.2f, r=%7.2f\n",
			   curTime - base,
			   RADtoDEG(orientation[0]),
			   RADtoDEG(orientation[1]),
			   RADtoDEG(orientation[2]),
			   RADtoDEG(angularVelocity[0]),
			   RADtoDEG(angularVelocity[1]),
			   RADtoDEG(angularVelocity[2]));
		printf("%.2f %.1f %.1f %.1f %.1f %.1f %.1f\n",
			   curTime - base,
			   RADtoDEG(orientation[0]),
			   RADtoDEG(orientation[1]),
			   RADtoDEG(orientation[2]),
			   RADtoDEG(angularVelocity[0]),
			   RADtoDEG(angularVelocity[1]),
			   RADtoDEG(angularVelocity[2]));
#endif
	}
}

static void
disGetState(craft * p)
{
	static double base = -1;
	double    location[3];
	double    velocity[3];
	double    orientation[3];
	double    disHeading;
	VPoint    tmp;
	VMatrix   ABCtoXYZ;
	int       res;

	if (base < 0)
		base = curTime;

	res = dis_getEntityState(p->disId, location, velocity, orientation);
	/* if (res != 0) should not happen? */

	p->prevSg = p->Sg;
	p->Sg.x = location[0];
	p->Sg.y = location[1];
	p->Sg.z = location[2];

	DISGeocentricToWorldCoordinates((void *) & p->Sg, &p->w);
	GenerateWorldToLocalMatrix(&p->w, &p->XYZtoNED);

	tmp.x = M_TO_FT(velocity[0]);
	tmp.y = M_TO_FT(velocity[1]);
	tmp.z = M_TO_FT(velocity[2]);
	VTransform_(&tmp, &p->XYZtoNED, &p->Cg);

/*
 *  Compute the "ABCtoNED" trihedral from the DIS euler angles
 */

	disHeading = pm_normalize_yaw( orientation[0] );
	if (p->w.latitude > (M_PI_2 * 0.97) || p->w.latitude < (-M_PI_2 * 0.97) ) {
		fprintf(stderr, "disGetState %.6f deg heading ignored (near pole)\n",
			RADtoDEG(disHeading));
	} else p->curHeading = disHeading;
	fprintf(stderr, "disGetState curHeading is %.6f deg\n", RADtoDEG(p->curHeading));
	p->curPitch = pm_normalize_pitch( orientation[1] );
	p->curRoll = pm_normalize_roll( orientation[2] );

	eulerToMatrix(p->curRoll, p->curPitch, p->curHeading,
					 &ABCtoXYZ);
	VMatrixMultByRank(&ABCtoXYZ, &p->XYZtoNED, &p->trihedral, 3);

/*
 *  Now derive NED heading, pitch and roll from adjusted trihedral
 */

	euler(p);

#ifdef notdef
	printf("%d: h=%.0f, p=%.0f, r=%.0f  ",
		   p->disId,
		   RADtoDEG(p->curHeading),
		   RADtoDEG(p->curPitch),
		   RADtoDEG(p->curRoll));
	printf("%.2f %.1f %.1f %.1f\n",
		   curTime - base,
		   RADtoDEG(orientation[0]),
		   RADtoDEG(orientation[1]),
		   RADtoDEG(orientation[2]));
#endif
}

#endif

suseconds_t	/* returns usecs since last call (0 if first call) */
updateSystemClock ()
{
	static time_t todStarted = 0;
	static struct timeval todCur, todPrev;
	time_t secs;
	suseconds_t usecs;

	if ( todStarted != 0 ) {
		todPrev.tv_sec = curTodSec;
		todPrev.tv_usec = curTodUsec;
	}
	gettimeofday(&todCur, NULL);
	if ( todStarted == 0 ) {
		todStarted = todCur.tv_sec;
		todPrev.tv_sec = todCur.tv_sec;
		todPrev.tv_usec = todCur.tv_usec;
	}
	curTodSec = todCur.tv_sec;
	curTodUsec = todCur.tv_usec;
	secs = todCur.tv_sec - todPrev.tv_sec;
	usecs = todCur.tv_usec - todPrev.tv_usec;
	if ( todCur.tv_usec < todPrev.tv_usec ) {
		secs--;
		usecs += 1000000;
	}
	usecs += (secs * 1000000);
	fprintf(stderr, "%05.5u.%06.6u elapsed usecs: %12lu\n",
		curTodSec % 100000, curTodUsec, usecs);
}

void
updateSimTimeFromSystemClock ()
/*
	Update the variable curTime with the current time in seconds
	with the 0 set as the time the program was started (this to
	prevent the number to become too large, so loosing in precision).
*/
{
	static time_t offset = 0;
	suseconds_t diff;

	diff = updateSystemClock();
	if ( offset == 0 ) {
		offset = curTodSec;
		curTime = 0.0;
	}
	curTime = (double) (curTodSec - offset) + (double) curTodUsec / 1000000.0;
	fprintf(stderr, "%05.5u.%06.6u curTime set to %8.2f\n",
		curTodSec % 100000, curTodUsec, curTime);
}


int
update_redraw(void)
{

	int       i, gameEmpty;
	craft    *p;
	char     *killReason;

	updateSimTimeFromSystemClock ();


	if (lastTime == 0.0) {
		lastTime = curTime - 0.1;
	}

/*
 *  Check if "curTime" ever equals "lastTime", if so, warn that REAL_DELTA_T
 *  probably should not be used on this system.
 */

	if (curTime == lastTime) {
		curTime += 0.0001;
		if (warned == 0) {
			warned = 1;
			fprintf(stderr,
					"update_redraw() Hmm. We seem to have performed a complete inner loop in zero elapsed time.\n\
This seems suspicious.  Perhaps gettimeofday() has a relatively coarse\n\
granularity on this system.\n\
This server should probably be recompiled without REAL_DELTA_T defined.\n");
		}
	}

	if (real_delta_t) {

		deltaT = curTime - lastTime;

/*
 *  If the amount of real elapsed time is unreasonably long, then
 *  make it one update interval.
 */

#ifdef FIXME
		if (deltaT > (double) ((UPDATE_INTERVAL / 1000000.0) * 4)) {
			deltaT = UPDATE_INTERVAL / 1000000.0;
		}
#endif

	}
	else {
		deltaT = update_interval_millisec / 1000.0;
	}

	lastTime = curTime;
	halfDeltaTSquared = 0.5 * deltaT * deltaT;

#ifdef HAVE_DIS
	if (disInUse) {
		if (disAbsoluteTime)
			dis_setTimeAbsolute();
		else
			dis_setTime(curTime);
		dis_receive();
	}
#endif

	gameEmpty = 1;

	events_window_keyb_buttons();

	for ((i = 0, p = ptbl); i < MAXPLAYERS; (++i, ++p)) {
		if ((p->type == CT_PLANE) && !(p->flags & FL_BLACK_BOX)) {
			gameEmpty = 0;

			killReason = flightCalculations(p);
			if (killReason != NULL) {
				killPlayer(p, killReason);
				continue;
			}
			fprintf(stderr, "post-fc: %s curHeading %12e radians, %.6f degrees\n",
				p->cinfo->name, p->curHeading, RADtoDEG(p->curHeading));
			doWeaponUpdate(p);
			flaps_update(p);
#ifdef HAVE_DIS
			if (disInUse)
				disBroadcast(p);
#endif
		}
	}

	for ((i = 0, p = ptbl); i < MAXPLAYERS; (++i, ++p)) {
		if ((p->type == CT_DRONE) && !(p->flags & FL_BLACK_BOX)) {

/*
 *  Determine what actions the drone will take
 */

			if( drone_calculations(p) ){
				killPlayer(p, "mission completed, self destroy");
				continue;
			}

/*
 *  Calling flightCalculations makes drones adhere to the flight model.
 */

			fprintf(stderr, "pre-fc:  %s curHeading %12e radians, %.6f degrees\n",
				p->cinfo->name, RADtoDEG(p->curHeading));
			killReason = flightCalculations(p);
			if (killReason != NULL) {
				killPlayer(p, killReason);
				continue;
			}
			fprintf(stderr, "post-fc: %s curHeading %12e radians, %.6f degrees\n",
				p->cinfo->name, p->curHeading, RADtoDEG(p->curHeading));

			doWeaponUpdate(p);
			flaps_update(p);

			doDroneRadar ( p );

#ifdef HAVE_DIS
			if (disInUse)
				disBroadcast(p);
#endif
		}
	}

#ifdef HAVE_DIS
	if (disInUse) {
		fprintf(stderr, "%s\n", "updateRedraw is calling disGetState");
		for ((i = 0, p = ptbl); i < MAXPLAYERS; (++i, ++p)) {
			if (p->type == CT_DIS_PLANE)
				disGetState(p);
		}
		for ((i = 0, p = mtbl); i < MAXPROJECTILES; (++i, ++p)) {
			if (p->type == CT_DIS_MUNITION)
				disGetState(p);
		}
	}
#endif

	blackBoxInput();

	for ((i = 0, p = mtbl); i < MAXPROJECTILES; (++i, ++p)) {
		if (p->type == CT_MISSILE) {
			if (missileCalculations(p) == 1) {
				killMissile(p, NULL);
			}
#ifdef HAVE_DIS
			else if (disInUse) {
				disBroadcast(p);
			}
#endif
		}
		else if (p->type == CT_CANNON || p->type == CT_DIS_CANNON) {
			if (cannonCalculations(p) == 1) {
				killCannonShell(p, NULL);
			}
		}
		else if(p->type == CT_BOMB) {
			if (ccip_bombCalculations(p) == 1) {
				ccip_killBomb(p);
			}
		}
		else if (p->type == CT_EXPLOSION) {
			if ( effects_explosion_calculations(p) ) {
				effects_kill_explosion(p);
			}
		}
	}

	lookForImpacts();

	if (real_delta_t) {
		doViews();
	}
	else if (cur++ % redraw_interval == 0) {
		doViews();
	}

	blackBoxOutput();

	alarm_update();

	if (watch_frame_rate) {
		sum += deltaT;
		frameCount++;
		if (frameCount % 100 == 0) {
			printf("rate is %lf frames/second\n", (double) frameCount / sum);
#ifdef HAVE_DIS
			if (disInUse)
				printf("DIS send rate is %lf packets/second\n",
				   (double) sendCount / sum);
			sendCount = 0;
#endif
			sum = deltaT;
			frameCount = 1;
		}
	}

	return 0;
}
