/*
 *  ACM-5.0-ico
 *  3-D draw utilities
 *  Copyright (C) 2008  Umberto Salsi
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 dated June, 1991.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program;  if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
 */

/*
	The module VPath implements collections of "tracks", every track being
	a collection of two or more VPoint. 3-D paths can then be transformed
	with a matrix and projected on the x-y plane of the screen.
*/

#ifndef _VPATH_H
#define _VPATH_H

#include <Alib.h>
#include <Vlib.h>

typedef struct _VPath VPath;

VPath * vpath_new(void);
/*
	Create new empty VPath. The drawing pen is located by default to the
	origin point (0,0,0).
*/

void vpath_moveTo(VPath *p, VPoint *to);
/*
	Move the drawing pen to the given absolute point, beginning a new
	track.
*/


void vpath_lineTo(VPath *p, VPoint *to);
/*
	Move the pen from the current pen position to the given destination
	point, drawing a segment.
*/

void vpath_arc(VPath *p, VPoint *center, VPoint *final, int n);
/*
	Starting from the current pen position as set by vpath_moveTo() or
	vpath_lineTo(), draw an arc of ellipse having the given center and the
	final point. n>=1 is the number of intermediate segments, higher values
	give a smoother line.
*/

void vpath_draw_string(VPath *path, char *str, int len, VMatrix *m);
/*
	Draw the string on the xy plane starting from the origin and
	proceeding along the x axis as baseline. Characters have the
	dimension 1.0x1.0.  m is the transformation matrix.  For example,
	this code will add to the VPath the string "Hello" centered around
	the point (h,v) of the screen:

	VMatrix m;
	double fw, fh;
	char *s = "Hello";

	fw = fh = 10.0;
	VIdentMatrix(&m);
	VScaleMatrix(&m, fw, fh, 1.0);
	VTranslate(&m, h - fw*strlen(s)/2, v + fh/2, 0.0);
	vpath_draw_string(path, s, strlen(s), m);
*/

void vpath_stroke(VPath *p, VMatrix *m, AWindow *w, Color color);
/*
	Translate the 3-D VPath into 2-D segments with a parallel projection
	from (x,y,z) 3-D coords to (x,y) 2-D coords in the screen frame
	(x=right, y=down, z=toward the screen). Every point get transformed as
	per the provided matrix. Segments are clipped to the z < 0 space (=
	front of the screen) so to ensure that only the wanted part of the
	drawing be visible either translate along the z axis or extrude setting
	z scale negative.
*/

void vpath_perspective_stroke(VPath *p, VMatrix *m, Viewport *v, Color color);
/*
	Perspective projection of the vpath on the screen. The matrix m is
	applied to every point of the vpath to obtain the point q=m*p to
	project on the given viewport, current clipping region. The unit of q
	is the meter, so q=(0,0,1) is the point exactly one meter in front to
	the eye of the observer, and q=(1,0,0) is one meter to the right.
*/

void vpath_free(VPath *p);
/*
	Release VPath
*/

#endif
